{$N+}
{$E+}
{
        Program Name : APM
        Package Name : Subroutine Library - Layer 1
        Purpose      : Application Performance Monitor
        Institution  : Conroy & Conroy Co.
        Date Written : 31-January-2005
        Written By   : Alan Conroy
        Version      : 1.0

	Copyright (C) 2005 by Alan Conroy.  Released to the Public Domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          The classes in this unit are used to provide an OO performance monitor
        for an application.
}

unit APM ;

interface

uses { Subroutine Library - Layer 0... }
     _DebugIn, { PDebug_Interface }

     { Subroutine Library - Layer 1... }
     _APM, { TSApplication_Performance_Monitor }
     Collect ; { TString_Collection }

{$I EDEFINES.INC}

{$IFDEF DELPHI}
type TStandard_Application_Performance_Monitor = class( TApplication_Performance_Monitor )
                                                     private
                                                         _Mult : double ;
                                                         Transactions : TString_Collection ;

                                                     public { Constructors and destructors... }
                                                         constructor Create ;
                                                         destructor Destroy ;
                                                            override ;

                                                     public { API... }
                                                         function Begin_Transaction : integer ;
                                                             override ;
                                                         procedure End_Transaction( TN, Max : integer ;
                                                             const Title : string ) ;
                                                             override ;
                                                         procedure Set_Multiplier( Value : double ) ;
                                                             override ;
                                                         function Get_Multiplier : double ;
                                                             override ;
                                                 end ;

type TStandard_COM_Application_Performance_Monitor = class( TCOM_Application_Performance_Monitor )
                                                         private
                                                             _Mult : double ;
                                                             Transactions : TString_Collection ;

                                                         public { Constructors and destructors... }
                                                             constructor Create ;
                                                             destructor Destroy ;
                                                                override ;

                                                         public { API... }
                                                             function Begin_Transaction : integer ;
                                                                 override ;
                                                             procedure End_Transaction( TN, Max : integer ;
                                                                 const Title : string ) ;
                                                                 override ;
                                                             procedure Set_Multiplier( Value : double ) ;
                                                                 override ;
                                                             function Get_Multiplier : double ;
                                                                 override ;
                                                     end ;
{$ENDIF}

type TSStandard_Application_Performance_Monitor = object( TSApplication_Performance_Monitor )
                                                      private
                                                          _Mult : double ;
                                                          Transactions : TString_Collection ;

                                                      public { Constructors and destructors... }
                                                          constructor Init ;
                                                          destructor Done ;
                                                             virtual ;

                                                      public { API... }
                                                          function Begin_Transaction : integer ;
                                                              virtual ;
                                                          procedure End_Transaction( TN, Max : integer ;
                                                              const Title : string ) ;
                                                              virtual ;
                                                          procedure Set_Multiplier( Value : double ) ;
                                                              virtual ;
                                                          function Get_Multiplier : double ;
                                                              virtual ;
                                                  end ;
     PSStandard_Application_Performance_Monitor = ^TSStandard_Application_Performance_Monitor ;

implementation

uses
{$IFDEF Delphi}
     { Borland... }
     Windows, // GetTickCount
     SysUtils, // inttostr
{$ENDIF}

     { Other... }
     _App ; { Application_Manager }

{ TSStandard_Application_Performance_Monitor methods... }

{ Constructors and destructors... }

constructor TSStandard_Application_Performance_Monitor.Init ;

begin
    inherited Init ;

    Transactions.Init( 0, 4 ) ;
end ;


destructor TSStandard_Application_Performance_Monitor.Done ;

begin
    Transactions.Done ;

    inherited Done ;
end ;


{ API... }

function TSStandard_Application_Performance_Monitor.Begin_Transaction : integer ;

begin
{$IFDEF Delphi}
    if( _Mult <= 0 ) then // Disabled
    begin
        Begin_Transaction := 0 ;
        exit ;
    end ;
    Transactions.Insert( inttostr( GetTickCount ) ) ;
{$ENDIF}

    Begin_Transaction := Transactions.Count - 1 ;
end ;


procedure TSStandard_Application_Performance_Monitor.End_Transaction( TN, Max : integer ;
    const Title : string ) ;

var Temp : double ;
    Timing : integer ;

begin
{$IFDEF Delphi}
    if( _Mult <= 0 ) then // Disabled
    begin
        exit ;
    end ;
    if( ( TN >= 0 ) and ( TN < Transactions.Count ) ) then
    begin
{$WARNINGS OFF}
  {$R-}
        Timing := GetTickCount - strtoint64( Transactions.At( TN ) ) ;
  {$R+}
{$WARNINGS ON}
        if( TN = Transactions.Count - 1 ) then
        begin
            Transactions.AtDelete( TN ) ;
        end ;
        Temp := Max ;
        if( _Mult > 0 ) then
        begin
            Temp := Temp * _Mult ;
        end ;
        if( ( Timing > trunc( Temp ) ) and ( Application_Manager <> nil ) ) then
        begin
            Application_Manager^.Popup_Message( Title + ' execution exceeds ' +
                inttostr( Timing ) + ' ms' ) ;
        end ;
    end ;
{$ENDIF}
end ;


procedure TSStandard_Application_Performance_Monitor.Set_Multiplier( Value : double ) ;

begin
    _Mult := Value ;
end ;


function TSStandard_Application_Performance_Monitor.Get_Multiplier : double ;

begin
    Get_Multiplier := _Mult ;
end ;



{$IFDEF DELPHI}
{ TStandard_COM_Application_Performance_Monitor methods... }

{ Constructors and destructors... }

constructor TStandard_COM_Application_Performance_Monitor.Create ;

begin
    inherited Create ;

    Transactions.Init( 0, 4 ) ;
end ;


destructor TStandard_COM_Application_Performance_Monitor.Destroy ;

begin
    Transactions.Done ;

    inherited Destroy ;
end ;


{ API... }

function TStandard_COM_Application_Performance_Monitor.Begin_Transaction : integer ;

begin
{$IFDEF Delphi}
    Transactions.Insert( inttostr( GetTickCount ) ) ;
{$ENDIF}

    Result := Transactions.Count ;
end ;


procedure TStandard_COM_Application_Performance_Monitor.End_Transaction( TN, Max : integer ;
    const Title : string ) ;

var Temp : double ;

begin
{$IFDEF Delphi}
    if( ( TN < 0 ) or ( TN >= Transactions.Count ) ) then
    begin
{$WARNINGS OFF}
        Temp := GetTickCount - strtoint( Transactions.At( TN ) ) ;
{$WARNINGS ON}
        Temp := Temp * _Mult ;
        if( ( trunc( Temp ) > Max ) and ( Application_Manager <> nil ) ) then
        begin
            Application_Manager^.Popup_Message( Title + ' execution exceeds ' + inttostr( trunc( Temp ) ) ) ;
        end ;
    end ;
{$ENDIF}
end ;


procedure TStandard_COM_Application_Performance_Monitor.Set_Multiplier( Value : double ) ;

begin
    _Mult := Value ;
end ;


function TStandard_COM_Application_Performance_Monitor.Get_Multiplier : double ;

begin
    Get_Multiplier := _Mult ;
end ;



{ TStandard_Application_Performance_Monitor methods... }

{ Constructors and destructors... }

constructor TStandard_Application_Performance_Monitor.Create ;

begin
    inherited Create ;

    Transactions.Init( 0, 4 ) ;
end ;


destructor TStandard_Application_Performance_Monitor.Destroy ;

begin
    Transactions.Done ;

    inherited Destroy ;
end ;


{ API... }

function TStandard_Application_Performance_Monitor.Begin_Transaction : integer ;

begin
{$IFDEF Delphi}
    Transactions.Insert( inttostr( GetTickCount ) ) ;
{$ENDIF}

    Result := Transactions.Count ;
end ;


procedure TStandard_Application_Performance_Monitor.End_Transaction( TN, Max : integer ;
    const Title : string ) ;

var Temp : double ;

begin
{$IFDEF Delphi}
    if( ( TN < 0 ) or ( TN >= Transactions.Count ) ) then
    begin
{$WARNINGS OFF}
        Temp := GetTickCount - strtoint( Transactions.At( TN ) ) ;
{$WARNINGS ON}
        Temp := Temp * _Mult ;
        if( ( trunc( Temp ) > Max ) and ( Application_Manager <> nil ) ) then
        begin
            Application_Manager^.Popup_Message( Title + ' execution exceeds ' + inttostr( trunc( Temp ) ) ) ;
        end ;
    end ;
{$ENDIF}
end ;


procedure TStandard_Application_Performance_Monitor.Set_Multiplier( Value : double ) ;

begin
    _Mult := Value ;
end ;


function TStandard_Application_Performance_Monitor.Get_Multiplier : double ;

begin
    Get_Multiplier := _Mult ;
end ;
{$ENDIF}



end.
