{
        Program Name : CDDLIB
        Package Name : LB
        Purpose      : Common Data Dictionary Utility
        Institution  : Conroy & Conroy Co.
        Date Written : 12-Jan-1997
        Written By   : Alan Conroy
        Version      : 1.0

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

	  This unit provides Common Data Dictionary support.
}

unit CDDLib ;

interface

{$I EDEFINES.INC}

uses { Conroy & Conroy Co. }
     Common, { TBase_Object }
     _UEHDefs, { tUEC }
     Files ; { TFind_Record }

const CDDErr_Facility = 13 ;
      CDDErr_Success = 0 ;
      CDDErr_No_Dictionary_Open = 1 ;
      CDDErr_Already_Defined = 2 ;
      CDDErr_Invalid_Syntax_Type = 3 ;
      CDDErr_Not_Defined = 4 ;

type CDD_Syntax_Type = ( CDDST_Low,
			 CDDST_API, { external procedure/function definitions }
			 CDDST_Array, { array declarations }
			 CDDST_Base_Class, { object definition when no ancestor }
			 CDDST_By_Reference, { parameter passed by reference }
			 CDDST_Class, { object definition }
			 CDDST_Constant, { constant declarations }
			 CDDST_Declaration, { variables within structures }
			 CDDST_Enumeration, { enumeration definition }
			 CDDST_Function, { function definition }
			 CDDST_Method, { object method definitions }
			 CDDST_Object_Ref, { object reference }
			 CDDST_Parameter, { function parameters }
			 CDDST_Private, { private methods }
			 CDDST_Procedure, { procedure definition }
			 CDDST_Protected, { protected methods }
			 CDDST_Public, { public methods }
			 CDDST_Structure, { record structure }
			 CDDST_Subscript, { array subscripts }
			 CDDST_Type, { type conversions }
			 CDDST_Variant, { variant records }
			 CDDST_High ) ;

type TCDD = object( TBase_Object )
		private { Instance data... }
		    Alternate : string ; { Name of alternate CDD file }
		    Nam : string ; { Name of CDD file }
		    Use_Alternate : boolean ; { True if to try Alternate on failures }

		public { Constructors and destructors... }
		    CDD_Error : tUEC ;

		    constructor Init ;

		private { Internal routines... }
		    { Copies from internal CDD file to external DOS file }
		    procedure Extract_File( N, O : string ) ;

		    { Returns true if file found }
		    function File_Found( N : string ) : boolean ;

		    { Copies from external DOS file to internal CDD file }
		    procedure Update_File( N, O : string ) ;

		public { Public methods... }
		    procedure Create_Data_Dictionary( N : string ) ;
		    { Creates a new Data Dictionary file. }

		    procedure Define_Language( N : string ) ;
		    { Defines a new language. }

		    procedure Define_Syntax( N, O : string ; T : CDD_Syntax_Type ) ;
		    { Defines a language syntax definition }

		    procedure Define_Type( N : string ) ;
		    { Defines a new data type. }

		    procedure Delete_Type( N : string ) ;
		    { Deletes specified data type }

		    function Dictionary_Name : string ; { Returns name of DD }

		    function Find_Language( N : string ;
			var Find_Record : TFind_Record ) : string ;

		    function Find_Type( N : string ;
			var Find_Record : TFind_Record ) : string ;

		    procedure Generate( Type_Name, Language, O : string ) ;
		    { Generate given type for given language, output to file O. }

		    procedure Get_Syntax( N, O : string ; T : CDD_Syntax_Type ) ;
		    { Extracts syntax defintion for language N, type T to
		      file O. }

		    procedure Get_Type_Definition( N, O : string ) ;
		    { Extracts type definition N to file O. }

		    procedure Open( N : string ) ;
		    { Defines Data Dictionary to use. }

		    procedure Open_Alternate( N : string ) ;
		    { Opens an alternate dictionary. }

		    procedure Set_Syntax( N, O : string ; T : CDD_Syntax_Type ) ;
		    { Sets syntax type T for language N from file O }

		    procedure Set_Type_Definition( N, O : string ) ;
		    { Sets type defintion N from file O. }
	    end ; { TCDD }
     PCDD = ^TCDD ;

implementation

uses Collect, { TString_Collection }
     CommonUt, { Edit }
     ErrMastr, { Register_Error_Interface }
     FSError,  { FS_Error }
     FSFH, { TFile_Processor }
     Standard, { ERT }
     UEHDefs ; { TStandard_Error_Interface }

{$IFDEF DELPHI}
type tCDD_EI = class( TStandard_Error_Interface )
                            public
                                function Get_Facility : longint ; override ; stdcall ;
                                function Get_Name : PChar ; override ; stdcall ;
                                function Translate_Error( Code : longint ) : PChar ; override ; stdcall ;
                                procedure Done ; override ; stdcall ;
                                function Facility_Version : longint ;
                                    override ; stdcall ;
                                function Version : longint ;
                                    override ; stdcall ;
                  end ;
     pCDD_EI = tCDD_EI ;
{$ELSE}
type tCDD_EI = object( TSStandard_Error_Interface )
                            public
                                function Get_Facility : longint ; virtual ;
                                function Get_Name : PChar ; virtual ;
                                function Translate_Error( Code : longint ) : PChar ; virtual ;
                                function Facility_Version : longint ;
                                    virtual ;
                                function Version : longint ;
                                    virtual ;
                  end ;
     pCDD_EI = ^tCDD_EI ;
{$ENDIF}

function tCDD_EI.Get_Facility : longint ;

begin
    Get_Facility := 13 ;
end ;


const CDD_Error_Name : PChar = 'CDD' ;

function tCDD_EI.Get_Name : PChar ;

begin
    Get_Name := CDD_Error_Name ;
end ;


var CDD_Error_Text : array[ 0..127 ] of char ;

function tCDD_EI.Translate_Error( Code : longint ) : PChar ;

var Error_Text : string ;

begin
    case Code of
	CDDErr_Success : Error_Text := 'Success' ;
	CDDErr_No_Dictionary_Open : Error_Text := 'No Data Dictionary is open' ;
	CDDErr_Already_Defined : Error_Text := 'Already defined' ;
	CDDErr_Invalid_Syntax_Type : Error_Text := 'Invalid syntax type' ;
	CDDErr_Not_Defined : Error_Text := 'Item not defined' ;
    end ;
    move( Error_Text[ 1 ], CDD_Error_Text[ 0 ], length( Error_Text ) ) ;
    Translate_Error := @CDD_Error_Text ;
end ;


{$IFDEF DELPHI}
procedure tCDD_EI.Done ;

begin
    { Done by unit }
end ;
{$ENDIF}


function tCDD_EI.Facility_Version : longint ;

begin
    Facility_Version := 0 ;
end ;


function tCDD_EI.Version : longint ;

begin
    Version := 0 ;
end ;


{ tCDD_EI methods... }

{ Internal support routines... }

function CDDST_To_String( T : CDD_Syntax_Type ) : string ;

begin
    case T of
	CDDST_API : CDDST_To_String := 'API' ;
	CDDST_Array : CDDST_To_String := 'ARRAY' ;
	CDDST_Base_Class : CDDST_To_String := 'OBJECT0' ;
	CDDST_By_Reference : CDDST_To_String := 'BY_REF' ;
	CDDST_Class : CDDST_To_String := 'OBJECT' ;
	CDDST_Constant : CDDST_To_String := 'CONSTANT' ;
	CDDST_Declaration : CDDST_To_String := 'DECLARATION' ;
	CDDST_Enumeration : CDDST_To_String := 'ENUMERATION' ;
	CDDST_Function : CDDST_To_String := 'FUNCTION' ;
	CDDST_Method : CDDST_To_String := 'METHOD' ;
	CDDST_Object_Ref : CDDST_To_String := 'OBJECT_REFERENCE' ;
	CDDST_Parameter : CDDST_To_String := 'PARAMETER' ;
	CDDST_Private : CDDST_To_String := 'PRIVATE' ;
	CDDST_Procedure : CDDST_To_String := 'PROCEDURE' ;
	CDDST_Protected : CDDST_To_String := 'PROTECTED' ;
	CDDST_Public : CDDST_To_String := 'PUBLIC' ;
	CDDST_Structure : CDDST_To_String := 'STRUCTURE' ;
	CDDST_Subscript : CDDST_To_String := 'SUBSCRIPT' ;
	CDDST_Type : CDDST_To_String := 'TYPE' ;
	CDDST_Variant : CDDST_To_String := 'VARIANT' ;
    end ;
end ;


function Temp_File_Name : string ;

var S : string ;

begin
    S := FileName( 'TEMP:CDD.TMP' ) ;
    Temp_File_Name := S ;
end ;


function Temp_File_Name0 : string ;

var S : string ;

begin
    S := Temp_File_Name ;
    Temp_File_Name0 := copy( S, 1, length( S ) - 4 ) + '0.TMP' ;
end ;


function Temp_File_Name1 : string ;

var S : string ;

begin
    S := Temp_File_Name ;
    Temp_File_Name1 := copy( S, 1, length( S ) - 4 ) + '1.TMP' ;
end ;



{ TCDD methods... }

{ Constructors and destructors... }

constructor TCDD.Init ;

begin
    inherited Init ;

{$IFDEF DELPHI}
    Initialize( Alternate ) ;
    Initialize( Nam ) ;
{$ELSE}
    Alternate := '' ;
    Nam := '' ;
{$ENDIF}
    CDD_Error.Code := CDDErr_Success ;
    CDD_Error.Facility := CDDErr_Facility ;
end ;


{ Internal routines... }

{ Copies from internal CDD file to external DOS file }
procedure TCDD.Extract_File( N, O : string ) ;

var Dummy : integer ;
    F : PSFile ;
    Out : text ;
    S : string ;

begin
    F := Files.Open( N, FSOM_Input or FSOM_Text, CDD_Error ) ;
    if( CDD_Error.Code = 0 ) then
    begin
	assign( Out, O ) ;
	{$I-}
	rewrite( Out ) ;
	{$I+}
	Dummy := IOResult ;
	if( Dummy <> 0 ) then
	begin
	    F^.Close ;
	    CDD_Error.Code := Dummy ;
	    CDD_Error.Facility := 1 ;
	    exit ;
	end ;
	while( not F^.EOF ) do
	begin
	    F^.Readln( S ) ;
	    if( F^.IO_Error.Code <> 0 ) then
	    begin
		CDD_Error := F^.IO_Error ;
		break ;
	    end ;
	    {$I-}
	    writeln( Out, S ) ;
	    {$I+}
	    Dummy := IOResult ;
	    if( Dummy <> 0 ) then
	    begin
		CDD_Error.Code := Dummy ;
		CDD_Error.Facility := 2 ;
		break ;
	    end ;
	end ;
	close( Out ) ;
	F^.Close ;
    end else
    begin
	if(
	    ( CDD_Error.Facility = CDDErr_Facility )
	    and
	    (
	      ( CDD_Error.Code = FSErr_FileNotFound )
	      or
	      ( CDD_Error.Code = FSErr_DirectoryNotFound )
	    )
	  ) then
	begin
	    CDD_Error.Code := CDDErr_Not_Defined ;
	    CDD_Error.Facility := CDDErr_Facility ;
	end ;
    end ;
end ; { TCDD.Extract_File }


function TCDD.File_Found( N : string ) : boolean ; { True if file found }

var F : PSFile ;
    UEC : tUEC ;

begin
    File_Found := False ;
    F := Files.Open( Dictionary_Name + '\' + N, FSOM_Input, UEC ) ;
    if( UEC.Code = 0 ) then
    begin
	F^.Close ;
	File_Found := True ;
    end else
    begin
	if( Use_Alternate ) then
	begin
	    F := Files.Open( Alternate + '\' + N, FSOM_Input, UEC ) ;
	    if( UEC.Code = 0 ) then
	    begin
		F^.Close ;
		File_Found := True ;
	    end ;
	end ;
    end ;
end ; { TCDD.File_Found }


{ Copies from external DOS file to internal CDD file }
procedure TCDD.Update_File( N, O : string ) ;

var Dummy : integer ;
    F : PSFile ;
    Out : text ;
    S : string ;

begin
    F := Files.Open( N, FSOM_Output or FSOM_Text, CDD_Error ) ;
    if( CDD_Error.Code = 0 ) then
    begin
	assign( Out, O ) ;
	{$I-}
	reset( Out ) ;
	{$I+}
	Dummy := IOResult ;
	if( Dummy <> 0 ) then
	begin
	    F^.Close ;
	    CDD_Error.Code := Dummy ;
	    CDD_Error.Facility := 2 ;
	    exit ;
	end ;
	while( not eof( Out ) ) do
	begin
	    {$I-}
	    readln( Out, S ) ;
	    {$I+}
	    Dummy := IOResult ;
	    if( Dummy <> 0 ) then
	    begin
		CDD_Error.Code := Dummy ;
		CDD_Error.Facility := 2 ;
		break ;
	    end ;
	    F^.writeln( S ) ;
	    if( F^.IO_Error.Code <> 0 ) then
	    begin
		CDD_Error := F^.IO_Error ;
		break ;
	    end ;
	end ;
	close( Out ) ;
	F^.Close ;
    end ;
end ;


{ Public methods... }

procedure TCDD.Create_Data_Dictionary( N : string ) ;
{ Creates a new Data Dictionary file }

var FS : PFile_System ;
    Res : TUEC ;

begin
    { Create the file system... }
    Initialize_File_System( N, 'Common Data Dictionary V1.0', 0, Res ) ;
    FS := Open_File_System( N, 0, CDD_Error ) ;
    if( CDD_Error.Code = 0 ) then
    begin
	FS^.MKDIR( N + '\TYPES' ) ;
	FS^.MKDIR( N + '\LANGUAGES' ) ;
    end ;
    dispose( FS, Done ) ;
end ;


procedure TCDD.Define_Language( N : string ) ; { Defines a new language. }

var Res : TUEC ;

begin
    { Be sure a dictionary is open }
    if( length( Dictionary_Name ) = 0 ) then
    begin
	CDD_Error.Code := CDDErr_No_Dictionary_Open ;
	CDD_Error.Facility := CDDErr_Facility ;
	exit ;
    end ;

    { Check for existing of same name }
    if( File_Found( 'LANGUAGES\' + N ) ) then
    begin
	CDD_Error.Code := CDDErr_Already_Defined ;
	CDD_Error.Facility := CDDErr_Facility ;
	exit ;
    end ;

    { Create new type name... }
    Res.Code := MKDIR( Dictionary_Name + '\LANGUAGES\' + N ) ;
end ; { Define_Language }


procedure TCDD.Define_Syntax( N, O : string ; T : CDD_Syntax_Type ) ;
{ Defines a language syntax definition. }

var F : PSFile ;

begin
    { Be sure a dictionary is open }
    if( length( Dictionary_Name ) = 0 ) then
    begin
	CDD_Error.Code := CDDErr_No_Dictionary_Open ;
	CDD_Error.Facility := CDDErr_Facility ;
	exit ;
    end ;

    if( ( T <= CDDST_Low ) or ( T >= CDDST_High ) ) then
    begin
	CDD_Error.Code := CDDErr_Invalid_Syntax_Type ;
	CDD_Error.Facility := CDDErr_Facility ;
	exit ;
    end ;

    { Check for existing of same name }
    if( File_Found( 'LANGUAGES\' + N + '\' + CDDST_To_String( T ) ) ) then
    begin
	CDD_Error.Code := CDDErr_Already_Defined ;
	CDD_Error.Facility := CDDErr_Facility ;
	exit ;
    end ;

    { Create new type name... }
    F := Files.Open( Dictionary_Name + '\LANGUAGES\' + N + '\' + CDDST_To_String( T ),
	FSOM_Output or FSOM_Text, CDD_Error ) ;
    if( CDD_Error.Code <> 0 ) then
    begin
	dispose( F, Done ) ;
	exit ;
    end ;
    F^.Writeln( O ) ;
    F^.Close ;
end ; { TCDD.Define_Syntax }


procedure TCDD.Define_Type( N : string ) ; { Defines a new data type. }

var F : PSFile ;

begin
    { Be sure a dictionary is open }
    if( length( Dictionary_Name ) = 0 ) then
    begin
	CDD_Error.Code := CDDErr_No_Dictionary_Open ;
	CDD_Error.Facility := CDDErr_Facility ;
	exit ;
    end ;

    { Check for existing of same name }
    if( File_Found( 'TYPES\' + N ) ) then
    begin
	CDD_Error.Code := CDDErr_Already_Defined ;
	CDD_Error.Facility := CDDErr_Facility ;
	exit ;
    end ;

    { Create new type name... }
    F := Files.Open( Dictionary_Name + '\TYPES\' + N, FSOM_Output or FSOM_Text, CDD_Error ) ;
    if( CDD_Error.Code <> 0 ) then
    begin
	dispose( F, Done ) ;
    end else
    begin
	F^.Close ;
    end ;
end ; { Define_Type }


procedure TCDD.Delete_Type( N : string ) ;
{ Deletes specified data type }

var F : PSFile ;

begin
    CDD_Error.Code := CDDErr_Success ;
    CDD_Error.Facility := CDDErr_Facility ;

    { Be sure a dictionary is open }
    if( length( Dictionary_Name ) = 0 ) then
    begin
	CDD_Error.Code := CDDErr_No_Dictionary_Open ;
	CDD_Error.Facility := CDDErr_Facility ;
	exit ;
    end ;

    F := Files.Open( Dictionary_Name + '\TYPES\' + N, FSOM_Input or FSOM_Text, CDD_Error ) ;
    if( CDD_Error.Code <> 0 ) then
    begin
	CDD_Error.Code := CDDErr_Not_Defined ;
	CDD_Error.Facility := CDDErr_Facility ;
	if( F <> nil ) then
	begin
	    dispose( F, Done ) ;
	end ;
	exit ;
    end ;
    F^.Erase( '\TYPES\' + N ) ;
    F^.Close ;
end ; { TCDD.Delete_Type }


function TCDD.Dictionary_Name : string ; { Returns name of the Data Dictionary }

begin
    Dictionary_Name := Nam ;
end ;


function TCDD.Find_Language( N : string ;
    var Find_Record : TFind_Record ) : string ;

var S : string ;

begin
    CDD_Error.Code := CDDErr_Success ;
    CDD_Error.Facility := CDDErr_Facility ;

    { Be sure a dictionary is open }
    if( length( Dictionary_Name ) = 0 ) then
    begin
	CDD_Error.Code := CDDErr_No_Dictionary_Open ;
	CDD_Error.Facility := CDDErr_Facility ;
	exit ;
    end ;

    if( length( N ) = 0 ) then { Find next }
    begin
	S := Files.Lookup( '', Find_Record ) ;
    end else
    begin
	S := Files.Lookup( Dictionary_Name + '\LANGUAGES\' + N, Find_Record ) ;
    end ;
    Find_Language := copy( S, length( Dictionary_Name ) + 12, length( S ) ) ;
end ; { Find_Language }


function TCDD.Find_Type( N : string ;
    var Find_Record : TFind_Record ) : string ;

var S : string ;

begin
    CDD_Error.Code := CDDErr_Success ;
    CDD_Error.Facility := CDDErr_Facility ;

    { Be sure a dictionary is open }
    if( length( Dictionary_Name ) = 0 ) then
    begin
	CDD_Error.Code := CDDErr_No_Dictionary_Open ;
	CDD_Error.Facility := CDDErr_Facility ;
	exit ;
    end ;

    if( length( N ) = 0 ) then { Find next }
    begin
	S := Files.Lookup( '', Find_Record ) ;
    end else
    begin
	S := Files.Lookup( Dictionary_Name + '\TYPES\' + N, Find_Record ) ;
    end ;
    Find_Type := copy( S, length( Dictionary_Name ) + 8, length( S ) ) ;
end ; { TCDD.Find_Type }


procedure Translate( var Target, Syntax, Values, Specs : TString_Collection ) ;
{ Translate specs and values via Syntax into one or more adds to the Target. }

var Count, Dummy : integer ;
    S, Spec : string ;
    Spec_Loop, Syntax_Loop : integer ;

begin
    for Syntax_Loop := 0 to Syntax.Count - 1 do
    begin
	S := Syntax.At( Syntax_Loop ) ;
	for Spec_Loop := 0 to Specs.Count - 1 do
	begin
	    Spec := Specs.At( Spec_Loop ) ;
	    Count := 0 ;
	    Dummy := pos( Spec, S ) ;
	    while( Dummy > 0 ) do { Found the spec }
	    begin
		inc( Count ) ;
		if( Count > 254 ) then { Too recursive }
		begin
		    exit ;
		end ;
		S := copy( S, 1, Dummy - 1 ) + Values.At( Spec_Loop ) +
			copy( S, Dummy + length( Spec ), length( S ) ) ;
		Dummy := pos( Spec, S ) ;
	    end ;
	end ; { for Spec_Loop := 0 to Specs.Count - 1 }
	Target.Insert( S ) ;
    end ; { for Syntax_Loop := 0 to Syntax.Count - 1 }
end ; { Translate }


procedure Translate2( var Target, Syntax : TString_Collection ;
    Spec1, Value1, Spec2, Value2 : string ) ;
{ Front-end to Translate() for 2 specs }

var Values : TString_Collection ;
    Specs : TString_Collection ;

begin
    Values.Init( 4, 4 ) ;
    Specs.Init( 4, 4 ) ;
    Values.Insert( Value1 ) ;
    Values.Insert( Value2 ) ;
    Specs.Insert( Spec1 ) ;
    Specs.Insert( Spec2 ) ;
    Translate( Target, Syntax, Values, Specs ) ;
    Values.Done ;
    Specs.Done ;
end ; { Translate2 }


procedure Translate3( var Target, Syntax : TString_Collection ;
    Spec1, Value1, Spec2, Value2, Spec3, Value3 : string ) ;
{ Front-end to Translate() for 2 specs }

var Values : TString_Collection ;
    Specs : TString_Collection ;

begin
    Values.Init( 4, 4 ) ;
    Specs.Init( 4, 4 ) ;
    Values.Insert( Value1 ) ;
    Values.Insert( Value2 ) ;
    Values.Insert( Value3 ) ;
    Specs.Insert( Spec1 ) ;
    Specs.Insert( Spec2 ) ;
    Specs.Insert( Spec3 ) ;
    Translate( Target, Syntax, Values, Specs ) ;
    Values.Done ;
    Specs.Done ;
end ; { Translate3 }


procedure Translate4( var Target, Syntax : TString_Collection ;
    Spec1, Value1, Spec2, Value2, Spec3, Value3, Spec4, Value4 : string ) ;
{ Front-end to Translate() for 2 specs }

var Values : TString_Collection ;
    Specs : TString_Collection ;

begin
    Values.Init( 4, 4 ) ;
    Specs.Init( 4, 4 ) ;
    Values.Insert( Value1 ) ;
    Values.Insert( Value2 ) ;
    Values.Insert( Value3 ) ;
    Values.Insert( Value4 ) ;
    Specs.Insert( Spec1 ) ;
    Specs.Insert( Spec2 ) ;
    Specs.Insert( Spec3 ) ;
    Specs.Insert( Spec4 ) ;
    Translate( Target, Syntax, Values, Specs ) ;
    Values.Done ;
    Specs.Done ;
end ; { Translate4 }


procedure Load_String_Collection( var List : TString_Collection ;
    File_Name : string ; var UEC : tUEC ) ;

var Dummy : integer ;
    f : text ;
    s : string ;

begin
    { Setup }
    UEC.Code := 0 ; { Assume success }

    { Open the file... }
    assign( F, File_Name ) ;
    {$I-}
    reset( F ) ;
    {$I+}
    Dummy := IOResult ;
    if( Dummy <> 0 ) then
    begin
	UEC.Code := Dummy ;
	UEC.Facility := 2 ;
	exit ;
    end ;

    { Load from file }
    while not eof( F ) do
    begin
	{$I-}
	readln( F, S ) ;
	{$I+}
	Dummy := IOResult ;
	if( Dummy <> 0 ) then
	begin
	    UEC.Code := Dummy ;
	    UEC.Facility := 2 ;
	    close( F ) ;
	    exit ;
	end ;
	List.Insert( S ) ;
    end ;

    { All done }
    close( F ) ;
end ; { Load_String_Collection }


type TDefinition_Type = ( TDT_Simple, TDT_Structure, TDT_Object, TDT_API ) ;

procedure TCDD.Generate( Type_Name, Language, O : string ) ;
{ Generate given type for given language, output to file O. }

var Translation : TString_Collection ;

    function Translate_Type( Original : string ) : string ;
    { Translate type specification in S into specification in Translation
      string collection and return that translation }

    var Dummy : integer ; { Position of tab separator }
	Last_Non_Number : char ;
	Loop : integer ; { Loop control }
	Maximum : integer ;
	S : string ;
	Temp : string ; { Current translation string }
	Work, Work1 : integer ; { Misc. }

    begin
	{ Normalize... }
	S := Edit( Original, 32 ) ;
	if( copy( S, 1, 8 ) = 'INTEGER ' ) then
	begin
	    { Default to SIGNED }
	    if( pos( copy( S, 9, 1 ), '0123456789' ) > 0 ) then
	    begin
		S := copy( S, 1, 8 ) + 'SIGNED ' +
			copy( S, 9, length( S ) ) ;
	    end ;
	end ;
	if( copy( S, 1, 7 ) = 'STRING ' ) then
	begin
	    { Default to STATIC }
	    if( pos( copy( S, 8, 1 ), '0123456789' ) > 0 ) then
	    begin
		S := copy( S, 1, 7 ) + 'STATIC ' +
			copy( S, 8, length( S ) ) ;
	    end ;
	end ;

	{ Now find a match }
	for Loop := 0 to Translation.Count - 1 do
	begin
	    { Get next type definition }
	    Temp := Translation.At( Loop ) ;

	    { Handle MAX... }
	    Dummy := pos( ' MAX ', Temp ) ;
	    if( Dummy > 0 ) then
	    begin
		val( copy( Temp, Dummy + 5, length( Temp ) ), Maximum, Work ) ;
		Temp := copy( Temp, 1, Dummy - 1 ) ; { Strip specification }
	    end else
	    begin
		Maximum := -1 ;
	    end ;

	    { Handle parameterized type definition... }
	    Work := pos( '~n', Temp ) ;
	    while( Work > 0 ) do
	    begin
		{ Search for numeric portion of passed type... }
		Dummy := 1 ;
		Last_Non_Number := ' ' ;
		while(
		       ( Dummy <= length( S ) )
		       and
		       (
			 ( pos( S[ Dummy ], '0123456789' ) = 0 )
			 or
			 ( Last_Non_Number <> ' ' )
		       )
		     ) do
		begin
		    if( pos( S[ Dummy ], '0123456789' ) = 0 ) then
		    begin
			Last_Non_Number := S[ Dummy ] ;
		    end ;
		    inc( Dummy ) ;
		end ;
		if( Dummy < length( S ) ) then { Found a number }
		begin
		    { Find end of number... }
		    Work1 := 1 ;
		    while(
			   ( Work1 <= length( S ) )
			   and
			   ( pos( S[ Dummy + Work1 ], '0123456789' ) > 0 )
			 ) do
		    begin
			inc( Work1 ) ;
		    end ;
		    { Substitute the number for "~n" and drop through to
		      normal processing }
		    Temp := copy( Temp, 1, Work - 1 ) +
			copy( S, Dummy, Work1 ) +
			copy( Temp, Work + 2, length( Temp ) ) ;
		end else
		begin
		    break ; { No number to substitute }
		end ;
		Work := pos( '~n', Temp ) ;
	    end ; { while( Work > 0 ) }

	    { Handle simple type definitions... }
	    Dummy := pos( #9, Temp ) ;
	    if( S = copy( Temp, Dummy + 1, length( Temp ) ) ) then
	    begin { Found a match }
		S := copy( Temp, 1, Dummy - 1 ) ;
		Translate_Type := S ;
		exit ;
	    end ;
	end ; { for Loop := 0 to Translation.Count - 1 }
	Translate_Type := Original ;
	{ Return type unchanged - no translation found }
    end ; { TCDD.Generate.Translate_Type }

var Declaration_Syntax, Definition_List : TString_Collection ;
    Definition_Type : TDefinition_Type ;
    Dummy : integer ;
    Object_Syntax : TString_Collection ;
    Output : text ; { Requested output file }
    S : string ;
    Structure_Syntax : TString_Collection ;
    Temp : text ; { Temporary file for building definition }
    Type_List : TString_Collection ;

    function Output_String_Collection( var List : TString_Collection ;
	Prefix : string ) : integer ;

    var Dummy : integer ;
	Loop : integer ;

    begin
	{ Substitute the definition... }
	Dummy := 0 ;
        for Loop := 0 to Definition_List.Count - 1 do
	begin
	    {$I-}
	    writeln( Output, Prefix, Definition_List.At( Loop ) ) ;
	    {$I+}
	    Dummy := IOResult ;
	    if( Dummy <> 0 ) then
	    begin
		CDD_Error.Code := Dummy ;
		CDD_Error.Facility := 2 ;
		break ;
	    end ;
	end ;
	Output_String_Collection := Dummy ;
    end ;


    var Parse_Line : string ;

    function Token : string ; { Return next token from Temp file }

    var Dummy : integer ;

    begin
	Token := '' ; { Assume EOF}
	while( length( Parse_Line ) = 0 ) do
	begin
	    if( eof( Temp ) ) then
	    begin
		exit ;
	    end ;
	    {$I-}
	    readln( Temp, Parse_Line ) ;
	    {$I+}
	    Dummy := IOResult ;
	    if( Dummy <> 0 ) then
	    begin
		exit ;
	    end ;
	    Parse_Line := Edit( Parse_Line, 8 or 16 or 128 ) ;
	end ;
	while( copy( Parse_Line, 1, 1 ) = ' ' ) do
	begin
	    Parse_Line := copy( Parse_Line, 2, length( Parse_Line ) ) ;
	end ;
	Dummy := 1 ;
	while(
		( pos( Parse_Line[ Dummy ], 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789_' ) > 0 )
		and
		( Dummy <= length( Parse_Line ) )
	     ) do
	begin
	    inc( Dummy ) ;
	end ;
	if( Dummy = 1 ) then
	begin
	    Dummy := 2 ; { Need at least one character }
	end ;
	Token := copy( Parse_Line, 1, Dummy - 1 ) ;
	Parse_Line := copy( Parse_Line, Dummy, length( Parse_Line ) ) ;
    end ; { TCDD.Generate.Token }


var By_Ref_List, Object_Ref_List : TString_Collection ;
    Object_Support : boolean ;

    procedure Parse_Object_Or_API( API : boolean ; var UEC : tUEC ) ;

    var By_Ref_Syntax : string ;
	By_Reference : boolean ;
	Dummy : integer ;
	Function_Syntax, Procedure_Syntax : TString_Collection ;
	Method_Name, Scope, Param_Name, Param_Type, Params, This_Param : string ;
	Object_Ref_Syntax : string ;
	Parameter_Syntax, Target : TString_Collection ;

	procedure Handle_Object_Refs( var Param_Name, Param_Type : string ) ;

	var Check : text ;
	    Dummy : integer ;

	begin { Parse_Object_Or_API.Handle_Object_Refs }
	    Get_Type_Definition( Param_Type, Temp_File_Name0 ) ;
	    if( CDD_Error.Code = 0 ) then { Found type definition }
	    begin
		assign( Check, Temp_File_Name0 ) ;
		{$I-}
		reset( Check ) ;
		{$I+}
		Dummy := IOResult ;
		if( Dummy <> 0 ) then
		begin
		    exit ;
		end ;
		{$I-}
		readln( Check, S ) ;
		{$I+}
		Dummy := IOResult ;
		if( Dummy <> 0 ) then
		begin
		    close( Check ) ;
		    exit ;
		end ;
		{$I-}
		readln( Check, S ) ;
		{$I+}
		Dummy := IOResult ;
		if( Dummy <> 0 ) then
		begin
		    close( Check ) ;
		    exit ;
		end ;
		if( S = 'OBJECT' ) then
		begin
		    if( Object_Support ) then
		    begin
			Dummy := pos( '~name', Object_Ref_Syntax ) ;
			if( Dummy > 0 ) then
			begin
			    Param_Name := copy( Object_Ref_Syntax, 1, Dummy - 1 ) +
				Param_Name +
				copy( Object_Ref_Syntax, Dummy + 5, length( Object_Ref_Syntax ) ) ;
			end ;
			Dummy := pos( '~type', Object_Ref_Syntax ) ;
			if( Dummy > 0 ) then
			begin
			    Param_Type := copy( Object_Ref_Syntax, 1, Dummy - 1 ) +
				Param_Type +
				copy( Object_Ref_Syntax, Dummy + 5, length( Object_Ref_Syntax ) ) ;
			end ;
		    end else
		    begin
			Param_Type := Translate_Type( 'POINTER' ) ;
		    end ;
		end ;
		close( Check ) ;
	    end ;
	end ; { Parse_Object_Or_API.Handle_Object_Refs }

    begin { Parse_Object_Or_API }
	{ Setup... }
	UEC.Code := 0 ; { Assume no errors }
	By_Ref_Syntax := '' ;
	for Dummy := 0 to By_Ref_List.Count - 1 do
	begin
	    By_Ref_Syntax :=
		By_Ref_Syntax + By_Ref_List.At( Dummy ) + ' ' ;
	end ;
	Object_Ref_Syntax := '' ;
	for Dummy := 0 to Object_Ref_List.Count - 1 do
	begin
	    Object_Ref_Syntax :=
		Object_Ref_Syntax + Object_Ref_List.At( Dummy ) + ' ' ;
	end ;
	Get_Syntax( Language, Temp_File_Name, CDDST_Parameter ) ;
	if( CDD_Error.Code <> 0 ) then
	begin
	    exit ;
	end ;
	Parameter_Syntax.Init( 4, 4 ) ;
	Load_String_Collection( Parameter_Syntax, Temp_File_Name, UEC ) ;
	if( UEC.Code <> 0 ) then
	begin
	    Parameter_Syntax.Done ;
	    exit ;
	end ;

	Get_Syntax( Language, Temp_File_Name, CDDST_Function ) ;
	if( CDD_Error.Code <> 0 ) then
	begin
	    Parameter_Syntax.Done ;
	    exit ;
	end ;
	Function_Syntax.Init( 4, 4 ) ;
	Load_String_Collection( Function_Syntax, Temp_File_Name, UEC ) ;
	if( UEC.Code <> 0 ) then
	begin
	    Function_Syntax.Done ;
	    Parameter_Syntax.Done ;
	    exit ;
	end ;

	Get_Syntax( Language, Temp_File_Name, CDDST_Procedure ) ;
	if( CDD_Error.Code <> 0 ) then
	begin
	    Function_Syntax.Done ;
	    Parameter_Syntax.Done ;
	    exit ;
	end ;
	Procedure_Syntax.Init( 4, 4 ) ;
	Load_String_Collection( Procedure_Syntax, Temp_File_Name, UEC ) ;
	if( UEC.Code <> 0 ) then
	begin
	    Function_Syntax.Done ;
	    Parameter_Syntax.Done ;
	    Procedure_Syntax.Done ;
	    exit ;
	end ;

	{ Parse the definition... }
	Parse_Line := '' ;
	S := Token ;
	while( length( S ) > 0 ) do
	begin
	    { State 1 - scope }
	    Scope := Edit( S, 32 ) ;
	    while(
		   ( length( Parse_Line ) > 0 )
		   and
		   ( copy( Parse_Line, 1, 1 ) <> ' ' ) { Terminated on non space }
		 ) do { API scopes can have non alphanumeric characters }
	    begin
		S := Token ;
		Scope := Scope + S ;
	    end ;

	    { State 2 - method name }
	    Method_Name := Token ;

	    { State 3 - parameters... }
	    Params := '' ;
	    if( length( Parse_Line ) > 0 ) then
	    begin
		S := Token ;
		if( S = '(' ) then
		begin
		    S := Token ;
		    while( S <> ')' ) do
		    begin
			{ Get parameter name }
			Param_Name := S ;

			{ Get parameter type... }
			Param_Type := Token ;
			if( Param_Type = 'STRING' ) then
			begin
			    Param_Type := Param_Type + ' ' + Token ;
			end ;
			if(
			    ( Param_Type = 'STRING STATIC' )
			    or
			    ( Param_Type = 'INTEGER' )
			    or
			    ( Param_Type = 'BOOLEAN' )
			    or
			    ( Param_Type = 'REAL' )
			  ) then
			begin
			    while( S <> 'BYTES' ) do
			    begin
				S := Token ;
				Param_Type := Param_Type + ' ' + S ;
			    end ;
			end ;

			{ Now handle "BY REFERENCE"... }
			By_Reference := False ;
			if(
			    ( copy( Parse_Line, 1, 3 ) = 'BY ' )
			    or
			    ( copy( Parse_Line, 1, 4 ) = ' BY ' )
			  ) then
			begin
			    S := Token ; { Get "BY" }
			    S := Token ; { Get "REFERENCE" }
			    By_Reference := True ;

			    Dummy := pos( '~name', By_Ref_Syntax ) ;
			    if( Dummy > 0 ) then
			    begin
				Param_Name :=
					copy( By_Ref_Syntax, 1, Dummy - 1 ) +
					Param_Name +
					copy( By_Ref_Syntax, Dummy + 5, length( By_Ref_Syntax ) ) ;
			    end ;
			    Dummy := pos( '~type', By_Ref_Syntax ) ;
			    if( Dummy > 0 ) then
			    begin
				Param_Type :=
					copy( By_Ref_Syntax, 1, Dummy - 1 ) +
					Param_Type +
					copy( By_Ref_Syntax, Dummy + 5, length( By_Ref_Syntax ) ) ;
			    end ;
			end ;

			{ Translate parameter syntax... }
			Target.Init( 1, 1 ) ;
			Param_Type := Translate_Type( Param_Type ) ;
			if(
			    ( not Object_Support )
			    or
			    (
			      ( Object_Ref_Syntax <> '~type' )
			      and
			      ( not By_Reference )
			    )
			  ) then
			begin
			    Handle_Object_Refs( Param_Name, Param_Type ) ;
			end ;
			Translate2( Target, Parameter_Syntax,
			    '~name', Param_Name, '~type', Param_Type ) ;
			This_Param := '' ;
			for Dummy := 0 to Target.Count - 1 do
			begin
			    This_Param := This_Param + Target.At( Dummy ) + ' ' ;
			end ;
			Target.Done ;

			if( By_Reference and ( pos( '~parameter', By_Ref_Syntax ) > 0 ) ) then
			begin
			    Dummy := pos( '~parameter', By_Ref_Syntax ) ;
			    This_Param := copy( By_Ref_Syntax, 1, Dummy - 1 ) +
				This_Param +
				copy( By_Ref_Syntax, Dummy + 10, length( By_Ref_Syntax ) ) ;
			end ;
			Params := Params + This_Param ;

			S := Token ;
			if( S = ',' ) then { Ignore commas }
			begin
			    S := Token ;
			end ;
		    end ;

		    { Trim last non-space character from parameter list... }
		    Dummy := length( Params ) ;
		    while( ( Dummy > 0 ) and ( Params[ Dummy ] = ' ' ) ) do
		    begin
			dec( Dummy ) ;
		    end ;
		    Params := copy( Params, 1, Dummy - 1 ) ;
		end else
		begin
		    Parse_Line := edit( S + ' ' + Parse_Line, 16 or 128 ) ;
		    { Function return type }
		end ;
	    end ; { if( length( Parse_Line ) > 0 ) }

	    { State 4 - result type... }
	    Target.Init( 1, 1 ) ;
	    if( length( Parse_Line ) > 0 ) then
	    begin
		{ Translate function... }
		Parse_Line := Translate_Type( edit( Parse_Line, 8 or 128 ) ) ;
		if( Object_Ref_Syntax <> '~type' ) then
		begin
		    Handle_Object_Refs( Param_Name, Parse_Line ) ;
		end ;
		if( API ) then
		begin
		    Translate4( Target, Function_Syntax, '~name', Method_Name,
			'~parameters', Params, '~type', Parse_Line,
			'~ref', Scope ) ;
		end else
		begin
		    Translate3( Target, Function_Syntax, '~name', Method_Name,
			'~parameters', Params, '~type', Parse_Line ) ;
		end ;
		Parse_Line := '' ;
	    end else
	    begin
		{ Translate procedure }
		if( API ) then
		begin
		    Translate3( Target, Procedure_Syntax, '~name', Method_Name,
			'~parameters', Params, '~ref', Scope ) ;
		end else
		begin
		    Translate2( Target, Procedure_Syntax, '~name', Method_Name,
			'~parameters', Params ) ;
		end ;
	    end ;
	    Params := Scope + ' ' ;
	    for Dummy := 0 to Target.Count - 1 do
	    begin
		Params := Params + Target.At( Dummy ) + ' ' ;
	    end ;
	    Definition_List.Insert( Params ) ;
	    Target.Done ;

	    S := Token ;
	end ; { while( length( S ) > 0 ) }
	Parameter_Syntax.Done ;
	Procedure_Syntax.Done ;
	Function_Syntax.Done ;
    end ; { TCDD.Generate.Parse_Object_Or_API }


    procedure Do_Work ; { TCDD.Generate.Do_Work }

    var Ancestor : string ; { Ancestor class(es) }
	Current : integer ; { Current position in current method scope list }
	D : string ;
	Last_Scope : string ; { Current method scope }
	Loop : integer ; { Loop control }
	Prefix : string ;
	Private_List, Protected_List, Public_List : TString_Collection ;
	Remaining : string ; { Remaining string of current position in method scope }
	S : string ;
	Structure : integer ; { Loop control }


	procedure Do_API ; { TCDD.Generate.Do_Work.Do_API }

	var API, Loop : integer ; { Loop controls }
	    Temp : integer ;

	begin
	    for Loop := 0 to Definition_List.Count - 1 do
	    begin
		D := Definition_List.At( Loop ) ;
		Temp := pos( ' ', D ) ;

		for API := 0 to Object_Syntax.Count - 1 do
		begin
		    S := Object_Syntax.At( API ) ;

		    Dummy := pos( '~ref', S ) ;
		    if( Dummy > 0 ) then
		    begin
			S := copy( S, 1, Dummy - 1 ) + copy( D, 1, Temp - 1 ) +
			    copy( S, Dummy + 4, length( S ) ) ;
		    end ;

		    Dummy := pos( '~definition', S ) ;
		    if( Dummy > 0 ) then
		    begin
			S := copy( S, 1, Dummy - 1 ) +
			    copy( D, Temp + 1, length( D ) ) +
			    copy( S, Dummy + 11, length( S ) ) ;
		    end ;
		    {$I-}
		    writeln( Output, S ) ;
		    {$I+}
		    Dummy := IOResult ;
		    if( Dummy <> 0 ) then
		    begin
			CDD_Error.Code := Dummy ;
			CDD_Error.Facility := 2 ;
			exit ;
		    end ;
		end ;
	    end ;
	end ; { TCDD.Generate.Do_Work.Do_API }


	procedure Do_Object ;

	var Loop, Structure : integer ; { Loop controls }
	    Temp : integer ;
	    This_Scope : PString_Collection ;

	begin { TCDD.Generate.Do_Work.Do_Object }
	    for Structure := 0 to Object_Syntax.Count - 1 do
	    begin
		S := Object_Syntax.At( Structure ) ;

		Dummy := pos( '~name', S ) ;
		while( Dummy <> 0 ) do
		begin
		    S := copy( S, 1, Dummy - 1 ) + Type_Name +
			copy( S, Dummy + 5, length( S ) ) ;
		    Dummy := pos( '~name', S ) ;
		end ;

		Dummy := pos( '~ancestor', S ) ;
		while( Dummy <> 0 ) do
		begin
		    S := copy( S, 1, Dummy - 1 ) + Ancestor +
			copy( S, Dummy + 9, length( S ) ) ;
		    Dummy := pos( '~ancestor', S ) ;
		end ;

		Dummy := pos( '~definition', S ) ;
		if( Dummy = 0 ) then { No definition substitution on this line }
		begin
		    {$I-}
		    writeln( Output, S ) ;
		    {$I+}
		    Dummy := IOResult ;
		    if( Dummy <> 0 ) then
		    begin
			CDD_Error.Code := Dummy ;
			CDD_Error.Facility := 2 ;
			exit ;
		    end ;
		end else
		begin
		    { Determine prefix for each line... }
		    Prefix := copy( S, 1, Dummy - 1 ) ;
		    S := copy( S, Dummy + 11, length( S ) ) ;
		    Dummy := length( S ) ;
		    while( ( Dummy > 0 ) and ( pos( Prefix[ Dummy ], #32#9 ) > 0 ) ) do
		    begin
			dec( Dummy ) ;
		    end ;
		    {$I-}
		    write( Output, copy( Prefix, 1, Dummy ) ) ;
		    {$I+}
		    Prefix := copy( Prefix, Dummy + 1, length( Prefix ) ) ;
		    Dummy := IOResult ;
		    if( Dummy <> 0 ) then
		    begin
			CDD_Error.Code := Dummy ;
			CDD_Error.Facility := 2 ;
			exit ;
		    end ;

		    { Output the methods... }
		    Last_Scope := '' ;
                    This_Scope := nil ;
		    Current := 0 ;
		    for Loop := 0 to Definition_List.Count - 1 do
		    begin
			D := Definition_List.At( Loop ) ;
			Temp := pos( ' ', D ) ;
			if( copy( D, 1, Temp - 1 ) <> Last_Scope ) then
			begin { Scope change }
			    if( length( Last_Scope ) > 0 ) then { close up last }
			    begin
				This_Scope := nil ;
                                {$I-}
				writeln( Output, Prefix, Remaining ) ;
				{$I+}
				Dummy := IOResult ;
				if( Dummy <> 0 ) then
				begin
				    CDD_Error.Code := Dummy ;
				    CDD_Error.Facility := 2 ;
				    exit ;
				end ;
				inc( Current ) ;
				while( Current < This_Scope^.Count ) do
				begin
				    Remaining := This_Scope^.At( Current ) ;
				    {$I-}
				    writeln( Output, Prefix, Remaining ) ;
				    {$I+}
				    Dummy := IOResult ;
				    if( Dummy <> 0 ) then
				    begin
					CDD_Error.Code := Dummy ;
					CDD_Error.Facility := 2 ;
					exit ;
				    end ;
				    inc( Current ) ;
				end ;
				Remaining := '' ;
				Current := 0 ;
			    end ;
			    Last_Scope := copy( D, 1, Temp - 1 ) ;
			    if( Last_Scope = 'PUBLIC' ) then
			    begin
				This_Scope := @Public_List ;
			    end ;
			    if( Last_Scope = 'PROTECTED' ) then
			    begin
				This_Scope := @Protected_List ;
			    end ;
			    if( Last_Scope = 'PRIVATE' ) then
			    begin
				This_Scope := @Private_List ;
			    end ;
			    Remaining := This_Scope^.At( Current ) ;
			    while( pos( '~definition', Remaining ) = 0 ) do
			    begin
				{$I-}
				writeln( Output, Remaining ) ;
				{$I+}
				Dummy := IOResult ;
				if( Dummy <> 0 ) then
				begin
				    CDD_Error.Code := Dummy ;
				    CDD_Error.Facility := 2 ;
				    exit ;
				end ;
				inc( Current ) ;
				if( Current >= This_Scope^.Count ) then
				begin
				    Remaining := '' ;
				    break ;
				end ;
				Remaining := This_Scope^.At( Current ) ;
			    end ;
			    Dummy := pos( '~definition', Remaining ) ;
			    if( Dummy > 0 ) then
			    begin
				{$I-}
				write( Output, Prefix, copy( Remaining, 1, Dummy - 1 ) ) ;
				{$I+}
				Remaining := copy( Remaining, Dummy + 11, length( Remaining ) ) ;
				Dummy := IOResult ;
				if( Dummy <> 0 ) then
				begin
				    CDD_Error.Code := Dummy ;
				    CDD_Error.Facility := 2 ;
				    exit ;
				end ;
			    end ;
			end ;
			D := copy( D, Temp + 1, length( D ) ) ;
			{$I-}
			writeln( Output, Prefix, D ) ;
			{$I+}
			Dummy := IOResult ;
			if( Dummy <> 0 ) then
			begin
			    CDD_Error.Code := Dummy ;
			    CDD_Error.Facility := 2 ;
			    exit ;
			end ;
		    end ;
		    if( length( Last_Scope ) > 0 ) then { close up last }
		    begin
			{$I-}
			writeln( Output, Remaining ) ;
			{$I+}
			Dummy := IOResult ;
			if( Dummy <> 0 ) then
			begin
			    CDD_Error.Code := Dummy ;
			    CDD_Error.Facility := 2 ;
			    exit ;
			end ;
			inc( Current ) ;
			while( Current < This_Scope^.Count ) do
			begin
			    Remaining := This_Scope^.At( Current ) ;
			    {$I-}
			    writeln( Output, Prefix, Remaining ) ;
			    {$I+}
			    Dummy := IOResult ;
			    if( Dummy <> 0 ) then
			    begin
				CDD_Error.Code := Dummy ;
				CDD_Error.Facility := 2 ;
				exit ;
			    end ;
			    inc( Current ) ;
			end ;
		    end ;

		    {$I-}
		    writeln( Output, S ) ;
		    {$I+}
		    Dummy := IOResult ;
		    if( Dummy <> 0 ) then
		    begin
			CDD_Error.Code := Dummy ;
			CDD_Error.Facility := 2 ;
			exit ;
		    end ;
		end ;
	    end ;
	end ; { TCDD.Generate.Do_Work.Do_Object }

var UEC : tUEC ;

begin { TCDD.Generate.Do_Work }
    { Setup - load Translation collection... }
    Get_Syntax( Language, Temp_File_Name, CDDST_Type ) ;
    if( CDD_Error.Code <> 0 ) then
    begin
	exit ;
    end ;
    assign( Temp, Temp_File_Name ) ;
    {$I-}
    reset( Temp ) ;
    {$I+}
    Dummy := IOResult ;
    if( Dummy <> 0 ) then
    begin
	CDD_Error.Code := Dummy ;
	CDD_Error.Facility := 2 ;
	exit ;
    end ;
    while( not eof( Temp ) ) do
    begin
	readln( Temp, S ) ;
	Translation.Insert( S ) ;
    end ;

    { Determine type of type... }
    Get_Type_Definition( Type_Name, Temp_File_Name1 ) ;
    if( CDD_Error.Code <> 0 ) then
    begin
	exit ;
    end ;
    assign( Temp, Temp_File_Name1 ) ;
    {$I-}
    reset( Temp ) ;
    {$I+}
    Dummy := IOResult ;
    if( Dummy <> 0 ) then
    begin
	CDD_Error.Code := Dummy ;
	CDD_Error.Facility := 2 ;
	exit ;
    end ;
    {$I-}
    readln( Temp, Type_Name ) ; { Get name in proper case }
    {$I+}
    Dummy := IOResult ;
    if( Dummy <> 0 ) then
    begin
	CDD_Error.Code := Dummy ;
	CDD_Error.Facility := 2 ;
	close( Temp ) ;
	exit ;
    end ;
    {$I-}
    readln( Temp, S ) ;
    {$I+}
    Dummy := IOResult ;
    if( Dummy <> 0 ) then
    begin
	CDD_Error.Code := Dummy ;
	CDD_Error.Facility := 2 ;
	close( Temp ) ;
	exit ;
    end ;
    Get_Syntax( Language, Temp_File_Name, CDDST_Declaration ) ;
    if( CDD_Error.Code <> 0 ) then
    begin
	Declaration_Syntax.Done ;
	exit ;
    end ;
    Declaration_Syntax.Init( 4, 4 ) ;
    Load_String_Collection( Declaration_Syntax, Temp_File_Name, CDD_Error ) ;
    if( CDD_Error.Code <> 0 ) then
    begin
	Declaration_Syntax.Done ;
	exit ;
    end ;
    Definition_List.Init( 4, 4 ) ;
    Type_List.Init( 4, 4 ) ;
    if( S = 'STRUCTURE' ) then
    begin
	Definition_Type := TDT_Structure ;
    end else
    if( S = 'API' ) then
    begin
	Definition_Type := TDT_API ;
    end else
    if( S = 'OBJECT' ) then
    begin
	Definition_Type := TDT_Object ;
	Dummy := pos( ' ', Type_Name + ' ' ) ;
	Ancestor := copy( Type_Name, Dummy + 1, length( Type_Name ) ) ;
	Type_Name := copy( Type_Name, 1, Dummy - 1 ) ;
    end else
    begin
	Definition_Type := TDT_Simple ;
	Dummy := pos( #9, S ) ;
	Definition_List.Insert( copy( S, 1, Dummy - 1 ) ) ;
	Type_List.Insert( Translate_Type( copy( S, Dummy + 1, length( S ) ) ) ) ;
    end ;

    { See if language support objects... }
    Object_Support := True ;
    Get_Syntax( Language, Temp_File_Name, CDDST_Base_Class ) ;
    if( CDD_Error.Code <> 0 ) then
    begin
	Get_Syntax( Language, Temp_File_Name, CDDST_Class ) ;
	if( CDD_Error.Code <> 0 ) then
	begin
	    Object_Support := False ;
	end ;
    end ;

    { Read definition into string collection... }
    if( ( Definition_Type = TDT_Object ) or ( Definition_Type = TDT_API ) ) then
    begin
	By_Ref_List.Init( 4, 4 ) ;
	Get_Syntax( Language, Temp_File_Name, CDDST_By_Reference ) ;
	if( CDD_Error.Code <> 0 ) then
	begin
	    By_Ref_List.Insert( '~parameter' ) ;
	    CDD_Error.Code := 0 ;
	    CDD_Error.Facility := CDDErr_Facility ;
	end else
	begin
	    Load_String_Collection( By_Ref_List, Temp_File_Name, CDD_Error ) ;
	    if( CDD_Error.Code <> 0 ) then
	    begin
		close( Temp ) ;
		By_Ref_List.Done ;
		Definition_List.Done ;
		Type_List.Done ;
		Declaration_Syntax.Done ;
		exit ;
	    end ;
	end ;
	Object_Ref_List.Init( 4, 4 ) ;
	Get_Syntax( Language, Temp_File_Name, CDDST_Object_Ref ) ;
	if( CDD_Error.Code <> 0 ) then
	begin
	    Object_Ref_List.Insert( '~type' ) ;
	    CDD_Error.Code := 0 ;
	    CDD_Error.Facility := CDDErr_Facility ;
	end else
	begin
	    Load_String_Collection( Object_Ref_List, Temp_File_Name, CDD_Error ) ;
	    if( CDD_Error.Code <> 0 ) then
	    begin
		close( Temp ) ;
		By_Ref_List.Done ;
		Object_Ref_List.Done ;
		Definition_List.Done ;
		Type_List.Done ;
		Declaration_Syntax.Done ;
		exit ;
	    end ;
	end ;
	Parse_Object_Or_API( Definition_Type = TDT_API, UEC ) ;
	By_Ref_List.Done ;
	Object_Ref_List.Done ;
    end else
    begin
	while( not eof( Temp ) ) do
	begin
	    {$I-}
	    readln( Temp, S ) ;
	    {$I+}
	    Dummy := IOResult ;
	    if( Dummy <> 0 ) then
	    begin
		CDD_Error.Code := Dummy ;
		CDD_Error.Facility := 2 ;
		close( Temp ) ;
		Definition_List.Done ;
		Type_List.Done ;
		Declaration_Syntax.Done ;
		exit ;
	    end ;
	    S := Edit( S, 8 or 16 or 128 ) ;
	    Dummy := pos( ' ', S ) ;
	    Translate2( Definition_List, Declaration_Syntax, '~name',
		copy( S, 1, Dummy - 1 ), '~type',
		Translate_Type( copy( S, Dummy + 1, length( S ) ) ) ) ;
	end ;
    end ;
    close( Temp ) ;
    Declaration_Syntax.Done ;
    
    { Create output file... }
    assign( Output, O ) ;
    {$I-}
    rewrite( Output ) ;
    {$I+}
    Dummy := IOResult ;
    if( Dummy <> 0 ) then
    begin
	CDD_Error.Code := Dummy ;
	CDD_Error.Facility := 2 ;
	Definition_List.Done ;
	Type_List.Done ;
	exit ;
    end ;

    { Generate the definition... }
    case Definition_Type of
	TDT_API :
	    begin
		Get_Syntax( Language, Temp_File_Name, CDDST_API ) ;
		Object_Syntax.Init( 4, 4 ) ;
		if( CDD_Error.Code <> 0 ) then { Language has no special API syntax }
		begin
		    Object_Syntax.Insert( '~definition' ) ;
		    CDD_Error.Code := 0 ;
		    CDD_Error.Facility := CDDErr_Facility ;
		end else
		begin
		    Load_String_Collection( Object_Syntax, Temp_File_Name, CDD_Error ) ;
		end ;
		if( CDD_Error.Code = 0 ) then
		begin
		    Do_API ;
		end ;
		Object_Syntax.Done ;
	    end ;

	TDT_Object :
	    begin
		if( length( Ancestor ) = 0 ) then
		begin
		    Get_Syntax( Language, Temp_File_Name, CDDST_Base_Class ) ;
		    if( CDD_Error.Code <> 0 ) then
		    begin
			Get_Syntax( Language, Temp_File_Name, CDDST_Class ) ;
			if( CDD_Error.Code <> 0 ) then
			begin
			    close( Output ) ;
			    Definition_List.Done ;
			    Type_List.Done ;
			    exit ;
			end ;
		    end ;
		end else
		begin
		    Get_Syntax( Language, Temp_File_Name, CDDST_Class ) ;
		    if( CDD_Error.Code <> 0 ) then
		    begin
			close( Output ) ;
			Definition_List.Done ;
			Type_List.Done ;
			exit ;
		    end ;
		end ;
		Object_Syntax.Init( 4, 4 ) ;
		Load_String_Collection( Object_Syntax, Temp_File_Name, CDD_Error ) ;
		if( CDD_Error.Code = 0 ) then
		begin
		    Public_List.Init( 4, 4 ) ;
		    Private_List.Init( 4, 4 ) ;
		    Protected_List.Init( 4, 4 ) ;
		    Get_Syntax( Language, Temp_File_Name, CDDST_Public ) ;
		    if( CDD_Error.Code <> 0 ) then
		    begin
			Public_List.Insert( '~definition' ) ;
			CDD_Error.Code := 0 ;
			CDD_Error.Facility := CDDErr_Facility ;
		    end else
		    begin
			Load_String_Collection( Public_List, Temp_File_Name, CDD_Error ) ;
			if( CDD_Error.Code <> 0 ) then
			begin
			    close( Output ) ;
			    Public_List.Done ;
			    Private_List.Done ;
			    Protected_List.Done ;
			    Definition_List.Done ;
			    Type_List.Done ;
			    exit ;
			end ;
		    end ;
		    Get_Syntax( Language, Temp_File_Name, CDDST_Protected ) ;
		    if( CDD_Error.Code <> 0 ) then
		    begin
			Protected_List.Insert( '~definition' ) ;
			CDD_Error.Code := 0 ;
			CDD_Error.Facility := CDDErr_Facility ;
		    end else
		    begin
			Load_String_Collection( Protected_List, Temp_File_Name, CDD_Error ) ;
			if( CDD_Error.Code <> 0 ) then
			begin
			    close( Output ) ;
			    Public_List.Done ;
			    Private_List.Done ;
			    Protected_List.Done ;
			    Definition_List.Done ;
			    Type_List.Done ;
			    exit ;
			end ;
		    end ;
		    Get_Syntax( Language, Temp_File_Name, CDDST_Private ) ;
		    if( CDD_Error.Code <> 0 ) then
		    begin
			Private_List.Insert( '~definition' ) ;
			CDD_Error.Code := 0 ;
			CDD_Error.Facility := CDDErr_Facility ;
		    end else
		    begin
			Load_String_Collection( Private_List, Temp_File_Name, CDD_Error ) ;
			if( CDD_Error.Code <> 0 ) then
			begin
			    close( Output ) ;
			    Public_List.Done ;
			    Private_List.Done ;
			    Protected_List.Done ;
			    Definition_List.Done ;
			    Type_List.Done ;
			    exit ;
			end ;
		    end ;
		    Do_Object ;
		    Public_List.Done ;
		    Private_List.Done ;
		    Protected_List.Done ;
		end ;
		Object_Syntax.Done ;
	    end ;
	TDT_Structure:
	    begin
		Get_Syntax( Language, Temp_File_Name, CDDST_Structure ) ;
		if( CDD_Error.Code <> 0 ) then
		begin
		    close( Output ) ;
		    Definition_List.Done ;
		    Type_List.Done ;
		    exit ;
		end ;
		Structure_Syntax.Init( 4, 4 ) ;
		Load_String_Collection( Structure_Syntax, Temp_File_Name, CDD_Error ) ;
		if( CDD_Error.Code = 0 ) then
		begin
		    for Structure := 0 to Structure_Syntax.Count - 1 do
		    begin
			S := Structure_Syntax.At( Structure ) ;
	
			Dummy := pos( '~name', S ) ;
			if( Dummy <> 0 ) then
			begin
			    S := copy( S, 1, Dummy - 1 ) + Type_Name +
				copy( S, Dummy + 5, length( S ) ) ;
			end ;

			Dummy := pos( '~definition', S ) ;
			if( Dummy = 0 ) then { No definition substitution on this line }
			begin
			    {$I-}
			    writeln( Output, S ) ;
			    {$I+}
			    Dummy := IOResult ;
			    if( Dummy <> 0 ) then
			    begin
				CDD_Error.Code := Dummy ;
				CDD_Error.Facility := 2 ;
				break ;
			    end ;
			end else
			begin
			    { Determine prefix for each line... }
			    Prefix := copy( S, 1, Dummy - 1 ) ;
			    S := copy( S, Dummy + 11, length( S ) ) ;
			    Dummy := length( S ) ;
			    while( ( Dummy > 0 ) and ( pos( Prefix[ Dummy ], #32#9 ) > 0 ) ) do
			    begin
				dec( Dummy ) ;
			    end ;
			    {$I-}
			    write( Output, copy( Prefix, 1, Dummy ) ) ;
			    {$I+}
			    Prefix := copy( Prefix, Dummy + 1, length( Prefix ) ) ;
			    Dummy := IOResult ;
			    if( Dummy <> 0 ) then
			    begin
				CDD_Error.Code := Dummy ;
				CDD_Error.Facility := 2 ;
				break ;
			    end ;

			    Dummy := Output_String_Collection( Definition_List, Prefix ) ;
			    if( Dummy <> 0 ) then
			    begin
				break ;
			    end ;
			    {$I-}
			    writeln( Output, S ) ;
			    {$I+}
			    Dummy := IOResult ;
			    if( Dummy <> 0 ) then
			    begin
				CDD_Error.Code := Dummy ;
				CDD_Error.Facility := 2 ;
				break ;
			    end ;
			end ;
		    end ;
		end ;
		Structure_Syntax.Done ;
	    end ;
	else
	    begin
		for Loop := 0 to Definition_List.Count - 1 do
		begin
		    {$I-}
		    writeln( Output, Definition_List.At( Loop ) ) ;
		    {$I+}
		    Dummy := IOResult ;
		    if( Dummy <> 0 ) then
		    begin
			CDD_Error.Code := Dummy ;
			CDD_Error.Facility := 2 ;
			break ;
		    end ;
		end ;
	    end ;
    end ; { case Definition_Type }

    { Finish up... }
    close( Output ) ;
    Definition_List.Done ;
    Type_List.Done ;
end ; { TCDD.Generate.Do_Work }

begin { TCDD.Generate }
    Use_Alternate := True ;
    Translation.Init( 4, 4 ) ;

    Do_Work ;

    Use_Alternate := False ;
    Translation.Done ;
end ; { TCDD.Generate }


procedure TCDD.Get_Syntax( N, O : string ; T : CDD_Syntax_Type ) ;
{ Extracts syntax definition for language N, type T to file O. }

begin
    if( ( T <= CDDST_Low ) or ( T >= CDDST_High ) ) then
    begin
	CDD_Error.Code := CDDErr_Invalid_Syntax_Type ;
	CDD_Error.Facility := CDDErr_Facility ;
	exit ;
    end ;
    Extract_File( Dictionary_Name + '\LANGUAGES\' + N + '\' +
	CDDST_To_String( T ), O ) ;
    if(
	( CDD_Error.Facility = CDDErr_Facility )
	and
	( CDD_Error.Code = CDDErr_Not_Defined )
      ) then
    begin
	if( Use_Alternate ) then
	begin
	    Extract_File( Alternate + '\LANGUAGES\' + N + '\' +
		CDDST_To_String( T ), O ) ;	    
	    if( CDD_Error.Code <> 0 ) then
	    begin
		CDD_Error.Code := CDDErr_Not_Defined ;
		CDD_Error.Facility := CDDErr_Facility ;
	    end ;
	end ;
    end ;
end ; { TCDD.Get_Syntax }


procedure TCDD.Get_Type_Definition( N, O : string ) ;

begin
    Extract_File( Dictionary_Name + '\TYPES\' + N, O ) ;
    if(
	( CDD_Error.Facility = CDDErr_Facility )
	and
	( CDD_Error.Code = CDDErr_Not_Defined )
      ) then
    begin
	if( Use_Alternate ) then
	begin
	    Extract_File( Alternate + '\TYPES\' + N, O ) ;
	    if( CDD_Error.Code <> 0 ) then
	    begin
		CDD_Error.Facility := CDDErr_Facility ;
		CDD_Error.Code := CDDErr_Not_Defined ;
	    end ;
	end ;
    end ;
end ; { TCDD.Get_Type_Definition }


procedure TCDD.Open( N : string ) ;
{ Defines Data Dictionary to use. }

begin
    Nam := N ;
    CDD_Error.Code := CDDErr_Success ;
    CDD_Error.Facility := CDDErr_Facility ;
end ;


procedure TCDD.Open_Alternate( N : string ) ;
{ Opens an alternate dictionary. }

begin
    Alternate := N ;
    CDD_Error.Code := CDDErr_Success ;
    CDD_Error.Facility := CDDErr_Facility ;
end ;


procedure TCDD.Set_Syntax( N, O : string ; T : CDD_Syntax_Type ) ;
{ Sets syntax type T for language N from file O }

begin
    if( ( T <= CDDST_Low ) or ( T >= CDDST_High ) ) then
    begin
	CDD_Error.Code := CDDErr_Invalid_Syntax_Type ;
	CDD_Error.Facility := CDDErr_Facility ;
	exit ;
    end ;
    Update_File( Dictionary_Name + '\LANGUAGES\' + N + '\' +
	CDDST_To_String( T ), O ) ;
end ;


procedure TCDD.Set_Type_Definition( N, O : string ) ;

begin
    Update_File( Dictionary_Name + '\TYPES\' + N, O ) ;
end ; { TCDD.Set_Type_Definition }


var CDD_EI : pCDD_EI ;

procedure Unit_Initialization ;

begin
{$IFDEF DELPHI}
    CDD_EI := tCDD_EI.Create ;
    Register_Error_Interface( PSError_Interface( CDD_EI ) ) ;
{$ELSE}
    new( CDD_EI, Init ) ;
    Register_Error_Interface( @CDD_EI ) ;
{$ENDIF}
end ;


procedure Unit_Terminate ;

begin
{$IFDEF DELPHI}
    Unregister_Error_Interface( CDD_EI.Get_Facility ) ;
    CDD_EI.Free ;
{$ENDIF}
end ;


{$IFDEF DELPHI}
initialization
{$ELSE}
begin
{$ENDIF}
    Unit_Initialization ;

{$IFDEF DELPHI}
finalization
    Unit_Terminate ;
{$ENDIF}
end.
