{
        Program Name : Compress
        Package Name : LB
        Purpose      : Compression/decompression routiens
        Institution  : Conroy & Conroy Co.
        Date Written : 21-June-2000
        Written By   : Alan Conroy
        Version      : 1.0

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

	  This unit contains various compression/decompression routines.
}

unit Compress ;

interface

{$I EDEFINES.INC}

uses _Streams ; { PSStream }

{ Run-Length-Number compression... }
{ RLN compression stores numeric data using 7 bits per byte, with the highest
  bit reserved as a flag indicating whether another byte is required.  Thus, a
  number of any size can be compressed.  Small numbers are stored most
  efficiently, with larger numbers being less efficient to a minimum efficiency
  of 7/8 that of a non-compressed value.  Using RLN, the following number of
  bytes are required for storing the specified range of integers:

  Bytes     Significant bits    Range
  -----     ----------------    -----
  1         7                   -64 to 63
  2         14                  -8,192 to 8,191
  3         21                  -1,048,576 to 1,048,575
  4         28                  -134,217,728 to 134,217,727
  5         35                  -17,179,869,184 to 17,179,869,183
  6         42                  -2,199,023,255,552 to 2,199,023,255,551
  7         49                  -281,474,976,710,656 to 281,474,976,710,655
  8         56                  -36,028,797,018,963,968 to 36,028,797,018,963,967
  9         64                  -9,223,372,036,854,775,808 to 9,223,372,036,854,775,807

  The data is stored with the highest 7 bits first.
  }
{$IFDEF DELPHI}
procedure RLN_Compress( Value : int64 ; Stream : PSStream ) ;
function RLN_Decompress( Stream : PSStream ) : int64 ;
{$ELSE}
procedure RLN_Compress( Value : longint ; Stream : PSStream ) ;
function RLN_Decompress( Stream : PSStream ) : longint ;
{$ENDIF}

implementation

{$IFDEF DELPHI}
procedure RLN_Compress( Value : int64 ; Stream : PSStream ) ;

var Work : int64 ;
{$ELSE}
procedure RLN_Compress( Value : longint ; Stream : PSStream ) ;

var Work : longint ;
{$ENDIF}

var Siz : integer ;

begin
    { Find highest significant bit to determine number of bits to write... }
    Work := Value ;
    Siz := sizeof( Value ) * 8 ;
    if( Work < 0 ) then
    begin
        while( ( Work < 0 ) and ( Siz > 0 ) ) do
        begin
            Work := Work shl 1 ;
            dec( Siz ) ;
        end ;
    end else
    begin
        while( ( Work >= 0 ) and ( Siz > 0 ) ) do
        begin
            Work := Work shl 1 ;
            dec( Siz ) ;
        end ;
    end ;
    inc( Siz ) ; { Need one sign bit }
    Siz := ( Siz + 6 ) div 7 ; { Number of 7-bit segments to save }
    while( Siz > 0 ) do
    begin
        Work := ( Value shr ( ( Siz - 1 ) * 7 ) ) and 127 ;
        Siz := Siz - 1 ;
        if( Siz > 0 ) then
        begin
            Work := Work or 128 ; { Value is continued in next byte }
        end ;
        Stream^.Write( Work, 1 ) ;
    end ;
end ;


{$IFDEF DELPHI}
function RLN_Decompress( Stream : PSStream ) : int64 ;

var Num : int64 ;
{$ELSE}
function RLN_Decompress( Stream : PSStream ) : longint ;

var Num : longint ;
{$ENDIF}

var L : integer ;
    X : byte ;

begin
    { Setup... }
    Num := 0 ;
    L := 1 ;
    Stream^.Read( X, L ) ;
    if( ( X and 64 ) = 64 ) then { Negative sign }
    begin
        Num := -1 ;
    end ;

    { Decompress into Num }
    while( X > 127 ) do
    begin
        Num := ( Num shl 7 ) or ( X and 127 ) ;
        L := 1 ;
        Stream^.Read( X, L ) ;
    end ;
    Num := ( Num shl 7 ) or X ;

    { Return decompressed value }
    RLN_Decompress := Num ;
end ;



end.

