{
        Program Name : DBList
        Package Name : LB
        Purpose      : Database implemented as a simple list/array
        Institution  : Conroy & Conroy Co.
        Date Written : 11-April-2002
        Written By   : Alan Conroy
        Version      : 1.0

	Copyright (C) 2002 by Alan Conroy.  All rights reserved.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          TList_Database is a Database_Table object with a simple structure.
        It therefore tends to be slower than other structures, and is
        unsorted.
}

unit DBList ;

interface

uses Collect, { TCollection }
     Database ; { TDatabase }

const DB_ListErr_Facility = 50 ;
      DB_ListErr_Success = 0 ;
      DB_ListErr_RecordNotFound = 1 ;

type TList_Database = object( TDatabase_Table )
                          private
                              Last_Record_Index : longint ;
                              List : TCollection ; { Collection of records in database }
                              Last_Context : longint ;

                          public { Constructors and destructors... }
                              constructor Init ;
                              destructor Done ; virtual ;

                          public { API... }
                              function Add( Rec : PDatabase_Record ) : PDatabase_Record ;
                                  virtual ;

                              procedure Delete( Index : longint ;
                                  const Key : string ) ; virtual ;

                              function Get( Index : longint ;
                                  const Key : string ) : PDatabase_Record ;
                                  virtual ;

                              function Next( Index : longint ;
                                  Rec : PDatabase_Record ) : PDatabase_Record ;
                                  virtual ;
                                  
                              function Previous( Index : longint ;
                                  Rec : PDatabase_Record ) : PDatabase_Record ;
                                  virtual ;

                              procedure Zero ; virtual ;
                      end ; { TList_Database }
     PList_Database = ^TList_Database ;
     
implementation

{ TList_Database methods... }


{ Constructors and destructors... }

constructor TList_Database.Init ;

begin
    inherited Init ;

    List.Init( 4, 4 ) ;
end ;


var In_Destructor : boolean ;

destructor TList_Database.Done ;

begin
    Zero ;
    List.Done ;

    In_Destructor := True ;
    inherited Done ;
    In_Destructor := False ;
end ;


{ Internal use only... }

function TList_Database.Next( Index : longint ;
    Rec : PDatabase_Record ) : PDatabase_Record ;

begin
    if( List.Count = 0 ) then
    begin
        Next := nil ;
        exit ;
    end ;
    if( Rec = nil ) then
    begin
        Next := PDatabase_Record( List.At( 0 ) ) ;
        Last_Context := 0 ;
        exit ;
    end ;
    if(
        ( Last_Context >= List.Count - 1 )
        or
        ( Rec <> List.At( Last_Context ) )
      ) then
    begin
        Last_Context := List.IndexOf( Rec ) ;
        if( Last_Context = -1 ) then
        begin
            Next := nil ;
            Last_Context := 0 ;
        end else
        begin
            if( Last_Context + 1 >= List.Count ) then
            begin
                Next := nil ;
            end else
            begin
                Next := PDatabase_Record( List.At( Last_Context + 1 ) )
            end ;
        end ;
    end else
    begin
        inc( Last_Context ) ;
        Next := PDatabase_Record( List.At( last_Context ) ) ;
    end ;
end ;


function TList_Database.Previous( Index : longint ;
    Rec : PDatabase_Record ) : PDatabase_Record ;

begin
    if( List.Count = 0 ) then
    begin
        Previous := nil ;
        exit ;
    end ;
    if( Rec = nil ) then
    begin
        Previous := PDatabase_Record( List.At( 0 ) ) ;
        Last_Context := 0 ;
        exit ;
    end ;
    if(
        ( Last_Context = 0 )
        or
        ( Last_Context >= List.Count )
        or
        ( Rec <> List.At( Last_Context ) )
      ) then
    begin
        Last_Context := List.IndexOf( Rec ) ;
        if( Last_Context = -1 ) then
        begin
            Previous := nil ;
            Last_Context := 0 ;
        end else
        begin
            if( Last_Context = 0 ) then
            begin
                Previous := nil ;
            end else
            begin
                Previous := PDatabase_Record( List.At( Last_Context - 1 ) )
            end ;
        end ;
    end else
    begin
        dec( Last_Context ) ;
        Previous := PDatabase_Record( List.At( last_Context ) ) ;
    end ;
end ;


{ API... }

function TList_Database.Add( Rec : PDatabase_Record ) : PDatabase_Record ;

begin
    Add := nil ;
    Set_Error( DB_ListErr_Facility, 0 ) ;
    List.Insert( Rec ) ;
end ;


procedure TList_Database.Delete(  Index : integer ; const Key : string ) ;

var Rec : PDatabase_Record ;

begin
    Rec := Get( Index, Key ) ;
    if( Rec <> nil ) then
    begin
	dispose( Rec, Done ) ;
        List.AtDelete( Last_Record_Index ) ;
    end ;
end ;


function TList_Database.Get( Index : integer ; const Key : string ) : PDatabase_Record ;

var Dummy : integer ;
    Rec : PDatabase_Record ;

begin
    Get := nil ; { Assume failure }
    Last_Record_Index := -1 ;
    for Dummy := 0 to List.Count - 1 do
    begin
        Rec := PDatabase_Record( List.At( Dummy ) ) ;
        if( Rec^.Key_Match( Index, Key ) ) then
        begin
            Get := Rec ;
            Set_Error( DB_ListErr_Facility, 0 ) ;
            Last_Record_Index := Dummy ;
            exit ;
        end ;
    end ;
    Set_Error( DB_ListErr_Facility, DB_ListErr_RecordNotFound ) ;
end ;


procedure TList_Database.Zero ;

var Dummy : integer ;

begin
    if( In_Destructor ) then
    begin
        exit ; { Do nothing if in destructor }
    end ;

    { Delete all records... }
    for Dummy := 0 to List.Count - 1 do
    begin
        dispose( PDatabase_Record( List.At( Dummy ) ), Done ) ;
    end ;

    { Reset list... }
    List.Done ;
    List.Init( 4, 4 ) ;
end ;



end.
