{$N+}
{$E+}
unit Heaps ;
{
        Program Name : Heaps
        Package Name : Subroutine Library - Layer 1
        Purpose      : Heap objects
        Institution  : Conroy & Conroy Co.
        Date Written : 1999
        Written By   : Alan Conroy
        Version      : 1.1B

        Copyright (C) 1999-2005 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *  1-Nov-2002    EAC         Use TObject from Common.       *
        * 13-Nov-2002    EAC         Rename Tallocator to           *
        *                            TManaged_Store.                *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          THeap is a descendant of TManaged_Store and is a wrapper around the
        Borland Pascal/Delphi memory heap.  TBig_Heap (Delphi only) extends
        THeap to handle alot of allocations by allocating big chunks of memory
        and suballocating from them.
}

interface

{$I EDEFINES.INC}

{$IFDEF DELPHI}
uses Classes, { TList }
{$ELSE}
uses Objects,
{$ENDIF}
     { Subroutine Library - Layer 0... }
     _DebugIn, { PSDebug_Interface }
     Common, { TBase_Object }

     { Subroutine Library - Layer 1... }
     _Stores, { TSStore }
     Cache, { TCache }
     MStores, { TManaged_Store }
     UEHDefs ; { TUEC }

type TSHeap = object( TSManaged_Store )
                    public { Instance data }
			Root : pointer ;
			Resolution : integer ; { Minimum allocation }

                    public { Constructors and destructors... }
			constructor Init ;

                    public { API... }
                        function Min_Storage : TStore_Address ; virtual ;

			function Allocate( Size : TStore_Address ) : TStore_Address ;
                            virtual ;
                        function Allocate_At( Offset, Size : TStore_Address ) : boolean ;
                            virtual ;
                        procedure Copy( Source, Destination : TStore_Address ;
                            Count : integer ) ; virtual ; { Copy data }
			procedure Deallocate( PTR, Size : TStore_Address ) ;
                            virtual ;

			{ Deallocate space }
                        procedure Fill( PTR : TStore_Address ; Count : integer ;
                            Value : byte ) ; virtual ;
			procedure Read( PTR, Size : TStore_Address ;
                            var Buffer ) ; virtual ; { Read data }
			procedure Read_Heap( PTR, Size : TStore_Address ;
                            var Buffer ) ; virtual ; { Read data }
                        function Reallocate( PTR, Old, New : TStore_Address ) : TStore_Address ;
                            virtual ;
			procedure Write( PTR, Size : TStore_Address ;
                            var Buffer ) ; virtual ; { Write data }
			procedure Write_Heap( PTR, Size : TStore_Address ;
                            var Buffer ) ; virtual ; { Write data }
			function SpaceAvail : TStore_Address ; virtual ;
			function XSpaceAvail : TStore_Address ; virtual ;
			function MaxSpace : TStore_Address ; virtual ;
			procedure Get_Root( var Buf ) ; virtual ;
                        function Debugger : PSDebug_Interface ; virtual ;
		  end ; { TSHeap }
     PSHeap = ^TSHeap ;

{$IFDEF DELPHI}
type TSHeap64 = object( TSManaged_Store64 )
                    public { Instance data }
			Root : pointer ;
			Resolution : integer ; { Minimum allocation }

                    public { Constructors and destructors... }
			constructor Init ;

                    public { API... }
                        function Min_Storage : TStore_Address64 ; virtual ;

			function Allocate( Size : TStore_Address64 ) : TStore_Address64 ;
                            virtual ;
                        function Allocate_At( Offset, Size : TStore_Address64 ) : boolean ;
                            virtual ;
                        procedure Copy( Source, Destination : TStore_Address64 ;
                            Count : integer ) ; virtual ; { Copy data }
			procedure Deallocate( PTR, Size : TStore_Address64 ) ;
                            virtual ;
			{ Deallocate space }
                        procedure Fill( PTR : TStore_Address64 ; Count : integer ;
                            Value : byte ) ; virtual ;
			procedure Read( PTR, Size : TStore_Address64 ;
                            var Buffer ) ; virtual ; { Read data }
			procedure Read_Heap( PTR, Size : TStore_Address64 ;
                            var Buffer ) ; virtual ; { Read data }
                        function Reallocate( PTR, Old, New : TStore_Address64 ) : TStore_Address64 ;
                            virtual ;
			procedure Write( PTR, Size : TStore_Address64 ;
                            var Buffer ) ; virtual ; { Write data }
			procedure Write_Heap( PTR, Size : TStore_Address64 ;
                            var Buffer ) ; virtual ; { Write data }
			function SpaceAvail : TStore_Address64 ; virtual ;
			function XSpaceAvail : TStore_Address64 ; virtual ;
			function MaxSpace : TStore_Address64 ; virtual ;
			procedure Get_Root( var Buf ) ; virtual ;
                        function Debugger : PSDebug_Interface ; virtual ;
		  end ; { TSHeap64 }
     PSHeap64 = ^TSHeap64 ;

type THeap = class( TManaged_Store )
                   public { Constructors and destructors... }
                        destructor Destroy ; override ;

                   public { Instance data... }
                       _Heap : PSHeap ; { non-nil if we are a wrapper for this }

                   protected { Internal utility routines... }
                       function Heap : PSHeap ;

                   protected { Property handlers... }
                       function _Get_Root : pointer ;
                       function Get_Resolution : integer ;
                       procedure _Set_Root( Value : pointer ) ;
                       procedure Set_Resolution( Value : integer ) ;

                   public { API... }
                       function Min_Storage : TStore_Address ; override ;

	     	       function Allocate( Size : TStore_Address ) : TStore_Address ;
                           override ;
                       function Allocate_At( Offset, Size : TStore_Address ) : boolean ;
                           override ;
                       procedure Copy( Source, Destination : TStore_Address ;
                           Count : integer ) ; override ; { Copy data }
	     	       procedure Deallocate( PTR, Size : TStore_Address ) ;
                           override ; { Deallocate space }
                       procedure Fill( PTR : TStore_Address ; Count : integer ;
                           Value : byte ) ; override ;
	     	       procedure Read( PTR, Size : TStore_Address ; var Buffer ) ;
	     	           override ; { Read data }
	     	       procedure Read_Heap( PTR, Size : TStore_Address ; var Buffer ) ;
	     	           virtual ; { Read data }
                       function Reallocate( PTR, Old, New : TStore_Address ) : TStore_Address ;
                           override ;
	    	       procedure Write( PTR, Size : TStore_Address ;
                           var Buffer ) ; override ; { Write data }
	    	       procedure Write_Heap( PTR, Size : TStore_Address ;
                           var Buffer ) ; virtual ; { Write data }
	    	       function SpaceAvail : TStore_Address ; override ;
	    	       function XSpaceAvail : TStore_Address ; override ;
	    	       function MaxSpace : TStore_Address ; override ;
	    	       procedure Get_Root( var Buf ) ; override ;
                       function Debugger : TCOM_Debug_Interface ; override ;

	     		property Root : pointer
                            read _Get_Root
                            write _Set_Root ;
	     		property Resolution : integer
                            read Get_Resolution
                            write Set_Resolution ;
	    	  end ;

type THeap64 = class( TManaged_Store64 )
                   public { Constructors and destructors... }
	     		destructor Destroy ; override ;

                   public { Instance data... }
                       _Heap : PSHeap64 ; { non-nil if we are a wrapper for this }

                   protected { Internal utility routines... }
                       function Heap : PSHeap64 ;

                   protected { Property handlers... }
                       function _Get_Root : pointer ;
                       function Get_Resolution : integer ;
                       procedure _Set_Root( Value : pointer ) ;
                       procedure Set_Resolution( Value : integer ) ;

                   public { API... }
                       function Min_Storage : TStore_Address64 ; override ;

	     	       function Allocate( Size : TStore_Address64 ) : TStore_Address64 ;
                           override ;
                       function Allocate_At( Offset, Size : TStore_Address64 ) : boolean ;
                           override ;
                       procedure Copy( Source, Destination : TStore_Address64 ;
                           Count : integer ) ; override ; { Copy data }
	     	       procedure Deallocate( PTR, Size : TStore_Address64 ) ;
                           override ; { Deallocate space }
                       procedure Fill( PTR : TStore_Address64 ; Count : integer ;
                           Value : byte ) ; override ;
	     	       procedure Read( PTR, Size : TStore_Address64 ;
                           var Buffer ) ; override ; { Read data }
	     	       procedure Read_Heap( PTR, Size : TStore_Address64 ;
                           var Buffer ) ; virtual ; { Read data }
                       function Reallocate( PTR, Old, New : TStore_Address64 ) : TStore_Address64 ;
                           override ;
	    	       procedure Write( PTR, Size : TStore_Address64 ;
                           var Buffer ) ; override ; { Write data }
	    	       procedure Write_Heap( PTR, Size : TStore_Address64 ;
                           var Buffer ) ; virtual ; { Write data }
	    	       function SpaceAvail : TStore_Address64 ; override ;
	    	       function XSpaceAvail : TStore_Address64 ; override ;
	    	       function MaxSpace : TStore_Address64 ; override ;
	    	       procedure Get_Root( var Buf ) ; override ;
                       function Debugger : TCOM_Debug_Interface ; override ;

	     		property Root : pointer
                            read _Get_Root
                            write _Set_Root ;
	     		property Resolution : integer
                            read Get_Resolution
                            write Set_Resolution ;
	    	  end ;

type TCOM_Heap = class( TCOM_Managed_Store )
                     public { Constructors and destructors... }
                         destructor Destroy ; override ;

                     public { Instance data... }
                         _Heap : PSHeap ; { non-nil if we are a wrapper for this }

                     protected { Internal utility routines... }
                         function Heap : PSHeap ;

                     protected { Property handlers... }
                         function _Get_Root : pointer ;
                         function Get_Resolution : integer ;
                         procedure _Set_Root( Value : pointer ) ;
                         procedure Set_Resolution( Value : integer ) ;

                     public { API... }
                         function Min_Storage : TStore_Address ; override ;

	     	         function Allocate( Size : TStore_Address ) : TStore_Address ;
                             override ;
                         function Allocate_At( Offset, Size : TStore_Address ) : boolean ;
                             override ;
                         procedure Copy( Source, Destination : TStore_Address ;
                             Count : integer ) ; override ; { Copy data }
	     	         procedure Deallocate( PTR, Size : TStore_Address ) ;
                             override ; { Deallocate space }
                         procedure Fill( PTR : TStore_Address ; Count : integer ;
                             Value : byte ) ; override ;
	     	         procedure Read( PTR, Size : TStore_Address ; var Buffer ) ;
	     	             override ; { Read data }
	     	         procedure Read_Heap( PTR, Size : TStore_Address ; var Buffer ) ;
	     	             virtual ; { Read data }
                         function Reallocate( PTR, Old, New : TStore_Address ) : TStore_Address ;
                             override ;
	    	         procedure Write( PTR, Size : TStore_Address ;
                             var Buffer ) ; override ; { Write data }
	    	         procedure Write_Heap( PTR, Size : TStore_Address ;
                             var Buffer ) ; virtual ; { Write data }
	    	         function SpaceAvail : TStore_Address ; override ;
	    	         function XSpaceAvail : TStore_Address ; override ;
	    	         function MaxSpace : TStore_Address ; override ;
	    	         procedure Get_Root( var Buf ) ; override ;
                         function Debugger : TCOM_Debug_Interface ; override ;

	     		 property Root : pointer
                             read _Get_Root
                             write _Set_Root ;
	     		 property Resolution : integer
                             read Get_Resolution
                             write Set_Resolution ;
	    	 end ;

type TCOM_Heap64 = class( TCOM_Managed_Store64 )
                       public { Constructors and destructors... }
                            destructor Destroy ; override ;

                       public { Instance data... }
                           _Heap : PSHeap64 ; { non-nil if we are a wrapper for this }

                       protected { Internal utility routines... }
                           function Heap : PSHeap64 ;

                       protected { Property handlers... }
                           function _Get_Root : pointer ;
                           function Get_Resolution : integer ;
                           procedure _Set_Root( Value : pointer ) ;
                           procedure Set_Resolution( Value : integer ) ;

                       public { API... }
                           function Min_Storage : TStore_Address64 ; override ;

                           function Allocate( Size : TStore_Address64 ) : TStore_Address64 ;
                               override ;
                           function Allocate_At( Offset, Size : TStore_Address64 ) : boolean ;
                               override ;
                           procedure Copy( Source, Destination : TStore_Address64 ;
                               Count : integer ) ; override ; { Copy data }
                           procedure Deallocate( PTR, Size : TStore_Address64 ) ;
                               override ; { Deallocate space }
                           procedure Fill( PTR : TStore_Address64 ; Count : integer ;
                               Value : byte ) ; override ;
                           procedure Read( PTR, Size : TStore_Address64 ;
                               var Buffer ) ; override ; { Read data }
                           procedure Read_Heap( PTR, Size : TStore_Address64 ;
                               var Buffer ) ; virtual ; { Read data }
                           function Reallocate( PTR, Old, New : TStore_Address64 ) : TStore_Address64 ;
                               override ;
                           procedure Write( PTR, Size : TStore_Address64 ;
                               var Buffer ) ; override ; { Write data }
                           procedure Write_Heap( PTR, Size : TStore_Address64 ;
                               var Buffer ) ; virtual ; { Write data }
                           function SpaceAvail : TStore_Address64 ; override ;
                           function XSpaceAvail : TStore_Address64 ; override ;
                           function MaxSpace : TStore_Address64 ; override ;
                           procedure Get_Root( var Buf ) ; override ;
                           function Debugger : TCOM_Debug_Interface ; override ;

                            property Root : pointer
                                read _Get_Root
                                write _Set_Root ;
                            property Resolution : integer
                                read Get_Resolution
                                write Set_Resolution ;
	    	   end ;

     TSBig_Heap = object( TSHeap )
                      private
                          Blocks : TList ; { Pointers to blocks }
                          Sizes : TList ; { Size of each block, in bytes  }

                      public
                          Allocation_Size : longint ; { How much is requested from Delphi at a time }

      			  constructor Init ;
                          destructor Done ; virtual ;
	  		  function Allocate( Size : TStore_Address ) : TStore_Address ;
                              virtual ; { Allocate space }
	      		  procedure Deallocate( PTR, Size : TStore_Address ) ;
                              virtual ; { Deallocate space }
		  	  procedure Read( PTR, Size : TStore_Address ;
                              var Buffer ) ; virtual ; { Read data }
		  	  procedure Read_Heap( PTR, Size : TStore_Address ;
                              var Buffer ) ; virtual ; { Read data }
		      	  procedure Write( PTR, Size : TStore_Address ;
                              var Buffer ) ; virtual ; { Write data}
		      	  procedure Write_Heap( PTR, Size : TStore_Address ;
                              var Buffer ) ; virtual ; { Write data}
			  function SpaceAvail : TStore_Address ; virtual ;
			  function XSpaceAvail : TStore_Address ; virtual ;
			  function MaxSpace : TStore_Address ; virtual ;
                   end ;
     PSBig_Heap = ^TSBig_Heap ;
{$ENDIF}

{$IFDEF Test}
procedure Test_Unit ;
{$ENDIF}

implementation

uses { Layer 0... }
     _App, { Application_Manager }
     Compatib, { MemAvail }
     Num1s, { Num1 }
{$IFDEF DELPHI}
     SysUtils, { Allocmem }
{$ENDIF}
     DebugInt ; { TSText_Debugger }

{$IFDEF DELPHI}
type TSub_Heap = object
                      private { Instance data... }
                          AT_Offset : longint ;
                          AT_Size : longint ;
                          Buffer : PChar ;
                          Buffer_Size : longint ;

                      public { API... }
                          SpaceAvail, MaxSpace : longint ;
                          Resolution : longint ;

                      public { Constructors and destructors... }
                          constructor Init( Size, Res : longint ) ;

                      public { API... }
                          destructor Done ; virtual ;
                          function Allocate( Size : longint ) : longint ;
                          procedure Deallocate( PTR, Size : longint ) ;
                    end ;
     PSub_Heap = ^TSub_Heap ;


type THeap_Statistics_Debugger = class( TText_Debugger )
                                     private
                                         _Instance : PSHeap ;
                                         _Instance64 : PSHeap64 ;
                                         S : string ;

                                     public
                                         constructor Create( I : PSHeap ) ;
                                         constructor Create64( I : PSHeap64 ) ;

                                         function Count : longint ; override ;

                                         function Child( Ordinal : longint ) : TCOM_Debug_Interface ;
                                             override ;
                                 end ;
     PHeap_Statistics_Debugger = THeap_Statistics_Debugger ;

{ THeap_Statistics_Debugger... }

constructor THeap_Statistics_Debugger.Create( I : PSHeap ) ;

begin
    inherited Create ;

    _Instance := I ;
    _Instance64 := nil ;
end ;


constructor THeap_Statistics_Debugger.Create64( I : PSHeap64 ) ;

begin
    inherited Create ;

    _Instance := nil ;
    _Instance64 := I ;
end ;


function THeap_Statistics_Debugger.Count : longint ;

begin
    Count := 18 ;
end ;


function THeap_Statistics_Debugger.Child( Ordinal : longint ) : TCOM_Debug_Interface ;

var R : PText_Debugger ;

begin
    R := TText_Debugger.Create ;
    if( _Instance64 = nil ) then
    begin
        case Ordinal of
            0 : S := 'Allocations = ' + Num1( _Instance.Statistics.Allocations ) ;
            1 : S := 'Bytes_Allocated = ' + Num1( _Instance.Statistics. Bytes_Allocated) ;
            2 : S := 'Bytes_Copied = ' + Num1( _Instance.Statistics.Bytes_Copied ) ;
            3 : S := 'Bytes_Deallocated = ' + Num1( _Instance.Statistics.Bytes_Deallocated ) ;
            4 : S := 'Bytes_Filled = ' + Num1( _Instance.Statistics.Bytes_Filled ) ;
            5 : S := 'Bytes_Read = ' + Num1( _Instance.Store_Statistics.Bytes_Read ) ;
            6 : S := 'Bytes_Written = ' + Num1( _Instance.Store_Statistics.Bytes_Written ) ;
            7 : S := 'Copies = ' + Num1( _Instance.Statistics.Copies ) ;
            8 : S := 'Deallocations = ' + Num1( _Instance.Statistics.Deallocations ) ;
            9 : S := 'Error_Count = ' + Num1( _Instance.Store_Statistics.Error_Count ) ;
            10 : S := 'Extensions = ' + Num1( _Instance.Statistics.Extensions ) ;
            11 : S := 'Fills = ' + Num1( _Instance.Statistics.Fills ) ;
            12 : S := 'Reads = ' + Num1( _Instance.Store_Statistics.Reads ) ;
            13 : S := 'Reallocations = ' + Num1( _Instance.Statistics.Reallocations ) ;
            14 : S := 'Reallocations_In_Place = ' + Num1( _Instance.Statistics.Reallocations_In_Place ) ;
            15 : S := 'Root_Reads = ' + Num1( _Instance.Statistics.Root_Reads ) ;
            16 : S := 'Root_Writes = ' + Num1( _Instance.Statistics.Root_Writes ) ;
            17 : S := 'Writes = ' + Num1( _Instance.Store_Statistics.Writes ) ;
        end ;
    end else
    begin
        case Ordinal of
            0 : S := 'Allocations = ' + Num1( _Instance64.Statistics.Allocations ) ;
            1 : S := 'Bytes_Allocated = ' + Num1( _Instance64.Statistics. Bytes_Allocated) ;
            2 : S := 'Bytes_Copied = ' + Num1( _Instance64.Statistics.Bytes_Copied ) ;
            3 : S := 'Bytes_Deallocated = ' + Num1( _Instance64.Statistics.Bytes_Deallocated ) ;
            4 : S := 'Bytes_Filled = ' + Num1( _Instance64.Statistics.Bytes_Filled ) ;
            5 : S := 'Bytes_Read = ' + Num1( _Instance64.Store_Statistics.Bytes_Read ) ;
            6 : S := 'Bytes_Written = ' + Num1( _Instance64.Store_Statistics.Bytes_Written ) ;
            7 : S := 'Copies = ' + Num1( _Instance64.Statistics.Copies ) ;
            8 : S := 'Deallocations = ' + Num1( _Instance64.Statistics.Deallocations ) ;
            9 : S := 'Error_Count = ' + Num1( _Instance64.Store_Statistics.Error_Count ) ;
            10 : S := 'Extensions = ' + Num1( _Instance64.Statistics.Extensions ) ;
            11 : S := 'Fills = ' + Num1( _Instance64.Statistics.Fills ) ;
            12 : S := 'Reads = ' + Num1( _Instance64.Store_Statistics.Reads ) ;
            13 : S := 'Reallocations = ' + Num1( _Instance64.Statistics.Reallocations ) ;
            14 : S := 'Reallocations_In_Place = ' + Num1( _Instance64.Statistics.Reallocations_In_Place ) ;
            15 : S := 'Root_Reads = ' + Num1( _Instance64.Statistics.Root_Reads ) ;
            16 : S := 'Root_Writes = ' + Num1( _Instance64.Statistics.Root_Writes ) ;
            17 : S := 'Writes = ' + Num1( _Instance64.Store_Statistics.Writes ) ;
        end ;
    end ;
    R.Title := PChar( S ) ;
    Child := TCOM_Debug_Interface( R ) ;
end ;


{$ELSE}
type PSHeap_Statistics_Debugger = ^TSHeap_Statistics_Debugger ;
     TSHeap_Statistics_Debugger = object( TSText_Debugger )
                          private
                              _Instance : PSHeap ;
                              S : string[ 255 ] ;

                          public
			      constructor Init( I : PSHeap ) ;

                              function Count : longint ; virtual ;

                              function Child( Ordinal : longint ) : PSDebug_Interface ;
                                  virtual ;
                      end ;

constructor TSHeap_Statistics_Debugger.Init( I : PSHeap ) ;

begin
    inherited Init ;

    _Instance := I ;
end ;


function TSHeap_Statistics_Debugger.Count : longint ;

begin
    Count := 18 ;
end ;


function TSHeap_Statistics_Debugger.Child( Ordinal : longint ) : PSDebug_Interface ;

var R : PSText_Debugger ;

begin
    new( R, Init ) ;
    R^.Set_Title( @S[ 1 ] ) ;
    Child := PSDebug_Interface( R ) ;
end ;
{$ENDIF}


{ THeap_Debugger... }

{$IFDEF DELPHI}
type THeap_Debugger = class( TText_Debugger )
                          private
                              _Instance : PSHeap ;
                              _Instance64 : PSHeap64 ;
                              S : string ;

                          public
                              constructor Create( I : PSHeap ) ;
                              constructor Create64( I : PSHeap64 ) ;

                              function Count : longint ; override ;

                              function Child( Ordinal : longint ) : TCOM_Debug_Interface ;
                                  override ;
                      end ;
     PHeap_Debugger = THeap_Debugger ;

constructor THeap_Debugger.Create( I : PSHeap ) ;

begin
    inherited Create ;

    _Instance := I ;
    _Instance64 := nil ;
end ;


constructor THeap_Debugger.Create64( I : PSHeap64 ) ;

begin
    inherited Create ;

    _Instance64 := I ;
    _Instance := nil ;
end ;


function THeap_Debugger.Count : longint ;

begin
    Count := 3 ;
end ;


function THeap_Debugger.Child( Ordinal : longint ) : TCOM_Debug_Interface ;

var R : PText_Debugger ;

begin
    R := TText_Debugger.Create ;
    if( _Instance64 = nil ) then
    begin
        case Ordinal of
            0 : S := 'Root = ' + Num1( longint( _Instance.Root ) ) ;
            1 : S := 'Resolution = ' + Num1( _Instance.Resolution ) ;
            2 : begin
                    R.Free ;
                    R := THeap_Statistics_Debugger.Create( _Instance ) ;
                    S := 'Statistics' ;
                end ;
        end ;
    end else
    begin
        case Ordinal of
            0 : S := 'Root = ' + Num1( longint( _Instance64.Root ) ) ;
            1 : S := 'Resolution = ' + Num1( _Instance64.Resolution ) ;
            2 : begin
                    R.Free ;
                    R := THeap_Statistics_Debugger.Create64( _Instance64 ) ;
                    S := 'Statistics' ;
                end ;
        end ;
    end ;
    R.Set_Title( @S[ 1 ] ) ;
    Child := TCOM_Debug_Interface( R ) ;
end ;
{$ELSE}
type PSHeap_Debugger = ^TSHeap_Debugger ;
     TSHeap_Debugger = object( TSText_Debugger )
                          private
                              _Instance : PSHeap ;
                              S : string[ 255 ] ;

                          public
                              constructor Init( I : PSHeap ) ;
                              function Count : longint ; virtual ;

                              function Child( Ordinal : longint ) : PSDebug_Interface ;
                                  virtual ;
                      end ;

constructor TSHeap_Debugger.Init( I : PSHeap ) ;

begin
    inherited Init ;

    _Instance := I ;
end ;


function TSHeap_Debugger.Count : longint ;

begin
    Count := 3 ;
end ;


function TSHeap_Debugger.Child( Ordinal : longint ) : PSDebug_Interface ;

var R : PSText_Debugger ;
    R1 : PSHeap_Statistics_Debugger ;

begin
    new( R, Init ) ;
    case Ordinal of
        0 : S := 'Root = ' + Num1( longint( _Instance^.Root ) ) ;
        1 : S := 'Resolution = ' + Num1( _Instance^.Resolution ) ;
        2 : begin
                dispose( R, done ) ;
                new( R1, Init( _Instance ) ) ;
		R := R1 ;
                S := 'Statistics' ;
            end ;
    end ;
    R^.Set_Title( @S[ 1 ] ) ;
    Child := PSDebug_Interface( R ) ;
end ;
{$ENDIF}


{ TSHeap methods... }

{ Constructors and destructors... }

constructor TSHeap.Init ;

begin
    inherited Init ;

    Resolution := 1 ;
    Root := nil ;
end ;


{ API... }

function TSHeap.Min_Storage : TStore_Address ;

begin
    Min_Storage := 16 ; { Heap doesn't allocate in smaller chunks than this }
end ;


function TSHeap.Allocate( Size : TStore_Address ) : TStore_Address ;

var Dummy : pointer ;

begin
    Getmem( Dummy, Size ) ;
    Allocate := longint( Dummy ) ;

    { Update statistics... }
    inc( Statistics.Allocations ) ;
    inc( Statistics.Bytes_Allocated, Size ) ;
{$IFDEF Debug}
    Log_Allocation( longint( Dummy ), Size ) ;
{$ENDIF}
end ;


function TSHeap.Allocate_At( Offset, Size : TStore_Address ) : boolean ;

begin
    Allocate_At := False ; { We don't allow allocations at specific offsets }
end ;


procedure TSHeap.Copy( Source, Destination : TStore_Address ; Count : integer ) ;
{ Copy data }

begin
{$IFDEF Debug}
    if( not Check_Access( Source, Count ) ) then
    begin
        halt ;
    end ;
    if( not Check_Access( Destination, Count ) ) then
    begin
        halt ;
    end ;
{$ENDIF}
    move( pointer( Source )^, pointer( Destination )^, Count ) ;

    { Update statistics... }
    inc( Statistics.Copies ) ;
    inc( Statistics.Bytes_Copied, Count ) ;
end ;


procedure TSHeap.Deallocate( PTR, Size : TStore_Address ) ;

begin
    if( ( PTR = 0 ) or ( Size = 0 ) ) then
    begin
        if( Application_Manager = nil ) then
        begin
            halt ;
        end else
        begin
            Application_Manager^.Fatal_Exception( 'Heap deallocation error' ) ;
        end ;
    end ;
{$IFDEF Debug}
    Log_Deallocation( PTR, Size ) ;
{$ENDIF}
    if( PTR = integer( Root ) ) then
    begin
        Root := nil ;
    end ;
    Freemem( pointer( PTR ), Size ) ;

    { Update statistics... }
    inc( Statistics.Bytes_Deallocated, Size ) ;
    inc( Statistics.Deallocations ) ;
end ;


procedure TSHeap.Fill( PTR : TStore_Address ; Count : integer ; Value : byte ) ;

begin
{$IFDEF Debug}
    if( not Check_Access( PTR, Count ) ) then
    begin
        halt ;
    end ;
{$ENDIF}
    fillchar( pointer( PTR )^, Count, Value ) ;

    { Update statistics... }
    inc( Statistics.Fills ) ;
    inc( Statistics.Bytes_Filled, Count ) ;
end ;


procedure TSHeap.Read( PTR, Size : TStore_Address ; var Buffer ) ;

begin
    Read_Heap( PTR, Size, Buffer ) ;
end ;


procedure TSHeap.Read_Heap( PTR, Size : TStore_Address ; var Buffer ) ;

begin
{$IFDEF Debug}
    if( not Check_Access( PTR, Size ) ) then
    begin
        if( Application_Manager = nil ) then
        begin
            halt ;
        end else
        begin
            Application_Manager^.Fatal_Exception( 'Heap access error' ) ;
        end ;
    end ;
{$ENDIF}
    move( pointer( PTR )^, Buffer, Size ) ;

    { Update statistics... }
    inc( Store_Statistics.Bytes_Read, Size ) ;
    inc( Store_Statistics.Reads ) ;
end ;


function TSHeap.Reallocate( PTR, Old, New : TStore_Address ) : TStore_Address ;

var P : pointer ;
    _Res : TStore_Address ;
    
begin
  {$IFDEF Debug}
    if( Allocations.Indexof( pointer( PTR ) ) = -1 ) then
    begin
        if( Application_Manager = nil ) then
        begin
            halt ;
        end else
        begin
            Application_Manager^.Fatal_Exception( 'Heap access error' ) ;
        end ;
    end ;
  {$ENDIF}
{$IFDEF DELPHI}
    P := pointer( PTR ) ;
    reallocmem( P, Old, New ) ;
    Result := longint( P ) ;
    _Res := Result ;
{$ELSE}
    _Res := inherited Reallocate( PTR, Old, New ) ;
    Reallocate := _Res ;
{$ENDIF}

    { Update statistics... }
    if( PTR = _Res ) then
    begin
        inc( Statistics.Reallocations_In_Place ) ;
  {$IFDEF Debug}
        _Sizes[ Allocations.Indexof( pointer( PTR ) ) ] := New ;
  {$ENDIF}
    end else
    begin
        inc( Statistics.Bytes_Allocated, New ) ;
        inc( Statistics.Bytes_Deallocated, Old ) ;
        inc( Statistics.Allocations ) ;
        inc( Statistics.Deallocations ) ;
        inc( Statistics.Reallocations ) ;
  {$IFDEF Debug}
        Log_Deallocation( PTR, Old ) ;
        Log_Allocation( longint( P ), New ) ;
  {$ENDIF}
        if( Root = pointer( PTR ) ) then
        begin
            Root := P ;
        end ;
    end ;
end ;


procedure TSHeap.Write( PTR, Size : TStore_Address ; var Buffer ) ;

begin
    Write_Heap( PTR, Size, Buffer ) ;
end ;


procedure TSHeap.Write_Heap( PTR, Size : TStore_Address ; var Buffer ) ;

begin
{$IFDEF Debug}
    if( not Check_Access( PTR, Size ) ) then
    begin
        if( Application_Manager = nil ) then
        begin
            halt ;
        end else
        begin
            Application_Manager^.Fatal_Exception( 'Heap access error' ) ;
        end ;
    end ;
{$ENDIF}
    move( Buffer, pointer( PTR )^, Size ) ;

    { Update statistics... }
    inc( Store_Statistics.Bytes_Written, Size ) ;
    inc( Store_Statistics.Writes ) ;
end ;


function TSHeap.SpaceAvail : TStore_Address ;

begin
    SpaceAvail := MemAvail ;
end ;


function TSHeap.XSpaceAvail : TStore_Address ;

begin
    XSpaceAvail := SpaceAvail ;
end ;


function TSHeap.MaxSpace : TStore_Address ;

begin
    MaxSpace := MaxAvail ;
end ;


procedure TSHeap.Get_Root( var Buf ) ;

begin
    if( Root = nil ) then
    begin
	fillchar( Buf, Resolution, 0 ) ;
    end else
    begin
	move( Root^, Buf, Resolution ) ;
    end ;

    { Update statistics... }
    inc( Store_Statistics.Bytes_Read, Resolution ) ;
    inc( Statistics.Root_Reads ) ;
end ;


function TSHeap.Debugger : PSDebug_Interface ;

{$IFNDEF DELPHI}
var P : PSHeap_Debugger ;
{$ENDIF}

begin
{$IFDEF DELPHI}
    Result := PSDebug_Interface( THeap_Debugger.Create( @self ) ) ;
{$ELSE}
    new( P, Init( @self ) ) ;
    Debugger := P ;
{$ENDIF}
end ;


{$IFDEF DELPHI}
{ TSHeap64 methods... }

{ Constructors and destructors... }

constructor TSHeap64.Init ;

begin
    inherited Init ;

    Resolution := 1 ;
    Root := nil ;
end ;


{ API... }

function TSHeap64.Min_Storage : TStore_Address64 ;

begin
    Min_Storage := 16 ; { Heap doesn't allocate in smaller chunks than this }
end ;


function TSHeap64.Allocate( Size : TStore_Address64 ) : TStore_Address64 ;

var Dummy : pointer ;

begin
    Getmem( Dummy, Size ) ;
    Allocate := longint( Dummy ) ;

    { Update statistics... }
    inc( Statistics.Allocations ) ;
    inc( Statistics.Bytes_Allocated, Size ) ;
{$IFDEF Debug}
    Log_Allocation( longint( Dummy ), Size ) ;
{$ENDIF}
end ;


function TSHeap64.Allocate_At( Offset, Size : TStore_Address64 ) : boolean ;

begin
    Allocate_At := False ; { We don't allow allocations at specific offsets }
end ;


procedure TSHeap64.Copy( Source, Destination : TStore_Address64 ;
    Count : integer ) ; { Copy data }

begin
{$IFDEF Debug}
    if( not Check_Access( Source, Count ) ) then
    begin
        halt ;
    end ;
    if( not Check_Access( Destination, Count ) ) then
    begin
        halt ;
    end ;
{$ENDIF}
    move( pointer( Source )^, pointer( Destination )^, Count ) ;

    { Update statistics... }
    inc( Statistics.Copies ) ;
    inc( Statistics.Bytes_Copied, Count ) ;
end ;


procedure TSHeap64.Deallocate( PTR, Size : TStore_Address64 ) ;

begin
    if( ( PTR = 0 ) or ( Size = 0 ) ) then
    begin
        exit ;
    end ;
{$IFDEF Debug}
    Log_Deallocation( PTR, Size ) ;
{$ENDIF}
    Freemem( pointer( PTR ), Size ) ;
    if( Root = pointer( PTR ) ) then
    begin
        Root := nil ;
    end ;

    { Update statistics... }
    inc( Statistics.Bytes_Deallocated, Size ) ;
    inc( Statistics.Deallocations ) ;
end ;


procedure TSHeap64.Fill( PTR : TStore_Address64 ; Count : integer ;
    Value : byte ) ;

begin
{$IFDEF Debug}
    if( not Check_Access( PTR, Count ) ) then
    begin
        halt ;
    end ;
{$ENDIF}
    fillchar( pointer( PTR )^, Count, Value ) ;

    { Update statistics... }
    inc( Statistics.Fills ) ;
    inc( Statistics.Bytes_Filled, Count ) ;
end ;


procedure TSHeap64.Read( PTR, Size : TStore_Address64 ; var Buffer ) ;

begin
    Read_Heap( PTR, Size, Buffer ) ;
end ;


procedure TSHeap64.Read_Heap( PTR, Size : TStore_Address64 ; var Buffer ) ;

begin
{$IFDEF Debug}
    if( not Check_Access( PTR, Size ) ) then
    begin
        halt ;
    end ;
{$ENDIF}
    move( pointer( PTR )^, Buffer, Size ) ;

    { Update statistics... }
    inc( Store_Statistics.Bytes_Read, Size ) ;
    inc( Store_Statistics.Reads ) ;
end ;


function TSHeap64.Reallocate( PTR, Old, New : TStore_Address64 ) : TStore_Address64 ;

var P : pointer ;

begin
  {$IFDEF Debug}
    if( not Check_Access( PTR, Old ) ) then
    begin
        halt ;
    end ;
  {$ENDIF}
{$IFDEF DELPHI}
    P := pointer( PTR ) ;
    reallocmem( P, Old, New ) ;
    Result := longint( P ) ;
{$ELSE}
    Result := inherited Reallocate( PTR, Old, New ) ;
{$ENDIF}

    { Update statistics... }
    if( PTR = Result ) then
    begin
        inc( Statistics.Reallocations_In_Place ) ;
    end else
    begin
        inc( Statistics.Bytes_Allocated, New ) ;
        inc( Statistics.Bytes_Deallocated, Old ) ;
        inc( Statistics.Allocations ) ;
        inc( Statistics.Deallocations ) ;
        inc( Statistics.Reallocations ) ;
        if( Root = pointer( PTR ) ) then
        begin
            Root := P ;
        end ;
    end ;
  {$IFDEF Debug}
    Log_Deallocation( PTR, Old ) ;
    Log_Allocation( longint( P ), New ) ;
  {$ENDIF}
end ;


procedure TSHeap64.Write( PTR, Size : TStore_Address64 ; var Buffer ) ;

begin
    Write_Heap( PTR, Size, Buffer ) ;
end ;


procedure TSHeap64.Write_Heap( PTR, Size : TStore_Address64 ; var Buffer ) ;

begin
    move( Buffer, pointer( PTR )^, Size ) ;

    { Update statistics... }
    inc( Store_Statistics.Bytes_Written, Size ) ;
    inc( Store_Statistics.Writes ) ;
end ;


function TSHeap64.SpaceAvail : TStore_Address64 ;

begin
    SpaceAvail := MemAvail ;
end ;


function TSHeap64.XSpaceAvail : TStore_Address64 ;

begin
    XSpaceAvail := SpaceAvail ;
end ;


function TSHeap64.MaxSpace : TStore_Address64 ;

begin
    MaxSpace := MaxAvail ;
end ;


procedure TSHeap64.Get_Root( var Buf ) ;

begin
    if( Root = nil ) then
    begin
	fillchar( Buf, Resolution, 0 ) ;
    end else
    begin
	move( Root^, Buf, Resolution ) ;
    end ;

    { Update statistics... }
    inc( Store_Statistics.Bytes_Read, Resolution ) ;
    inc( Statistics.Root_Reads ) ;
end ;


function TSHeap64.Debugger : PSDebug_Interface ;

begin
    Result := PSDebug_Interface( THeap_Debugger.Create64( @self ) ) ;
end ;



{ THeap methods... }

{ Constructors and destructors... }

destructor THeap.Destroy ;

begin
    if( _Heap <> nil ) then
    begin
        _Heap^.Detach ;
    end ;

    inherited Destroy ;
end ;


{ Internal utility routines... }

function THeap.Heap : PSHeap ;

begin
    if( _Heap = nil ) then
    begin
        _Heap := new( PSHeap, Init ) ;
        _Heap^.Attach ;
        Resolution := 1 ;
    end ;
    Result := _Heap ;
end ;


{ Property handlers... }

function THeap._Get_Root : pointer ;

begin
    Result := Heap^.Root ;
end ;


function THeap.Get_Resolution : integer ;

begin
    Result := Heap^.Resolution ;
end ;


procedure THeap._Set_Root( Value : pointer ) ;

begin
    _Heap^.Root := Value ;
end ;


procedure THeap.Set_Resolution( Value : integer ) ;

begin
    Heap^.Resolution := Value ;
end ;


{ API... }

function THeap.Min_Storage : TStore_Address ;

begin
    Result := Heap^.Min_Storage ;
end ;


function THeap.Allocate( Size : TStore_Address ) : TStore_Address ;

begin
    Result := Heap^.Allocate( Size ) ;
end ;


function THeap.Allocate_At( Offset, Size : TStore_Address ) : boolean ;

begin
    Result := Heap.Allocate_At( Offset, Size ) ;
end ;


procedure THeap.Copy( Source, Destination : TStore_Address ; Count : integer ) ;

begin
    Heap^.Copy( Source, Destination, Count ) ;
end ;


procedure THeap.Deallocate( PTR, Size : TStore_Address ) ;

begin
    Heap^.Deallocate( PTR, Size ) ;
end ;


procedure THeap.Fill( PTR : TStore_Address ; Count : integer ; Value : byte ) ;

begin
    Heap^.Fill( PTR, Count, Value ) ;
end ;


procedure THeap.Read( PTR, Size : TStore_Address ; var Buffer ) ;

begin
    Heap^.Read( PTR, Size, Buffer ) ;
end ;


procedure THeap.Read_Heap( PTR, Size : TStore_Address ; var Buffer ) ;

begin
    Heap^.Read_Heap( PTR, Size, Buffer ) ;
end ;


function THeap.Reallocate( PTR, Old, New : TStore_Address ) : TStore_Address ;

begin
    Result := Heap^.Reallocate( PTR, Old, New ) ;
end ;


procedure THeap.Write( PTR, Size : TStore_Address ; var Buffer ) ;

begin
    Heap^.Write( PTR, Size, Buffer ) ;
end ;


procedure THeap.Write_Heap( PTR, Size : TStore_Address ; var Buffer ) ;

begin
    Heap^.Write_Heap( PTR, Size, Buffer ) ;
end ;


function THeap.SpaceAvail : TStore_Address ;

begin
    Result := Heap^.SpaceAvail ;
end ;


function THeap.XSpaceAvail : TStore_Address ;

begin
    Result := Heap^.XSpaceAvail ;
end ;


function THeap.MaxSpace : TStore_Address ;

begin
    Result := Heap^.MaxSpace ;
end ;


procedure THeap.Get_Root( var Buf ) ;

begin
    Heap^.Get_Root( Buf ) ;
end ;


function THeap.Debugger : TCOM_Debug_Interface ;

begin
    Result := Delphi_Debug_Wrapper( Heap^.Debugger ) ;
end ;



{ THeap64 methods... }

{ Constructors and destructors... }

destructor THeap64.Destroy ;

begin
    if( _Heap <> nil ) then
    begin
        _Heap^.Detach ;
        _Heap := nil ;
    end ;

    inherited Destroy ;
end ;


{ Internal utility routines... }

function THeap64.Heap : PSHeap64 ;

begin
    if( _Heap = nil ) then
    begin
        _Heap := new( PSHeap64, Init ) ;
        _Heap^.Attach ;
        Resolution := 1 ;
    end ;
    Result := _Heap ;
end ;


{ Property handlers... }

function THeap64._Get_Root : pointer ;

begin
    Result := Heap^.Root ;
end ;


function THeap64.Get_Resolution : integer ;

begin
    Result := Heap^.Resolution ;
end ;


procedure THeap64._Set_Root( Value : pointer ) ;

begin
    _Heap^.Root := Value ;
end ;


procedure THeap64.Set_Resolution( Value : integer ) ;

begin
    Heap^.Resolution := Value ;
end ;


{ API... }

function THeap64.Min_Storage : TStore_Address64 ;

begin
    Result := Heap^.Min_Storage ;
end ;


function THeap64.Allocate( Size : TStore_Address64 ) : TStore_Address64 ;

begin
    Result := Heap^.Allocate( Size ) ;
end ;


function THeap64.Allocate_At( Offset, Size : TStore_Address64 ) : boolean ;

begin
    Result := Heap.Allocate_At( Offset, Size ) ;
end ;


procedure THeap64.Copy( Source, Destination : TStore_Address64 ; Count : integer ) ;

begin
    Heap^.Copy( Source, Destination, Count ) ;
end ;


procedure THeap64.Deallocate( PTR, Size : TStore_Address64 ) ;

begin
    Heap^.Deallocate( PTR, Size ) ;
end ;


procedure THeap64.Fill( PTR : TStore_Address64 ; Count : integer ; Value : byte ) ;

begin
    Heap^.Fill( PTR, Count, Value ) ;
end ;


procedure THeap64.Read( PTR, Size : TStore_Address64 ; var Buffer ) ;

begin
    Heap^.Read( PTR, Size, Buffer ) ;
end ;


procedure THeap64.Read_Heap( PTR, Size : TStore_Address64 ; var Buffer ) ;

begin
    Heap^.Read_Heap( PTR, Size, Buffer ) ;
end ;


function THeap64.Reallocate( PTR, Old, New : TStore_Address64 ) : TStore_Address64 ;

begin
    Result := Heap^.Reallocate( PTR, Old, New ) ;
end ;


procedure THeap64.Write( PTR, Size : TStore_Address64 ; var Buffer ) ;

begin
    Heap^.Write( PTR, Size, Buffer ) ;
end ;


procedure THeap64.Write_Heap( PTR, Size : TStore_Address64 ; var Buffer ) ;

begin
    Heap^.Write_Heap( PTR, Size, Buffer ) ;
end ;


function THeap64.SpaceAvail : TStore_Address64 ;

begin
    Result := Heap^.SpaceAvail ;
end ;


function THeap64.XSpaceAvail : TStore_Address64 ;

begin
    Result := Heap^.XSpaceAvail ;
end ;


function THeap64.MaxSpace : TStore_Address64 ;

begin
    Result := Heap^.MaxSpace ;
end ;


procedure THeap64.Get_Root( var Buf ) ;

begin
    Heap^.Get_Root( Buf ) ;
end ;


function THeap64.Debugger : TCOM_Debug_Interface ;

begin
    Result := Delphi_Debug_Wrapper( Heap^.Debugger ) ;
end ;



{ TCOM_Heap methods... }

{ Constructors and destructors... }

destructor TCOM_Heap.Destroy ;

begin
    if( _Heap <> nil ) then
    begin
        _Heap^.Detach ;
        _Heap := nil ;
    end ;

    inherited Destroy ;
end ;


{ Internal utility routines... }

function TCOM_Heap.Heap : PSHeap ;

begin
    if( _Heap = nil ) then
    begin
        _Heap := new( PSHeap, Init ) ;
        _Heap^.Attach ;
        Resolution := 1 ;
    end ;
    Result := _Heap ;
end ;


{ Property handlers... }

function TCOM_Heap._Get_Root : pointer ;

begin
    Result := Heap^.Root ;
end ;


function TCOM_Heap.Get_Resolution : integer ;

begin
    Result := Heap^.Resolution ;
end ;


procedure TCOM_Heap._Set_Root( Value : pointer ) ;

begin
    _Heap^.Root := Value ;
end ;


procedure TCOM_Heap.Set_Resolution( Value : integer ) ;

begin
    Heap^.Resolution := Value ;
end ;


{ API... }

function TCOM_Heap.Min_Storage : TStore_Address ;

begin
    Result := Heap^.Min_Storage ;
end ;


function TCOM_Heap.Allocate( Size : TStore_Address ) : TStore_Address ;

begin
    Result := Heap^.Allocate( Size ) ;
end ;


function TCOM_Heap.Allocate_At( Offset, Size : TStore_Address ) : boolean ;

begin
    Result := Heap.Allocate_At( Offset, Size ) ;
end ;


procedure TCOM_Heap.Copy( Source, Destination : TStore_Address ; Count : integer ) ;

begin
    Heap^.Copy( Source, Destination, Count ) ;
end ;


procedure TCOM_Heap.Deallocate( PTR, Size : TStore_Address ) ;

begin
    Heap^.Deallocate( PTR, Size ) ;
end ;


procedure TCOM_Heap.Fill( PTR : TStore_Address ; Count : integer ; Value : byte ) ;

begin
    Heap^.Fill( PTR, Count, Value ) ;
end ;


procedure TCOM_Heap.Read( PTR, Size : TStore_Address ; var Buffer ) ;

begin
    Heap^.Read( PTR, Size, Buffer ) ;
end ;


procedure TCOM_Heap.Read_Heap( PTR, Size : TStore_Address ; var Buffer ) ;

begin
    Heap^.Read_Heap( PTR, Size, Buffer ) ;
end ;


function TCOM_Heap.Reallocate( PTR, Old, New : TStore_Address ) : TStore_Address ;

begin
    Result := Heap^.Reallocate( PTR, Old, New ) ;
end ;


procedure TCOM_Heap.Write( PTR, Size : TStore_Address ; var Buffer ) ;

begin
    Heap^.Write( PTR, Size, Buffer ) ;
end ;


procedure TCOM_Heap.Write_Heap( PTR, Size : TStore_Address ; var Buffer ) ;

begin
    Heap^.Write_Heap( PTR, Size, Buffer ) ;
end ;


function TCOM_Heap.SpaceAvail : TStore_Address ;

begin
    Result := Heap^.SpaceAvail ;
end ;


function TCOM_Heap.XSpaceAvail : TStore_Address ;

begin
    Result := Heap^.XSpaceAvail ;
end ;


function TCOM_Heap.MaxSpace : TStore_Address ;

begin
    Result := Heap^.MaxSpace ;
end ;


procedure TCOM_Heap.Get_Root( var Buf ) ;

begin
    Heap^.Get_Root( Buf ) ;
end ;


function TCOM_Heap.Debugger : TCOM_Debug_Interface ;

begin
    Result := Delphi_Debug_Wrapper( Heap^.Debugger ) ;
end ;



{ TCOM_Heap64 methods... }

{ Constructors and destructors... }

destructor TCOM_Heap64.Destroy ;

begin
    if( _Heap <> nil ) then
    begin
        _Heap^.Detach ;
        _Heap := nil ;
    end ;

    inherited Destroy ;
end ;


{ Internal utility routines... }

function TCOM_Heap64.Heap : PSHeap64 ;

begin
    if( _Heap = nil ) then
    begin
        _Heap := new( PSHeap64, Init ) ;
        _Heap^.Attach ;
        Resolution := 1 ;
    end ;
    Result := _Heap ;
end ;


{ Property handlers... }

function TCOM_Heap64._Get_Root : pointer ;

begin
    Result := Heap^.Root ;
end ;


function TCOM_Heap64.Get_Resolution : integer ;

begin
    Result := Heap^.Resolution ;
end ;


procedure TCOM_Heap64._Set_Root( Value : pointer ) ;

begin
    _Heap^.Root := Value ;
end ;


procedure TCOM_Heap64.Set_Resolution( Value : integer ) ;

begin
    Heap^.Resolution := Value ;
end ;


{ API... }

function TCOM_Heap64.Min_Storage : TStore_Address64 ;

begin
    Result := Heap^.Min_Storage ;
end ;


function TCOM_Heap64.Allocate( Size : TStore_Address64 ) : TStore_Address64 ;

begin
    Result := Heap^.Allocate( Size ) ;
end ;


function TCOM_Heap64.Allocate_At( Offset, Size : TStore_Address64 ) : boolean ;

begin
    Result := Heap.Allocate_At( Offset, Size ) ;
end ;


procedure TCOM_Heap64.Copy( Source, Destination : TStore_Address64 ; Count : integer ) ;

begin
    Heap^.Copy( Source, Destination, Count ) ;
end ;


procedure TCOM_Heap64.Deallocate( PTR, Size : TStore_Address64 ) ;

begin
    Heap^.Deallocate( PTR, Size ) ;
end ;


procedure TCOM_Heap64.Fill( PTR : TStore_Address64 ; Count : integer ; Value : byte ) ;

begin
    Heap^.Fill( PTR, Count, Value ) ;
end ;


procedure TCOM_Heap64.Read( PTR, Size : TStore_Address64 ; var Buffer ) ;

begin
    Heap^.Read( PTR, Size, Buffer ) ;
end ;


procedure TCOM_Heap64.Read_Heap( PTR, Size : TStore_Address64 ; var Buffer ) ;

begin
    Heap^.Read_Heap( PTR, Size, Buffer ) ;
end ;


function TCOM_Heap64.Reallocate( PTR, Old, New : TStore_Address64 ) : TStore_Address64 ;

begin
    Result := Heap^.Reallocate( PTR, Old, New ) ;
end ;


procedure TCOM_Heap64.Write( PTR, Size : TStore_Address64 ; var Buffer ) ;

begin
    Heap^.Write( PTR, Size, Buffer ) ;
end ;


procedure TCOM_Heap64.Write_Heap( PTR, Size : TStore_Address64 ; var Buffer ) ;

begin
    Heap^.Write_Heap( PTR, Size, Buffer ) ;
end ;


function TCOM_Heap64.SpaceAvail : TStore_Address64 ;

begin
    Result := Heap^.SpaceAvail ;
end ;


function TCOM_Heap64.XSpaceAvail : TStore_Address64 ;

begin
    Result := Heap^.XSpaceAvail ;
end ;


function TCOM_Heap64.MaxSpace : TStore_Address64 ;

begin
    Result := Heap^.MaxSpace ;
end ;


procedure TCOM_Heap64.Get_Root( var Buf ) ;

begin
    Heap^.Get_Root( Buf ) ;
end ;


function TCOM_Heap64.Debugger : TCOM_Debug_Interface ;

begin
    Result := Delphi_Debug_Wrapper( Heap^.Debugger ) ;
end ;



{ TSub_Heap methods... }
{ Note that this uses an entire byte per Resolution bytes of allocated space, in
  the Allocation Table - not the most efficient.  Perhaps someday it can make
  use of bit-mapping to take 1/8 as much space. }

constructor TSub_Heap.Init( Size, Res : longint ) ;

begin
    Buffer := Allocmem( Size ) ;
    if( Res = 0 ) then
    begin
        Res := 16 ;
    end ;
    if( ( Res <> 2 ) and ( Res <> 4 ) and ( Res <> 8 ) and ( Res <> 16 ) and
        ( Res <> 32 ) and ( Res <> 64 ) and ( Res <> 128 ) and ( Res <> 256 )
        and ( Res <> 512 ) and ( Res <> 1024 ) and ( Res <> 2048 ) and
        ( Res <> 4096 ) and ( Res <> 8192 ) and ( Res <> 16384 ) and
        ( Res <> 32768 ) ) then
    begin
        if( Application_Manager = nil ) then
        begin
            halt ; { UNTESTED }
        end else
        begin
            Application_Manager^.Fatal_Exception( 'Illegal resolution size' ) ;
        end ;
    end ;
    Resolution := Res ;
    Buffer_Size := Size ;
    AT_Size := Size div ( Resolution + 1 ) ; { Number of bytes needed for Allocation Table }
    SpaceAvail := Size - AT_Size ; { Actual space available for allocations }
    AT_Offset := SpaceAvail ; { First byte offset of allocation table }
end ;


destructor TSub_Heap.Done ;

begin
    freemem( Buffer, Buffer_Size ) ;
    Buffer := nil ;
end ;


function TSub_Heap.Allocate( Size : longint ) : longint ;

var Count : longint ;
    Ending : longint ; { Ending offset of AT }
    Offset : longint ; { Current offset in AT }
    Start : integer ; { Start of current free segment }

begin
    { Round size and quick space check... }
    Size := ( Size + Resolution - 1 ) and ( not ( Resolution - 1 ) ) ;
    if( Size > SpaceAvail ) then
    begin
        Allocate := 0 ;  { Failed to allocate }
        exit ;
    end ;

    { See if we can find enough contiguous space... }
    Count := 0 ;
    Offset := AT_Offset ;
    Ending := AT_Size + AT_Offset - 1 ;
    Start := 0 ;
    while( Offset <= Ending ) do
    begin
        if( Buffer[ Offset ] = #0 ) then
        begin
            if( Count = 0 ) then
            begin
                Start := Offset ;
            end ;
            Count := Count + Resolution ;
            if( Count >= Size ) then
            begin
                fillchar( Buffer[ Start ], Size div Resolution, 255 ) ;
                SpaceAvail := SpaceAvail - Size ;
                Allocate := longint( Buffer + ( Start - AT_Offset ) * Resolution ) ;
                exit ;
            end ;
        end else
        begin
            Count := 0 ; { Reset - ran out of free space too soon }
        end ;
        inc( Offset ) ;
    end ;
    Allocate := 0 ; { No room }
end ; { TSub_Heap.Allocate }


procedure TSub_Heap.Deallocate( PTR, Size : longint ) ;

begin
    { Round size and quick sanity check... }
    Size := ( Size + Resolution - 1 ) and ( not ( Resolution - 1 ) ) ;
    PTR := PTR - longint( Buffer ) ; { Make relative to buffer }
    if( ( PTR < 0 ) or ( PTR + Size >= AT_Offset ) ) then
    begin
        exit ; { Not within our range of addresses }
    end ;
    if( ( PTR and ( Resolution - 1 ) ) <> 0 ) then { Invalid address }
    begin
        if( Application_Manager = nil ) then
        begin
            halt ; { UNTESTED }
        end else
        begin
            Application_Manager^.Fatal_Exception( 'Invalid address' ) ;
        end ;
    end ;

    { Deallocate the space... }
    SpaceAvail := SpaceAvail + Size ;
    Size := Size div Resolution ; { Number of bytes to clear in AT }
    PTR := ( PTR div Resolution ) + AT_Offset ; { Starting byte of AT to clear }
    fillchar( Buffer[ PTR ], Size, 0 ) ; { Make the space available }
end ;



{ TSBig_Heap methods... }

constructor TSBig_Heap.Init ;

begin
    inherited Init ;

    Blocks := TList.Create ;
    Sizes := TList.Create ;
end ;


destructor TSBig_Heap.Done ;

begin
    Blocks.Free ;
    Sizes.Free ;

    inherited Done ;
end ;


function TSBig_Heap.Allocate( Size : TStore_Address ) : TStore_Address ; { Allocate space }

begin
    Allocate := 0 ;
    { Try to suballocate from existing blocks... }

    { Need to allocate another block... }
    {~~~}
end ;


procedure TSBig_Heap.Deallocate( PTR, Size : TStore_Address ) ; { Deallocate space }

begin
    {~~~}
end ;


procedure TSBig_Heap.Read( PTR, Size : TStore_Address ; var Buffer ) ;

begin
    Read_Heap( PTR, Size, Buffer ) ;
end ;


procedure TSBig_Heap.Read_Heap( PTR, Size : TStore_Address ; var Buffer ) ; { Read data }

begin
    {~~~}
end ;


procedure TSBig_Heap.Write( PTR, Size : TStore_Address ; var Buffer ) ;

begin
    Write_Heap( PTR, Size, Buffer ) ;
end ;


procedure TSBig_Heap.Write_Heap( PTR, Size : TStore_Address ; var Buffer ) ; { Write data }

begin
    {~~~}
end ;


function TSBig_Heap.SpaceAvail : TStore_Address ;

var Count, Dummy : longint ;

begin
    Count := 0 ;
    for Dummy := 0 to Blocks.Count - 1 do
    begin
        Count := Count + PSub_Heap( Blocks[ Dummy ] ).SpaceAvail ;
    end ;
    SpaceAvail := Count ;
end ;


function TSBig_Heap.XSpaceAvail : TStore_Address ;

begin                      
{$WARNINGS OFF}
    XSpaceAvail := SpaceAvail + GetHeapStatus.TotalFree ;
{$WARNINGS ON}
end ;


function TSBig_Heap.MaxSpace : TStore_Address ;

var C, Count : TStore_Address ;
    Dummy : longint ;

begin
    Count := 0 ;
    for Dummy := 0 to Blocks.Count - 1 do
    begin
        C := PSub_Heap( Blocks[ Dummy ] ).MaxSpace ;
        if( C > Count ) then
        begin
            Count := C ;
        end ;
    end ;
    MaxSpace := Count ;
end ;
{$ENDIF}


{$IFDEF Test}
procedure Error ;

begin
    writeln( 'Error' ) ;
    halt ;
end ;


procedure Test_Static_Heap ;

var Address : TStore_Address ;
    I : integer ;
    Test : TSHeap ;

begin
    Test.Init ;
    Test.Debugger ;
    if( Test.SpaceAvail <> Test.XSpaceAvail ) then
    begin
        Error ;
    end ;
    if( Test.MaxSpace < 1024 ) then
    begin
        Error ;
    end ;
    if( Test.Min_Storage <> 16 ) then
    begin
        Error ;
    end ;
    if( Test.Allocate_At( 0, 10 ) ) then
    begin
        Error ;
    end ;
    Address := Test.Allocate( 10 ) ;
    if( Address = 0 ) then
    begin
        Error ;
    end ;
    Test.Fill( Address, 10, 0 ) ;
    if( Address <> Test.Reallocate( Address, 10, 9 ) ) then
    begin
        Error ;
    end ;
    Test.Copy( Address, Address, 9 ) ;
    I := 0 ;
    Test.Get_Root( I ) ;
    if( I <> 0 ) then
    begin
        Error ;
    end ;
    I := -1 ;
    Test.Write( Address, Sizeof( I ), I ) ;
    I := 0 ;
    Test.Read( Address, Sizeof( I ), I ) ;
    if( I <> -1 ) then
    begin
        Error ;
    end ;
    Test.Write_Heap( Address, Sizeof( I ), I ) ;
    I := 0 ;
    Test.Read_Heap( Address, Sizeof( I ), I ) ;
    if( I <> -1 ) then
    begin
        Error ;
    end ;
    Test.Root := pointer( Address ) ;
    Test.Resolution := sizeof( I ) ;
    Test.Get_Root( I ) ;
    Test.Deallocate( Address, 15 ) ;
    Test.Deallocate( Address, 0 ) ;
    Address := Test.Allocate( 10 ) ;
    Test.Root := pointer( Address ) ;
    Address := Test.ReAllocate( Address, 10, 1024000 ) ;
    Test.Deallocate( Address, 1024000 ) ;
    Test.Done ;
end ;


procedure Test_Heap ;

var Address : TStore_Address ;
    I : integer ;
    P : PSHeap ;
    Test : THeap ;

begin
    Test := THeap.Create ;
    Test.Debugger ;
    if( Test.SpaceAvail <> Test.XSpaceAvail ) then
    begin
        Error ;
    end ;
    if( Test.MaxSpace < 1024 ) then
    begin
        Error ;
    end ;
    if( Test.Min_Storage <> 16 ) then
    begin
        Error ;
    end ;
    if( Test.Allocate_At( 0, 10 ) ) then
    begin
        Error ;
    end ;
    Address := Test.Allocate( 10 ) ;
    if( Address = 0 ) then
    begin
        Error ;
    end ;
    Test.Fill( Address, 10, 0 ) ;
    if( Address <> Test.Reallocate( Address, 10, 9 ) ) then
    begin
        Error ;
    end ;
    Test.Copy( Address, Address, 9 ) ;
    I := 0 ;
    Test.Get_Root( I ) ;
    if( I <> 0 ) then
    begin
        Error ;
    end ;
    I := -1 ;
    Test.Write( Address, Sizeof( I ), I ) ;
    I := 0 ;
    Test.Read( Address, Sizeof( I ), I ) ;
    if( I <> -1 ) then
    begin
        Error ;
    end ;
    Test.Write_Heap( Address, Sizeof( I ), I ) ;
    I := 0 ;
    Test.Read_Heap( Address, Sizeof( I ), I ) ;
    if( I <> -1 ) then
    begin
        Error ;
    end ;
    Test.Root := pointer( Address ) ;
    Test.Resolution := sizeof( I ) ;
    Test.Get_Root( I ) ;
    Test.Deallocate( Address, 15 ) ;
    Test.Deallocate( Address, 0 ) ;
    Address := Test.Allocate( 10 ) ;
    Address := Test.ReAllocate( Address, 10, 1024000 ) ;
    Test.Deallocate( Address, 1024000 ) ;
    if( Test.Root <> nil ) then
    begin
        Error ;
    end ;
    if( Test.Resolution <> 4 ) then
    begin
        Error ;
    end ;
    Test.Free ;
    Test := THeap.Create ;
    Test.Free ;
    P := new( PSHeap, Init ) ;
    P^.Attach ;
    Test := THeap.Create ;
    Test._Heap := P ;
    Test.Free ;
end ;


procedure Test_COM_Heap ;

var Address : TStore_Address ;
    I : integer ;
    P : PSHeap ;
    Test : TCOM_Heap ;

begin
    Test := TCOM_Heap.Create ;
    Test.Debugger ;
    if( Test.SpaceAvail <> Test.XSpaceAvail ) then
    begin
        Error ;
    end ;
    if( Test.MaxSpace < 1024 ) then
    begin
        Error ;
    end ;
    if( Test.Min_Storage <> 16 ) then
    begin
        Error ;
    end ;
    if( Test.Allocate_At( 0, 10 ) ) then
    begin
        Error ;
    end ;
    Address := Test.Allocate( 10 ) ;
    if( Address = 0 ) then
    begin
        Error ;
    end ;
    Test.Fill( Address, 10, 0 ) ;
    if( Address <> Test.Reallocate( Address, 10, 9 ) ) then
    begin
        Error ;
    end ;
    Test.Copy( Address, Address, 9 ) ;
    I := 0 ;
    Test.Get_Root( I ) ;
    if( I <> 0 ) then
    begin
        Error ;
    end ;
    I := -1 ;
    Test.Write( Address, Sizeof( I ), I ) ;
    I := 0 ;
    Test.Read( Address, Sizeof( I ), I ) ;
    if( I <> -1 ) then
    begin
        Error ;
    end ;
    Test.Write_Heap( Address, Sizeof( I ), I ) ;
    I := 0 ;
    Test.Read_Heap( Address, Sizeof( I ), I ) ;
    if( I <> -1 ) then
    begin
        Error ;
    end ;
    Test.Root := pointer( Address ) ;
    Test.Resolution := sizeof( I ) ;
    Test.Get_Root( I ) ;
    Test.Deallocate( Address, 15 ) ;
    Test.Deallocate( Address, 0 ) ;
    Address := Test.Allocate( 10 ) ;
    Address := Test.ReAllocate( Address, 10, 1024000 ) ;
    Test.Deallocate( Address, 1024000 ) ;
    if( Test.Root <> nil ) then
    begin
        Error ;
    end ;
    if( Test.Resolution <> 4 ) then
    begin
        Error ;
    end ;
    Test.Free ;
    Test := TCOM_Heap.Create ;
    Test.Free ;
    P := new( PSHeap, Init ) ;
    P^.Attach ;
    Test := TCOM_Heap.Create ;
    Test._Heap := P ;
    Test.Free ;
end ;


procedure Test_Static_Heap64 ;

var Address : TStore_Address ;
    I : integer ;
    Test : TSHeap64 ;

begin
    Test.Init ;
    Test.Debugger ;
    if( Test.SpaceAvail <> Test.XSpaceAvail ) then
    begin
        Error ;
    end ;
    if( Test.MaxSpace < 1024 ) then
    begin
        Error ;
    end ;
    if( Test.Min_Storage <> 16 ) then
    begin
        Error ;
    end ;
    if( Test.Allocate_At( 0, 10 ) ) then
    begin
        Error ;
    end ;
    Address := Test.Allocate( 10 ) ;
    if( Address = 0 ) then
    begin
        Error ;
    end ;
    Test.Fill( Address, 10, 0 ) ;
    if( Address <> Test.Reallocate( Address, 10, 9 ) ) then
    begin
        Error ;
    end ;
    Test.Copy( Address, Address, 9 ) ;
    I := 0 ;
    Test.Get_Root( I ) ;
    if( I <> 0 ) then
    begin
        Error ;
    end ;
    I := -1 ;
    Test.Write( Address, Sizeof( I ), I ) ;
    I := 0 ;
    Test.Read( Address, Sizeof( I ), I ) ;
    if( I <> -1 ) then
    begin
        Error ;
    end ;
    Test.Write_Heap( Address, Sizeof( I ), I ) ;
    I := 0 ;
    Test.Read_Heap( Address, Sizeof( I ), I ) ;
    if( I <> -1 ) then
    begin
        Error ;
    end ;
    Test.Root := pointer( Address ) ;
    Test.Resolution := sizeof( I ) ;
    Test.Get_Root( I ) ;
    Test.Deallocate( Address, 15 ) ;
    Test.Deallocate( Address, 0 ) ;
    Address := Test.Allocate( 10 ) ;
    Address := Test.ReAllocate( Address, 10, 1024000 ) ;
    Test.Deallocate( Address, 1024000 ) ;
    Test.Done ;
end ;


procedure Test_Heap64 ;

var Address : TStore_Address ;
    I : integer ;
    P : PSHeap64 ;
    Test : THeap64 ;

begin
    Test := THeap64.Create ;
    Test.Debugger ;
    if( Test.SpaceAvail <> Test.XSpaceAvail ) then
    begin
        Error ;
    end ;
    if( Test.MaxSpace < 1024 ) then
    begin
        Error ;
    end ;
    if( Test.Min_Storage <> 16 ) then
    begin
        Error ;
    end ;
    if( Test.Allocate_At( 0, 10 ) ) then
    begin
        Error ;
    end ;
    Address := Test.Allocate( 10 ) ;
    if( Address = 0 ) then
    begin
        Error ;
    end ;
    Test.Fill( Address, 10, 0 ) ;
    if( Address <> Test.Reallocate( Address, 10, 9 ) ) then
    begin
        Error ;
    end ;
    Test.Copy( Address, Address, 9 ) ;
    I := 0 ;
    Test.Get_Root( I ) ;
    if( I <> 0 ) then
    begin
        Error ;
    end ;
    I := -1 ;
    Test.Write( Address, Sizeof( I ), I ) ;
    I := 0 ;
    Test.Read( Address, Sizeof( I ), I ) ;
    if( I <> -1 ) then
    begin
        Error ;
    end ;
    Test.Write_Heap( Address, Sizeof( I ), I ) ;
    I := 0 ;
    Test.Read_Heap( Address, Sizeof( I ), I ) ;
    if( I <> -1 ) then
    begin
        Error ;
    end ;
    Test.Root := pointer( Address ) ;
    Test.Resolution := sizeof( I ) ;
    Test.Get_Root( I ) ;
    Test.Deallocate( Address, 15 ) ;
    Test.Deallocate( Address, 0 ) ;
    Address := Test.Allocate( 10 ) ;
    Address := Test.ReAllocate( Address, 10, 1024000 ) ;
    Test.Deallocate( Address, 1024000 ) ;
    if( Test.Root <> nil ) then
    begin
        Error ;
    end ;
    if( Test.Resolution <> 4 ) then
    begin
        Error ;
    end ;
    Test.Free ;
    Test := THeap64.Create ;
    Test.Free ;
    P := new( PSHeap64, Init ) ;
    P^.Attach ;
    Test := THeap64.Create ;
    Test._Heap := P ;
    Test.Free ;
end ;


procedure Test_COM_Heap64 ;

var Address : TStore_Address ;
    I : integer ;
    P : PSHeap64 ;
    Test : TCOM_Heap64 ;

begin
    Test := TCOM_Heap64.Create ;
    Test.Debugger ;
    if( Test.SpaceAvail <> Test.XSpaceAvail ) then
    begin
        Error ;
    end ;
    if( Test.MaxSpace < 1024 ) then
    begin
        Error ;
    end ;
    if( Test.Min_Storage <> 16 ) then
    begin
        Error ;
    end ;
    if( Test.Allocate_At( 0, 10 ) ) then
    begin
        Error ;
    end ;
    Address := Test.Allocate( 10 ) ;
    if( Address = 0 ) then
    begin
        Error ;
    end ;
    Test.Fill( Address, 10, 0 ) ;
    if( Address <> Test.Reallocate( Address, 10, 9 ) ) then
    begin
        Error ;
    end ;
    Test.Copy( Address, Address, 9 ) ;
    I := 0 ;
    Test.Get_Root( I ) ;
    if( I <> 0 ) then
    begin
        Error ;
    end ;
    I := -1 ;
    Test.Write( Address, Sizeof( I ), I ) ;
    I := 0 ;
    Test.Read( Address, Sizeof( I ), I ) ;
    if( I <> -1 ) then
    begin
        Error ;
    end ;
    Test.Write_Heap( Address, Sizeof( I ), I ) ;
    I := 0 ;
    Test.Read_Heap( Address, Sizeof( I ), I ) ;
    if( I <> -1 ) then
    begin
        Error ;
    end ;
    Test.Root := pointer( Address ) ;
    Test.Resolution := sizeof( I ) ;
    Test.Get_Root( I ) ;
    Test.Deallocate( Address, 15 ) ;
    Test.Deallocate( Address, 0 ) ;
    Address := Test.Allocate( 10 ) ;
    Address := Test.ReAllocate( Address, 10, 1024000 ) ;
    Test.Deallocate( Address, 1024000 ) ;
    if( Test.Root <> nil ) then
    begin
        Error ;
    end ;
    if( Test.Resolution <> 4 ) then
    begin
        Error ;
    end ;
    Test.Free ;
    Test := TCOM_Heap64.Create ;
    Test.Free ;
    P := new( PSHeap64, Init ) ;
    P^.Attach ;
    Test := TCOM_Heap64.Create ;
    Test._Heap := P ;
    Test.Free ;
end ;


procedure Test_Sub_Heap ;

var Address : longint ;
    OK : boolean ;
    Test : TSub_Heap ;

begin
    Test.Init( 1024, 16 ) ;
    Address := Test.Allocate( 10 ) ;
    if( Address = 0 ) then
    begin
        Error ;
    end ;
    if( Test.Allocate( 2048 ) <> 0 ) then
    begin
        Error ;
    end ;
    Test.Deallocate( 0, 10 ) ;
    Test.Deallocate( Address, 10 ) ;
    {~~~}
    Test.Done ;
    Test.Init( 1024, 0 ) ;
    Test.Allocate( 10 ) ;
    Address := Test.Allocate( 32 ) ;
{$IFDEF Delphi}
    try
        Test.Deallocate( Address + 1, 1 ) ;
        OK := False ;
    except
        OK := True ;
    end ;
    if( not OK ) then
    begin
        Error ;
    end ;
{$ENDIF}
    Test.Done ;
{$IFDEF Delphi}
    try
        Test.Init( 1024, 5 ) ;
        OK := False ;
    except
        OK := True ;
    end ;
    if( not OK ) then
    begin
        Error ;
    end ;
{$ENDIF}
    Test.Done
end ;


procedure Test_Unit ;

begin
    Test_Static_Heap ;
    Test_Heap ;
    Test_COM_Heap ;
    Test_Static_Heap64 ;
    Test_Heap64 ;
    Test_COM_Heap64 ;
    Test_Sub_Heap ;
end ;
{$ENDIF}



end.
