{
              Program Name : INSTR
              Package Name : LB
              Purpose      : Search within a string functions
              Institution  : Conroy & Conroy
              Date Written : 30-Nov-1988
              Last Changed : 25-Aug-2004
              Written By   : Alan Conroy
              Version      : 2.0B

              Copyright (C) 1988 by Alan Conroy.  Released to the Public Domain.

              TO THE GLORY OF GOD THROUGH JESUS CHRIST.

        *********************************************************
        *                                                       *
        *        M O D I F I C A T I O N   H I S T O R Y        *
        *                                                       *
        *     DATE      BY             REASON                   *
        *                                                       *
        *  1-Nov-2002   EAC            Use const parameters in  *
	*                              Delphi.                  *
        *                                                       *
        * 25-Aug-2004   EAC            Added RInstr.            *
        *                                                       *
        *********************************************************

        *********************************************************
        *                                                       *
        *           P R O G R A M   P U R P O S E               *
        *                                                       *
        *********************************************************

          These functions return the position within a string of a substring,
        starting at a specified character.  Format: X=Instr(start,string,sub),
        where X is the returned value, string is the string to search, and sub
        is the string to search for.
}

unit Instrs ;

interface

{$I EDEFINES.INC}

function Instr( Position : Integer ; const C, Search_String : string ) : Integer ;
function RInstr( Position : Integer ; const C, Search_String : string ) : Integer ;

implementation

function RInstr( Position : Integer ; const C, Search_String : string ) : Integer ;

var Dummy : integer ;

begin
    Dummy := Position - length( C ) ;
    while( ( Dummy > 0 ) and ( copy( Search_String, Dummy, length( C ) ) <> C ) ) do
    begin
        dec( Dummy ) ;
    end ;
    RInstr := Dummy ;
end ;


{$IFDEF DELPHI}
procedure Finstr( var D : integer ; Position : Integer ; const C, Search_String : string ) ;

label Start ;

var C_Offset : integer ; { String offsets }
{$ELSE}
procedure FInstr( var D, Position : Integer ; const C, Search_String ) ;
{$ENDIF}

begin
    if( Position < 1 ) then
    begin
        Position := 1 ;
    end ;
{$IFDEF DELPHI}
Start:
    if( length( Search_String ) + Position - 1 > length( C ) ) then
    begin
        D := 0 ;
        exit ;
    end ;
    C_Offset := 1 ;
    D := Position ;
    while(
           ( Position <= length( C ) )
           and
           ( C_Offset <= length( Search_String ) )
         ) do
    begin
        if( C[ Position ] <> Search_String[ C_Offset ] ) then
        begin
            Position := Position - C_Offset + 2 ;
            goto Start ;
        end ;
        inc( Position ) ;
        inc( C_Offset ) ;
    end ;
{$ELSE}
    D := 0 ;
    move( C, D, 1 ) ;
    if( Position > D ) then
    begin
	D := 0 ;
	exit ;
    end ;
    inline(
        {0000} $1E /                  { PUSH DS ; Save DS and BP }
        {0001} $55 /                  { PUSH BP }
        {0002} $C5 / $9E / Position / { LDS BX, [BP].Position }
        {0006} $8B / $37 /            { MOV SI, [BX] ; Get starting position }
        {0008} $83 / $Fe / $01 /      { CMP SI, 1 ; Starting position < 1? }
        {000B} $7D / $03 /            { JGE $10 ; No }
        {000D} $BE / > $0001 /        { MOV SI, 1 ; Yes - start at position 1 }
        {0010}                   { $10: }
        {0010} $C5 / $9E / C /        { LDS BX, [BP].C ; Starting position in string }
        {0014} $03 / $F3 /            { ADD SI, BX }
        {0016} $C5 / $9E / Search_String / { LDS BX, [BP].Search_String ; Get address of search string }
        {001A} $8B / $FB /            { MOV DI, BX ; Current position in search string }
        {001C} $47 /                  { INC DI }
        {001D} $BA / > $0000 /        { MOV DX, 0 ; Offset result }
        {0020} $8B / $Cb /            { MOV CX, BX ; Address of search string }
        {0022} $8A / $27 /            { MOV AH, [BX] ; Length of search string }
        {0024} $C5 / $9E / C /        { LDS BX, [BP].C ; Address of string }
        {0028} $8A / $07 /            { MOV AL, [BX] ; Length of string }
        {002A} $3A / $C4 /            { CMP AL, AH }
        {002C} $72 / $3D /            { JB $0 ; If search string is larger than string, exit }
        {002E} $8A / $C4 /            { MOV AL, AH ; Length of search string }
        {0030} $B4 / $00 /            { MOV AH, 0 }
        {0032} $03 / $C8 /            { ADD CX, AX ; Get maximum address of search string }
        {0034} $50 /                  { PUSH AX ; Save search string length }
        {0035} $C5 / $9E / C /        { LDS BX, [BP].C ; Address of string }
        {0039} $8A / $07 /            { MOV AL, [BX] ; Length of string }
        {003B} $03 / $D8 /            { ADD BX, AX ; Get maximum address of string }
        {003D} $58 /                  { POP AX ; Retrieve length of search string }
        {003E} $2B / $D8 /            { SUB BX, AX }
        {0040} $43 /                  { INC BX ; BX now contains the maximum address in string }
                                      { ; to search up to }
        {0041} $8B / $D6 /            { MOV DX, SI }
        {0043}                   { $1: }
        {0043} $8A / $04 /            { MOV AL, [SI] ; Get characters for comparison }
        {0045} $8A / $25 /            { MOV AH, [DI] }
        {0047} $3A / $C4 /            { CMP AL, AH ; Equal? }
        {0049} $75 / $08 /            { JNE $20 ; nope }
        {004B} $47 /                  { INC DI ; Check next character }
        {004C} $3B / $Cf /            { CMP CX, DI }
        {004E} $7E / $15 /            { JLE $00 ; Done }
        {0050} $46 /                  { INC SI }
        {0051} $Eb / $F0 /            { JMP $1 }
        {0053}                   { $20: }
        {0053} $C5 / $Be / Search_String / { LDS DI, [BP].Search_String }
        {0057} $47 /                  { INC DI }
        {0058} $42 /                  { INC DX }
        {0059} $8B / $F2 /            { MOV SI, DX }
        {005B} $3B / $F3 /            { CMP SI, BX }
        {005D} $7E / $E4 /            { JLE $1 ; Not done yet }
        {005F} $Ba / > $0000 /        { MOV DX, 0 ; Search failure - return 0 }
        {0062} $Eb / $07 / $90 /      { JMP $0 }
        {0065}                   { $00: ; We're done - but first convert from address to position }
        {0065} $8B / $9E / C /        { MOV BX, [BP].C }
        {0069} $2B / $D3 /            { SUB DX, BX }
        {006B}                   { $0: ; End of routine }
        {006B} $C5 / $9E / D /        { LDS BX, [BP].D }
        {006F} $89 / $17 /            { MOV [BX], DX ; Return result }
        {0071} $5D /                  { POP BP ; Pop BP and DS }
        {0072} $1F                    { POP DS }
    )
{$ENDIF}
end ;


function Instr( Position : Integer ; const C, Search_String : string ) : Integer ;

var D : Integer ;

begin
    Finstr( D, Position, C, Search_String ) ;
    Instr := D
end ;


end.
