{
        Program Name : Integers
        Package Name : LB
        Purpose      : Multiple-precision integer object
        Institution  : Conroy & Conroy Co.
        Date Written : 11-Nov-1998
        Written By   : Alan Conroy
        Version      : 1.0

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

	  TInteger is a multiple-precision integer object.  It is a
	2's-complement signed binary integer number.
}


unit Integers ;

interface

type TInteger = object
		    private { Instance data }
			_Data : PChar ;
			_Division_Error : boolean ;
                        _Overflow : boolean ; { Last math operation overflowed? }
			_Size : integer ;

		    public { Constructors and destructors... }
			constructor Init ; 
			destructor Done ; virtual ;

		    public { API... }
			{ Logical (bit) operations... }
			procedure Op_Not ;
			procedure Op_SHL( X : integer ) ;
			procedure Op_SHR( X : integer ) ;
			procedure Op_Or( X : PChar ) ;
			procedure Op_And( X : PChar ) ;
			procedure Op_Xor( X : PChar ) ;
			procedure Op_Nor( X : PChar ) ;
			procedure Op_Nand( X : Pchar ) ;
			procedure Op_XNor( X : PChar ) ;

			{ Math functions... }
			procedure Op_Neg ;
			procedure Op_Add( X : PChar ) ;
			procedure Op_Sub( X : PChar ) ;
			procedure Op_Multiply( X : PChar ) ;
                        procedure Op_Divide( X : PChar ) ;

			{ General functions... }
                        function Division_Error : boolean ;
			function Overflow : boolean ;
			function Sign : integer ;
			function Size : integer ;
			procedure Set_Size( X : integer ) ;
			procedure Set_Value( X : PChar ) ;
			function Value : PChar ;
		end ;

implementation

{ Internal data types... }

type CVT = record
	       case integer of
		    0 : ( B0, B1, B2, B3, B4, B5, B6, B7 : byte ) ;
		    1 : ( W0, W1, W2, W3 : word ) ;
		    2 : ( L0, L1 : longint ) ;
		    3 : ( C : Comp ) ;
	   end ;

{ TInteger methods... }

{ Constructors and destructors... }

constructor TInteger.Init ;

begin
    _Size := 0 ;
    _Overflow := False ;
end ;


destructor TInteger.Done ;

begin
    if( _Size <> 0 ) then
    begin
	freemem( _Data, _Size ) ;
    end ;
end ;



{ API... }

procedure TInteger.Op_Not ;

var Loop : integer ;

begin
    for Loop := 0 to _Size - 1 do
    begin
	_Data[ Loop ] := chr( not ord( _Data[ Loop ] ) ) ;
    end ;
end ;


procedure TInteger.Op_SHL( X : integer ) ;

    procedure Do_SHL( X : integer ) ;

    var Carry : word ;
	Loop : integer ;
	Temp : CVT ;

    begin
	Temp.L0 := 0 ;
	for Loop := 0 to ( ( _Size + 1 ) div 2 ) - 1 do
	begin
	    Carry := Temp.W1 ;
            Temp.W1 := 0 ;
	    move( _Data[ Loop * 2 ], Temp.W0, 2 ) ;
	    Temp.L0 := Temp.L0 shl X ;
	    Temp.W0 := Temp.W0 or Carry ;
	    move( Temp.W0, _Data[ Loop * 2 ], 2 ) ;
	end ;
    end ;

begin
    while( X > 16 ) do
    begin
	Do_SHL( 16 ) ;
	X := X - 16 ;
    end ;
    Do_SHL( X ) ;
end ;


procedure TInteger.Op_SHR( X : integer ) ;

    procedure Do_SHR( X : integer ) ;

    var Carry : word ;
	Loop : integer ;
	Temp : CVT ;

    begin
	Temp.L0 := 0 ;
	for Loop := ( ( _Size + 1 ) div 2 ) - 1 downto 0 do
	begin
	    Carry := Temp.W0 ;
	    Temp.W0 := 0 ;
	    move( _Data[ Loop * 2 ], Temp.W1, 2 ) ;
	    Temp.L0 := Temp.L0 shr X ;
	    Temp.W1 := Temp.W1 or Carry ;
	    move( Temp.W1, _Data[ Loop * 2 ], 2 ) ;
	end ;
    end ;

begin
    while( X > 16 ) do
    begin
	Do_SHR( 16 ) ;
	X := X - 16 ;
    end ;
    Do_SHR( X ) ;
end ;


procedure TInteger.Op_Or( X : PChar ) ;

var Loop : integer ;

begin
    for Loop := 0 to _Size - 1 do
    begin
	_Data[ Loop ] := chr( ord( _Data[ Loop ] ) or ord( X[ Loop ] ) ) ;
    end ;
end ;


procedure TInteger.Op_And( X : PChar ) ;

var Loop : integer ;

begin
    for Loop := 0 to _Size - 1 do
    begin
	_Data[ Loop ] := chr( ord( _Data[ Loop ] ) and ord( X[ Loop ] ) ) ;
    end ;
end ;


procedure TInteger.Op_Xor( X : PChar ) ;

var Loop : integer ;

begin
    for Loop := 0 to _Size - 1 do
    begin
	_Data[ Loop ] := chr( ord( _Data[ Loop ] ) xor ord( X[ Loop ] ) ) ;
    end ;
end ;


procedure TInteger.Op_Nor( X : PChar ) ;

var Loop : integer ;

begin
    for Loop := 0 to _Size - 1 do
    begin
	_Data[ Loop ] := chr( not( ord( _Data[ Loop ] ) or ord( X[ Loop ] ) ) ) ;
    end ;
end ;


procedure TInteger.Op_Nand( X : PChar ) ;

var Loop : integer ;

begin
    for Loop := 0 to _Size - 1 do
    begin
	_Data[ Loop ] := chr( not( ord( _Data[ Loop ] ) and ord( X[ Loop ] ) ) ) ;
    end ;
end ;


procedure TInteger.Op_Xnor( X : PChar ) ;

var Loop : integer ;

begin
    for Loop := 0 to _Size - 1 do
    begin
	_Data[ Loop ] := chr( not( ord( _Data[ Loop ] ) xor ord( X[ Loop ] ) ) ) ;
    end ;
end ;


procedure TInteger.Op_Neg ;

var X : PChar ;

begin
    { Setup... }
    if( _Size = 0 ) then
    begin
	exit ;
    end ;
    _Overflow := False ;
    _Division_Error := False ;

    { Create an appropriately-sized value of 1... }
    getmem( X, _Size ) ;
    fillchar( X^, _Size, 0 ) ;
    X[ 0 ] := #1 ;

    { Negate ourself... }
    Op_Sub( X ) ;
    Op_Not ;

    { Cleanup }
    freemem( X, _Size ) ;
end ;


procedure TInteger.Op_Add( X : PChar ) ;

var Carry : integer ;
    Loop : integer ;
    V : integer ;

begin
    Carry := 0 ;
    for Loop := 0 to _Size - 1 do
    begin
	V := ord( _Data[ Loop ] ) + ord( X[ Loop ] ) + Carry ;
	Carry := V shr 8 ;
	_Data[ Loop ] := chr( V ) ;
    end ;
    _Overflow := ( Carry <> 0 ) ;
    _Division_Error := False ;
end ;


procedure TInteger.Op_Sub( X : PChar ) ;

var Borrow : integer ;
    Loop : integer ;
    Work : TInteger ;
    U, V : integer ;
    Y : PChar ;

begin
    { Create a TInteger for the passed value... }
    Work.Init ;
    Work.Set_Size( _Size ) ;
    Work.Set_Value( X ) ;
    Y := Work.Value ;

    { Setup }
    Borrow := 0 ;

    { Do the subtraction... }
    for Loop := 1 to _Size do
    begin
        U := ord( _Data[ Loop - 1 ] ) ;
        V := ord( Y[ Loop - 1 ] ) ;
        U := U - V - Borrow ;
        if( U < 0 ) then
        begin
            Borrow := 1 ;
        end else
        begin
            Borrow := 0 ;
        end ;
        _Data[ Loop - 1 ] := chr( U and 255 ) ;
    end ;
    _Overflow := ( Borrow <> 0 ) ;
    _Division_Error := False ;

    { Clean up... }
    Work.Done ;
end ;


procedure TInteger.Op_Multiply( X : PChar ) ;

var Carry : integer ;
    Loop, Loop1 : integer ;
    New_Sign : integer ;
    Res : PChar ;
    Temp : longint ;
    Work : TInteger ;

begin
    { Setup... }
    _Overflow := False ;
    _Division_Error := False ;
    Work.Init ;
    Work.Set_Size( _Size ) ;
    Work.Set_Value( X ) ;
    getmem( Res, _Size * 2 ) ;
    fillchar( Res^, _Size * 2, 0 ) ;

    { Determine sign of result... }
    New_Sign := Sign * Work.Sign ;

    { Make sure both values are non-negative... }
    if( Sign < 0 ) then
    begin
	Op_Neg ;
    end ;
    if( Work.Sign < 0 ) then
    begin
	Work.Op_Neg ;
    end ;

    { Do multiplication... }
    for Loop := 0 to _Size - 1 do
    begin
	Carry := 0 ;
	for Loop1 := 0 to _Size - 1 do
	begin
	    Temp := ord( _Data[ Loop1 ] ) * ord( X[ Loop ] ) +
                ord( Res[ Loop + Loop1 ] ) + Carry ;
	    Res[ Loop + Loop1 ] := chr( Temp and 255 ) ;
	    Carry := Temp div 256 ;
	end ;
	if( Loop < _Size - 1 ) then
        begin
            Res[ Loop + _Size ] := chr( Carry ) ;
        end ;
    end ;

    { Set result }
    Set_Value( Res ) ;

    { Reapply sign, if necessary }
    if( New_Sign < 0 ) then
    begin
	Op_Neg ; { Make us negative }
    end ;

    { Cleanup... }
    freemem( Res, _Size * 2 ) ;
    Work.Done ;
end ; { TInteger.Op_Multiply }


procedure TInteger.Op_Divide( X : PChar ) ;

var Div_By_1 : boolean ;
    Loop : integer ; { Main loop control }
    New_Sign : integer ;
    Res : PChar ;
    Significant_Size, X_Significant_Size : integer ;
    Temp : longint ;
    Work : TInteger ;
    Y : TInteger ;

begin
    { Setup... }
    _Overflow := False ;
    _Division_Error := False ;
    Work.Init ;
    Work.Set_Size( _Size ) ;
    Work.Set_Value( X ) ;

    { Sanity check }
    if( Work.Sign = 0 ) then
    begin
        Work.Done ;
        _Division_Error := True ;
        exit ;
    end ;

    { Determine sign of result... }
    New_Sign := Sign ;
    if( New_Sign = 0 ) then { Treat 0 as any positive }
    begin
        New_Sign := 1 ;
    end ;
    if( New_Sign <> Work.Sign ) then
    begin
        New_Sign := -1 ;
    end else
    begin
        New_Sign := 1 ;
    end ;

    { Make sure both values are non-negative... }
    if( Sign < 0 ) then
    begin
	Op_Neg ;
    end ;
    if( Work.Sign < 0 ) then
    begin
	Work.Op_Neg ;
    end ;

    { Special case... }
    if( Work.Value[ 0 ] = #1 ) then
    begin
        Div_By_1 := True ;
        for Loop := 2 to _Size - 1 do
        begin
            if( Work.Value[ Loop ] <> #0 ) then
            begin
                Div_By_1 := False ;
                break ;
            end ;
        end ;
        if( Div_By_1 ) then { Do nothing on division by 1 }
        begin
            Work.Done ;
            exit ;
        end ;
    end ;

    Significant_Size := 0 ;
    for Loop := _Size - 1 downto 0 do { Find most significant non-zero byte }
    begin
        if( _Data[ Loop ] <> #0 ) then
        begin
            Significant_Size := Loop ;
            break ;
        end ;
    end ;

    X_Significant_Size := 0 ;
    for Loop := _Size - 1 downto 0 do { Find most significant non-zero byte }
    begin
        if( Work.Value[ Loop ] <> #0 ) then
        begin
            X_Significant_Size := Loop ;
            break ;
        end ;
    end ;

    if(
        ( X_Significant_Size > Significant_Size )
        or
        (
          ( X_Significant_Size = Significant_Size )
          and
          ( X[ Significant_Size ] > _Data[ Significant_Size ] )
        )
      ) then { Dividing by larger value }
    begin
        fillchar( _Data^, _Size, 0 ) ; { Return zero (would be fractional) }
        Work.Done ;
        exit ;
    end ;

    { Prepare result... }
    getmem( Res, Significant_Size - X_Significant_Size + 1 ) ;
    fillchar( Res^, Significant_Size - X_Significant_Size + 1, 0 ) ;

    { Normalize... }
    Temp := 256 div ( ord( Work.Value[ X_Significant_Size ] ) + 1 ) ;
    Y.Init ;
    Y.Set_Size( 4 ) ;
    Y.Set_Value( @Temp ) ;
    Y.Set_Size( _Size ) ;
    Op_Multiply( Y.Value ) ;
    Work.Op_Multiply( Y.Value ) ;
    X := Work.Value ;

    { Do the division... }
    for Loop := Significant_Size - X_Significant_Size downto 0 do
    begin
        if( _Data[ Loop ] = X[ Significant_Size - 1 ] ) then
        begin
            Temp := 255 ;
        end else
        begin
            Temp :=
                ( ord( _Data[ Loop ] ) * 256
                +
                ord( _Data[ Loop + 1 ] ) ) div ord( X[ X_Significant_Size ] ) ;
        end ;
        if( X_Significant_Size > 0 ) then
        begin
             while(
                    ord( X[ Significant_Size - 1 ] ) * Temp
                    >
                    (
                      (
                        ord( _Data[ Loop ] ) * 256 + ord( _Data[ Loop - 1 ] )
                        -
                        Temp * ord( X[ X_Significant_Size ] )
                      )
                      * 256 + ord( _Data[ Loop - 2 ] )
                    )
                 ) do
            begin
                Temp := Temp - 1 ;
            end ;
        end ;
        
        { Subtract Temp * X... }
        Y.Set_Size( 4 ) ;
        Y.Set_Value( @Temp ) ;
        Y.Set_Size( X_Significant_Size + 1 ) ;
        Y.Op_Multiply( X ) ;
        Op_Sub( Y.Value ) ;

        Res[ Loop ] := chr( Temp ) ;

        if( _Data[ _Size ] > #127 ) then { Negative }
        begin
            { Convert to 256's-complement... }
            Set_Size( _Size + 1 ) ;
            Y.Set_Size( 4 ) ;
            Temp := 0 ;
            Y.Set_Value( @Temp ) ;
            Y.Set_Size( _Size ) ;
            Y.Value[ _Size + 1 ] := #1 ;
            Op_Add( Y.Value ) ;
            Set_Size( _Size - 1 ) ;

            Res[ Loop ] := chr( ord( Res[ Loop ] ) - 1 ) ;

            { Add X * 256... }
            Y.Set_Value( X ) ;
            Y.Op_SHL( 8 ) ;
            Op_Add( Y.Value ) ;
        end ;
    end ;
    (*
    d := 256 div ( v[ 1 ] + 1 ) ;
    u := u * d ;
    v := v * d ;

    for j := 0 to m do
    begin
        if( u[ j ] = v[ 1 ] ) then
        begin
            q := b - 1 ;
        end else
        begin
            q := ( u[ j ] * b + u[ j + 1 ] ) div v[ 1 ] ;
        end ;
        while( v[ 2 ] * q > ( u[ j ] * b + u[ j + 1 ] - q * v[ 1 ] ) * b + u[ j + 2 ] ) do
        begin
            q := q - 1 ;
        end ;

        u := u - q * v ;

        q[ j ] := q ;

        if u[ _Size ] > 127 then { Negative }
        begin
            u := u + b^( n + 1 )
            q[ j ] := q[ j ] - 1 ;
            u := u + v * b ;
        end ;
    end ;
    *)

    { Set result... }
    fillchar( _Data^, _Size, 0 ) ;
    move( Res^, _Data^, Significant_Size - X_Significant_Size + 1 ) ;
 
    { Reapply sign, if necessary }
    if( New_Sign < 0 ) then
    begin
	Op_Neg ; { Make us negative }
    end ;

    { Cleanup... }
    freemem( Res, _Size ) ;
    Work.Done ;
    Y.Done ;
end ; { TInteger.Op_Divide }


function TInteger.Overflow : boolean ;

begin
    Overflow := _Overflow ;
end ;


function TInteger.Division_Error : boolean ;

begin
    Division_Error := _Division_Error ;
end ;


function TInteger.Sign : integer ;

var Loop : integer ;
    Zero : boolean ;

begin
    { Setup }
    Sign := 0 ;

    { Special check }
    if( _Size = 0 ) then
    begin
	exit ;
    end ;

    { Determine sign... }
    Zero := True ;
    for Loop := 0 to _Size - 1 do
    begin
	if( _Data[ Loop ] <> #0 ) then
	begin
	    Zero := False ;
	    break ;
	end ;
    end ;
    if( Zero ) then
    begin
	exit ;
    end ;

    if( _Data[ _Size - 1 ] > #127 ) then
    begin
	Sign := -1 ;
    end else
    begin
	Sign := 1 ;
    end ;
end ;


function TInteger.Size : integer ;

begin
    Size := _Size ;
end ;


procedure TInteger.Set_Size( X : integer ) ;

var V : PChar ;

begin
    { Allocate new space... }
    V := nil ;
    if( X <> 0 ) then
    begin
	getmem( V, X ) ;
    end ;
    fillchar( V^, X, 0 ) ;

    { Save as much of old value as possible }
    if( ( X <> 0 ) and ( _Size <> 0 ) ) then
    begin
	if( X < _Size ) then
	begin
	    move( _Data^, V^, X ) ;
	end else
	begin
	    move( _Data^, V^, _Size ) ;
	end ;
    end ;

    { Get rid of old space }
    if( _Size <> 0 ) then
    begin
	freemem( _Data, _Size ) ;
    end ;

    { Finish up... }
    _Size := X ;
    _Data := V ;
end ; { TInteger.Set_Size }


procedure TInteger.Set_Value( X : PChar ) ;

begin
    move( X^, _Data^, _Size ) ;
end ;


function TInteger.Value : PChar ;

var V : PChar ;

begin
    Value := nil ;
    if( _Size <> 0 ) then
    begin
	getmem( V, _Size ) ;
	move( _Data^, V^, _Size ) ;
	Value := V ;
    end ;
end ;



end.
