{$N+}
{$E+}
{
              Program Name : Maths
              Package Name : LB
              Purpose      : Various math routines
              Institution  : Conroy & Conroy Company
              Original Vers: 01-Apr-1987
              Date Written : 25-Jan-2002
              Written By   : Alan Conroy
              Version      : 2.0C

              Copyright (C) 1987-2003 by Alan Conroy.  Released to the public domain.

              TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *********************************************************
        *                                                       *
        *        M O D I F I C A T I O N   H I S T O R Y        *
        *                                                       *
        *     DATE      BY             REASON                   *
        *                                                       *
        *  4-Sep-2002   EAC            Added GCD.               *
        *                                                       *
        *  1-Nov-2002   EAC            Allow operation on both	*
	*                              systems with and without *
	*			       FPUs.                    *
	*			       Added numerous other math*
	*			       functions.		*
        *  20-Jan-2003  EAC            Added Min and Max.       *
        *                                                       *
        *********************************************************

        *********************************************************
        *                                                       *
        *           P R O G R A M   P U R P O S E               *
        *                                                       *
        *********************************************************

          These functions perform various math functions.
}

unit Maths ;

interface

function Max( X, Y : integer ) : integer ;
function Min( X, Y : integer ) : integer ;

function Root( X : Extended ; R : longint ) : Extended ;
function Sgn( X : real ) : real ;
function ISgn( X : longint ) : integer ;
function Power( X, N : Real ; var _Result : Integer ) : Real ;
function GCD( N, D : longint ) : longint ;

function Factorial( X : extended ) : extended ;
function Pi : Extended ;
function Tan( X : Extended ) : Extended ; { Return tangent of a value }
function Log10( X : Extended ) : Extended ; { Return logarithm, base 10 }
function Rad( X : extended ) : extended ; { Degree to Radian conversion }
function Deg( X : extended ) : extended ; { Radian to Degree conversion }
function CoT( X : extended ) : extended ; { Cotangent }
function CSc( X : extended ) : extended ; { Cosecant }
function Sec( X : extended ) : extended ; { Secant }
function Fib( X : extended ) : extended ; { Fibonacci sequence }
function ArcCos( X : extended ) : extended ; { Arc Cosine }
function ArcCoT( X : extended ) : extended ; { Arc Cotangent }
function ArcCSc( X : extended ) : extended ; { Arc Cosecant }
function ArcSec( X : extended ) : extended ; { Arc Secant }
function ArcSin( X : extended ) : extended ; { Arc Sine }
function HypCos( X : extended ) : extended ; { Hyperbolic Cosine }
function HypCoT( X : extended ) : extended ; { Hyperbolic Cotangent }
function HypCSc( X : extended ) : extended ; { Hyperbolic Cosecant }
function HypSec( X : extended ) : extended ; { Hyperbolic secant }
function HypSin( X : extended ) : extended ; { Hyperbolic Sine }
function HypTan( X : extended ) : extended ; { Hyperbolic Tangent }
function HypArcCos( X : extended ) : extended ; { Hyperbolic Arc Cosine }
function HypArcCoT( X : extended ) : extended ; { Hyperbolic Arc Cotangent }
function HypArcCsc( X : extended ) : extended ; { Hyperbolic Arc Cosecant }
function HypArcSec( X : extended ) : extended ; { Hyperbolic Arc Secant }
function HypArcSin( X : extended ) : extended ; { Hyperbolic Arc Sine }
function HypArcTan( X : extended ) : extended ; { Hyperbolic Arc Tangent }

implementation

uses { Subroutine Library - Layer 0... }
     _Types, { Small_String }

     { Subroutine Library - Layer 1 }
     _App, { Application_Manager }
     CommonUt ; { Edit }


function Max( X, Y : integer ) : integer ;

begin
    if( X > Y ) then
    begin
        Max := X ;
    end else
    begin
        Max := Y ;
    end ;
end ;


function Min( X, Y : integer ) : integer ;

begin
    if( X < Y ) then
    begin
        Min := X ;
    end else
    begin
        Min := Y ;
    end ;
end ;


function Factorial( X : extended ) : extended ;

var Y : extended ;

begin
    Y := X - 1 ;
    while( Y > 1 ) do
    begin
	X := X * Y ;
	Y := Y - 1 ;
    end ;
    Factorial := X ;
end ;


function Pi : Extended ;

begin
    Pi := 3.1415926535897932385 ;
end ;


function Tan( X : Extended ) : Extended ; { Return tangent of a value }

begin
    Tan := Sin( X ) / Cos( X )
end ;


function Log10( X : Extended ) : Extended ; { Return logarithm, base 10 }

begin
    Log10 := Ln( X ) / Ln( 10 )
end ;


function Rad( X : extended ) : extended ; { RAD - Degree to Radian conversion }

begin
    Rad := X / ( 180. / Pi ) ;
end ;


function Deg( X : extended ) : extended ; { DEG - Radian to Degree conversion }

begin
    Deg := X * ( 180. / Pi ) ;
end ;


function CoT( X : extended ) : extended ; { COT - Cotangent }

begin
    CoT := 1 / Tan( X ) ;
end ;


function CSc( X : extended ) : extended ; { CSC - Cosecant }

begin
    CSc := 1. / Sin( X ) ;
end ;


function Sec( X : extended ) : extended ; { SEC - Secant }

begin
    Sec := 1. / Cos( X ) ;
end ;


function Fib( X : extended ) : extended ; { Fibonacci sequence }

var A1, A2, A3 : extended ;
    Dummy : integer ;

begin
    X := Int( Abs( X ) ) ;
    if( X < 2 ) then
    begin
        Fib := 1 ;
    end else
    begin
        A1 := 1 ;
        A2 := 1 ;
        for Dummy := 2 to Trunc( X ) do
        begin
            A3 := A2 ;
            A2 := A1 ;
            A1 := A2 + A3 ;
        end ;
        Fib := A1 + A2 ;
    end
end ;


function GCD( N, D : longint ) : longint ;
{ Return greatest common divisor of passed numbers }

var X : longint ;

begin
    if( N = D ) then
    begin
	GCD := D ;
	exit ;
    end ;
    if( N < D ) then
    begin
	GCD := 0 ; { Greatest common divisor is non-integer }
	exit ;
    end ;
    X := N mod D ;
    while( X <> 0 ) do
    begin
	N := D ;
	D := X ;
	X := N mod D ;
    end ;
    GCD := D ;
end ;


function ArcCos( X : extended ) : extended ; { Arc Cosine }

var E : integer ;
    Temp : extended ;

begin
    Temp := Power( X, 2, E ) ;
    Temp := 1 - Temp ;
    ArcCos := Arctan( Sqrt( Temp ) / X )
end ;


function ArcCoT( X : extended ) : extended ; { Arc Cotangent }

begin
    ArcCoT := Arctan( 1. / X )
end ;


function ArcCSc( X : extended ) : extended ; { Arc Cosecant }

begin
    ArcCSc := Arctan( 1. / Sqrt( X ) ) ;
end ;


function ArcSec( X : extended ) : extended ; { Arc Secant }

var E : integer ;

begin
    X := Power( X, 2, E ) - 1 ;
    ArcSec := Arctan( Sqrt( X ) ) ;
end ;


function ArcSin( X : extended ) : extended ; { Arc Sine }

var E : integer ;
    Temp : extended ;

begin
    Temp := Power( X, 2, E ) ;
    Temp := 1 - Temp ;
    ArcSin := Arctan( X / Sqrt( Temp ) )
end ;


function HypCos( X : extended ) : extended ; { Hyperbolic Cosine }

begin
    HypCos := ( Exp( X ) + Exp( - X ) ) / 2.
end ;


function HypCoT( X : extended ) : extended ; { Hyperbolic Cotangent }

var Temp : extended ;

begin
    Temp := Exp( X ) - Exp( - X ) ;
    HypCoT := ( Exp( X ) + Exp( -X ) ) / Temp ;
end ;


function HypCSc( X : extended ) : extended ; { Hyperbolic Cosecant }

begin
    X := Exp( X ) - Exp( -X ) ;
    HypCSc := 2. / X ;
end ;


function HypSec( X : extended ) : extended ; { Hyperbolic secant }

begin
    X := Exp( X ) + Exp( -X ) ;
    HypSec := 2. / X ;
end ;


function HypSin( X : extended ) : extended ; { Hyperbolic Sine }

begin
    HypSin := ( Exp( X ) - Exp( -X ) ) / 2.
end ;


function HypTan( X : extended ) : extended ; { Hyperbolic Tangent }

var Temp : extended ;

begin
    Temp := Exp( X ) + Exp( -X ) ;
    HypTan := ( Exp( X ) - Exp( -X ) ) / Temp ;
end ;


function HypArcCos( X : extended ) : extended ; { Hyperbolic Arc Cosine }

var E : integer ;
    Temp : extended ;

begin
    Temp := Power( X, 2, E ) ;
    Temp := Temp - 1 ;
    X := X + Sqrt( Temp ) ;
    HypArcCos := Ln( X )
end ;


function HypArcCoT( X : extended ) : extended ; { Hyperbolic Arc Cotangent }

begin
    HypArcCoT := ( Ln( X + 1 ) - Ln( X - 1 ) ) / X ;
end ;


function HypArcCsc( X : extended ) : extended ; { Hyperbolic Arc Cosecant }

var E : integer ;
    Temp : extended ;

begin
    Temp := Power( X, 2, E ) ;
    HypArcCsc := Ln( ( 1. / X ) + Sqrt( 1 / Temp + 1 ) )
end ;


function HypArcSec( X : extended ) : extended ; { Hyperbolic Arc Secant }

var E : integer ;
    Temp : extended ;

begin
    Temp := Power( X, 2, E ) ;
    Temp := 1 / Temp - 1 ;
    HypArcSec := Ln( ( 1. / X ) + Sqrt( Temp ) ) ;
end ;


function HypArcSin( X : extended ) : extended ; { Hyperbolic Arc Sine }

var E : integer ;
    Temp : extended ;

begin
    Temp := Power( X, 2, E ) + 1 ;
    X := X + Sqrt( Temp ) ;
    HypArcSin := Ln( X )
end ;


function HypArcTan( X : extended ) : extended ; { Hyperbolic Arc Tangent }

begin
    HypArcTan := ( Ln( ( 1. + X ) / ( 1. - X ) ) ) / 2. ;
end ;


function Root( X : Extended ; R : longint ) : Extended ;
{ This function returns the Rth root of X.  For example, Root( 27, 3 ) returns 3. }

var Dummy : integer ;
    Low, High, Mid, Intermediate : Extended ;

begin
    { Check invalid results }
    if( X < 0 ) then { Negative value }
    begin
	if( ( R and 1 ) = 0 ) then { Even number }
	begin
	    Application_Manager^.Fatal_Exception( '?Imaginary number' ) ;
	    { Cannot get a negative value from an even number of multiplies }
	end ;
    end ;

    { Determine initial range }
    High := R ;
    High := X / High ;
    if( High < 2 ) then
    begin
	High := 2 ;
    end ;
    Low := 1.0 ;

    { Bisearch for the solution }
    while( Low <> High ) do
    begin
	Mid := Low + ( ( High - Low ) / 2.0 ) ;
	if( ( Low = Mid ) or ( High = Mid ) ) then
	begin
	    break ;
	end ;
	Intermediate := Mid ;
	for Dummy := 1 to R - 1 do
	begin
	    Intermediate := Intermediate * Mid ;
	end ;
	if( Intermediate = X ) then
	begin
	    Low := Mid ;
	    break ;
	end else
	begin
	    if( Intermediate < X ) then
	    begin
		Low := Mid ;
	    end else
	    begin
		High := Mid ;
	    end ;
	end ;
    end ;

    { Return result }
    Root := Low ;
end ;


function Sgn( X : Real ) : real ;
{This function emulates the BASIC-Plus SGN function.  It returns the following:
           for SGN(n):
              n<0 returns -1
              n=0 returns 0
              n>0 returns 1
}

begin
    if (X = 0 ) then
    begin
        Sgn := 0 ;
    end else
    if( X < 0 ) then
    begin
        Sgn := -1 ;
    end else
    begin
        Sgn := 1 ;
    end ;
end ;


function ISgn( X : longint ) : integer ;
{ This function emulates the BASIC-Plus SGN function.  It returns the following:
           for SGN(n):
              n<0 returns -1
              n=0 returns 0
              n>0 returns 1
}

begin
    if( X = 0 ) then
    begin
        ISgn := 0 ;
    end else
    if( X < 0 ) then
    begin
        ISgn := -1 ;
    end else
    begin
        ISgn := 1 ;
    end ;
end ;


function Power( X, N : Real ; var _Result : Integer ) : Real ;
{ This function returns the result of one number to the power of another.
  Based on C code from Computer Language magazine. }

var Size_X : Integer ;
    Temp : Small_string ;

    function Pwr( N : Integer ) : Real ;

    var Temp1 : Real ;

    begin
        Pwr := -1 ; { Assume failure }
        if( _Result <> 0 ) then
        begin
            Pwr := 1 ;
            Exit
        end ;
        if( N = 1 ) then
        begin
            Pwr := X ;
        end else
        begin
            Temp1 := Pwr( N div 2 ) ;
            Str( Temp1 : 40 : 0, Temp ) ;
            Temp := Edit( Temp, 2 ) ;
            if( Odd( N ) ) then
            begin
                if( Size_X + Length( Temp ) * 2 < 32 ) then
                begin
                    Pwr := Sqr( Temp1 ) * X ;
                end else
                begin
                    _Result := 7 ;
                    Exit
                end
            end else
            if( Length( Temp ) * 2 < 32 ) then
            begin
                Pwr := Sqr( Temp1 ) ;
            end else
            begin
                _Result := 7 ;
            end ;
        end
    end ;

begin { Power }
    _Result := 0 ;
    Power := 0 ;
    if( X <> 0 ) then
    begin
        if( N = 0 ) then
        begin
            Power := 1 ;
        end else
        if( Frac( N ) <> 0 ) then
        begin
            if ( X < 0 ) then
            begin
                _Result := 11 ;
                Exit ;
            end ;
            if ( N * Ln( X ) > 88 ) or ( N * Ln( X ) < -88 ) then
            begin
                _Result := 12 ; { Error }
                Exit ;
            end ;
            Power := Exp( N * Ln( X ) )
        end else
        begin
            Str( X : 40 : 0, Temp ) ;
            Temp := Edit( Temp, 2 ) ;
            Size_X := Length( Temp ) ;
            if( ( N > 32767 ) or ( N < -32767 ) ) then
            begin
                _Result := 7 ;
            end else
            if( N < 0 ) then
            begin
                Power := 1 / Pwr( -Trunc( N ) ) ;
            end else
            begin
                Power := Pwr( Trunc( N ) ) ;
            end ;
        end ;
    end ;
end ; { Power }



end.
