{$N+}
{
              Program Name : NUM1S
              Package Name : Subroutine Library - Layer 1
              Purpose      : Convert number to string
              Institution  : Conroy & Conroy Company
              Date Written : 03-Apr-87
              Written By   : Alan Conroy
              Version      : 1.0B

              Copyright (C) 1987 by Alan Conroy.  Released to the public domain.

              TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *********************************************************
        *                                                       *
        *        M O D I F I C A T I O N   H I S T O R Y        *
        *                                                       *
        *     DATE      BY             REASON                   *
        *                                                       *
        *  28-Mar-2003  EAC            Fixed bug in Num2 that   *
        *                              returned random length   *
        *                              strings if the value was *
        *                              too large to fit in non- *
        *                              E-Notation.              *
        *                                                       *
        *  8-Oct-2003   EAC            Fixed Num1 bug where     *
        *                              FP-to-string conversion  *
        *                              inserted fractional value*
        *                              for integers.            *
        *                                                       *
        *********************************************************

        *********************************************************
        *                                                       *
        *           P R O G R A M   P U R P O S E               *
        *                                                       *
        *********************************************************

          These functions perform numeric formatting.

          LEVEL 2 CERTIFICATION ON 3-Sep-2005 BY EAC
}

unit Num1s ;

interface

function INum( Variable : real ) : string ;
{ This function takes a number and returns a string, consisting of the number
  passed, without the fractional portion. }

function Num( X : Longint ) : string ;
function Num1( X : Real ) : string ;
function Num2( X : Extended ) : string ;
function XVal( S : string ; var F : Integer ) : Extended ;
{ Return extended value converted from string S }

{$IFDEF Test}
procedure Test_Unit ;
{$ENDIF}

implementation

uses { Layer 1... }
     CommonUt ; { FEdit }

function INum( Variable : real ) : string ;
{ This function takes a number and returns a string, consisting of the number
  passed. }

var Res : String ;
    A : integer ;

begin
    STR( Variable:10:1, Res ) ;
    FEdit( Res, Res, -1 ) ;
    A := POS( '.', Res + '.' ) ;
    INum := copy( Res, 1, A - 1 ) ;
end ;


function Num( X : Longint ) : string ;

var A : string ;

begin
    Str( X:1, A ) ; { Convert }
    Num := A ; { Return result }
end ;


function Num1( X : Real ) : string ;

var A : string ;
    B : Integer ;

begin
    Str( X:1:11, A ) ; { Convert }
    if( X = int( X ) ) then { Integer value passed in }
    begin
        { Sometimes Str converts an integer real to a string with a fractional
          value.  In this case, we ensure no fraction is included. }
        B := pos( '.', A + '.' ) ;
        A := copy( A, 1, B - 1 ) ; { Trim fractional part }
    end ;
    B := length( A ) ;
    if( pos( '.', A ) > 0 ) then
    begin
        while( copy( A, B, 1 ) = '0' ) do
        begin
            B := B - 1 ;
        end ;
        if( copy( A, B, 1 ) = '.' ) then
        begin
            B := B - 1 ; { UNTESTED }
        end ;
    end ;
    Num1 := copy( A, 1, B ) ; { Return result }
end ;


function NUM2( X : Extended ) : string ;
{ Converted extended floating-point value to ASCII string representation }

var A, E : string ;
    B, D, P, V : Integer ;
    N : Boolean ;

begin
    STR( X, A ) ; { Convert }
    FEdit( A, A, 8 or 128 ) ; { Trim spaces }
    P := Pos( 'E', A ) ;
    if( P = 0 ) then
    begin
        { UNTESTED... }
	B := length( A ) ;
	while( copy( A, B, 1 ) = '0' ) do
        begin
            B := B - 1 ;
        end ;
	if( copy( A, B, 1 ) = '.' ) then
        begin
            B := B - 1 ;
        end ;
    end else { Handle E-Notation }
    begin
	E := Copy( A, P + 2, 254 ) ; { Get number }
	Val( E, V, D ) ;
	if( V < 253 ) then { Can fit into non-E-Notation } { V >=253 UNTESTED }
	begin
	    if( A[ P + 1 ] = '-' ) then
            begin
		V := -V ;
	    end ;
            if( A[ 1 ] = '-' ) then
            begin
                N := true ;
                A := COPY( A, 2, 254 ) ; { Strip sign }
                Dec( P ) ; { Shift things left one character }
            end else
            begin
                N := false ;
            end ;
	    E := COPY( A, 1, P - 1 ) ; { Strip E+/-nnnn }
	    E := E[ 1 ] + COPY( E, 3, 254 ) ; { Strip decimal }

	    { Determine placement of decimal and zeroes }
	    if( V >= 0 ) then { Move decimal right }
            begin
                while( Length( E ) <= V ) do
                begin
                    E := E + '0' ; { UNTESTED }
                end ;
                E := Copy( E, 1, V + 1 ) + '.' + Copy( E, V + 2, 254 ) ;
            end else
            begin { Move decimal left }
                while( V < -1 ) do
                begin
                    Inc( V ) ;
                    E := '0' + E ;
                end ;
                E := '.' + E ;
            end ;

	    { Trim leading zeroes }
	    while( Copy( E, 1, 2 ) = '00' ) do
            begin
                E := Copy( E, 2, 254 ) ; { UNTESTED }
            end ;

	    if N then { If negative }
            begin
	        E := '-' + E ; { Add sign back on }
            end ;
	    P := Pos( '.', E ) ;
	    if( P > 0 ) then { P < 0 UNTESTED }
            begin
		while( ( Length( E ) > P ) and ( E[ Length( E ) ] = '0' ) ) do
                begin
		    E := Copy( E, 1, Length( E ) - 1 ) ; { Trim trailing 0 }
                end ;
	    end ;
            if( E[ Length( E ) ] = '.' ) then
	    begin
		E := Copy( E, 1, Length( E ) - 1 ) ; { Trim trailing . }
            end ;
            if( copy( E, 1, 1 ) = '.' ) then
            begin
                E := '0' + E ;
            end ;
	    NUM2 := E ;
	    exit ;
	end ; { if( V < 253 ) }
        B := length( A ) ; { Just copy whole string, as is }
    end ; { if( P = 0 ) }
    { UNTESTED: following code }
    A := copy( A, 1, B ) ;
    if( copy( A, 1, 1 ) = '.' ) then
    begin
        A := '0' + A ;
    end ;
    Num2 := A ; { Return result }
end ; { Num2 }


procedure Parse_Real( Value : string ; var Sign, Int, Decimal, Fractional, E,
    E_Sign, Scale : string ) ; { Parse real number into string segments }

var D, D1 : Integer ;

begin
    FEDIT( Value, Value, -1 ) ;
    Sign := '' ;
    Int := '';
    Decimal := '';
    Fractional := '' ;
    E := '' ;
    E_Sign := '' ;
    Scale := '' ;
    if( copy( Value, 1, 1 ) = '-' ) then
    begin
	Value := Copy( Value, 2, 254 ) ;
	Sign := '-' ;
    end ;
    D := Pos( '.', Value + '.' ) ;
    D1 := Pos( 'E', Value + 'E' ) ;
    if( D1 < D ) then { Means no decimal, but E-Notation }
    begin
	D := D1 ; { Should never happen due to callling code } { UNTESTED }
    end ;
    Int := Copy( Value, 1, D - 1 ) ;
    while( ( Length( Int ) > 0 ) and ( Int[ 1 ] = '0' ) ) do
    begin
	Int := Copy( Int, 2, 254 ) ; { Remove leading zeroes }
    end ;
    if( Length( Int ) = 0 ) then
    begin
	Int := '0' ;
    end ;
    Value := Copy( Value, D, 254 ) ;
    if( Copy( Value, 1, 1 ) = '.' ) then
    begin
	Value := Copy( Value, 2, 254 ) ;
	Decimal := '.' ;
	D := Pos( 'E', Value + 'E' ) ;
	Fractional := Copy( Value, 1, D - 1 ) ;
	while(
               ( Length( Fractional ) > 1 )
               and
	       ( Fractional[ Length( Fractional ) ] = '0' )
             ) do
        begin
	    Fractional := Copy( Fractional, 1, Length( Fractional ) - 1 ) ;
	    { Remove trailing zeroes }
        end ;
	Value := Copy( Value, D, 254 ) ;
    end ;
    if( Copy( Value, 1, 1 ) = 'E' ) then
    begin
        { UNTESTED }
	E := 'E' ;
	E_Sign := Copy( Value, 2, 1 ) ;
	Scale := Copy( Value, 3, 254 ) ;
    end ;
end ; { Parse_Real }


function XVal( S : string ; var F : Integer ) : Extended ;
{ Return extended value converted from string S }

var D : Integer ;
    X : Extended ;
    Sgn, SI, SD, SF, SE, SES, SS : string ;

begin
    FEdit( S, S, -1 ) ;
    if( ( Length( S ) > 19 ) and ( Pos( 'E', S ) = 0 ) ) then
    begin { We need to massage it before VAL tries it }
        Parse_Real( S, Sgn, SI, SD, SF, SE, SES, SS ) ;
        { Parse real number into string segments }

        if( SI = '0' ) then { Significance is in fractional part }
        begin
            D := 0 ;
            while( ( Length( SF ) > 1 ) and ( SF[ 1 ] = '0' ) ) do
            begin { Trim leading zeroes and count how many there are }
                Inc( D ) ;
                SF := Copy( SF, 2, 254 ) ;
            end ;
            if( Length( SF ) > 20 ) then { Get rid of excess digits } { else condition UNTESTED }
            begin
                Set_Length( SF, 20 ) ; { Trim fractional to 20 digits }
            end ;
            SE := 'E' ;
            SES := '-' ;
            SS := NUM( D ) ;
        end else { Significance is in the integer portion }
        begin
            if( Length( SF ) > 20 ) then { Trim excess characters }
            begin
                Set_Length( SF, 20 ) ; { Trim fractional to 20 digits }
            end ;
            if( Length( SI ) > 20 ) then { Trim excess }
            begin
                SS := NUM( Length( SI ) - 20 - length( SF ) ) ;
                SE := 'E' ;
                if( copy( SS, 1, 1 ) = '-' ) then
                begin
                    SES := '' ;
                end else
                begin
                    SES := '+' ;
                end ;
                Set_Length( SI, 20 ) ; { Trim fractional to 20 digits }
                SI := SI + SF ;
                SF := '' ;
                SD := '' ;
            end ;
        end ; { if( SI = '0' ) }
        if( SS = '0' ) then { No exponent }
        begin
            SE := '' ;
            SES := '' ;
            SS := '' ;
        end ;
        S := Sgn + SI + SD + SF + SE + SES + SS ; { Build string }
    end ; { if( ( Length( S ) > 19 ) and ( Pos( 'E', S ) = 0 ) ) }
    Val( S, X, F ) ;
    XVal := X ;
end ; { XVal }



{$IFDEF Test}
procedure Error ;

begin
    writeln( 'Test failed' ) ;
    halt ;
end ;


procedure Test_Unit ;

var F : integer ;
    L : longint ;
    R : real ;
    E : extended ;

begin
    { INum tests... }
    R := 0.0 ;
    if( INum( R ) <> '0' ) then
    begin
        Error ;
    end ;
    R := -1.0 ;
    if( INum( R ) <> '-1' ) then
    begin
        Error ;
    end ;
    R := 1.0E+10 ;
    if( INum( R ) <> '10000000000' ) then
    begin
        Error ;
    end ;
    R := 3.141592 ;
    if( INum( R ) <> '3' ) then
    begin
        Error ;
    end ;

    { Num tests... }
    L := 0 ;
    if( Num( L ) <> '0' ) then
    begin
        Error ;
    end ;
    L := -32768 ;
    if( Num( L ) <> '-32768' ) then
    begin
        Error ;
    end ;
    L := 1234567890 ;
    if( Num( L ) <> '1234567890' ) then
    begin
        Error ;
    end ;

    { Num1 tests... }
    R := 0.0 ;
    if( Num1( R ) <> '0' ) then
    begin
        Error ;
    end ;
    R := -1.0 ;
    if( Num1( R ) <> '-1' ) then
    begin
        Error ;
    end ;
    R := 1.0E+10 ;
    if( Num1( R ) <> '10000000000' ) then
    begin
        Error ;
    end ;
    R := 1.0E-10 ;
    if( Num1( R ) <> '0.0000000001' ) then
    begin
        Error ;
    end ;
    R := 3.141592 ;
    if( Num1( R ) <> '3.141592' ) then
    begin
        Error ;
    end ;

    { Num2 tests... }
    E := 0.0 ;
    if( Num2( E ) <> '0' ) then
    begin
        Error ;
    end ;
    E := -1.0 ;
    if( Num2( E ) <> '-1' ) then
    begin
        Error ;
    end ;
    E := 1.0E+10 ;
    if( Num2( E ) <> '10000000000' ) then
    begin
        Error ;
    end ;
    E := 1.0E-10 ;
    if( Num2( E ) <> '0.0000000001' ) then
    begin            
        Error ;
    end ;
    E := 3.141592 ;
    if( Num2( E ) <> '3.141592' ) then
    begin
        Error ;
    end ;
    E := 3.141592E+2 ;
    if( Num2( E ) <> '314.1592' ) then
    begin
        Error ;
    end ;

    { XVal tests... }
    if( XVal( '32768', F ) <> 32768 ) then
    begin
        Error ;
    end ;
    if( F <> 0 ) then
    begin
        Error ;
    end ;
    if( XVal( '0.32768', F ) <> 0.32768 ) then
    begin
        Error ;
    end ;
    if( F <> 0 ) then
    begin
        Error ;
    end ;
    if( XVal( '3.2E+5', F ) <> 320000 ) then
    begin
        Error ;
    end ;
    if( F <> 0 ) then
    begin
        Error ;
    end ;
    if( XVal( '-3.2E+5', F ) <> -320000 ) then
    begin
        Error ;
    end ;
    if( XVal( '-12345678901234567890', F ) <> -12345678901234567890.0 ) then
    begin
        Error ;
    end ;
    if( F <> 0 ) then
    begin
        Error ;
    end ;
    if( XVal( '123456789012345678901', F ) <> 123456789012345678901.0 ) then
    begin
        Error ;
    end ;
    if( F <> 0 ) then
    begin
        Error ;
    end ;
    if( XVal( '-123456789012345678901.123456789012345678901', F ) <> -123456789012345678901.0 ) then
    begin
        Error ;
    end ;
    if( F <> 0 ) then
    begin
        Error ;
    end ;
    if( XVal( '-123456789012345678901', F ) <> -123456789012345678901.0 ) then
    begin
        Error ;
    end ;
    if( F <> 0 ) then
    begin
        Error ;
    end ;
    if( XVal( '0123456789012345678901', F ) <> 123456789012345678901.0 ) then
    begin
        Error ;
    end ;
    if( F <> 0 ) then
    begin
        Error ;
    end ;
    if( XVal( '123456789012345678901.123', F ) <> 123456789012345678901.0 ) then
    begin
        Error ;
    end ;
    if( F <> 0 ) then
    begin
        Error ;
    end ;
    if( XVal( '0.0123456789012345678901', F ) <> 0.0123456789012345678901 ) then
    begin
        Error ;
    end ;
    if( F <> 0 ) then
    begin
        Error ;
    end ;
    if( XVal( '-123456789012345678901.123456789012345678901E+1', F ) <> -123456789012345678901.123456789012345678901E+1 ) then
    begin
        Error ;
    end ;
    if( F <> 0 ) then
    begin
        Error ;
    end ;
    if( Num2( XVal( '.1234567890123456789010', F ) ) <> '0.123456789012345' ) then
    begin
        Error ;
    end ;
    if( F <> 0 ) then
    begin
        Error ;
    end ;
    if( Num2( XVal( '.123456789012345678901', F ) ) <> '0.123456789012345' ) then
    begin
        Error ;
    end ;
    if( F <> 0 ) then
    begin
        Error ;
    end ;
end ;
{$ENDIF}



end.
