{
        Program Name : Output_Panel
        Package Name : LB
        Purpose      : TPanel for output
        Institution  : Conroy & Conroy Co.
        Date Written : 12-Jan-1999
        Written By   : Alan Conroy
        Version      : 1.0

        Copyright (C) 1999 by Alan Conroy.  Released to the public domain.
        
        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          TOutput_Panel provides an output-only display.
}

unit Output_Panel ;

interface

uses // Borland...
     Windows,
     Messages,
     SysUtils,
     Classes, // TNotifyEvent
     Graphics,
     Controls,
     Forms,
     Dialogs,
     StdCtrls, // TListBox
     ExtCtrls ; // TPanel

type TOutput_Panel = class( TPanel )
         public // Constructors and destructors...
             constructor Create( AOwner : TComponent ) ; override ;

         private // Instance data...
             LB : TListBox ;
             Line, Column : integer ;

         public // Overrides...
             procedure CB_Draw_Line( Control : TWinControl ; Index : Integer ;
                 Rect : TRect ; State : TOwnerDrawState ) ;
             procedure CB_LB_Click( Sender : TObject ) ;

         public // API...
             Margin : integer ;

             procedure Output( const S : string ; Color : TColor = clWhite ) ;
             procedure Replace( Starting, Ending : integer ;
                 const Value : string ) ;

         published // API...
             property CCPos : integer
                 read Column ;
             property CLPos : integer
                 read Line ;
     end ;

implementation

uses // C&C...
     ASCIIDef ; // CR

// TOutput_Panel methods...

// Constructors and destructors...

constructor TOutput_Panel.Create( AOwner : TComponent ) ;

begin
    inherited Create( AOwner ) ;

    Margin := 2 ;
    Column := 1 ;
    
    LB := TListBox.Create( self ) ;
    LB.Parent := self ;
    LB.Align := alClient ;
    LB.ParentColor := True ;
    LB.ParentShowHint := True ;
    LB.ParentFont := True ;
    LB.ParentBiDiMode := True ;
    LB.OnDrawItem := CB_Draw_Line ;
    LB.BorderStyle := bsNone ;
    LB.Style := lbOwnerDrawFixed ;
    LB.Ctl3D := False ;
    LB.OnClick := CB_LB_Click ;
end ;


// Overrides...

procedure TOutput_Panel.CB_Draw_Line( Control : TWinControl ; Index : Integer ;
    Rect : TRect ; State : TOwnerDrawState ) ;

var R : TRect ;

begin
    LB.Canvas.Font.Color := TColor( LB.Items.Objects[ Index ] ) ;
    R := Rect ;
    R.Left := -1 ;
    R.Top := R.Top - 1 ;
    R.Bottom := R.Bottom + 1 ;
    FillRect( Canvas.Handle, R, LB.Brush.Handle ) ;
    SetBKMode( Canvas.Handle, OPAQUE ) ;
    LB.Canvas.TextOut( Rect.Left + Margin, Rect.Top, LB.Items[ Index ] ) ;
end ;


procedure TOutput_Panel.CB_LB_Click( Sender : TObject ) ;

begin
    if( assigned( OnEnter ) ) then
    begin
        OnEnter( self ) ;
    end ;
end ;


// API...

procedure TOutput_Panel.Output( const S : string ; Color : TColor ) ;

var Before, After : string ;
    Loop : integer ;
    Work : char ;

begin
    if( Line >= LB.Count ) then
    begin
        LB.Items.AddObject( '', pointer( Color ) ) ;
    end ;
    if( length( LB.Items[ Line ] ) = 0 ) then
    begin
        LB.Items.Objects[ Line ] := pointer( Color ) ;
    end ;
    Before := copy( LB.Items[ Line ], 1, Column - 1 ) ;
    After := copy( LB.Items[ Line ], Column, length( LB.Items[ Line ] ) ) ;
    for Loop := 1 to length( S ) do
    begin
        Work := S[ Loop ] ;
        if( Work = CR ) then
        begin
            Column := 1 ;
        end else
        if( Work = LF ) then
        begin
            LB.Items[ Line ] := Before ;
            LB.Items.AddObject( '', pointer( Color ) ) ;
            inc( Line ) ;
            Before := copy( LB.Items[ Line ], 1, Column - 1 ) ;
            After := copy( LB.Items[ Line ], Column, length( LB.Items[ Line ] ) ) ;
        end else
        begin
            Before := Before + Work ;
            Column := Column + 1 ;
        end ;
    end ; // for Loop := 1 to length( S )
    LB.Items[ Line ] := Before + After ;
    LB.Invalidate ;
end ;



procedure TOutput_Panel.Replace( Starting, Ending : integer ;
    const Value : string ) ;

begin
    if( Ending = -1 ) then
    begin
        Ending := length( LB.Items[ Line ] ) ;
    end ;
    LB.Items[ Line ] := copy( LB.Items[ Line ], 1, Starting - 1 ) + Value +
        copy( LB.Items[ Line ], Ending + 1, length( LB.Items[ Line ] ) ) ;
end ;


end.
