unit R2D ;

interface

uses // Borland...
     Windows,   // TPoint

     Classes,   // TList
     Graphics,  // TColor
     
     // C&C...
     TypeDefs ;    // PInteger_Array

     // Points...
type TPoint_Array = array[ 0..Max_Memory div sizeof( TPoint ) - 1 ] of TPoint ;
     PPoint_Array = ^TPoint_Array ;

     TTexture = class
                    public // Instance data
                        Height, Width : integer ; // Image dimensions
                        _Array : PInteger_Array ; // Image

                    protected // Property handlers...
                        function Get_Pixels( X, Y : integer ) : integer ;
                        procedure Set_Pixels( X, Y, Value : integer ) ;

                    public // API...
                        destructor Destroy ; override ;

                        procedure Blank ;
                        procedure Clear ;
                        procedure Fill( Color : integer ) ;
                        procedure Init( X, Y : integer ) ;
                        procedure Load_From_Bitmap( Bitmap : TBitmap ) ;
                        procedure Load_From_File( S : string ) ;
                        procedure Set_Alpha( RGB : integer ) ;
                        procedure Set_Buffer( P : pointer ) ;
                        procedure Stretch ;
                        property Pixels[ X, Y : integer ] : integer
                            read Get_Pixels
                            write Set_Pixels ;
                end ;

     TXY_List = class
                    private // Instance data
                        X_List : TList ;
                        Y_List : TList ;

                    protected // Property handlers
                        function Get_X( Index : integer ) : TPoint ;

                        procedure Set_X( Index : integer ; Value : TPoint ) ;

                    public // API
                        constructor Create ; virtual ;
                        destructor Destroy ; override ;

                        function Count : integer ;
                        procedure Clear ; virtual ;
                        procedure Generate( A, B : TPoint ) ;

                        property X[ Index : integer ] : TPoint
                                     read Get_X
                                     write Set_X ;
                                     default ;
                end ;

     TPolygon = class( TXY_List )
                    private // Instance data
                        Bitmap_X_List : TList ;
                        Bitmap_Y_List : TList ;
                        Vertex_X_List : TList ;
                        Vertex_Y_List : TList ;
			Lines_Left_X : TList ;
			Lines_Y : TList ;
			Lines_Right_X : TList ;
                        Lines_Bitmap_X_Right : TList ;
                        Lines_Bitmap_Y_Right : TList ;
                        Lines_Bitmap_X_Left : TList ;
                        Lines_Bitmap_Y_Left : TList ;

                    protected // Property handlers
                        function Get_X( Index : integer ) : TPoint ;

                        procedure Set_X( Index : integer ; Value : TPoint ) ;

                    public // API
                        constructor Create ; override ;
                        destructor Destroy ; override ;

                        procedure Add( X : TPoint ) ;
                        procedure Clear ; override ;
                        procedure Generate ;
                        procedure Generate_Texture_Line( Bitmap, Other : integer ;
                            B_X : boolean ; A, B : TPoint ) ;
                        procedure Generate_With_Texture( Bitmap : TTexture ) ;
			function Line_Count : integer ;
			function Line_Left( Index : integer ) : TPoint ;
			function Line_Right( Index : integer ) : TPoint ;
                        function Line_Left_Texture( Index : integer ) : TPoint ;
                        function Line_Right_Texture( Index : integer ) : TPoint ;
			procedure MinMax ;
                        function Vertex_Count : integer ;

                        property Vertex[ Index : integer ] : TPoint
                                     read Get_X
                                     write Set_X ;
                end ;

procedure Draw_Line( C : TTexture ; A, B : TPoint ; Color : TColor ;
    Pixel_Map : pInteger_Array ; O : integer ) ;
procedure Polygon( C : TTexture ; Points : PPoint_Array ; Count : integer ; Color : TColor ) ;
procedure Filled_Polygon( C : TTexture ; Points : PPoint_Array ;
    Count : integer ; Color : TColor ; Pixel_Map : pInteger_Array ;
    O : integer ) ;
procedure Draw_Polygon( C : TTexture ; Bitmap : TTexture ;
    Points : PPoint_Array ; Count : integer ; Pixel_Map : pInteger_Array ;
    O : integer ) ;
procedure Rotate( Points : PPoint_Array ; Count : integer ; Angle : real ) ;

implementation

uses // Borland...
     SysUtils ; // Allocmem

procedure Rotate( Points : PPoint_Array ; Count : integer ; Angle : real ) ;
{ Do a 2-D rotate operation.  The passed points are rotated around point (0,0)
  in a counterclockwise direction.  Thus, (5,-5) rotated 90 degrees would become
  (5,5). }

var ZRotate : real ;
    Sin_XRotate : real ;
    Sin_YRotate : real ;
    Sin_ZRotate : real ;
    Cos_XRotate : real ;
    Cos_YRotate : real ;
    Cos_ZRotate : real ;
     X1, X2, X3, Y1, Y2, Y3, Z1 : real ;

begin
    if( Angle = 0 ) then
    begin
        exit ;
    end ;
    ZRotate := Angle * ( pi / 180.0 ) ;

    Sin_XRotate := sin( 0 ) ;
    Sin_YRotate := Sin_XRotate ;
    Sin_ZRotate := sin( ZRotate ) ;
    Cos_XRotate := cos( 0 ) ;
    Cos_YRotate := Cos_XRotate ;
    Cos_ZRotate := cos( ZRotate ) ;

    dec( Count ) ;
    while( Count >= 0 ) do
    begin
        X1 := Points^[ Count ].X ;
        Y1 := Points^[ Count ].Y ;
        Y1 := Y1 * cos_XRotate  ;
        Z1 := Points^[ Count ].Y ;
        Z1 := Z1 * sin_XRotate ;

        X2 := Z1 * sin_YRotate + X1 * cos_YRotate ;
        Y2 := Y1 ;

        X3 := X2 * cos_ZRotate - Y2 * sin_ZRotate ;
        Y3 := X2 * sin_ZRotate + Y2 * cos_ZRotate ;

        Points^[ Count ].X := trunc( X3 ) ;
        Points^[ Count ].Y := trunc( Y3 ) ;
        dec( Count ) ;
    end ;
end ; // Rotate


procedure Draw_Line( C : TTexture ; A, B : TPoint ; Color : TColor ;
    Pixel_Map : pInteger_Array ; O : integer ) ;

var i : integer ;
    List : TXY_List ;
    Temp : integer ;
    X, Y : integer ;

begin
    List := TXY_List.Create ;
    List.Generate( A, B ) ;
    for i := 0 to List.Count - 1 do
    begin
        X := List[ i ].X ;
        Y := List[ i ].Y ;
        if( ( X >= 0 ) and ( Y >= 0 ) and ( X < C.Width ) and ( Y < C.Height ) ) then
        begin
            Temp := X + Y * C.Width ;
            C._Array[ Temp ] := integer( Color ) ;
            if( Pixel_Map <> nil ) then
            begin
                Pixel_Map[ Temp ] := O ;
            end ;
        end ;
    end ;
    List.Free ;
end ;


procedure Polygon( C : TTexture ; Points : PPoint_Array ; Count : integer ; Color : TColor ) ;

var Loop : integer ;
    P : TPolygon ;
    X, Y : integer ;

begin
    P := TPolygon.Create ;
    for Loop := 0 to Count - 1 do
    begin
        P.Add( Points[ Loop ] ) ;
    end ;
    P.Generate ;
    for Loop := 0 to P.Count - 1 do
    begin
        X := P[ Loop ].X ;
        Y := P[ Loop ].Y ;
        if( ( X >= 0 ) and ( Y >= 0 ) and ( X < C.Width ) and ( Y < C.Height ) ) then
        begin
            C._Array[ X + Y * C.Width ] := integer( Color ) ;
        end ;
    end ;
    P.Free ;
end ; // Polygon


procedure Filled_Polygon( C : TTexture ; Points : PPoint_Array ;
    Count : integer ; Color : TColor ; Pixel_Map : PInteger_Array ; O : integer ) ;

var Loop : integer ;
    P : TPolygon ;

begin
    P := TPolygon.Create ;
    for Loop := 0 to Count - 1 do
    begin
        P.Add( Points[ Loop ] ) ;
    end ;
    P.Generate ;
    P.MinMax ;
    for Loop := 0 to P.Line_Count - 1 do
    begin
        Draw_Line( C, P.Line_Left( Loop ), P.Line_Right( Loop ), Color,
            Pixel_Map, O ) ;
    end ;
    P.Free ;
end ; // Filled_Polygon


// Draw a textured polygon
procedure Draw_Polygon( C : TTexture ; Bitmap : TTexture ;
    Points : PPoint_Array ; Count : integer ; Pixel_Map : pInteger_Array ;
    O : integer ) ;

var alpha : integer ; // Current alpha value
    B_X, B_Y : integer ; // Current texture bitmap positions
    B_Left, B_Right : TPoint ; // Endpoints of current texture path
    Color : integer ; // Current bit color
    Delta_X, Delta_Y : integer ; // Deltas to add to texture positions
    Line_Length : integer ;
    Loop : integer ;
    Max_X_Delta, Max_Y_Delta : integer ;
    P : TPolygon ;
    Pixel : integer ;
    Pixel_Offset : integer ;
    Pt_Left, Pt_Right: TPoint ;
    Low_X, Low_Y, High_X, High_Y : integer ;
    Temp : integer ;

begin
    // Create a polygon to do the work...
    P := TPolygon.Create ;
    Low_X := 32767 ;
    High_X := -32767 ;
    Low_Y := 32767 ;
    High_Y := -32767 ;
    for Loop := 0 to Count - 1 do
    begin
        P.Add( Points[ Loop ] ) ;
        if( Points[ Loop ].X < Low_X ) then
        begin
            Low_X := Points[ Loop ].X ;
        end ;
        if( Points[ Loop ].Y < Low_Y ) then
        begin
            Low_Y := Points[ Loop ].Y ;
        end ;
        if( Points[ Loop ].X > High_X ) then
        begin
            High_X := Points[ Loop ].X ;
        end ;
        if( Points[ Loop ].Y > High_Y ) then
        begin
            High_Y := Points[ Loop ].Y ;
        end ;
    end ;
    Max_X_Delta := abs( High_X - Low_X ) ;
    while( Max_X_Delta > Bitmap.Width shl 1 ) do
    begin
        Bitmap.Stretch ;
    end ;
    Max_Y_Delta := abs( High_Y - Low_Y ) ;
    while( Max_Y_Delta > Bitmap.Height shl 1 ) do
    begin
        Bitmap.Stretch ;
    end ;

    // Generate the boundaries of the polygon...
    P.Generate_With_Texture( Bitmap ) ;
    P.MinMax ;

    // Draw the polygon
    for Loop := 0 to P.Line_Count - 1 do // For each line
    begin
        // Get line left/right end points...
        Pt_Right := P.Line_Right( Loop ) ;
        Pt_Left := P.Line_Left( Loop ) ;
        if( Pt_Left.Y < 0 ) then // Line is outside of visible area
        begin
            continue ;
        end ;

        // Get texture path end points...
        B_Left := P.Line_Left_Texture( Loop ) ;
        B_Right := P.Line_Right_Texture( Loop ) ;

        // Determine X and Y deltas through the texture...
        Delta_X := abs( B_Left.X - B_Right.X ) ;
        Delta_Y := abs( B_Left.Y - B_Right.Y ) ;
        Line_Length := abs( Pt_Right.X - Pt_Left.X + 1 ) ;
        Delta_X := ( Delta_X shl 16 ) div Line_Length ;
        Delta_Y := ( Delta_Y shl 16 ) div Line_Length ;
        if( B_Left.X > B_Right.X ) then
        begin
            Delta_X := -Delta_X ;
        end ;
        if( B_Left.Y > B_Right.Y ) then
        begin
            Delta_Y := -Delta_Y ;
        end ;
        B_X := B_Left.X shl 16 ;
        B_Y := B_Left.Y shl 16 ;

        // Draw the line
        if(
            ( Pt_Right.X >= 0 ) // At least partially visible
            and
            ( Pt_Left.Y < C.Height ) // Make sure we are within the texture's range
          ) then
        begin
            Pixel_Offset := Pt_Left.Y * C.Width ;
            for Pixel := Pt_Left.X to Pt_Right.X do
            begin
                if( ( Pixel >= 0 ) and ( Pixel < C.Width ) ) then // Inside output texture
                begin
                    Color := Bitmap._Array[ B_X shr 16 + ( B_Y shr 16 ) * Bitmap.Width ] ;
                    Alpha := Color shr 24 ;
                    if( Alpha <> 0 ) then // Not transparent
                    begin
                        Temp := Pixel + Pixel_Offset ;
                        // Do alpha blending here ~~~
                        C._Array[ Temp ] := Color ;
                        if( Pixel_Map <> nil ) then
                        begin
                            Pixel_Map^[ Temp ] := O ;
                        end ;
                    end ;
                end ; // if( ( Pixel >= 0 ) and ( Pixel < C.Width ) )
                B_X := B_X + Delta_X ;
                B_Y := B_Y + Delta_Y ;
            end ; // for Pixel := Pt_Left.X to Pt_Right.X
        end ;
    end ; // for Loop := 0 to P.Line_Count - 1

    // Done with the polygon object
    P.Free ;
end ; // Draw_Polygon


// TTexture methods...

// API...

destructor TTexture.Destroy ;

begin
    Clear ;

    inherited Destroy ;
end ;


// Property handlers...

function TTexture.Get_Pixels( X, Y : integer ) : integer ;

begin
    if( _Array = nil ) then
    begin
        Result := 0 ;
    end else
    begin
        Result := _Array[ X * Width + Y ] ;
    end ;
end ;


procedure TTexture.Set_Pixels( X, Y, Value : integer ) ;

begin
    _Array[ X * Width + Y ] := Value ;
end ;


procedure TTexture.Blank ;

begin
    if( _Array <> nil ) then
    begin
        fillchar( _Array^, Height * Width * sizeof( integer ), 0 ) ;
    end ;
end ;


procedure TTexture.Fill( Color : integer ) ;

var W, H : integer ;

begin
    if( _array <> nil ) then
    begin
        for H := 0 to Height - 1 do
        begin
            for W := 0 to Width - 1 do
            begin
                _Array^[ H * Width + W ] := Color ;
            end ;
        end ;
    end ;
end ;


procedure TTexture.Clear ;

begin
    if( _Array <> nil ) then
    begin
        freemem( _Array ) ;
        Height := 0 ;
        Width := 0 ;
    end ;
end ;


procedure TTexture.Init( X, Y : integer ) ;

begin
    Clear ;
    Height := Y ;
    Width := X ;
    _Array := Allocmem( Height * Width * sizeof( integer ) ) ;
end ;


procedure TTexture.Load_From_Bitmap( Bitmap : TBitmap ) ;

var I, X, Y : integer ;

begin
    // Setup...
    Clear ;
    Height := Bitmap.Height ;
    Width := Bitmap.Width ;
    _Array := Allocmem( Height * Width * sizeof( integer ) ) ;

    // Load from bitmap...
    I := 0 ;
    for Y := 0 to Height - 1 do
    begin
        for X := 0 to Width - 1 do
        begin
{$R-}
{$WARNINGS OFF}
            _Array[ I ] := integer( Bitmap.Canvas.Pixels[ X, Y ] ) or $FF000000 ;
{$WARNINGS ON}
{$R+}
            inc( I ) ;
        end ;
    end ;
end ; // TTexture.Load_From_Bitmap


procedure TTexture.Load_From_File( S : string ) ;

var Header : TBitmapFileHeader ;
    Bits : integer ;
    BitmapInfoHeader : TBitmapInfoHeader ;
    BitmapCoreHeader : TBitmapCoreHeader ;
    Byte_Width : integer ;
    Fil : file ;
    Dummy : integer ;
    Bottom_Up : boolean ;
    Buffer : PChar ;
    C, I, D, Src : integer ;
    Padding : integer ; // Number of pad bytes at end of each line
    PEL : array[ 0..255 ] of TPaletteEntry ;
    PEL_Size : integer ;
    Temp : integer ; // Temporary value for expanding 5-bit color values
    X : integer ; // Current pixel on current line

begin
    assign( Fil, S ) ;
    reset( Fil, 1 ) ;
    blockread( Fil, Header, sizeof( Header ) ) ;
    if( Header.bfType <> 19778 ) then
    begin
        raise Exception.Create( 'Invalid bitmap file format' ) ;
    end ;
    blockread( Fil, Dummy, 4 ) ;
{$WARNINGS OFF}
    PEL_Size := Header.bfOffBits - Dummy - sizeof( Header ) ;
{$WARNINGS ON}
    Bottom_Up := False ;
    if( Dummy = sizeof( TBitmapInfoHeader ) ) then
    begin
        BitmapInfoHeader.biSize := Dummy ;
        blockread( Fil, BitmapInfoHeader.biWidth, Dummy - 4 ) ;
        if( BitmapInfoHeader.biCompression <> 0 ) then
        begin
            raise Exception.Create( 'Cannot read compressed bitmap files' ) ;
        end ;
        Bits := BitmapInfoHeader.biBitCount ;
        Width := BitmapInfoHeader.biWidth ;
        Height := BitmapInfoHeader.biHeight ;
        if( Height < 0 ) then
        begin
            Height := -Height ;
        end else
        begin
            Bottom_Up := True ;
        end ;
        blockread( Fil, PEL, Pel_Size ) ;
        Byte_Width := ( ( Width * Bits ) + 7 ) div 8 ;
        Padding := Byte_Width and 3 ; // Bytes left-over per line
        if( Padding > 0 ) then
        begin
            Padding := 4 - Padding ;
        end ;
    end else
    begin
        BitmapCoreHeader.bcSize := Dummy ;
        blockread( Fil, BitmapCoreHeader.bcWidth, Dummy - 4 ) ;
        Bits := BitmapCoreHeader.bcBitCount ;
        Width := BitmapCoreHeader.bcWidth ;
        Height := BitmapCoreHeader.bcHeight ;
        PEL_Size := PEL_Size div 3 ;
        D := 0 ;
        while( PEL_Size > 0 ) do
        begin
            fillchar( PEL[ D ], 0, 4 ) ;
            blockread( Fil, PEL[ D ], 3 ) ;
            inc( D ) ;
            dec( PEL_Size ) ;
        end ;
        Padding := 0 ;
        Byte_Width := ( ( Width * Bits ) + 7 ) div 8 ;
    end ;
    _Array := Allocmem( Height * Width * sizeof( integer ) ) ;

    // Get bitmap pixels...
    I := Height * ( Byte_Width + Padding ) ;
    Buffer := Allocmem( I ) ;
    blockread( Fil, Buffer^, I ) ;
    // Load texture from bitmap...
    if( Bottom_Up ) then
    begin
        Src := ( Height - 1 ) * ( Byte_Width + Padding ) ;
    end else
    begin
        Src := 0 ;
    end ;
    X := 0 ;
    D := 0 ; // Current position in texture
    while( I > 0 ) do
    begin
        // 16-color
        if( Bits = 4 ) then
        begin
            C := ord( Buffer[ Src ] ) shr 4 ;
{$R-}
{$WARNINGS OFF}
            _Array[ D ] := integer( PEL[ C ] ) or $FF000000 ;
{$WARNINGS ON}
{$R+}
            inc( D ) ;
            inc( X ) ;

            if( X < Width ) then
            begin
                C := ord( Buffer[ Src ] ) and $F ;
{$R-}
{$WARNINGS OFF}
                _Array[ D ] := integer( PEL[ C ] ) or $FF000000 ;
{$WARNINGS ON}
{$R+}
                inc( D ) ;
                inc( X ) ;
            end ;

            inc( Src ) ;
            if( X = Width ) then
            begin
                X := 0 ;
                Src := Src + Padding ; // Goto end of line
                if( Bottom_Up ) then
                begin
                    Src := Src - ( Byte_Width + Padding ) * 2 ; // Go to start of previous line
                end ;
                I := I - Padding ;
            end ;
            dec( I ) ;
            continue ;
        end ;

        // Monochrome
        if( Bits = 1 ) then
        begin
            C := ord( Buffer[ Src ] ) ;
            for Dummy := 0 to 7 do
            begin
{$R-}
{$WARNINGS OFF}
                if( ( C and 128 ) = 0 ) then
                begin
                    _Array[ D ] := $FF000000 ;
                end else
                begin
                    _Array[ D ] := $FFFFFFFF ;
                end ;
{$WARNINGS ON}
{$R+}
                inc( D ) ;
                inc( X ) ;
                if( X = Width ) then
                begin
                    break ;
                end ;
                C := C shl 1 ;
            end ; //  for Dummy := 0 to 7

            inc( Src ) ;
            if( X = Width ) then
            begin
                X := 0 ;
                Src := Src + Padding ; // Goto end of line
                if( Bottom_Up ) then
                begin
                    Src := Src - ( Byte_Width + Padding ) * 2 ; // Go to start of previous line
                end ;
                I := I - Padding ;
            end ;
            dec( I ) ;
            continue ;
        end ;

        // 1 to 4 byte color....
        C := 0 ;
        move( Buffer[ Src ], C, Bits div 8 ) ;
        if( Bits = 8 ) then // 256 color
        begin
{$R-}
{$WARNINGS OFF}
            _Array[ D ] := integer( PEL[ C ] ) or $FF000000 ;
{$WARNINGS ON}
{$R+}
        end else
        begin
            if( Bits = 16 ) then
            begin
                Dummy := 0 ;
                Temp := C and $1F ;
                C := C shr 5 ;
                if( Temp = $1F ) then
                begin
                    Temp := $FF ;
                end else
                if( Temp = 0 ) then
                begin
                    Temp := 0 ;
                end else
                begin
                    Temp := ( ( ( Temp * 100 ) div 32 ) * 256 ) div 100 ;
                end ;
                Dummy := Dummy or Temp ;
                Temp := C and $1F ;
                C := C shr 5 ;
                if( Temp = $1F ) then
                begin
                    Temp := $FF ;
                end else
                if( Temp = 0 ) then
                begin
                    Temp := 0 ;
                end else
                begin
                    Temp := ( ( ( Temp * 100 ) div 32 ) * 256 ) div 100 ;
                end ;
                Dummy := Dummy or ( Temp shl 8 ) ;
                Temp := C and $1F ;
                C := C shr 5 ;
                if( Temp = $1F ) then
                begin
                    Temp := $FF ;
                end else
                if( Temp = 0 ) then
                begin
                    Temp := 0 ;
                end else
                begin
                    Temp := ( ( ( Temp * 100 ) div 32 ) * 256 ) div 100 ;
                end ;
                C := Dummy or ( Temp shl 16 ) ;
            end ;
{$R-}
{$WARNINGS OFF}
            _Array[ D ] := C or $FF000000 ;
{$WARNINGS ON}
{$R+}
        end ;
        inc( D ) ;
        inc( X ) ;
        inc( Src, Bits div 8 ) ;
        if( X = Width ) then
        begin
            X := 0 ;
            Src := Src + Padding ; // Goto end of line
            if( Bottom_Up ) then
            begin
                Src := Src - ( Byte_Width + Padding ) * 2 ; // Go to start of previous line
            end ;
            I := I - Padding ;
        end ;
        dec( I, Bits div 8 ) ;
    end ; // while( I > 0 )

    // Finish up...
    freemem( Buffer ) ;
    closefile( Fil ) ;
end ;


{ Finds all pixels with RGB matching what is passed, and sets their alpha to
  what is passed. }
procedure TTexture.Set_Alpha( RGB : integer ) ;

var Loop : integer ;

begin
    for Loop := 0 to Width * Height - 1 do // Process all pixels
    begin
{$R-}
        if( ( _Array[ Loop ] and $FFFFFF ) = ( RGB and $FFFFFF ) ) then
{$R+}
        begin
            _Array[ Loop ] := RGB ;
        end ;
    end ;
end ;


procedure TTexture.Set_Buffer( P : pointer ) ;

begin
    if( _Array <> nil ) then
    begin
        freemem( _Array ) ;
    end ;
    _Array := PInteger_Array( P ) ;
end ;


// Stretch the texture to twice its X and Y dimensions.
procedure TTexture.Stretch ;

var Count : integer ;
    Temp : PInteger_Array ;
    Row_D, Column_D, Row_S, Column_S : integer ;
    X : integer ;

begin
    // Safety check...
    Count := Width * Height ; // Number of integers to move
    if( Count * sizeof( integer ) * 4 > Max_Memory ) then
    begin
        exit ; // Array would be too large
    end ;

    // Allocate new memory and do the copy...
    Temp := Allocmem( Count * sizeof( integer ) * 4 ) ;
    Row_D := 0 ;
    for Row_S := 0 to Height - 1 do
    begin
        Column_D := 0 ;
        for Column_S := 0 to Width - 1 do
        begin
            X := _Array[ Row_S * Width + Column_S ] ;
            Temp[ ( Row_D * Width ) shl 1 + Column_D ] := X ;
            Temp[ ( ( Row_D + 1 ) * Width ) shl 1 + Column_D ] := X ;
            inc( Column_D ) ;
            Temp[ ( Row_D * Width ) shl 1 + Column_D ] := X ;
            Temp[ ( ( Row_D + 1 ) * Width ) shl 1 + Column_D ] := X ;
            inc( Column_D ) ;
        end ;
        inc( Row_D, 2 ) ;
    end ;

    // Adjust size...
    Width := Width * 2 ;
    Height := Height * 2 ;
    freemem( _Array ) ;
    _Array := Temp ;
end ;



// TXY_List methods...

// Property handlers

function TXY_List.Get_X( Index : integer ) : TPoint ;

begin
    Result.X := integer( X_List[ Index ] ) ;
    Result.Y := integer( Y_List[ Index ] ) ;
end ;


procedure TXY_List.Set_X( Index : integer ; Value : TPoint ) ;

begin
    X_List[ Index ] := pointer( Value.X ) ;
    Y_List[ Index ] := pointer( Value.Y ) ;
end ;


// API

constructor TXY_List.Create ;

begin
    inherited Create ;

    X_List := TList.Create ;
    Y_List := TList.Create ;
end ;


destructor TXY_List.Destroy ;

begin
    X_List.Free ;
    Y_List.Free ;
    inherited Destroy ;
end ;


function TXY_List.Count : integer ;

begin
    Result := X_List.Count ;
end ;


procedure TXY_List.Clear ;

begin
    // Clear out old points...
    X_List.Clear ;
    Y_List.Clear ;
end ;


procedure TXY_List.Generate( A, B : TPoint ) ;

var C : TPoint ;
    New_X, New_Y : integer ;
    Count : integer ; // Number of steps in dependant access for each step in the independant
    i : integer ; // Loop count
    X_Delta, Y_Delta : integer ;
    Scaled : integer ;

begin
    // Setup...
    X_Delta := abs( A.X - B.X ) ;
    Y_Delta := abs( A.Y - B.Y ) ;
    if( X_Delta > Y_Delta ) then
    begin
        Count := ( Y_Delta shl 16 ) div ( X_Delta + 1 ) ;
        X_List.Capacity := X_List.Capacity + X_Delta + 1 ;
        Y_List.Capacity := X_List.Capacity ;
        if( A.X > B.X ) then // If inverted, swap the points
        begin
            C := B ;
            B := A ;
            A := C ;
        end ;
        if( A.Y > B.Y ) then
        begin
            Count := -Count ;
        end ;
        New_Y := A.Y shl 16 ;//+ 32768 ;
        for i := A.X to B.X do
        begin
            X_List.Add( pointer( i ) ) ;
{$R-}
            Scaled := New_Y shr 16 ;
            if( New_Y < 0 ) then
            begin
{$WARNINGS OFF}
                Scaled := Scaled or $FFFF0000 ; // Sign-extend
{$WARNINGS ON}
            end ;
{$R+}
            Y_List.Add( pointer( Scaled ) ) ;
            New_Y := New_Y + Count ;
        end ;
    end else
    begin
        Count := ( X_Delta shl 16 ) div ( Y_Delta + 1 ) ;
        X_List.Capacity := X_List.Capacity + Y_Delta + 1 ;
        Y_List.Capacity := X_List.Capacity ;
        if( A.Y > B.Y ) then // If inverted, swap the points
        begin
            C := B ;
            B := A ;
            A := C ;
        end ;
        if( A.X > B.X ) then
        begin
            Count := -Count ;
        end ;
        New_X := A.X shl 16 ;//+ 32768 ; // Add 32K (.5) so we are in "middle" of pixel
        for i := A.Y to B.Y do
        begin
            Y_List.Add( pointer( i ) ) ;
{$R-}
            Scaled := New_X shr 16 ;
            if( New_X < 0 ) then
            begin
{$WARNINGS OFF}
                Scaled := Scaled or $FFFF0000 ; // Sign-extend
{$WARNINGS ON}
            end ;
{$R+}
            X_List.Add( pointer( Scaled ) ) ;
            New_X := New_X + Count ;
        end ;
    end ;
end ;



// TPolygon methods...

// Property handlers

function TPolygon.Get_X( Index : integer ) : TPoint ;

begin
    Result.X := integer( Vertex_X_List[ Index ] ) ;
    Result.Y := integer( Vertex_Y_List[ Index ] ) ;
end ;


procedure TPolygon.Set_X( Index : integer ; Value : TPoint ) ;

begin
    Vertex_X_List[ Index ] := pointer( Value.X ) ;
    Vertex_Y_List[ Index ] := pointer( Value.Y ) ;
end ;


// API

constructor TPolygon.Create ;

begin
    inherited Create ;

    Vertex_X_List := TList.Create ;
    Vertex_Y_List := TList.Create ;
    Bitmap_X_List := TList.Create ;
    Bitmap_Y_List := TList.Create ;
end ;


destructor TPolygon.Destroy ;

begin
    Vertex_X_List.Free ;
    Vertex_Y_List.Free ;
    Bitmap_X_List.Free ;
    Bitmap_Y_List.Free ;
    if( Lines_Y <> nil ) then
    begin
        Lines_Left_X.Free ;
        Lines_Y.Free ;
        Lines_Right_X.Free ;
    end ;
    if( Lines_Bitmap_X_Right <> nil ) then
    begin
        Lines_Bitmap_X_Right.Free ;
        Lines_Bitmap_Y_Right.Free ;
        Lines_Bitmap_X_Left.Free ;
        Lines_Bitmap_Y_Left.Free ;
    end ;

    inherited Destroy ;
end ;


procedure TPolygon.Add( X : TPoint ) ;

begin
    Vertex_X_List.Add( pointer( X.X ) ) ;
    Vertex_Y_List.Add( pointer( X.Y ) ) ;
end ;


procedure TPolygon.Clear ;

begin
    Vertex_X_List.Clear ;
    Vertex_Y_List.Clear ;
end ;


procedure TPolygon.Generate ;

var Loop : integer ;

begin
    X_List.Clear ;
    Y_List.Clear ;

    if( Vertex_Count < 2 ) then
    begin
        exit ;
    end ;

    for Loop := 0 to Vertex_Count - 2 do
    begin
        inherited Generate( Vertex[ Loop ], Vertex[ Loop + 1 ] ) ;
    end ;
    if( Vertex_Count > 2 ) then
    begin
        inherited Generate( Vertex[ Vertex_Count - 1 ], Vertex[ 0 ] ) ;
    end ;
end ;


{ Generate a line of points with matching texture X,Y positions. Bitmap is the
  size of the side of the texture we are traversing.  Other is a constant.  B_X
  is true if we are traversing the X-axis of the bitmap.  If Bitmap is negative,
  we are traversing from the lower right, otherwise from the upper left.  If
  B_X is true, Bitmap is the Width otherwise it is the Height of the texture.
  Other is the maximum of the other dimension (0, Height, or Width).  A and B
  are the end points for this line in the polygon. }
procedure TPolygon.Generate_Texture_Line( Bitmap, Other : integer ;
    B_X : boolean ; A, B : TPoint ) ;

var C : TPoint ;
    New_X, New_Y : integer ;
    Count : integer ; // Number of steps in dependant access for each step in the independant
    Bit : integer ; // Current bitmap position
    B_Count : integer ; // Number of steps across passed bitmap range
    i : integer ; // Loop count
    X_Delta, Y_Delta : integer ;
    Scaled : integer ; // Scaled value

begin
    // Setup...
    X_Delta := abs( A.X - B.X ) ;
    Y_Delta := abs( A.Y - B.Y ) ;
    if( Bitmap >= 0 ) then
    begin
        Bit := 0 ;
    end else
    begin
        Bit := ( -Bitmap ) shl 16 ;
    end ;

    // Calculate the points
    if( X_Delta > Y_Delta ) then
    begin
        B_Count := ( abs( Bitmap ) shl 16 ) div X_Delta ;
        if( Bitmap < 0 ) then // Counting backwards
        begin
            B_Count := -B_Count ;
        end ;
        Count := ( Y_Delta shl 16 ) div X_Delta  ;
        X_List.Capacity := X_List.Capacity + X_Delta + 1 ;
        Y_List.Capacity := X_List.Capacity ;
        Bitmap_X_List.Capacity := X_List.Capacity ;
        Bitmap_Y_List.Capacity := X_List.Capacity ;
        if( A.X > B.X ) then // If inverted, swap the points
        begin
            C := B ;
            B := A ;
            A := C ;
            if( Bit = 0 ) then
            begin
                Bit := ( abs( Bitmap ) - 1 ) shl 16 ;
            end else
            begin
                Bit := 0 ;
            end ;
            B_Count := -B_Count ;
        end ;
        if( A.Y > B.Y ) then
        begin
            Count := -Count ;
        end ;
        New_Y := A.Y shl 16 ;
        for i := A.X to B.X do
        begin
            X_List.Add( pointer( i ) ) ;
{$R-}
            Scaled := New_Y shr 16 ;
            if( New_Y < 0 ) then
            begin
{$WARNINGS OFF}
                Scaled := Scaled or $FFFF0000 ; // Sign-extend
{$WARNINGS ON}
            end ;
{$R+}
            Y_List.Add( pointer( Scaled ) ) ;
            New_Y := New_Y + Count ;
            if( B_X ) then // Stepping across the X-axis of the bitmap
            begin
                Bitmap_X_List.Add( pointer( Bit shr 16 ) ) ;
                Bitmap_Y_List.Add( pointer( Other ) ) ;
            end else
            begin
                Bitmap_X_List.Add( pointer( Other ) ) ;
                Bitmap_Y_List.Add( pointer( Bit shr 16 ) ) ;
            end ;
            Bit := Bit + B_Count ;
            if( Bit < 0 ) then
            begin
                Bit := 0 ;
            end ;
        end ; // for i := A.X to B.X
    end else
    begin
        B_Count := ( abs( Bitmap ) shl 16 ) div Y_Delta ;
        if( Bitmap < 0 ) then // Counting backwards
        begin
            B_Count := -B_Count ;
        end ;
        Count := ( X_Delta shl 16 ) div Y_Delta ;
        X_List.Capacity := X_List.Capacity + Y_Delta + 1 ;
        Y_List.Capacity := X_List.Capacity ;
        Bitmap_X_List.Capacity := X_List.Capacity ;
        Bitmap_Y_List.Capacity := X_List.Capacity ;
        if( A.Y > B.Y ) then // If inverted, swap the points
        begin
            C := B ;
            B := A ;
            A := C ;
            if( Bit = 0 ) then
            begin
                Bit := ( abs( Bitmap ) - 1 ) shl 16 ;
            end else
            begin
                Bit := 0 ;
            end ;
            B_Count := -B_Count ;
        end ;
        if( A.X > B.X ) then
        begin
            Count := -Count ;
        end ;
        New_X := A.X shl 16 ;
        for i := A.Y to B.Y do
        begin
            Y_List.Add( pointer( i ) ) ;
{$R-}
            Scaled := New_X shr 16 ;
            if( New_X < 0 ) then
            begin
{$WARNINGS OFF}
                Scaled := Scaled or $FFFF0000 ; // Sign-extend
{$WARNINGS ON}
            end ;
{$R+}
            X_List.Add( pointer( Scaled ) ) ;
            New_X := New_X + Count ;
            if( B_X ) then // Stepping across the X-axis of the bitmap
            begin
                Bitmap_X_List.Add( pointer( Bit shr 16 ) ) ;
                Bitmap_Y_List.Add( pointer( Other ) ) ;
            end else
            begin
                Bitmap_X_List.Add( pointer( Other ) ) ;
                Bitmap_Y_List.Add( pointer( Bit shr 16 ) ) ;
            end ;
            Bit := Bit + B_Count ;
            if( Bit < 0 ) then
            begin
                Bit := 0 ;
            end ;
        end ; // for i := A.Y to B.Y
    end ; // if( X_Delta > Y_Delta )
end ; // TPolygon.Generate_Texture_Line



procedure TPolygon.Generate_With_Texture( Bitmap : TTexture ) ;

var Loop : integer ;

begin
    X_List.Clear ;
    Y_List.Clear ;

    if( Vertex_Count < 2 ) then
    begin
        exit ;
    end ;

    // Do upper-left to upper-right (top line)
    Generate_Texture_Line( Bitmap.Width, 0, True, Vertex[ 0 ], Vertex[ 1 ] ) ;
    if( Vertex_Count > 2 ) then // Not just a line
    begin
        // Do right-hand side
        Generate_Texture_Line( Bitmap.Height, Bitmap.Width - 1, False, Vertex[ 1 ], Vertex[ 2 ] ) ;
        if( Vertex_Count > 3 ) then // More sides than a triangle
        begin
            if( Vertex_Count > 4 ) then // More sides than a rectangle
            begin
                for Loop := 4 to Vertex_Count - 2 do
                begin
                    Generate_Texture_Line( Bitmap.Height, Bitmap.Width - 1, False, Vertex[ Loop ], Vertex[ Loop + 1 ] ) ;
                end ; // for Loop := 0 to Vertex_Count - 2
            end else
            begin
                // Do bottom of rectangle
                Generate_Texture_Line( -Bitmap.Width, Bitmap.Height - 1, True, Vertex[ 2 ], Vertex[ 3 ] ) ;
                // Do left-hand side of rectangle
                Generate_Texture_Line( -Bitmap.Height, 0, False, Vertex[ 3 ], Vertex[ 0 ] ) ;
            end ;
        end else
        begin
            // Do "left" side of triangle
            Generate_Texture_Line( -Bitmap.Height, 0, False, Vertex[ 2 ], Vertex[ 0 ] ) ;
        end ; // if( Vertex_Count > 2 )
    end ; // if( Vertex_Count > 2 )
end ; // TPolygon.Generate_With_Texture


function TPolygon.Line_Count : integer ;

begin
    Result := Lines_Left_X.Count ;
end ;


function TPolygon.Line_Left( Index : integer ) : TPoint ;

begin
    Result.X := integer( Lines_Left_X.List[ Index ] ) ;
    Result.Y := integer( Lines_Y.List[ Index ] ) ;
end ;


function TPolygon.Line_Right( Index : integer ) : TPoint ;

begin
    Result.X := integer( Lines_Right_X.List[ Index ] ) ;
    Result.Y := integer( Lines_Y.List[ Index ] ) ;
end ;


function TPolygon.Line_Left_Texture( Index : integer ) : TPoint ;

begin
    Result.X := integer( Lines_Bitmap_X_Left.List[ Index ] ) ;
    Result.Y := integer( Lines_Bitmap_Y_Left.List[ Index ] ) ;
end ;


function TPolygon.Line_Right_Texture( Index : integer ) : TPoint ;

begin
    Result.X := integer( Lines_Bitmap_X_Right.List[ Index ] ) ;
    Result.Y := integer( Lines_Bitmap_Y_Right.List[ Index ] ) ;
end ;


procedure TPolygon.MinMax ;

var Lines_Valid : TList ;
    Loop, Max_Y, Min_Y, This_Y : integer ;
    P : TPoint ;

begin
    // Initialize the matched X,Y lists
    if( Lines_Left_X = nil ) then
    begin
	Lines_Left_X := TList.Create ;
	Lines_Y := TList.Create ;
	Lines_Right_X := TList.Create ;
        if( Bitmap_X_List.Count > 0 ) then
        begin
            Lines_Bitmap_X_Right := TList.Create ;
            Lines_Bitmap_Y_Right := TList.Create ;
            Lines_Bitmap_X_Left := TList.Create ;
            Lines_Bitmap_Y_Left := TList.Create ;
        end ;
    end else
    begin
	Lines_Left_X.Clear ;
	Lines_Y.Clear ;
	Lines_Right_X.Clear ;
        if( Bitmap_X_List.Count > 0 ) then
        begin
            Lines_Bitmap_X_Right.Clear ;
            Lines_Bitmap_Y_Right.Clear ;
            Lines_Bitmap_X_Left.Clear ;
            Lines_Bitmap_Y_Left.Clear ;
        end ;
    end ;

    // Determine minimum and maximum Y values...
    Min_Y := 32767 ;
    Max_Y := 0 ;
    for Loop := 0 to Vertex_Count - 1 do
    begin
	This_Y := Vertex[ Loop ].Y ;
	if( This_Y < Min_Y ) then
	begin
	    Min_Y := This_Y ;
	end ;
	if( This_Y > Max_Y ) then
	begin
	    Max_Y := This_Y ;
	end ;
    end ;

    // Initialize line lists...
    Lines_Left_X.Capacity := Max_Y - Min_Y + 1 ;
    Lines_Y.Capacity := Lines_Left_X.Capacity ;
    Lines_Right_X.Capacity := Lines_Left_X.Capacity ;
    if( Bitmap_X_List.Count > 0 ) then
    begin
        Lines_Bitmap_X_Right.Capacity := Lines_Left_X.Capacity ;
        Lines_Bitmap_Y_Right.Capacity := Lines_Left_X.Capacity ;
        Lines_Bitmap_X_Left.Capacity := Lines_Left_X.Capacity ;
        Lines_Bitmap_Y_Left.Capacity := Lines_Left_X.Capacity ;
    end ;
    Lines_Valid := TList.Create ;
    for Loop := 0 to Max_Y - Min_Y do
    begin
	Lines_Left_X.Add( pointer( 32767 ) ) ;
	Lines_Y.Add( pointer( Loop + Min_Y ) ) ;
	Lines_Right_X.Add( nil ) ;
        Lines_Valid.Add( nil ) ;
        if( Bitmap_X_List.Count > 0 ) then
        begin
            Lines_Bitmap_X_Left.Add( nil ) ;
            Lines_Bitmap_Y_Left.Add( nil ) ;
            Lines_Bitmap_X_Right.Add( nil ) ;
            Lines_Bitmap_Y_Right.Add( nil ) ;
        end ;
    end ;

    // Now build list
    for Loop := 0 to Count - 1 do
    begin
	P := X[ Loop ] ;
	This_Y := P.Y - Min_Y ;
        Lines_Valid.List[ This_Y ] := pointer( -1 ) ;
	if( integer( Lines_Left_X.List[ This_Y ] ) > P.X ) then
	begin
	    Lines_Left_X.List[ This_Y ] := pointer( P.X ) ;
            if( Bitmap_X_List.Count > 0 ) then
            begin
                Lines_Bitmap_X_Left.List[ This_Y ] := Bitmap_X_List.List[ Loop ] ;
                Lines_Bitmap_Y_Left.List[ This_Y ] := Bitmap_Y_List.List[ Loop ] ;
            end ;
	end ;
	if( integer( Lines_Right_X.List[ This_Y ] ) < P.X ) then
	begin
	    Lines_Right_X.List[ This_Y ] := pointer( P.X ) ;
            if( Bitmap_X_List.Count > 0 ) then
            begin
                Lines_Bitmap_X_Right.List[ This_Y ] := Bitmap_X_List.List[ Loop ] ;
                Lines_Bitmap_Y_Right.List[ This_Y ] := Bitmap_Y_List.List[ Loop ] ;
            end ;
	end ;
    end ;

    { Finally, make sure each line is a valid line.  Sometimes the line drawing
      routine does not hit exactly on a vertex pixel, due to rounding.  This
      loop makes sure that the lines are all valid. }
    Loop := 0 ;
    while( Loop < Lines_Left_X.Count ) do
    begin
        if( Lines_Valid.List[ Loop ] = nil ) then // An invalid line
        begin
            Lines_Left_X.Delete( Loop ) ;
            Lines_Right_X.Delete( Loop ) ;
	    Lines_Y.Delete( Loop ) ;
            Lines_Valid.Delete( Loop ) ;
            if( Bitmap_X_List.Count > 0 ) then
            begin
                Lines_Bitmap_X_Left.Delete( Loop ) ;
                Lines_Bitmap_Y_Left.Delete( Loop ) ;
                Lines_Bitmap_X_Right.Delete( Loop ) ;
                Lines_Bitmap_Y_Right.Delete( Loop ) ;
            end ;
        end else
        begin
            inc( Loop ) ;
        end ;
    end ;
end ; // TPolygon.MinMax


function TPolygon.Vertex_Count : integer ;

begin
    Result := Vertex_X_List.Count ;
end ;



end.
