{
              Program Name : R3D
              Package Name : LB
              Purpose      : Three-D rendering package
              Institution  : Conroy & Conroy Company
              Original Vers: 01-Dec-1996
              Date Written : 01-Dec-1996
              Written By   : Alan Conroy
              Version      : 2.0

              TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *********************************************************
        *                                                       *
        *        M O D I F I C A T I O N   H I S T O R Y        *
        *                                                       *
        *     DATE      BY             REASON                   *
        *                                                       *
        *********************************************************

        *********************************************************
        *                                                       *
        *           P R O G R A M   P U R P O S E               *
        *                                                       *
        *********************************************************

          This unit provides 3-D rendering.
}

unit R3D ;

interface

uses // Borland...
     Windows,   // TRect

     Classes,   // TList
     Graphics,  // TColor

     // C&C...
     R2D,
     TypeDefs ; // Max_Memory

const Rad : single = pi / 180.0 ;

type // Cubes
     TCube = record
                 Low_X, Low_Y, Low_Z : integer ;
                 High_X, High_Y, High_Z : integer ;
             end ;

     // Vertecies...
     TVertex = record
                   X, Y, Z : single ; // Vertex model coordinates
                   _X, _Y, _Z : single ; // Rotated model coordinates (or yaw, pitch, and roll for camera)
                   Vertex : integer ; // Vertex to connect to
                   Equivalent_Object : integer ; // Equivalent object
                   Equivalent_Vertex : integer ; // Equivalent Vertex
               end ;
     PVertex = ^TVertex ;
     TVertex_Array = array[ 0..Max_Memory div sizeof( TVertex ) - 1 ] of TVertex ;
     PVertex_Array = ^TVertex_Array ;

     // 2-D objects...
     T_Object = record
                    Average_Z : integer ; // Averaged Z-depth of object
                    Color : TColor ; // Color of object
                    Defined : boolean ; // True if object has been defined
                    Vertex_Count : integer ; // Number of vertecies
                    Vertecies : PVertex_Array ; // Array of vertecies
                    Texture : TTexture ; // Texture to use (use Color if nil)
                    Drawn : boolean ; // Internal use only (for drawing)
                end ;
     P_Object = ^T_Object ;
     T_Object_Array = array[ 0..Max_Memory div sizeof( T_Object ) - 1 ] of T_Object ;
     P_Object_Array = ^T_Object_Array ;

     // 3-D objects...
     TSolid = class
                  private // Instance data...
                      _Bounding_Cube : TCube ; // Rotated bounding cube
                      _Object_Count : integer ;
                      _Objects : P_Object_Array ;
                      _X, _Y, _Z : integer ; // Rotated coordinates of center position
                      Points : PPoint_Array ;
                      Point_Count : integer ; // Number of point in Points[]
                      X_Rotation, Y_Rotation, Z_Rotation : single ; // Current self-relative rotation

                  public // Public data...
                      X, Y, Z : integer ; // World coordinates of center position

                  public // Constructors and destructors...
                      // Construct the object.
                      constructor Create( Count : integer ) ;

                      // Destruct the object.
                      destructor Destroy ; override ;

                  private // Internal utility routines...
                      procedure Optimize ;
                      // Optimizes vertecies

                  protected // Property hanlders...
                      function Get_Objects( Index : integer ) : T_Object ;

                      procedure Set_Objects( Index : integer ; Value : T_Object ) ;

                  public // API...
                      procedure Adjust_Distance( X, Y, Z : single ) ;
                      // Adjust object vertecies for viewing from specified position.

                      // Bounding rectangle of current rotation.
                      function Bounding_Rect : TRect ;

                      function Bounds_Check_X( Low, High : integer ) : boolean ;
                      // Return True if any rotated X Vertex exceeds specified bounds.

                      function Bounds_Check_Y( Low, High : integer ) : boolean ;
                      // Return True if any rotated Y Vertex exceeds specified bounds.

                      procedure Draw( C : TTexture ; Camera_Z : integer ;
                          Display_Buffer : TTexture ;
                          Pixel_Map : PInteger_Array ) ;
                      // Draw rotated object in buffer.

                      property Object_Count : integer
                          read _Object_Count ;

                      property Objects[ Index : integer ] : T_Object
                          read Get_Objects
                          write Set_Objects ;

                      procedure Rotate( XRotate, YRotate, ZRotate : single ) ;
                      // Rotate object by specirfied X, Y, Z angles

                      procedure Set_Rotation( XRotate, YRotate, ZRotate : single ) ;
                      // Set object's self-relative rotation

              end ;
     PSolid = ^TSolid ;

     T3D_World = class
                     private // Instance data...
                         Camera : TVertex ;
                         Display_Bitmap : HBitmap ;
                         Display_Bits : pointer ;
                         Display_Buffer : TTexture ;
                         Display_Buffer_DC : HDC ;
                         _Inhibit_Display : boolean ;
                         Pixel_Map : pInteger_Array ;
                         _Objects : TList ;
                         Update_Rect : TRect ;

                     public // Constructors and destructors...
                         destructor Destroy ; override ;

                     protected // Property handlers...
                         function Get_Map_Pixels : boolean ;
                         function Get_Object( Index : integer ) : TObject ;

                         procedure Set_Map_Pixels( Value : boolean ) ;
                         procedure Set_Object( Index : integer ; Value : TObject ) ;

                     protected // Internal utility routines...
                         procedure Prepare_Display( Height, Width : integer ) ;

                     public // API...
                         function Add_Object( X : TSolid ) : integer ;
                         // Add a new object to the world.  Returns index of it.

                         procedure Adjust_Distance( _Object : TSolid ) ;
                         // Adjust object vertecies based on camera distance

                         procedure Clear ; // Delete all objects

                         procedure Erase_Buffer ;
                         // Erases display buffer.

                         function Get_Camera_Position : TVertex ;
                         // Return camera position.

                         property Inhibit_Display : boolean
                             read _Inhibit_Display
                             write _Inhibit_Display ;

                         property Map_Pixels : boolean
                             read Get_Map_Pixels
                             write Set_Map_Pixels ;

                         function Object_At( X, Y : integer ;
                             var Surface : integer ) : TSolid ;

                         property Objects[ Index : integer ] : TObject
                             read Get_Object
                             write Set_Object ;

                         procedure Render( DC : HDC ) ;
                         // Render current camera view onto device context.

                         procedure Rotate( XRotate, YRotate, ZRotate : single ) ;
                         // Rotate objects.

                         procedure Set_Camera_Position( V : TVertex ) ;
                         // Sets camera position.
                 end ;


type TDisplay_Mode = ( Display_Full, Display_Lines, Display_Points, Display_Outline ) ;

var BM : TBitmapInfo ;
    Distance_Factor : integer ;
    Global_Display_Mode : TDisplay_Mode ;

function Create_Object( Color : TColor ; Count : integer ) : T_Object ;
function Cube( LX, LY, LZ, HX, HY, HZ : integer ) : TCube ;
// Returns a cube record

function Merge_Rects( Rect1, Rect2 : TRect ) : TRect ;
// Merge two TRects into one which contains both.

function Vertex( X, Y, Z : single ; V : integer ) : TVertex ;

implementation

uses // Borland...
     SysUtils ; // Exception

function tan( X : single ) : single ;

begin
    tan := sin( X ) / cos( X ) ;
end ;


function Cube( LX, LY, LZ, HX, HY, HZ : integer ) : TCube ;
// Returns a cube record

begin
    Result.Low_X := LX ;
    Result.Low_Y := LY ;
    Result.Low_Z := LZ ;
    Result.High_X := HX ;
    Result.High_Y := HY ;
    Result.High_Z := HZ ;
end ;


function Rect( L, T, R, B : integer ) : TRect ;

begin
    Result.Left := L ;
    Result.Top := T ;
    Result.Right := R ;
    Result.Bottom := B ;
end ;


function Create_Object( Color : TColor ; Count : integer ) : T_Object ;

begin
    if( Count < 1 ) then
    begin
        raise Exception.Create( 'Objects require a minimum of 1 Vertex' ) ;
    end ;
    Result.Texture := nil ;
    Result.Color := Color ;
    Result.Vertex_Count := Count ;
    Result.vertecies := Allocmem( Count * sizeof( TVertex ) ) ;
    fillchar( Result.Vertecies^, Count * sizeof( TVertex ), 0 ) ;
    Result.Defined := True ;
end ;


function Merge_Rects( Rect1, Rect2 : TRect ) : TRect ;
// Merge two TRects into one which contains both.

begin
    Result := Rect1 ;
    if( Rect2.Left < Rect1.Left ) then
    begin
        Result.Left := Rect2.Left ;
    end ;
    if( Rect2.Top < Rect1.Top ) then
    begin
        Result.Top := Rect2.Top ;
    end ;
    if( Rect2.Right > Rect1.Right ) then
    begin
        Result.Right := Rect2.Right ;
    end ;
    if( Rect2.Bottom > Rect1.Bottom ) then
    begin
        Result.Bottom := Rect2.Bottom ;
    end ;
end ;


function Vertex( X, Y, Z : single ; V : integer ) : TVertex ;

begin
    Result.x := X ;
    Result.y := Y ;
    Result.z := Z ;
    Result.Vertex := V ;
    Result._X := X ;
    Result._Y := Y ;
    Result._Z := Z ;
end ;


// TSolid methods...

// Constructors and destructors...

// Construct the object.
constructor TSolid.Create( Count : integer ) ;

begin
    if( Count < 1 ) then
    begin
        raise Exception.Create( 'Solids require a minimum of 1 object' ) ;
    end ;
    _Object_Count := Count ;
    _Objects := Allocmem( Count * sizeof( T_Object ) ) ;
    fillchar( _Objects^, Count * sizeof( T_Object ), 0 ) ;
end ;


// Destruct the object.
destructor TSolid.Destroy ;

var Loop : integer ;

begin
    for Loop := 0 to _Object_Count - 1 do
    begin
        freemem( _Objects^[ Loop ].Vertecies ) ;
    end ;
    freemem( _Objects ) ;

    inherited Destroy ;
end ;


// Internal utility routines...

// Optimizes vertecies by setting equivelance for vertecies which occupy the same points
procedure TSolid.Optimize ;

var O : integer ; // Object being checked
    O1 : integer ; // Objects being compared against
    V : integer ; // Vertex being checked
    V1 : integer ; // vertecies being compared against
    P : integer ; // Count of verticies

begin
    P := 0 ;
    for O := 0 to _Object_Count - 1 do
    begin
        if( _Objects^[ O ].Defined ) then
        begin
            P := P + _Objects^[ O ].Vertex_Count ;
            for V := 0 to _Objects^[ O ].Vertex_Count - 1 do
            begin
                _Objects^[ O ].vertecies^[ V ].Equivalent_Object := O ;
                _Objects^[ O ].vertecies^[ V ].Equivalent_Vertex := V ;
                for O1 := 0 to O - 1 do
                begin
                    if( _Objects^[ O1 ].Defined ) then
                    begin
                        for V1 := 0 to _Objects^[ O1 ].Vertex_Count - 1 do
                        begin
                            if(
                                ( _Objects^[ O1 ].vertecies^[ V1 ].X = _Objects^[ O ].vertecies^[ V ].X )
                                and
                                ( _Objects^[ O1 ].vertecies^[ V1 ].Y = _Objects^[ O ].vertecies^[ V ].Y )
                                and
                                ( _Objects^[ O1 ].vertecies^[ V1 ].Z = _Objects^[ O ].vertecies^[ V ].Z )
                              ) then
                            begin
                                _Objects^[ O ].vertecies^[ V ].Equivalent_Object := O1 ;
                                _Objects^[ O ].vertecies^[ V ].Equivalent_Vertex := V1 ;
                            end ;
                        end ; // for V1 := 0 to _Objects^[ O1 ].Vertex_Count - 1
                    end ; // if( _Objects^[ O1 ].Defined )
                end ; // for O1 := 0 to O - 1
            end ;
        end ;
    end ;

    // Make sure we have a large-enough point array
    if( Point_Count < P + 1 ) then
    begin
        if( Points <> nil ) then // Old buffer needs to be deallocated
        begin
            freemem( Points ) ;
        end ;
        Points := Allocmem( ( P + 1 ) * sizeof( TPoint ) ) ; // Create new one
        Point_Count := P ;
    end ;
end ; // TSolid.Optimize


// Property handlers...

function TSolid.Get_Objects( Index : integer ) : T_Object ;

begin
    Get_Objects := _Objects^[ Index ] ;
end ;


procedure TSolid.Set_Objects( Index : integer ; Value : T_Object ) ;

begin
    _Objects^[ Index ] := Value ;
    Optimize ;
end ;


// Public API...

procedure TSolid.Adjust_Distance( X, Y, Z : single ) ;
// Adjust object vertecies for viewing from specified Z.

var Distance : single ;
    Dummy, Dummy1 : integer ;
    _Object : T_Object ;
    Other_Object, Other_Vertex : integer ;
    _X_, _Y_, _Z_ : integer ;
    X3, Y3, Z3 : single ;

begin
    // Invalidate bounding cube
    _Bounding_Cube := Cube( 32767, 32767, 32767, 0, 0, 0 ) ;

    // Process all objects...
    for Dummy1 := 0 to Object_Count - 1 do
    begin
        _Object := _Objects[ Dummy1 ] ;
        _Object.Average_Z := 0 ;
        for Dummy := 0 to _Object.Vertex_Count - 1 do
        begin
            Other_Object := _Object.Vertecies[ Dummy ].Equivalent_Object ;
            Other_Vertex := _Object.Vertecies[ Dummy ].Equivalent_Vertex ;
            if(
                ( Other_Object <> Dummy1 )
                or
                ( Other_Vertex <> Dummy )
              ) then // This Vertex is equivalent to an earlier one
            begin
                _Object.Vertecies[ Dummy ]._X :=
                    _Objects[ Other_Object ].vertecies[ Other_Vertex ]._X ;
                _Object.Vertecies[ Dummy ]._Y :=
                    _Objects[ Other_Object ].vertecies[ Other_Vertex ]._Y ;
                _Object.Vertecies[ Dummy ]._Z :=
                    _Objects[ Other_Object ].vertecies[ Other_Vertex ]._Z ;
                _Object.Average_Z := _Object.Average_Z +
                    trunc( _Object.Vertecies[ Dummy ]._Z ) ;
                continue ; // Skip calculation
            end ;

            // Get vertecies...
            X3 := _Object.Vertecies[ Dummy ]._X ;
            Y3 := _Object.Vertecies[ Dummy ]._Y ;
            Z3 := _Object.Vertecies[ Dummy ]._Z ;

            // Adjust for distance...
            Distance := SQRT( ( Z - Z3 ) * ( Z - Z3 ) + ( Y - Y3 ) * ( Y - Y3 ) + ( X - X3 ) * ( X - X3 ) ) ;
            if( Distance <> 0 ) then
            begin
                Distance := Distance / Distance_Factor ; // Convert to proper units

                if( Distance > 0 ) then
                begin
                    //Distance := SQR( Distance ) ;
                end ;

                X3 := X3 / Distance ;
                Y3 := Y3 / Distance ;
                Z3 := Z3 / Distance ;
            end ; // if( Distance <> 0 )

            // Adjust the vertecies...
            _Object.Vertecies[ Dummy ]._X := X3 ;
            _Object.Vertecies[ Dummy ]._Y := Y3 ;
            _Object.Vertecies[ Dummy ]._Z := Z3 ;

            // Update bounding cube...
            _X_ := trunc( X3 ) ;
            _Y_ := trunc( Y3 ) ;
            _Z_ := trunc( Z3 ) ;
            _Object.Average_Z := _Object.Average_Z + _Z_ ;
            if( _X_ < _Bounding_Cube.Low_X ) then
            begin
                _Bounding_Cube.Low_X := _X_ ;
            end ;
            if( _Y_ < _Bounding_Cube.Low_Y ) then
            begin
                _Bounding_Cube.Low_Y := _Y_ ;
            end ;
            if( _X_ > _Bounding_Cube.High_X ) then
            begin
                _Bounding_Cube.High_X := _X_ ;
            end ;
            if( _Y_ > _Bounding_Cube.High_Y ) then
            begin
                _Bounding_Cube.High_Y := _Y_ ;
            end ;
            if( _Z_ < _Bounding_Cube.Low_Z ) then
            begin
                _Bounding_Cube.Low_Z := _Z_ ;
            end ;
            if( _Z_ > _Bounding_Cube.High_Z ) then
            begin
                _Bounding_Cube.High_Z := _Z_ ;
            end ;
        end ; // for Dummy := 0 to _Object.Vertex_Count - 1
        _Object.Average_Z := _Object.Average_Z div
            _Object.Vertex_Count ;
    end ; // for Dummy1 := 0 to Object_Count - 1
end ; // TSolid.Adjust_Distance


// Bounding rectangle of current rotation.
function TSolid.Bounding_Rect : TRect ;

begin
    Result.Top := _Bounding_Cube.Low_Y + Y ;
    Result.Bottom := _Bounding_Cube.High_Y + Y ;
    Result.Left := _Bounding_Cube.Low_X + X ;
    Result.Right := _Bounding_Cube.High_X + X ;
end ; // TSolid.Bounding_Rect


function TSolid.Bounds_Check_X( Low, High : integer ) : boolean ;
// Return True if any rotated X Vertex exceeds specified bounds.

begin
    if(
        ( _Bounding_Cube.Low_X + X < Low )
        or
        ( _Bounding_Cube.High_X + X > High )
      ) then
    begin
        Bounds_Check_X := True ;
    end else
    begin
        Bounds_Check_X := False ;
    end ;
end ;


function TSolid.Bounds_Check_Y( Low, High : integer ) : boolean ;
// Return True if any rotated Y Vertex exceeds specified bounds.

begin
    if(
        ( _Bounding_Cube.Low_Y + Y < Low )
        or
        ( _Bounding_Cube.High_Y + Y > High )
      ) then
    begin
        Bounds_Check_Y := True ;
    end else
    begin
        Bounds_Check_Y := False ;
    end ;
end ;


function Clockwise( A, B, C : TVertex ) : boolean ;

begin
    Result := True ; // For now, draw all faces
end ; // Clockwise


procedure TSolid.Draw( C : TTexture ; Camera_Z : integer ;
    Display_Buffer : TTexture ; Pixel_Map : PInteger_Array ) ;
    // Draw rotated object in buffer.

var Color : TColor ;
    Temp : integer ;
    This_Object, This_Vertex : integer ; // Loop controls
    _Object : T_Object ; // Current object
    O, O1 : integer ;
    Visible : boolean ;

begin
    // Handle full-mode
    if(
        ( Global_Display_Mode = Display_Full )
        or
        ( Global_Display_Mode = Display_Outline )
      ) then
    begin
        // Clear object drawn flags
        for This_Object := 0 to Object_Count - 1 do
        begin
            _Objects[ This_Object ].Drawn := False ;
        end ;

        for This_Object := 0 to Object_Count - 1 do // Do all objects
        begin
            O := -1 ;
            for O1 := 0 to Object_Count - 1 do // Find next farthest back object
            begin
                if( not _Objects^[ O1 ].Drawn ) then
                begin
                    if(
                        ( O = -1 )
                        or
                        ( _Objects^[ O ].Average_Z < _Objects^[ O1 ].Average_Z )
                      ) then
                    begin
                        O := O1 ;
                    end ;
                end ;
            end ;
            _Object := _Objects^[ O ] ;
            _Objects[ O ].Drawn := True ;

            // Determine if object is at least partially in front of our field of view
            Visible := False ;
            for This_Vertex := 0 to _Object.Vertex_Count - 1 do
            begin
                if( _Object.Vertecies^[ This_Vertex ]._Z - Camera_Z >= 0 ) then
                begin
                    Visible := True ;
                    break ;
                end ;
            end ;

            if( Visible and ( _Object.Vertex_Count > 2 ) ) then // Cull back faces
            begin
                if( not Clockwise( _Object.Vertecies^[ 0 ], _Object.Vertecies^[ 1 ], _Object.Vertecies^[ 2 ] ) ) then
                begin
                    Visible := False ;
                end ;
            end ;

            // Draw the object
            if( Visible ) then
            begin
                for This_Vertex := 0 to _Object.Vertex_Count - 1 do
                begin
                    Points[ This_Vertex ].X := trunc( _Object.Vertecies^[ This_Vertex ]._X ) +
                        _X + Display_Buffer.Width div 2 ;
                    Points[ This_Vertex ].Y := trunc( _Object.Vertecies^[ This_Vertex ]._Y ) +
                        _Y + Display_Buffer.Height div 2 ;
                end ;
                if( _Object.Texture <> nil ) then
                begin
                    Draw_Polygon( Display_Buffer, _Object.Texture, Points,
                        _Object.Vertex_Count, Pixel_Map, integer( self ) ) ;
                end else
                begin
                    Filled_Polygon( Display_Buffer, Points,
                        _Object.Vertex_Count, _Object.Color, Pixel_Map,
                        integer( self ) ) ;
                end ;
                if( Global_Display_Mode = Display_Outline ) then
                begin
                    for This_Vertex := 0 to _Object.Vertex_Count - 1 do
                    begin
                        Points[ This_Vertex ].X := trunc( _Object.Vertecies^[ This_Vertex ]._X ) +
                            _X + Display_Buffer.Width div 2 ;
                        Points[ This_Vertex ].Y := trunc( _Object.Vertecies^[ This_Vertex ]._Y ) +
                            _Y + Display_Buffer.Height div 2 ;
                    end ;
                    if( _Object.Color = clBlack ) then
                    begin
                        Color := clWhite ;
                    end else
                    begin
                        Color := clBlack ;
                    end ;
                    Polygon( Display_Buffer, Points, _Object.Vertex_Count, Color ) ;
                end ;
            end ;
        end ;

        exit ;
    end ; // if( Global_Display_Mode = Display_Full )

    for This_Object := 0 to Object_Count - 1 do
    begin
        // Setup...
        _Object := _Objects[ This_Object ] ;

        // Handle point/line modes
        if( Global_Display_Mode = Display_Points ) then
        begin
            for This_Vertex := 0 to _Object.Vertex_Count - 1 do
            begin
                Temp := trunc( _Object.Vertecies^[ This_Vertex ]._X ) + _X + Display_Buffer.Width div 2 +
                    ( trunc( _Object.Vertecies^[ This_Vertex ]._Y ) + _Y + Display_Buffer.Height div 2 ) * Display_Buffer.Width ;
                Display_Buffer._Array[ Temp ] := integer( _Object.Color ) ;
                if( Pixel_Map <> nil ) then
                begin
                    Pixel_Map^[ Temp ] := integer( self ) ;
                end ;
            end ;
        end else
        begin
            for This_Vertex := 0 to _Object.Vertex_Count - 1 do
            begin
                Points[ This_Vertex ].X := trunc( _Object.Vertecies^[ This_Vertex ]._X ) +
                    _X + Display_Buffer.Width div 2 ;
                Points[ This_Vertex ].Y := trunc( _Object.Vertecies^[ This_Vertex ]._Y ) +
                    _Y + Display_Buffer.Height div 2 ;
            end ;
            //TODO: Set pixel map
            Polygon( Display_Buffer, Points, _Object.Vertex_Count, _Object.Color ) ;
        end ;
    end ; // for This_Object := 0 to Object_Count - 1
end ; // TSolid.Draw


procedure TSolid.Rotate( XRotate, YRotate, ZRotate : single ) ;
// Rotate object by specified X, Y, Z angles.

var Distance : single ;
    Dummy, Dummy1 : integer ; // Loop controls
    Sin_XRotate, Sin_YRotate, Sin_ZRotate : single ;
    Cos_XRotate, Cos_YRotate, Cos_ZRotate : single ;
    X, Y, Z, X1, Y1, Z1, X2, Y2, Z2, X3, Y3, Z3 : single ;
    Other_Object, Other_Vertex : integer ;
    _X_, _Y_, _Z_ : integer ; // Temp variables

begin
    _Bounding_Cube := Cube( 32767, 32767, 32767, 0, 0, 0 ) ;

    XRotate := ( XRotate + X_Rotation ) * Rad ;
    YRotate := ( YRotate + Y_Rotation ) * Rad ;
    ZRotate := ( ZRotate + Z_Rotation ) * Rad ;

    Sin_XRotate := sin( XRotate ) ;
    Sin_YRotate := sin( YRotate ) ;
    Sin_ZRotate := sin( ZRotate ) ;
    Cos_XRotate := cos( XRotate ) ;
    Cos_YRotate := cos( YRotate ) ;
    Cos_ZRotate := cos( ZRotate ) ;

    for Dummy1 := 0 to Object_Count - 1 do
    begin
        _Objects[ Dummy1 ].Average_Z := 0 ;
        for Dummy := 0 to _Objects[ Dummy1 ].Vertex_Count - 1 do
        begin
            Other_Object := _Objects[ Dummy1 ].vertecies[ Dummy ].Equivalent_Object ;
            Other_Vertex := _Objects[ Dummy1 ].vertecies[ Dummy ].Equivalent_Vertex ;
            if(
                ( Other_Object <> Dummy1 )
                or
                ( Other_Vertex <> Dummy )
              ) then // This Vertex is equivalent to an earlier one
            begin
                _Objects[ Dummy1 ].vertecies[ Dummy ]._X :=
                    _Objects[ Other_Object ].vertecies[ Other_Vertex ]._X ;
                _Objects[ Dummy1 ].vertecies[ Dummy ]._Y :=
                    _Objects[ Other_Object ].vertecies[ Other_Vertex ]._Y ;
                _Objects[ Dummy1 ].vertecies[ Dummy ]._Z :=
                    _Objects[ Other_Object ].vertecies[ Other_Vertex ]._Z ;
                _Objects[ Dummy1 ].Average_Z :=
                    _Objects[ Dummy1 ].Average_Z +
                    trunc( _Objects[ Dummy1 ].vertecies[ Dummy ]._Z ) ;
                continue ; // Skip calculation
            end ;

            // Get current Vertex...
            X := _Objects[ Dummy1 ].vertecies[ Dummy ].X ;
            Y := _Objects[ Dummy1 ].vertecies[ Dummy ].Y ;
            Z := _Objects[ Dummy1 ].vertecies[ Dummy ].Z ;

            // Do rotation...
            X1 := X ;
            Y1 := Y * cos_XRotate - Z * sin_XRotate ;
            Z1 := Y * sin_XRotate + Z * cos_XRotate ;

            X2 := Z1 * sin_YRotate + X1 * cos_YRotate ;
            Y2 := Y1 ;
            Z2 := Z1 * cos_YRotate - X1 * sin_YRotate ;

            X3 := X2 * cos_ZRotate - Y2 * sin_ZRotate ;
            Y3 := X2 * sin_ZRotate + Y2 * cos_ZRotate ;
            Z3 := Z2 ;

            // Store rotated values...
            _Objects[ Dummy1 ].vertecies[ Dummy ]._X := X3 ;
            _Objects[ Dummy1 ].vertecies[ Dummy ]._Y := Y3 ;
            _Objects[ Dummy1 ].vertecies[ Dummy ]._Z := Z3 ;

            _X_ := trunc( X3 ) ;
            _Y_ := trunc( Y3 ) ;
            _Z_ := trunc( Z3 ) ;
            _Objects[ Dummy1 ].Average_Z := _Objects[ Dummy1 ].Average_Z +
                _Z_ ;
            if( _X_ < _Bounding_Cube.Low_X ) then
            begin
                _Bounding_Cube.Low_X := _X_ ;
            end ;
            if( _Y_ < _Bounding_Cube.Low_Y ) then
            begin
                _Bounding_Cube.Low_Y := _Y_ ;
            end ;
            if( _X_ > _Bounding_Cube.High_X ) then
            begin
                _Bounding_Cube.High_X := _X_ ;
            end ;
            if( _Y_ > _Bounding_Cube.High_Y ) then
            begin
                _Bounding_Cube.High_Y := _Y_ ;
            end ;
            if( _Z_ < _Bounding_Cube.Low_Z ) then
            begin
                _Bounding_Cube.Low_Z := _Z_ ;
            end ;
            if( _Z_ > _Bounding_Cube.High_Z ) then
            begin
                _Bounding_Cube.High_Z := _Z_ ;
            end ;
        end ; // for Dummy := 0 to _Objects[ Dummy1 ].Vertex_Count - 1
        _Objects[ Dummy1 ].Average_Z := _Objects[ Dummy1 ].Average_Z div
            _Objects[ Dummy1 ].Vertex_Count ;
    end ; // for Dummy1 := 0 to Object_Count - 1
end ; // TSolid.Rotate


procedure TSolid.Set_Rotation( XRotate, YRotate, ZRotate : single ) ;
// Set object's self-relative rotation

begin
    X_Rotation := XRotate ;
    Y_Rotation := YRotate ;
    Z_Rotation := ZRotate ;
end ;



// T3D_World methods...

// Constructors and destructors...

destructor T3D_World.Destroy ;

begin
    Clear ;

    inherited Destroy ;
end ;


// Property handlers...

function T3D_World.Get_Map_Pixels : boolean ;

begin
    Result := Pixel_Map <> nil ;
end ;


function T3D_World.Get_Object( Index : integer ) : TObject ;

begin
    Result := _Objects[ Index ] ;
end ;


procedure T3D_World.Set_Map_Pixels( Value : boolean ) ;

begin
    if( Value <> Map_Pixels ) then
    begin
        Freemem( Pixel_Map ) ;
        if( Value ) then
        begin
            if( Display_Buffer = nil ) then // Not quite set-up yet
            begin
                Pixel_Map := AllocMem( 4 ) ;
            end else
            begin
                Pixel_Map := AllocMem( sizeof( integer ) * Display_Buffer.Height * Display_Buffer.Width ) ;
            end ;
        end else
        begin
            Pixel_Map := nil ;
        end ;
    end ;
end ;


procedure T3D_World.Set_Object( Index : integer ; Value : TObject ) ;

begin
    _Objects[ Index ] := Value ;
end ;


// Internal utility routines...

procedure T3D_World.Prepare_Display( Height, Width : integer ) ;

begin
    Display_Buffer_DC := CreateCompatibleDC( 0 ) ;
    BM.bmiheader.biSize := sizeof( BM.bmiheader ) ;
    BM.bmiheader.biWidth := Width ;
    BM.bmiheader.biHeight := -Height ; // Top-down
    BM.bmiheader.biBitCount := 32 ;
    BM.bmiheader.biPlanes := 1 ;
    BM.bmiheader.biCompression := BI_RGB ;
    BM.bmiheader.biSizeImage := 0 ;
    BM.bmiheader.biXPelsPerMeter := 3456 ;
    BM.bmiheader.biYPelsPerMeter := 3456 ;
    BM.bmiheader.biClrUsed := 0 ;
    BM.bmiheader.biClrImportant := 0 ;
    Display_Bitmap := CreateDIBSection( Display_Buffer_DC, BM, DIB_RGB_Colors, Display_Bits, 0, 0 ) ;
    SelectObject( Display_Buffer_DC, Display_Bitmap ) ;
    Display_Buffer := TTexture.Create ;
    Display_Buffer.Init( Width, Height ) ;
    Display_Buffer.Set_Buffer( Display_Bits ) ;
    Display_Buffer.Blank ;
    if( Map_Pixels ) then
    begin
        Freemem( Pixel_Map ) ;
        Pixel_Map := AllocMem( sizeof( integer ) * Display_Buffer.Height * Display_Buffer.Width ) ;
    end ;
end ; // T3D_World.Prepare_Display


// Public API...

function T3D_World.Add_Object( X : TSolid ) : integer ;
// Add a new object to the world.  Returns index of it.

begin
    if( _Objects = nil ) then
    begin
        _Objects := TList.Create ;
    end ;
    Result := _Objects.Count ;
    _Objects.Add( X ) ;
end ;


procedure T3D_World.Adjust_Distance( _Object : TSolid ) ;
// Adjust object vertecies based on camera distance.

var Distance : single ;
    X3, Y3, Z3 : single ;

begin
    // Get vertecies...
    X3 := _Object._X ;
    Y3 := _Object._Y ;
    Z3 := _Object._Z ;

    // Adjust for distance...
    Distance := SQRT( ( Camera.Z - Z3 ) * ( Camera.Z - Z3 ) +
        ( Camera.Y - Y3 ) * ( Camera.Y - Y3 ) +
        ( Camera.X - X3 ) * ( Camera.X - X3 ) ) ;
    if( Distance <> 0 ) then
    begin
        Distance := Distance / Distance_Factor ; // Convert to proper units
        if( Distance > 0 ) then
        begin
            //Distance := SQR( Distance ) ;
        end ;
        X3 := X3 / Distance ;
        Y3 := Y3 / Distance ;
        Z3 := Z3 / Distance ;
    end ; // if( Distance <> 0 )

    // Adjust the vertecies...
    _Object._X := trunc( X3 ) ;
    _Object._Y := trunc( Y3 ) ;
    _Object._Z := trunc( Z3 ) ;
end ; // T3D_World.Adjust_Distance


procedure T3D_World.Clear ; // Delete all objects

var Loop : integer ;

begin
    if( _Objects <> nil ) then
    begin
        for Loop := 0 to _Objects.Count - 1 do
        begin
            TSolid( _Objects[ Loop ] ).Free ;
        end ;
        _Objects.Clear ;
    end ;
    Erase_Buffer ;
end ;


procedure T3D_World.Erase_Buffer ;
// Erases display buffer.

(*
var Dummy : integer ; // Object loop control
*)

begin
    if( ( _Objects = nil ) or ( _Objects.Count = 0 ) ) then
    begin
        exit ; // No objects defined
    end ;
(*
    Update_Rect := TSolid( Objects[ 0 ] ).Bounding_Rect ;
    for Dummy := 1 to _Objects.Count - 1 do
    begin
        Update_Rect :=
            Merge_Rects( Update_Rect, TSolid( Objects[ Dummy ] ).Bounding_Rect ) ;
    end ;
    inc( Update_Rect.Right ) ;
    inc( Update_Rect.Bottom ) ;
*)
    Display_Buffer.Blank ;
    if( Map_Pixels ) then
    begin
        fillchar( Pixel_Map,
            sizeof( integer ) * Display_Buffer.Height * Display_Buffer.Width,
            0 ) ;
    end ;
end ;


function T3D_World.Get_Camera_Position : TVertex ;
// Return camera position.

begin
    Result := Camera ;
end ;


procedure T3D_World.Render( DC : HDC ) ;
// Render current camera view onto device context.

var This_Object : integer ;
    _Object : TSolid ;
    Rect : TRect ;
    WH : HWND ;

begin
    // Setup...
    WH := WindowFromDC( DC ) ;
    GetClientRect( WH, Rect ) ;
    if( Display_Buffer = nil ) then // Not yet setup
    begin
        Display_Buffer := TTexture.Create ;
    end ;
    if(
        ( Rect.Bottom <> Display_Buffer.Height - 1 )
        or
        ( Rect.Right <> Display_Buffer.Width - 1 )
      ) then
    begin // Size change
        Prepare_Display( Rect.Bottom + 1, Rect.Right + 1 ) ;
    end else
    begin
        Display_Buffer.Blank ;
    end ;

    // Handle camera yaw, pitch, and roll
    Rotate( Camera._X, Camera._Y, Camera._Z ) ;

    // Process distances and draw all objects
    for This_Object := 0 to _Objects.Count - 1 do
    begin
        // Setup
        _Object := TSolid( _Objects[ This_Object ] ) ;

        // Determine minimum output rectangle (for performance)
        Update_Rect :=
            Merge_Rects( Update_Rect, _Object.Bounding_Rect ) ;

        // Adjust object's rotation for camera orientation
        _Object.Rotate( Camera._X, Camera._Y, Camera._Z ) ;

        // Adjust object for distance from camera
        Adjust_Distance( _Object ) ;
        _Object.Adjust_Distance( Camera.X, Camera.Y, Camera.Z ) ;

        // Draw object into buffer
        _Object.Draw( Display_Buffer, trunc( Camera.Z ), Display_Buffer,
            Pixel_Map ) ;

        // Determine minimum output rectangle (for performance)
        Update_Rect :=
            Merge_Rects( Update_Rect, _Object.Bounding_Rect ) ;
    end ; // for This_Object := 0 to _Objects.Count - 1

    // Output the buffer
    if( not Inhibit_Display ) then
    begin
        BitBlt( DC, 0, 0, Display_Buffer.Width, Display_Buffer.Height,
            Display_Buffer_DC, 0, 0, SRCCOPY ) ;
    end ;
end ; // T3D_World.Render


function T3D_World.Object_At( X, Y : integer ; var Surface : integer ) : TSolid ;

begin
    if(
        ( X < 0 )
        or
        ( Y < 0 )
        or
        ( X >= Display_Buffer.Width )
        or
        ( Y >= Display_Buffer.Height )
      ) then
    begin
        Result := nil ;
    end else
    begin
        Result := TSolid( Pixel_Map[ Y * Display_Buffer.Width + X ] ) ;
    end ;
end ;


procedure T3D_World.Rotate( XRotate, YRotate, ZRotate : single ) ;
// Rotate objects.

var This_Object : integer ; // Object loop control
    _Object : TSolid ;
    Sin_XRotate, Sin_YRotate, Sin_ZRotate : single ;
    Cos_XRotate, Cos_YRotate, Cos_ZRotate : single ;
    X, Y, Z, X1, Y1, Z1, X2, Y2, Z2, X3, Y3, Z3 : single ;

begin
    // Setup...
    XRotate := XRotate * Rad ;
    YRotate := YRotate * Rad ;
    ZRotate := ZRotate * Rad ;
    if( _Objects = nil ) then
    begin
        exit ;
    end ;

    Sin_XRotate := sin( XRotate ) ;
    Sin_YRotate := sin( YRotate ) ;
    Sin_ZRotate := sin( ZRotate ) ;
    Cos_XRotate := cos( XRotate ) ;
    Cos_YRotate := cos( YRotate ) ;
    Cos_ZRotate := cos( ZRotate ) ;

    // Rotate objects
    for This_Object := 0 to _Objects.Count - 1 do
    begin
        _Object := TSolid( Objects[ This_Object ] ) ;

        // Get current Vertex...
        X := _Object.X ;
        Y := _Object.Y ;
        Z := _Object.Z ;

        // Adjust to be relative to camera...
        X := X - Camera.X ;
        Y := Y - Camera.Y ;
        Z := Z - Camera.Z ;

        // Do rotation...
        X1 := X ;
        Y1 := Y * cos_XRotate - Z * sin_XRotate ;
        Z1 := Y * sin_XRotate + Z * cos_XRotate ;

        X2 := Z1 * sin_YRotate + X1 * cos_YRotate ;
        Y2 := Y1 ;
        Z2 := Z1 * cos_YRotate - X1 * sin_YRotate ;

        X3 := X2 * cos_ZRotate - Y2 * sin_ZRotate ;
        Y3 := X2 * sin_ZRotate + Y2 * cos_ZRotate ;
        Z3 := Z2 ;

        // Store rotated values...
        _Object._X := trunc( X3 ) ;
        _Object._Y := trunc( Y3 ) ;
        _Object._Z := trunc( Z3 ) ;
    end ; //  for This_Object := 0 to _Objects.Count - 1
end ; // T3D_World.Rotate


procedure T3D_World.Set_Camera_Position( V : TVertex ) ;
// Sets camera position.

begin
    Camera.X := V.X ;
    Camera.Y := V.Y ;
    Camera.Z := V.Z ;
    Camera._X := V._X ;
    Camera._Y := V._Y ;
    Camera._Z := V._Z ;
end ;



// Unit stuff...

procedure Init_Unit ;

begin
    Distance_Factor := GetSystemMetrics( SM_CXFULLSCREEN ) ;
end ;


initialization
    Init_Unit ;
end.
