{$V-}
{
          Program Name : SMU
          Package Name : LB
          Purpose      : Screen Management Utility routines
          Institution  : Conroy & Conroy
          Date Written : 18-Mar-1989
          Written By   : Alan Conroy
          Version      : 2.0C

          Copyright (C) 1989-2005 by Alan Conroy.  Released to the public domain.

          TO THE GLORY of GOD THROUGH JESUS CHRIST

	*********************************************************
	*                                                       *
	*        M O D I F I C A T I O N   H I S T O R Y        *
	*                                                       *
	*     DATE      BY             REASON                   *
	*                                                       *
	*  1-Nov-2002   EAC            Code cleanup.            *
	* 13-Nov-2002   EAC            Fix compilation error in *
        *                              test mode.               *
        *  4-Apr-2005   EAC            Added foreground and     *
        *                              background colors.       *
	*                                                       *
	*********************************************************

	*********************************************************
	*                                                       *
	*           P R O G R A M   P U R P O S E               *
	*                                                       *
	*********************************************************

	  These routines provide complete screen management.
}

unit SMU ;

interface

{$I EDEFINES.INC}

uses { Borland... }
{$IFDEF DELPHI}
     Windows, Graphics,
{$ELSE}
    {$IFDEF WINDOWS}
     WinTypes, WinProcs,
    {$ENDIF}
{$ENDIF}

     { C&C... }
     Common, { TBase_Object }
     Colors, { TRGB }
     SMUBIOS,
     TypeDefs ; { Max_Memory }

{$IFDEF DELPHI}
    {$IFNDEF CONSOLE}
        {$DEFINE WINDOWS}
    {$ENDIF}
{$ENDIF}

{$IFDEF DELPHI}
    {$DEFINE WINDOWSORDELPHI}
{$ENDIF}
{$IFDEF WINDOWS}
    {$DEFINE WINDOWSORDELPHI}
{$ENDIF}


type proc = procedure ;

var  KB_Buffer : string[ 255 ] ; { Input not related to a window }
     Scr : Text ;

type Window_Def = record
		      case Integer of
			  0 : ( Bytes : array[ 0..Max_Memory - 1 ] of Byte ) ;
			  1 : ( Chars : array[ 0..Max_Memory - 1 ] of Char ) ;
		  end ;
     Window_Def_Pointer = ^Window_Def ;
     PSWindow_Contents = ^TSWindow_Contents ;
     TSWindow_Contents = object( TBase_Object )
                             private { Instance data... }
                                 _Data_Pointer : Window_Def_Pointer ;
                                 _Attribute_Pointer : Window_Def_Pointer ;
                                 _Background_Colors_Pointer : Window_Colors_Pointer ;
                                 _Foreground_Colors_Pointer : Window_Colors_Pointer ;
                                 _Size : integer ; { Size of contents, in elements }

                             public { Constructors and destructors... }
                                 constructor Init ;
                                 destructor Done ; virtual ;

                             public { API... }
                                 { Access to individual contents... }
                                 function Get_Attribute( Offset : integer ) : integer ;
                                 procedure Set_Attribute( Offset, Value : integer ) ;
                                 function Get_Data_Byte( Offset : integer ) : integer ;
                                 procedure Set_Data_Byte( Offset, Value : integer ) ;
                                 function Get_Background_Color( Offset : integer ) : TRGB ;
                                 procedure Set_Background_Color( Offset : integer ;
                                     Value : TRGB ) ;
                                 function Get_Foreground_Color( Offset : integer ) : TRGB ;
                                 procedure Set_Foreground_Color( Offset : integer ;
                                     Value : TRGB ) ;

                             public { Fill routines... }
                                 procedure Fill_Data( Start, Size : longint ;
                                     Value : byte ) ;
                                 procedure Fill_Attributes( Start, Size : longint ;
                                     Value : byte ) ;
                                 procedure Fill_Background_Colors( Start, Size : longint ;
                                     Value : TRGB ) ;
                                 procedure Fill_Foreground_Colors( Start, Size : longint ;
                                     Value : TRGB ) ;
                                 procedure Copy( Source : PSWindow_Contents ;
                                     Source_Start, Destination_Start, Size : integer ) ;

                             public { Misc. }
                                 function Get_Size : longint ;
                                 procedure Set_Size( Size : longint ) ;
                         end ; { TSWindow_Contents }
     PWindow = ^Window ;
     Window_Record = record { Pointer to a window }
			Pointer : PWindow ;
			ID : integer ;
		     end ;
     Window = record { Data for window }
		  Level : Integer ; {Level of window}
		  Attribute : Byte ; {Current attribute value}
                  Border : byte ; { Size of border }
                  Background, Foreground : TRGB ; { Current colors }
		  Rows, Columns : integer ; {Window row and column count}
		  Map_Row, Map_Column, Map_Rows, Map_Columns : integer ;
		  {Row and column positions and sizes on physical screen}
		  View_Row, View_Column : integer ;
		  {Upper-left-hand corner of window visible in view-port}
		  Position_Row, Position_Column : integer ;
		  {Current position}
		  Low_Update, High_Update : Integer ;
		  {Low and high offsets defining what needs to be updated}
		  Size : Integer ; {Window size (columns*rows)}
		  Flags : Integer ; {Flags for window:
				    and   1 = No scroll - vertical wrap
				    and   2 = 100% occluded
				    and   4 = No wrap
				    and   8 = No auto-update
				    and  16 = 100% unoccluded
				    and  32 = Cursor is disabled in this window
				    and  64 = Graphics window
				    and 128 = NCM on input
				    and 256-32768 = reserved}
                  Contents : PSWindow_Contents ;
                  Text_Height, Text_Width : Integer ; {Size of font characters}
                  KB_Buffer : string[ 255 ] ;
		  Next_Window : PWindow ;
		  {Pointer to next screen in list}
		  ID : Integer ; {Window ID/handle}
{$IFDEF Debug}
                  Sanity : string[ 6 ] ;
{$ENDIF}
	      end ; { Window }

{$IFDEF WINDOWS}
function Create_Window_Wrapper( Handle : HWND ) : PWindow ;
type TRect = Windows.TRect ;
{$ELSE}
type TRect = record
		Top, Bottom, Left, Right : integer ;
	     end ;
{$ENDIF}

{ API... }
{$IFDEF WINDOWS}
procedure Calculate_Window_Extent( Window_P : PWindow ; var R, C : longint ) ;
{$ENDIF}
procedure Clear_Cursor( Vr, Vc : Integer ) ;
procedure Clear_Window( Window_P : PWindow ) ; {Clear a window}
function Create_Window( Row_Count, Column_Count, Flag : Integer ) :
    PWindow ; {Create a window}
function Current_Window : PWindow ; {Return current window pointer}
procedure Delete_Line( Window_P : PWindow ; Direction : Byte ) ;
procedure Delete_Window( Wp : PWindow ) ; {Delete a window}
procedure End_SMU ; {Nicely terminate SMU}
procedure Erase_EOL( Window_P : PWindow ) ;
procedure Erase_EOW( Window_P : PWindow ) ; {Erase to end of window}
procedure Get_Default_Window( var X : Window_Record ) ;
function Get_Model : Byte ; {Return system type}
function Get_Screen( Row, Col : Integer ) : Byte ;
function Get_Text_Extent( Window_P : PWindow; String_P : string; Count : integer ) : longint ;
function Get_Window_Attribute( R, C : Integer ) : Byte ;
function Get_Window_Char( R, C : Integer ) : Byte ;
procedure Get_Window_Size( Window_P : PWindow ; var R, C : integer ) ;
function GetTextExtent( Window_P : PWindow; String_P : PChar; Count : integer ) : longint ;
procedure Init_SMU( X, Y, Z, A : Integer ) ;
procedure Insert_Line( Window_P : PWindow ; Direction : Byte ) ;
procedure Invalidate_Window( Window_P : PWindow ) ;
procedure Load_Driver( X : string ) ;
procedure Map_Window( Window_P : PWindow ; Row, Row_Count, Column,
		     Column_Count, View_R, View_C, Level : Byte ) ;
procedure Move_To( Window_P : PWindow ; Row, Column : integer ) ;
procedure Output_Text( const S : string ) ;
procedure Refresh_Screen ; {Refresh physical screen}
procedure Refresh_Window( Window_P : PWindow ) ;
procedure Resize( var Window_P : PWindow ; New_Rows, New_Columns : integer ) ;
procedure Scroll_Window( Window_P : PWindow ; X, Y : integer ; Rect : TRect ) ;
procedure Select_Window( Window_P : PWindow ) ;
procedure Set_Background( const RGB : TRGB ) ;
procedure Set_Foreground( const RGB : TRGB ) ;
procedure Set_Cursor( Vr, Vc : Integer ) ;
procedure Set_Default_Window( X : Window_Record ) ;
procedure Set_Screen( Model, Printer_Model, Input_Buffers, Mode : Integer ) ;
procedure Set_Viewport( Row, Column : Byte ) ; {Set viewport row and column}
procedure Set_Window( Window_P : PWindow ; Row, Column, Flag,
    Attributes : Integer ) ;
procedure Text_Out( Window_P : PWindow; T : string; L : integer ) ;
procedure Unmap_Window( Window_P : PWindow ) ;
procedure Update_Screen ; {Update all windows}
procedure Update_Window( Window_P : PWindow ) ;
function Window_Async_Input_Line( Window_P : PWindow ; Flags : Integer ; Key : string ) : boolean ;
procedure Window_Input_Line( Window_P : PWindow ; Flags : Integer ) ;
function Window_Mapped( Window_P : PWindow ) : Boolean ;
procedure Window_Read ; {Read a line}
procedure Window_Read_Input( Flags : Integer ) ; {Read a line}
function Window_To_Screen_Row( Window : PWindow ; Row : integer ) : integer ;
function Window_To_Screen_Column( Window : PWindow ; Column : integer ) : integer ;


{$IFDEF Test}
procedure Test_Unit ;
{$ENDIF}

implementation

uses Compatib,
     O_S, { OS }
{$IFNDEF WINDOWS}
     GetChars,
{$ENDIF}
     ASCIIDef,
{$IFDEF DELPHI}
     SysUtils ;
{$ELSE}
    {$IFNDEF WINDOWS}
     Graph,
    {$ENDIF}
     DOS, { Registers }
     Strings ;
{$ENDIF}

const Max_Row = 50 ;
      Max_Column = 80 ;

Type Data_Rec = Record
		Case Integer of
		    0 : ( Chars : Array[ 1..Max_Row ] of Array[ 1..Max_Column ] of Char ) ;
		    1 : ( Bytes : Array[ 1..Max_Row ] of Array[ 1..Max_Column ] of Byte ) ;
	    end ;
     Screen_Def = Record { Physical screen definition }
		      Data : Data_Rec ;
		      Window : Data_Rec ;
		  end ;
     Input_P = ^Input_Record ;
     Input_Record = Record
			Last_Input, Next_Input : Input_P ;
			{ Pointers to next and previous lines }
			Input_Line : string[ 255 ] ; { Actual input }
		    end ;
     SMU_Root = record { Root data for screen management }
		    Driver : Input_P ; { Pointer to video driver.  nil if none }
		    Model : Byte ;
		    Default : integer ; { Current default window }
		    Default_P : PWindow ; { Pointer to default window }
		    Next_Window : PWindow ; { Pointer to first window }
		    Window_Count : Byte ; { Total windows defined }
		    Printer_Type : Byte ; { Type of printer
					  1 = Unknown
					  2 = LA50/LA75
					  3 = IBM Graphics printer }
		    Insert_Mode : Boolean ; { True if insert mode }
		    Printer_Mode : Byte ;
		    { Print mode: 0 = Off, 1 = On, 2 = Disabled }
		    Input_Buffer_Max : Byte ; { Number of lines to save }
		    Last_Input : Input_P ;
		    { Pointer to list of previous input lines }
		    Video_Mode : Byte ; { Type of video mode (a la PC) }
		    Foreground, Background : Byte ; { Fg/Bg colors }
		    Palatte : Byte ; { Color pallate selected }
		    Screen_Width : Integer ; { Maximum size of screen in columns }
		    Flags : Integer { General flags:
				      and 1-16 = reserved
				      and 32 = Disable cursor
				      and 64-32768 = reserved }
		end ;
     Cvt = Array[ 0..31 ] of Byte ; { First part of conversion table }
     Cvt1 = Array[ 127..254 ] of Byte ; { Second part of conversion table }
     Block = Array[ 0..255 ] of Byte ;
     Block_B = Array[ 0..255 ] of Boolean ;

 Const Rainbow_Cvt : Cvt = ( 0, 32, 32, 32, 1, 32, 32, 32, 32, 32,
			    32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
			    182, 167, 32, 32, 32, 32, 32, 32, 32, 32, 32, 216 ) ;
       Rainbow_Cvt1 : Cvt1 = ( 32, 199, 220,
			     233, 226, 228, 224, 229, 231, 234, 235, 232, 239,
			     238, 236, 196, 197, 201, 247, 198, 244, 246, 242,
			     251, 249, 221, 214, 219, 162, 30, 165, 32, 32,
			     225, 237, 243, 250, 241, 209, 170, 186, 129, 13,
			     12, 189, 188, 161, 171, 187, 2, 2, 2, 25,
			     22, 22, 22, 12, 12, 22, 25, 12, 11, 11,
			     11, 12, 14, 23, 24, 21, 18, 15, 21, 21,
			     14, 13, 23, 24, 21, 18, 15, 23, 23, 24,
			     24, 14, 14, 13, 13, 15, 15, 11, 13, 32,
			     32, 32, 32, 32, 32, 223, 32, 28, 32, 32,
			     181, 32, 32, 32, 32, 32, 32, 248, 32, 32,
			     32, 8, 27, 26, 32, 32, 32, 32, 7, 31,
			     31, 32, 32, 178, 46 ) ;
       La50_Cvt : Cvt = ( 0, 32, 32, 32, 96, 32, 32, 7, 8, 9,
			 10, 11, 12, 13, 32, 32, 32, 32, 32, 32,
			 182, 167, 32, 32, 32, 32, 32, 32, 32, 32, 32, 216 ) ;
       La50_Cvt1 : Cvt1 = ( 32, 199, 220,
			   233, 229, 228, 224, 226, 231, 234, 235, 232, 239,
			   238, 236, 196, 197, 201, 230, 198, 244, 246, 242,
			   251, 249, 221, 214, 251, 162, 163, 165, 32, 32,
			   225, 237, 243, 250, 241, 209, 170, 186, 191, 108,
			   107, 189, 188, 161, 171, 187, 97, 97, 97, 120,
			   117, 117, 117, 107, 107, 117, 120, 107, 106, 106,
			   106, 107, 109, 118, 119, 116, 113, 110, 116, 116,
			   109, 108, 118, 119, 116, 113, 110, 118, 118, 119,
			   119, 109, 109, 108, 108, 110, 110, 106, 108, 32,
			   32, 32, 32, 32, 32, 223, 32, 123, 32, 32,
			   181, 32, 32, 32, 32, 32, 32, 248, 32, 32,
			   32, 177, 122, 121, 32, 32, 32, 32, 176, 183,
			   183, 32, 32, 178, 46 ) ;
       La50_Cvt_F : Cvt = ( 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
			   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ) ;
       La50_Cvt1_F : Cvt1 = ( 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
				    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
				    1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
				    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
				    1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0,
				    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0,
				    0, 0, 0, 0, 0 ) ;
       {Characters which "print"}
       Print_Characters : Block_B = ( True, True, True, True, True, True, True, False,
			   {8-14}     False, True, False, True, True, False, True,
			   {15-22}    True, True, True, True, True, True, True, True,
			   {23-30}    True, True, True, True, True, True, True, True,
			   {31-38}    True, True, True, True, True, True, True, True,
			   {39-46}    True, True, True, True, True, True, True, True,
			   {47-54}    True, True, True, True, True, True, True, True,
			   {55-62}    True, True, True, True, True, True, True, True,
			   {63-70}    True, True, True, True, True, True, True, True,
			   {71-78}    True, True, True, True, True, True, True, True,
			   {79-86}    True, True, True, True, True, True, True, True,
			   {87-94}    True, True, True, True, True, True, True, True,
			   {95-102}   True, True, True, True, True, True, True, True,
			   {103-110}  True, True, True, True, True, True, True, True,
			   {111-118}  True, True, True, True, True, True, True, True,
			   {119-126}  True, True, True, True, True, True, True, True,
			   {127-134}  True, True, True, True, True, True, True, True,
			   {135-142}  True, True, True, True, True, True, True, True,
			   {143-150}  True, True, True, True, True, True, True, True,
			   {151-158}  True, True, True, True, True, True, True, True,
			   {159-166}  True, True, True, True, True, True, True, True,
			   {167-174}  True, True, True, True, True, True, True, True,
			   {175-182}  True, True, True, True, True, True, True, True,
			   {183-190}  True, True, True, True, True, True, True, True,
			   {191-198}  True, True, True, True, True, True, True, True,
			   {199-206}  True, True, True, True, True, True, True, True,
			   {207-214}  True, True, True, True, True, True, True, True,
			   {215-222}  True, True, True, True, True, True, True, True,
			   {223-230}  True, True, True, True, True, True, True, True,
			   {231-238}  True, True, True, True, True, True, True, True,
			   {239-246}  True, True, True, True, True, True, True, True,
			   {247-254}  True, True, True, True, True, True, True, False,
			   {255}      True ) ;
       {Characters which affect cursor position}
       Cursor_Characters : Block_B = ( False, False, False, False, False, False,
			   {6-12}     False, False, True, True, False, False, False,
			   {13-19}    True, False, False, False, False, False, False,
			   {20-25}    False, False, False, False, False, False,
			   {26-32}    False, False, False, False, False, False, True,
			   {33-40}    True, True, True, True, True, True, True, True,
			   {41-48}    True, True, True, True, True, True, True, True,
			   {49-56}    True, True, True, True, True, True, True, True,
			   {57-64}    True, True, True, True, True, True, True, True,
			   {65-72}    True, True, True, True, True, True, True, True,
			   {73-80}    True, True, True, True, True, True, True, True,
			   {81-88}    True, True, True, True, True, True, True, True,
			   {89-96}    True, True, True, True, True, True, True, True,
			   {97-104}   True, True, True, True, True, True, True, True,
			   {105-112}  True, True, True, True, True, True, True, True,
			   {113-120}  True, True, True, True, True, True, True, True,
			   {121-127}  True, True, True, True, True, True, False,
			   {128-133}  True, True, True, True, True, True,
			   {134-139}  True, True, True, True, True, True,
			   {140-145}  True, True, True, True, True, True,
			   {146-151}  True, True, True, True, True, True,
			   {152-157}  True, True, True, True, True, True,
			   {158-163}  False, False, True, True, True, True,
			   {164-169}  True, True, True, True, True, True,
			   {170-175}  True, True, True, True, True, True,
			   {176-181}  True, True, True, True, True, True,
			   {182-187}  True, True, True, True, True, True,
			   {188-193}  True, True, True, True, True, True,
			   {194-199}  True, True, True, True, True, True,
			   {200-205}  True, True, True, True, True, True,
			   {206-211}  True, True, True, True, True, True,
			   {212-217}  True, True, True, True, True, True,
			   {218-223}  True, False, False, False, False, False,
			   {224-229}  False, True, False, True, False, False,
			   {230-235}  True, False, False, False, False, False,
			   {236-241}  False, True, False, False, False, True,
			   {242-247}  True, True, False, False, False, False,
			   {248-253}  True, True, False, False, False, True,
			   {254-255}  True, False ) ;

var  Attribute_Default : Byte ; { Default attribute value }
     Conversion : ^Block ; { Screen conversion array }
     Deconversion : ^Block ; { Inverse of Conversion array }
     Printer0 : ^Block ; { Printer conversion array }
     Printer1 : ^Block ; { Printer conversion flags }
     Printer_File : Text ;
     Input_Row, Input_Column : Integer ; { Row and Column at beginning of input }
     Screen : Screen_Def ; { Define screen }
     Screen_Root : SMU_Root ;
     Attribute_String, Screen_String : string[ 255 ] ;
     Screen_Background_Offset, Screen_Foreground_Offset : longint ;
     Output_Buffer : string[ 255 ] ;
{$IFNDEF DELPHI}
     Row_Start : Array[ 1..Max_Row ] of Integer ; { Start of rows in memory }
{$ENDIF}

{ TSWindow_Contents methods... }

constructor TSWindow_Contents.Init ;

begin
    inherited Init ;

    _Size := 0 ;
    _Data_Pointer := nil ;
    _Attribute_Pointer := nil ;
    _Background_Colors_Pointer := nil ;
    _Foreground_Colors_Pointer := nil ;
end ;


destructor TSWindow_Contents.Done ;

begin
    Set_Size( 0 ) ; { Free all data }

    inherited Done ;
end ;


procedure TSWindow_Contents.Fill_Data( Start, Size : longint ; Value : byte ) ;

begin
{$IFDEF Debug}
    if( ( Start < 0 ) or ( Start + Size - 1 > _Size ) ) then
    begin
        writeln( 'Subscript out of range' ) ;
        halt ;
    end ;
{$ENDIF}
    fillchar( _Data_Pointer^.Bytes[ Start ], Size, Value ) ;
end ;


procedure TSWindow_Contents.Fill_Attributes( Start, Size : longint ;
    Value : byte ) ;

begin
{$IFDEF Debug}
    if( ( Start < 0 ) or ( Start + Size - 1 > _Size ) ) then
    begin
        writeln( 'Subscript out of range' ) ;
        halt ;
    end ;
{$ENDIF}
    fillchar( _Attribute_Pointer^.Bytes[ Start ], Size, Value ) ;
end ;


procedure TSWindow_Contents.Fill_Background_Colors( Start, Size : longint ;
    Value : TRGB ) ;

var Loop : integer ;

begin
{$IFDEF Debug}
    if( ( Start < 0 ) or ( Start + Size - 1 > _Size ) ) then
    begin
        writeln( 'Subscript out of range' ) ;
        halt ;
    end ;
{$ENDIF}
    for Loop := 0 to Size - 1 do
    begin
        _Background_Colors_Pointer^[ Loop ] := Value ;
    end ;
end ;


procedure TSWindow_Contents.Fill_Foreground_Colors( Start, Size : longint ;
    Value : TRGB ) ;

var Loop : integer ;

begin
{$IFDEF Debug}
    if( ( Start < 0 ) or ( Start + Size - 1 > _Size ) ) then
    begin
        writeln( 'Subscript out of range' ) ;
        halt ;
    end ;
{$ENDIF}
    for Loop := 0 to Size - 1 do
    begin
        _Foreground_Colors_Pointer^[ Loop ] := Value ;
    end ;
end ;


procedure TSWindow_Contents.Copy( Source : PSWindow_Contents ;
    Source_Start, Destination_Start, Size : integer ) ;

begin
{$IFDEF Debug}
    if( ( Destination_Start < 0 ) or ( Destination_Start + Size - 1 > _Size ) ) then
    begin
        writeln( 'Subscript out of range' ) ;
        halt ;
    end ;
    if( ( Source_Start < 0 ) or ( Source_Start + Size - 1 > Source^.Get_Size ) ) then
    begin
        writeln( 'Subscript out of range' ) ;
        halt ;
    end ;
{$ENDIF}
    move( Source^._Data_Pointer^.Bytes[ Source_Start ],
        _Data_Pointer^.Bytes[ Destination_Start ], Size ) ;
    move( Source^._Attribute_Pointer^.Bytes[ Source_Start ],
        _Attribute_Pointer^.Bytes[ Destination_Start ], Size ) ;
    move( Source^._Background_Colors_Pointer^[ Source_Start ],
        _Background_Colors_Pointer^[ Destination_Start ],
        Size * sizeof( TRGB ) ) ;
    move( Source^._Foreground_Colors_Pointer^[ Source_Start ],
        _Foreground_Colors_Pointer^[ Destination_Start ],
        Size * sizeof( TRGB ) ) ;
end ;


function TSWindow_Contents.Get_Attribute( Offset : integer ) : integer ;

begin
{$IFDEF Debug}
    if( ( Offset < 0 ) or ( Offset >= _Size ) ) then
    begin
        writeln( 'Subscript out of range' ) ;
        halt ;
    end ;
{$ENDIF}
    Get_Attribute := _Attribute_Pointer^.Bytes[ Offset ] ;
end ;


procedure TSWindow_Contents.Set_Attribute( Offset, Value : integer ) ;

begin
{$IFDEF Debug}
    if( ( Offset < 0 ) or ( Offset >= _Size ) ) then
    begin
        writeln( 'Subscript out of range' ) ;
        halt ;
    end ;
{$ENDIF}
    _Attribute_Pointer^.Bytes[ Offset ] := Value ;
end ;


function TSWindow_Contents.Get_Data_Byte( Offset : integer ) : integer ;

begin
{$IFDEF Debug}
    if( ( Offset < 0 ) or ( Offset >= _Size ) ) then
    begin
        writeln( 'Subscript out of range' ) ;
        halt ;
    end ;
{$ENDIF}
    Get_Data_Byte := _Data_Pointer^.Bytes[ Offset ] ;
end ;


procedure TSWindow_Contents.Set_Data_Byte( Offset, Value : integer ) ;

begin
{$IFDEF Debug}
    if( ( Offset < 0 ) or ( Offset >= _Size ) ) then
    begin
        writeln( 'Subscript out of range' ) ;
        halt ;
    end ;
{$ENDIF}
    _Data_Pointer^.Bytes[ Offset ] := Value ;
end ;


function TSWindow_Contents.Get_Background_Color( Offset : integer ) : TRGB ;

begin
{$IFDEF Debug}
    if( ( Offset < 0 ) or ( Offset >= _Size ) ) then
    begin
        writeln( 'Subscript out of range' ) ;
        halt ;
    end ;
{$ENDIF}
    Get_Background_Color := _Background_Colors_Pointer[ Offset ] ;
end ;


procedure TSWindow_Contents.Set_Background_Color( Offset : integer ;
    Value : TRGB ) ;

begin
{$IFDEF Debug}
    if( ( Offset < 0 ) or ( Offset >= _Size ) ) then
    begin
        writeln( 'Subscript out of range' ) ;
        halt ;
    end ;
{$ENDIF}
    _Background_Colors_Pointer[ Offset ] := Value ;
end ;


function TSWindow_Contents.Get_Foreground_Color( Offset : integer ) : TRGB ;

begin
{$IFDEF Debug}
    if( ( Offset < 0 ) or ( Offset >= _Size ) ) then
    begin
        writeln( 'Subscript out of range' ) ;
        halt ;
    end ;
{$ENDIF}
    Get_Foreground_Color := _Foreground_Colors_Pointer[ Offset ] ;
end ;


procedure TSWindow_Contents.Set_Foreground_Color( Offset : integer ;
    Value : TRGB ) ;

begin
{$IFDEF Debug}
    if( ( Offset < 0 ) or ( Offset >= _Size ) ) then
    begin
        writeln( 'Subscript out of range' ) ;
        halt ;
    end ;
{$ENDIF}
    _Foreground_Colors_Pointer[ Offset ] := Value ;
end ;


function TSWindow_Contents.Get_Size : longint ;

begin
    Get_Size := _Size ;
end ;


procedure TSWindow_Contents.Set_Size( Size : longint ) ;

begin
{$IFDEF Debug}
    if( Size < 0 ) then
    begin
        writeln( 'Size out of range' ) ;
        halt ;
    end ;
{$ENDIF}
    freemem( _Data_Pointer, _Size ) ;
    freemem( _Attribute_Pointer, _Size ) ;
    freemem( _Background_Colors_Pointer, _Size * sizeof( TRGB ) ) ;
    freemem( _Foreground_Colors_Pointer, _Size * sizeof( TRGB ) ) ;
    _Size := Size ;
    Getmem( _Data_Pointer, Size ) ; { Create window }
    Getmem( _Attribute_Pointer, Size ) ;
    Getmem( _Background_Colors_Pointer, Size * sizeof( TRGB ) ) ;
    Getmem( _Foreground_Colors_Pointer, Size * sizeof( TRGB ) ) ;
end ;



type pStatic_Input = ^tStatic_Input ;
     tStatic_Input = record
        B, Command, Command0 : string ;
        Backup : Char ; { Backup character to use }
        Ch1 : Char ;
        Cursor, Cursor1 : Byte ;
        Ord_B, Vc, Vr : Byte ;
        {Beginning column & row, ASCII value of current characters, etc.}
        Offset, Offset1 : Integer ;
        Old_Length : Byte ;
        Input_Pointer : Input_P ;
        Wid : integer ;
     end ;

var Static_Input : pStatic_Input ;
    Max_Saved_Input_Lines : integer ;

{$IFDEF Debug}
procedure Check_Window( Window : PWindow ) ;

begin
    if( Window = nil ) then
    begin
        writeln( 'Reference to nil window' ) ;
        halt ;
    end ;
    if( Window^.Sanity <> 'Window' ) then
    begin
        writeln( 'Invalid typecast or reference to deleted window' ) ;
        halt ;
    end ;
    if( Window^.Columns * Window^.Rows > Window^.Size ) then
    begin
        writeln( 'corruption' ) ;
        halt ;
    end ;
end ;
{$ENDIF}


procedure Fill_Data( Window_P : PWindow ; Start, Size : longint ; Value : byte ) ;

begin
    Window_P^.Contents^.Fill_Data( Start, Size, Value ) ;
end ;


procedure Fill_Attributes( Window_P : PWindow ; Start, Size : longint ;
    Value : byte ) ;

begin
    Window_P^.Contents^.Fill_Attributes( Start, Size, Value ) ;
end ;


function GetTextExtent( Window_P : PWindow ; String_P : PChar ;
    Count : integer ) : longint ;

{$IFDEF WINDOWSORDELPHI}
var
{$ENDIF}
{$IFDEF WINDOWS}
    DC : HDC ;
{$ENDIF}
{$IFDEF DELPHI}
    Size : TSize ;
{$ENDIF}

begin
{$IFDEF Debug}
    Check_Window( Window_P ) ;
{$ENDIF}
{$IFDEF WINDOWS}
    DC := GetDC( Window_P^.ID ) ;
    SelectObject( DC, GetStockObject( ANSI_FIXED_FONT ) ) ;
{$IFDEF DELPHI}
    Windows.GetTextExtentPoint32( DC, String_P, Count, Size ) ;
    GetTextExtent := Size.cy shl 16 or Size.cx ;
{$ELSE}
    GetTextExtent := WinProcs.GetTextExtent( DC, String_P, Count ) ;
{$ENDIF}
    ReleaseDC( Window_P^.ID, DC ) ;
{$ELSE}
    GetTextExtent := $10000 or Count ; { Height 1, length is count }
{$ENDIF}
end ;


function Get_Screen( Row, Col : Integer ) : Byte ;

begin
    Get_Screen := _Get_Screen( Row, Col, Screen.Data.Bytes[ Row, Col ] ) ;
end ;


function Get_Text_Extent( Window_P : PWindow ; String_P : string ;
    Count : integer ) : longint ;

var P : PChar ;

begin
    getmem( P, length( String_P ) + 1 ) ;
    strPCopy( P, String_P ) ;
    Get_Text_Extent := GetTextExtent( Window_P, P, Count ) ;
    freemem( P, length( String_P ) + 1 ) ;
end ;


function RGBs_Equal( const RGB1, RGB2 : TRGB ) : boolean ;

begin
    RGBs_Equal := ( RGB1.R = RGB2.R ) and ( RGB1.G = RGB2.G ) and ( RGB1.B = RGB2.B ) and ( RGB1.A = RGB2.A ) ;
end ;


procedure Put_Char( Window_P : PWindow; Offset : integer ;
    Data_Byte, Attribute : byte ) ;

begin
    if( Offset < 0 ) then
    begin
	exit ;
    end ;
{$IFDEF Debug}
    Check_Window( Window_P ) ;
{$ENDIF}
    if( ( Window_P^.Position_Row < 1 ) or
	( Window_P^.Position_Row > Window_P^.Rows ) ) then
    begin
	exit ;
    end ;
    if( ( Window_P^.Position_Column < 1 ) or
	( Window_P^.Position_Column > Window_P^.Columns ) ) then
    begin
	exit ;
    end ;
    if(
        ( Window_P^.Contents^.Get_Attribute( Offset ) <> Attribute )
        or
	( Window_P^.Contents^.Get_Data_Byte( Offset ) <> Data_Byte )
        or
	( not RGBs_Equal( Window_P^.Contents^.Get_Background_Color( Offset ), Window_P^.Background ) )
        or
	( not RGBs_Equal( Window_P^.Contents^.Get_Foreground_Color( Offset ), Window_P^.Foreground ) )
      ) then
    begin
	if( Offset < Window_P^.Low_Update ) then
	begin
	    Window_P^.Low_Update := Offset ;
	end ;
	if( Offset > Window_P^.High_Update ) then
	begin
	    Window_P^.High_Update := Offset ;
	end ;
    end ;
    Window_P^.Contents^.Set_Data_Byte( Offset, Data_Byte ) ;
    Window_P^.Contents^.Set_Attribute( Offset, Attribute ) ;
    Window_P^.Contents^.Set_Background_Color( Offset, Window_P^.Background ) ;
    Window_P^.Contents^.Set_Foreground_Color( Offset, Window_P^.Foreground ) ;
end ; { Put_Char }


procedure Get_Window_Size( Window_P : PWindow ; var R, C : integer ) ;

{$IFDEF WINDOWS}
var Rect : TRect ;
{$ENDIF}

begin
    if( Window_P = nil ) then { Return screen information }
    begin
{$IFDEF WINDOWS}
        R := GetSystemMetrics( SM_CYFULLSCREEN ) ;
        C := GetSystemMetrics( SM_CXFULLSCREEN ) ;
{$ELSE}
	R := Max_Row ;
	C := Max_Column ;
{$ENDIF}
    end else
    begin
{$IFDEF Debug}
        Check_Window( Window_P ) ;
{$ENDIF}
{$IFDEF WINDOWS}
        GetClientRect( Window_P^.ID, Rect ) ;
        R := Rect.Bottom ;
        C := Rect.Right ;
{$ELSE}
 	R := Window_P^.Rows ;
	C := Window_P^.Columns ;
{$ENDIF}
    end ;
end ;


function Current_Window ;

begin
    Current_Window := Screen_Root.Default_P ;
    { Return pointer for current window }
end ;


procedure Resize( var Window_P : PWindow ; New_Rows, New_Columns : integer ) ;
{ Resize window to specified size }

var New_Window : PWindow ;
    Max_Rows, Max_Columns : integer ; { Maximum data to copy }
    Current_Row : integer ; { Loop controls }
{$IFDEF WINDOWS}
    Size : integer ;
{$ENDIF}

begin
{$IFDEF Debug}
    Check_Window( Window_P ) ;
{$ENDIF}
    { Do nothing if no change }
    if( ( New_Rows = Window_P^.Rows ) and ( New_Columns = Window_P^.Columns ) ) then
    begin
        exit ;
    end ;

{$IFDEF WINDOWS}
    New_Window := Create_Window_Wrapper( Window_P^.ID ) ;
    Size := New_Rows * New_Columns ;
    New_Window^.Contents^.Set_Size( Size ) ;
    Fill_Data( New_Window, 0, Size, 32 ) ; { Space fill }
    Fill_Attributes( New_Window, 0, Size, Attribute_Default ) ; { Null attributes }
    New_Window^.Contents^.Fill_Background_Colors( 0, Size - 1, New_Window^.Background ) ;
    New_Window^.Contents^.Fill_Foreground_Colors( 0, Size - 1, New_Window^.Foreground ) ;
    New_Window^.Rows := New_Rows ;
    New_Window^.Columns := New_Columns ;
{$ELSE}
    New_Window := Create_Window( New_Rows, New_Columns, Window_P^.Flags ) ;
{$ENDIF}
    New_Window^.Level := Window_P^.Level ;
    New_Window^.Attribute := Window_P^.Attribute ;
    New_Window^.Map_Row := Window_P^.Map_Row ;
    New_Window^.Map_Column := Window_P^.Map_Column ;
    New_Window^.Map_Rows := Window_P^.Map_Rows ;
    New_Window^.Map_Columns := Window_P^.Map_Columns ;
    New_Window^.View_Row := Window_P^.View_Row ;
    New_Window^.View_Column := Window_P^.View_Column ;
    New_Window^.Flags := Window_P^.Flags ;
    if(
        ( Window_P^.Position_Row <= Window_P^.Rows )
        or
        ( Window_P^.Position_Column <= Window_P^.Columns )
      ) then
    begin
        New_Window^.Position_Column := Window_P^.Position_Column ;
        New_Window^.Position_Row := Window_P^.Position_Row ;
    end ;

    { Determine maximum amount of data to copy }
    if( New_Rows > Window_P^.Rows ) then
    begin
        Max_Rows := Window_P^.Rows ;
    end else
    begin
        Max_Rows := New_Rows ;
    end ;
    if( New_Columns > Window_P^.Columns ) then
    begin
        Max_Columns := Window_P^.Columns ;
    end else
    begin
        Max_Columns := New_Columns ;
    end ;

    { Copy old data to new window, then delete the old window... }
    for Current_Row := 1 to Max_Rows do
    begin
        New_Window^.Contents^.Copy( Window_P^.Contents,
            ( Current_Row - 1 ) * Window_P^.Columns,
            ( Current_Row - 1 ) * New_Window^.Columns, Max_Columns ) ;
    end ;
    Delete_Window( Window_P ) ;

    { Return new handle }
    Window_P := New_Window ;
end ; { Resize }


function Window_Mapped ;
{ Return TRUE/FALSE regarding whether or not window is mapped }

begin
{$IFDEF Debug}
    Check_Window( Window_P ) ;
{$ENDIF}
    if( Window_P^.Map_Row > 0 ) then
    begin
        Window_Mapped := True ;
    end else
    begin
        Window_Mapped := False ;
    end ;
end ;


procedure Get_Default_Window( var X : Window_Record ) ;

begin
    X.Pointer := Screen_Root.Default_P ;
    X.Id := Screen_Root.Default ; { Return default window }
end ;


procedure Set_Default_Window( X : Window_Record ) ;

begin
    Screen_Root.Default_P := X.Pointer ;
    Screen_Root.Default := X.Id ; {Set default window}
end ;


function Get_Model : Byte ; {Return system type}

begin
    Get_Model := Screen_Root.Model
end ;


procedure End_SMU ; { Nicely terminate SMU }

{$IFNDEF DELPHI}
var Reg : Registers ;
{$ENDIF}

begin
    if( Printer0 = nil ) then { Already terminated/never started }
    begin
        exit ;
    end ;

    { Position cursor to bottom of screen }
{$IFNDEF DELPHI}
    if( Screen_Root.Model = Model_Rainbow ) then {Rainbow}
    begin
	Write( '[24;1H' ) ; {Go to line 24, column 1}
	Reg.Di := 10 ;
	Intr( $18, Reg ) ; {Enable cursor}
	if Screen_Root.Video_Mode <> 7 then
	begin {In graphics mode}
	    Port[ $53 ] := $Bf ; {Access mode via indirect}
	    Port[ $51 ] := $31 ; {Done}
	    Port[ $0A ] := $83 {Enable VT102}
	end
    end else
{$ENDIF}
    begin
        if( Screen_Root.Model <> Model_Unknown ) then {IBM}
        begin
            IBM_Pos( Max_Row, 1 ) ;
        end ;
    end ;

    { Free allocated memory... }
    freemem( Printer0, Sizeof( Block ) ) ;
    freemem( Printer1, Sizeof( Block ) ) ;
    freemem( Conversion, Sizeof( Block ) ) ;
    freemem( Deconversion, Sizeof( Block ) ) ;
    Printer0 := nil ;
    Printer1 := nil ;
    Conversion := nil ;
    Deconversion := nil ;
end ; { End_SMU }


procedure Init_Screen ; { Initialize depending upon hardware model }

var R : Integer ;
{$IFNDEF DELPHI}
    Reg : Registers ;
{$ENDIF}

begin
     Screen_Root.Video_Mode := VM_Text ; {Text is default}
     Set_Video_Mode( VM_Text ) ;
     Screen_Root.Screen_Width := 80 ; {Default screen width}
     Screen_Root.Foreground := 0 ;
     Screen_Root.Background := 0 ;
     Screen_Root.Flags := 0 ;
     Attribute_Default := 7 ;
{$IFNDEF DELPHI}
     Case Screen_Root.Model of
	 Model_Rainbow : {Rainbow}
	     begin
		 Attribute_Default := 14 ;
		 for R := 1 To 24 do Row_Start[ R ] := ( R - 1 ) * 135 + 16 ;
		 Write( '[?3l' ) ;
		 {Reset screen to set up screen RAM properly}
		 Reg.Di := 8 ;
		 Intr( $18, Reg ) {Disable cursor}
	     end ;
	 Model_PC_Monochrome..Model_PC_Color_Text : {IBM PC in text mode}
		begin
		    Attribute_Default := $07 ;
		    for R := 1 To 24 do Row_Start[ R ] := ( R - 1 ) * 80 ;
		    Reg.Ax := $0500 ;
		    {Set page to 0 (which initializes and clears screen)}
		    Intr( $10, Reg ) ; {do BIOS call}
		    Reg.Ax := $0200 ;
		    Reg.Bx := 0 ;
		    Reg.Dx := 81 ;
		    Intr( $10, Reg ) ; {Disable cursor}
		end ;
     end ;
{$ENDIF}
     if Screen_Root.Model > 4 then {IBM PC graphics}
     begin
{$IFDEF DELPHI}
         Screen_Root.Video_Mode := VM_PC_Graphics ;
{$ELSE}
	 Reg.Ax := $0F00 ;
	 Intr( $10, Reg ) ; {Get video mode}
	 Screen_Root.Video_Mode := Lo( Reg.Ax ) ;
{$ENDIF}
         Set_Video_Mode( Screen_Root.Video_Mode ) ;
     end ;

     for R := 0 to 255 do Conversion^[ R ] := R ;
     if Screen_Root.Model = Model_Rainbow then {Rainbow}
	     begin
		 Conversion^[ 255 ] := 32 ;
		 for R := 32 To 126 do Conversion^[ R ] := R ;
		 for R := 1 To 31 do Conversion^[ R ] := Rainbow_Cvt[ R ] ;
		 for R := 127 To 254 do Conversion^[ R ] := Rainbow_Cvt1[ R ]
	     end ;
     for R := 0 To 254 do Deconversion^[ Conversion^[ R ] ] := R ;
     {Create reverse conversion array}
end ; { Init_Screen }


procedure Init_Printer ; { Initialize the printer }

var Dummy : Integer ;

begin
    for Dummy := 0 To 255 do Printer0^[ Dummy ] := Dummy ;
    for Dummy := 0 To 255 do Printer1^[ Dummy ] := 0 ; {Clear flags}
    Case Screen_Root.Printer_Type of
	2 : {LA50/LA75}
	    begin
		Write( Printer_File, '*<+0' ) ; {Setup the printer}
		for Dummy := 1 To 31 do Printer0^[ Dummy ] := La50_Cvt[ Dummy ] ;
		for Dummy := 127 To 254 do Printer0^[ Dummy ] := La50_Cvt1[ Dummy ] ;
		for Dummy := 1 To 31 do Printer1^[ Dummy ] := La50_Cvt_F[ Dummy ] ;
		for Dummy := 127 To 254 do Printer1^[ Dummy ] := La50_Cvt1_F[ Dummy ]
	    end ;
	3 : {IBM Graphics}
	    for Dummy := 0 To 255 do Printer0^[ Dummy ] := Dummy ;
    end ;
end ;


{$IFDEF WINDOWS}
procedure Calculate_Window_Extent( Window_P : PWindow ; var R, C : longint) ;

var Rect : TRect ;

begin
{$IFDEF Debug}
    Check_Window( Window_P ) ;
{$ENDIF}
    GetClientRect( Window_P^.Id, Rect ) ;
    Window_P^.Text_Width := Get_Text_Extent( Window_P, 'W', 1 ) and 32767 ;
    Window_P^.Text_Height := Get_Text_Extent( Window_P, 'W', 1 ) shr 16 ;
    C := Rect.Right div Window_P^.Text_Width ;
    R := Rect.Bottom div Window_P^.Text_Height ;
end ;


function Create_Window_Wrapper( Handle : HWND ) : PWindow ;
{ Create a SMU wrapper for a MS Windows window }

var Window_P : PWindow ;
    R, C : longint ;
    Size : Integer ;

begin
    Create_Window_Wrapper := nil ;
    if( Screen_Root.Window_Count = 255 ) then
    begin
	Messagebox( 0, 'SMU Error', '?Too many windows', MB_OK ) ;
	Exit ;
    end ;
    inc( Screen_Root.Window_Count ) ; {One more window}

    Getmem( Window_P, sizeof( Window_P^) ) ; { Create window header }
    fillchar( Window_P^, sizeof( Window_P^ ), 0 ) ;
{$IFDEF Debug}
    Window_P^.Sanity := 'Window' ;
{$ENDIF}
    Window_P^.Id := Handle ;
    Calculate_Window_Extent( Window_P, R, C ) ;
    Size := R * C ;
    Window_P^.Rows := R ;
    Window_P^.Columns := C ;
    Window_P^.Contents := new( PSWindow_Contents, Init ) ;
    Window_P^.Contents^.Set_Size( Size ) ;
    Window_P^.Level := 0 ;
    Window_P^.Map_Row := 1 ;
    Window_P^.Map_Column := 1 ;
    Window_P^.View_Row := 1 ;
    Window_P^.View_Column := 1 ;
    Window_P^.Border := 0 ;
    Window_P^.Background := RGB_Black ;
    Window_P^.Foreground := RGB_White ;
(*
    if( not IsWindowVisible( Handle ) ) then
    begin
        Window_P^.Flags := 2 ; {Occluded}
    end else
*)
    begin
        Window_P^.Flags := 0 ;
    end ;

    { Link into window list }
    Window_P^.Next_Window := Screen_Root.Next_Window ;
    Screen_Root.Next_Window := Window_P ; {Insert window in list}

    Window_P^.Position_Row := 1 ; {Current character position}
    Window_P^.Position_Column := 1 ;
    Window_P^.Low_Update := 32767 ;
    Window_P^.High_Update := 0 ;
    Window_P^.Size := Size ;
    Window_P^.Attribute := Attribute_Default ;
    Fill_Data( Window_P, 0, Size, 32 ) ; { Space fill }
    Fill_Attributes( Window_P, 0, Size, Attribute_Default ) ;
    Window_P^.Contents.Fill_Background_Colors( 0, Size - 1, Window_P^.Background ) ;
    Window_P^.Contents.Fill_Foreground_Colors( 0, Size - 1, Window_P^.Foreground ) ;
    { Now window is initialized }

    Create_Window_Wrapper := Window_P ;
end ;
{$ENDIF}


function Create_Window( Row_Count, Column_Count, Flag : Integer ) : PWindow ;
{ Create a window }

var Wid : Byte ;
{$IFNDEF WINDOWS}
    Window_P : PWindow ;
    A, Size : Integer ;
    Wida : Array[ 1..255 ] of Boolean ;
    Loop : longint ;
{$ENDIF}

begin
{$IFDEF WINDOWS}
    WID := CreateWindow( '', '', 0, 0, 0,
        Column_Count, Row_Count, 0, 0, HInstance, nil ) ;
    Create_Window := Create_Window_Wrapper( WID ) ;
{$ELSE}
    Create_Window := nil ;
    Size := Row_Count * Column_Count ;
    if( Screen_Root.Window_Count = 255 ) then
    begin
	Writeln( '?Too many windows' ) ;
	Exit ;
    end ;
    A := 0 ;
    if( ( Size < Row_Count ) or ( Size < Column_Count ) ) then
    begin
	Writeln( '?Window too large' ) ;
	Exit
    end ;
    fillchar( WIDA, sizeof( WIDA ), 0 ) ; { Prepare for loop }
    Window_P := Screen_Root.Next_Window ; {Start with first window in list}
    while( Window_P <> nil ) do {Find available/used WIDs}
    begin
	 Wida[ Window_P^.Id ] := True ;
	 Window_P := Window_P^.Next_Window {Go to next window}
     end ;
     Wid := 0 ;
     A := 0 ;
     while( Wid = 0 ) do { Search until unused WID assigned }
     begin
	 A := A + 1 ;
	 if( not Wida[ A ] ) then
         begin
             Wid := A ; { if unused WID }
         end ;
     end ;
     inc( Screen_Root.Window_Count ) ; {One more window}

     Getmem( Window_P, sizeof( Window_P^ ) ) ; { Create window header }
     fillchar( Window_P^, sizeof( Window_P^ ), 0 ) ;
     Window_P^.Contents := new( PSWindow_Contents, Init ) ;
     Window_P^.Contents.Set_Size( Size ) ;
     Window_P^.Rows := Row_Count ;
     Window_P^.Columns := Column_Count ;
     Window_P^.Flags := Flag or 2 ; { Occluded }
     Window_P^.Id := Wid ;
     Window_P^.Next_Window := Screen_Root.Next_Window ;
     Screen_Root.Next_Window := Window_P ; {Insert window in list}
     Window_P^.Position_Row := 1 ; {Current character position}
     Window_P^.Position_Column := 1 ;
     Window_P^.Low_Update := 32767 ;
     Window_P^.Size := Size ;
     Window_P^.Attribute := Attribute_Default ;
     Window_P^.Background := RGB_Black ;
     Window_P^.Foreground := RGB_White ;
     Window_P^.Text_Width := Get_Text_Extent( Window_P, 'W', 1 ) ;
     Window_P^.Text_Height := Get_Text_Extent( Window_P, 'W', 1 ) shr 16 ;
     Fill_Data( Window_P, 0, Size, 32 ) ; { Space fill }
     Fill_Attributes( Window_P, 0, Size, Attribute_Default ) ; { Null attributes }
     for Loop := 0 to Size - 1 do
     begin
         Window_P^.Contents^.Set_Background_Color( Loop, RGB_Black ) ;{Background_Color ) ;}
         Window_P^.Contents^.Set_Foreground_Color( Loop, RGB_White ) ;{Foreground_Color ) ;}
     end ;
     { Now window is initialized! }

     Create_Window := Window_P ;
{$ENDIF}
end ; { Create_Window }


procedure Do_Unmap( WP : PWindow ) ; { Physical unmap of a window }

var A, Offset : Integer ;
    Clear_Partial, Clear_Full : Boolean ;
    Last_Window : Byte ; {Last offset used in Window_* arrays}
    Level, Level_Window : Integer ;
    R, C : Integer ; {Row, column pointers}
    This_Window : PWindow ;
    Window_List : array[ 0..254 ] of PWindow ;
    Window_Flag : array[ 0..254 ] of Byte ;
    PWindow0 : PWindow ;

begin
{$IFDEF Debug}
    Check_Window( WP ) ;
{$ENDIF}
    { Setup... }
    fillchar( Window_List[ 0 ], sizeof( Window_List ), 0 ) ;
    fillchar( Window_Flag[ 0 ], sizeof( Window_Flag ), 0 ) ;

    { Build Window_List[]... }
    PWindow0 := Screen_Root.Next_Window ; { Start with first window defined }
    Last_Window := 0 ;
    while( PWindow0 <> nil ) do { Until done with all windows }
    begin
	if( PWindow0 <> Wp ) then { Not the same window we're unmapping }
	begin
	    if ( PWindow0^.Map_Row < Wp^.Map_Row + Wp^.Map_Rows ) and
		( PWindow0^.Map_Column < Wp^.Map_Column + Wp^.Map_Columns ) and
		( Wp^.Map_Row < PWindow0^.Map_Row + PWindow0^.Map_Rows )
		and ( Wp^.Map_Column <
		  PWindow0^.Map_Column + PWindow0^.Map_Columns ) then
	    begin { if window we are unmapping occludes another }
		Window_List[ Last_Window ] := PWindow0 ;
		Last_Window := Last_Window + 1 ;
		{ Build list of occluded windows }
	    end ;
	end ;
	PWindow0 := PWindow0^.Next_Window ;
	{ Follow pointer to next defined window }
    end ;

    { Unhide what unmapped window occluded }
    for R := Wp^.Map_Row To Wp^.Map_Row + Wp^.Map_Rows - 1 do {for all rows}
	for C := Wp^.Map_Column To Wp^.Map_Column + Wp^.Map_Columns - 1 do
	{ ...and columns }
	begin
	    Screen.Window.Bytes[ R, C ] := 0 ;
	    { Assume no window maps this cell anymore }
	    Level := 0 ;
	    Level_Window := -1 ;
	    if( Last_Window > 0 ) then
	      for A := 0 To Last_Window - 1 do {for all possible revealed windows}
	      begin
	       This_Window := Window_List[ A ] ; {Get pointer to next window}
	       if This_Window^.Level >= Level then {Closest to top so far}
		if This_Window^.Map_Row <= R then
		  if This_Window^.Map_Column <= C then
		    if This_Window^.Map_Column + This_Window^.Map_Columns > C then
		      if This_Window^.Map_Row + This_Window^.Map_Rows > R then
		      {and maps same row,column position}
		      begin
			Level := This_Window^.Level ;
			Screen.Window.Bytes[ R, C ] := This_Window^.Id ;
			Level_Window := A ;
		      end ;
	      end ;
	      {At this point, Level_Window is the window offset located at this
	       cell}
	    if Level_Window >= 0 then {if a window maps this cell}
	    begin
		Offset := ( R - Window_List[ Level_Window ]^.Map_Row +
		    Window_List[ Level_Window ]^.View_Row - 1 ) *
		    Window_List[ Level_Window ]^.Columns +
		    C - Window_List[ Level_Window ]^.Map_Column +
		    Window_List[ Level_Window ]^.View_Column - 1 ;
		if Window_List[ Level_Window ]^.Low_Update > Offset then
		    Window_List[ Level_Window ]^.Low_Update := Offset ;
		if Window_List[ Level_Window ]^.High_Update < Offset then
		    Window_List[ Level_Window ]^.High_Update := Offset ;
		Window_Flag[ Level_Window ] := 1 ; {Window occlusion has changed}
	    end ;
	end ;
	if( Last_Window = 0 ) then
	begin
	    exit ; { No revealed windows }
	end ;

	for A := 0 To Last_Window - 1 do if Window_Flag[ A ] > 0 then
	{Check for occulsion}
	begin
	    Window_List[ A ]^.Flags := Window_List[ A ]^.Flags or 18 ;
	    Clear_Partial := False ;
	    Clear_Full := False ;
	    for R := Window_List[ A ]^.Map_Row To
	      Window_List[ A ]^.Map_Row + Window_List[ A ]^.Map_Rows - 1 do
	      {for all rows}
		for C := Window_List[ A ]^.Map_Column To
		  Window_List[ A ]^.Map_Column + Window_List[ A ]^.Map_Columns - 1 do
		{and columns}
		    if Screen.Window.Bytes[ R, C ] <> Window_List[ A ]^.Id
			then
			    Clear_Partial := True {Not 100% unoccluded}
			else
			    Clear_Full := True ; {Not 100% occluded}
	    if Clear_Full then Window_List[ A ]^.Flags :=
		Window_List[ A ]^.Flags and ( Not 2 ) ;
	    if Clear_Partial then Window_List[ A ]^.Flags :=
		Window_List[ A ]^.Flags and ( Not 16 )
	end ;
end ; { Do_Unmap }


procedure Unmap_Window( Window_P : PWindow ) ;
{ Unmap a window from the physical screen }

begin
{$IFDEF Debug}
    Check_Window( Window_P ) ;
{$ENDIF}
    { Quick out... }
    if( ( Window_P^.Flags and 2 ) = 2 ) then
    begin
	exit ; { Window is 100% occluded }
    end ;

    { Clear flags... }
    Window_P^.Flags := Window_P^.Flags or 2 ; { Window is occluded }
    Window_P^.Flags := Window_P^.Flags and ( not 16 ) ; { Clear unoccluded flag }

    { Unmap it... }
    if( Window_P^.Map_Row > 0 ) then
    begin
	Do_Unmap( Window_P ) ; {Unmap, if needed}
    end ;
    if Screen_Root.Default_P = Window_P then {Unmapping default window}
    begin
	if( Screen_Root.Next_Window = nil ) then {No other windows that could be mapped}
	begin
	    Screen_Root.Default := 0 ;
	    Screen_Root.Default_P := nil ;
	end else {Map next window in list}
	begin
	    Screen_Root.Default := Screen_Root.Next_Window^.Id ;
	    Screen_Root.Default_P := Screen_Root.Next_Window ;
	end
    end ;
    Window_P^.Map_Row := 0 ;
    Window_P^.Map_Column := 0 ;
    Window_P^.Map_Rows := 0 ;
    Window_P^.Map_Columns := 0 ;
    Window_P^.View_Row := 0 ;
    Window_P^.View_Column := 0
    {Finish unmapping window}
end ;


procedure Map_Window( Window_P : PWindow ; Row, Row_Count, Column,
		     Column_Count, View_R, View_C, Level : Byte ) ;
{Map/remap a window to the physical screen}
{Row, Row_Count, Column, and Column_Count define the location of the viewport
 on the physical screen.  View_R and View_C define the origin of the viewport
 in the virtual screen (the row and column of the virtual screen that will
 appear in the upper-left-hand corner of the viewport)}

var C, R : Integer ;
    Collision, Occluded, Partial : Boolean ;
    Old_Default : Byte ;
    PWindow1 : PWindow ;
    Wid : Byte ;

begin
{$IFDEF Debug}
    Check_Window( Window_P ) ;
{$ENDIF}
    {The next section allows the current values to remain unchanged if 0 is
     passed for row, column, row and column counts, and/or view port stuff}
    if Row = 0 then Row := Window_P^.Map_Row ;
    if Column = 0 then Column := Window_P^.Map_Column ;
    if Row_Count = 0 then Row_Count := Window_P^.Map_Rows ;
    if Column_Count = 0 then Column_Count := Window_P^.Map_Columns ;
    if View_R = 0 then View_R := Window_P^.View_Row ;
    if View_C = 0 then View_C := Window_P^.View_Column ;

    Old_Default := Screen_Root.Default ;
    {Save default window for later comparison (unmap changes it if unmapping
     it)}
    Unmap_Window( Window_P ) ; {Unmap window, if it is mapped}
    Wid := Window_P^.Id ;
    if ( Screen_Root.Default = 0 ) or ( Old_Default = Wid ) then {Select this window}
    begin
	Screen_Root.Default := Wid ;
	Screen_Root.Default_P := Window_P ;
    end ;
    if ( Row > Max_Row ) or ( Column > Screen_Root.Screen_Width ) then
    begin
        Exit ; { Not on screen }
    end ;
    if Row + Row_Count > Max_Row + 1 then Row_Count := Max_Row - Row + 1 ;
    if Column + Column_Count > Screen_Root.Screen_Width + 1 then
	Column_Count := Screen_Root.Screen_Width - Column + 1 ;
    if View_R > Row_Count then View_R := Row_Count ;
    if View_C > Column_Count then View_C := Column_Count ;
    Collision := False ;

    Occluded := True ; {Assume this window is 100% occluded}
    Partial := False ;
    Window_P^.Flags := Window_P^.Flags or 18 ;

    for C := Column To Column + Column_Count - 1 do {for each column}
	for R := Row To Row + Row_Count - 1 do {for each row in column}
	    if Screen.Window.Bytes[ R, C ] = 0 then {if nothing else mapped}
	    begin
		Occluded := False ; {At least partially unoccluded}
		Screen.Window.Bytes[ R, C ] := Wid
	    end else {Something else mapped}
	    begin
		PWindow1 := Screen_Root.Next_Window ;
		while PWindow1^.Id <> Screen.Window.Bytes[ R, C ] do
		    PWindow1 := PWindow1^.Next_Window ;
		    {Locate window which is mapped to this cell}
		if PWindow1^.Level < Level
		    then {Newly mapped window occludes currently mapped}
			begin
			    Occluded := False ; {Not 100% occluded}
			    PWindow1^.Flags :=
				PWindow1^.Flags and ( Not 16 ) ;
				{Other window is partially occluded}
			    Screen.Window.Bytes[ R, C ] := Wid
			end
		    else {This window is occluded by something else}
			begin
			    Partial := True ; {At least partial occlusion}
			    if PWindow1^.Level = Level then
				Collision := True {Window collision}
			end
	    end ;
    if Collision then Writeln( '%Window collision' ) ;
    if Not Occluded then {Window is no longer 100% occluded}
    begin
    	Window_P^.Flags := Window_P^.Flags and ( Not 2 ) ;
    end ;
    Window_P^.Map_Row := Row ;
    Window_P^.Map_Column := Column ;
    Window_P^.Map_Rows := Row_Count ;
    Window_P^.Map_Columns := Column_Count ;
    Window_P^.View_Row := View_R ;
    Window_P^.View_Column := View_C ;
    Window_P^.Low_Update := 0 ;
    Window_P^.High_Update := Window_P^.Size - 1 ;
    Window_P^.Level := Level ;
    if Partial or Occluded then
    begin
        Window_P^.Flags := Window_P^.Flags and ( Not 16 ) ;
        { Not 100% unoccluded }
    end ;
end ;


procedure Delete_Window( Wp : PWindow ) ; { Delete a window }

var Last_Pointer, Window_P : PWindow ;

begin
    Unmap_Window( Wp ) ; { Unmap it }

    { Find window in window list... }
    Window_P := Screen_Root.Next_Window ;
    Last_Pointer := nil ;
    while( ( Window_P <> Wp ) and ( Window_P <> nil ) ) do
    begin
	Last_Pointer := Window_P ;
	Window_P := Window_P^.Next_Window ;
    end ;
    if( Window_P = nil ) then
    begin
        exit ;
    end ;

    { Remove window from list }
    if( Last_Pointer = nil ) then { First screen }
    begin
        Screen_Root.Next_Window := Window_P^.Next_Window ;
    end else
    begin
        Last_Pointer^.Next_Window := Window_P^.Next_Window ;
    end ;

    { Deallocate memory }
    dispose( Window_P^.Contents, Done ) ;
    Window_P^.Contents := nil ;
{$IFDEF Debug}
    Window_P^.Sanity := ' ' ;
{$ENDIF}
    Freemem( Window_P, sizeof( Window_P^ ) ) ; { Delete window header }
    dec( Screen_Root.Window_Count ) ;
end ; { Delete_Window }


procedure Write_Screen( L, C : Integer ) ;
{ Write text to physical display }

begin
    Clear_Cursor( 0, 0 ) ;
    _Write_Screen( Screen_Root.Default_P^.ID, L, C,
        Screen_Root.Default_P^.Text_Height, Screen_Root.Default_P^.Text_Width,
        Screen_Root.Default_P^.Border, Screen_Root.Driver,
        Screen_String, Attribute_String, Screen_Root.Default_P^.Contents^._Background_Colors_Pointer,
        Screen_Root.Default_P^.Contents^._Foreground_Colors_Pointer,
        Screen_Background_Offset, Screen_Foreground_Offset ) ;
    Set_Cursor( Window_To_Screen_Row( Screen_Root.Default_P, Screen_Root.Default_P^.Position_Row ),
        Window_To_Screen_Column( Screen_Root.Default_P, Screen_Root.Default_P^.Position_Column ) ) ;
end ;


procedure Update_Window( Window_P : PWindow ) ;
{ Update on screen what has changed on window }

label Done ;

var Columns : Integer ; {Columns in window}
    Contents : PSWindow_Contents ;
    Dummy, Dummy2 : Integer ;
    {Dummy variables used to speed up execution}
    Lr, Lc, R, C, R1, Vr, Vc : Integer ; {Row and column pointers}
    Low, High : Integer ; {Update range}
    Offset : Integer ; {Offset of current row - to reduce multiplies}
    Offset_Vc : Integer ; {Offset+VC}
    Unoccluded : Boolean ;
    Save_C, Save_R : integer ; { Saved row and column for cursor }
    Screen_Width : Byte ; {Width of screen}
    Stop : Integer ;
    Wid : integer ;
{
    Window_A_P, Window_D_P : Window_Def_Pointer ; {Used for speed}

begin
    if( Window_P = nil ) then
    begin
        Window_P := Screen_Root.Default_P ; { Get current window }
    end ;
{$IFDEF Debug}
    Check_Window( Window_P ) ;
{$ENDIF}
    if( ( Window_P^.Flags and 2 ) = 2 ) then
    begin
        Exit ; { Window is occluded }
    end ;
    Save_R := Window_P^.Position_Row ;
    Save_C := Window_P^.Position_Column ;
    Low := Window_P^.Low_Update ;
    High := Window_P^.High_Update ;
    if( Low > High ) then
    begin
        exit ; { Nothing to update }
    end ;
    Vr := Window_P^.View_Row ;
    Vc := Window_P^.View_Column ;
    Columns := Window_P^.Columns ; { Get offset to start of view row }
    Offset := ( Vr - 1 ) * Columns ;
    { Get offset to start of first (current) view port row }
    Offset_Vc := Offset + Vc - 1 ;
    { Offset to first (current) position in view port }

{$IFDEF Windows}
    LR := Window_P^.Rows ;
    C := Window_P^.Columns ;
    Screen_Width := C ;
    R := 1 ;
    Dummy := 1 ;
{$ELSE}
    R := Window_P^.Map_Row ; { First row in viewport }
    Screen_Width := Screen_Root.Screen_Width ;
    Lr := R + Window_P^.Map_Rows - 1 ; { Last row in viewport }
    if( Window_P^.Rows - Vr + 1 < Window_P^.Map_Rows ) then
    begin
        Lr := Window_P^.Rows - Vr + R ;
    end ;
    C := Window_P^.Map_Columns ; { Width }
    if( Window_P^.Columns - Vc + 1 < Window_P^.Map_Columns ) then
    begin
	C := Window_P^.Columns - Vc + Window_P^.Map_Column ;
	Screen_Width := C ;
        { Handle viewport which goes past edge of window }
    end ;
    Dummy := Window_P^.Map_Column ;
{$ENDIF}

    Screen_String[ 0 ] := Chr( C ) ;
    Attribute_String[ 0 ] := Chr( C ) ;
    Wid := Window_P^.Id ;
{$IFNDEF WINDOWS}
    if( ( ( Window_P^.Flags and 16 ) = 16 ) and ( Screen_Root.Model <> Model_Unknown ) )
	then
	    Unoccluded := True
	else
	    Unoccluded := False ;
{$ELSE}
    Unoccluded := True ;
{$ENDIF}
    Contents := Window_P^.Contents ;
{
    Window_D_P := Window_P^.Data_Pointer ;
    Window_A_P := Window_P^.Attribute_Pointer ;
}
    R1 := R ;
    while( R1 <= Lr ) do { for each row in view port }
    begin
        Screen_Background_Offset := Offset_VC ;
        Screen_Foreground_Offset := Offset_VC ;
	if Unoccluded then { 100% unoccluded and not ANSI.SYS }
        begin
            if( Offset_Vc > High ) then
            begin
                goto Done ; { if we're already past the last updated character }
            end ;
            if( Offset_Vc + Window_P^.Columns >= Low ) then
            { Past beginning of updated area }
            begin
                Move( Contents^._Data_Pointer^.Chars[ Offset_Vc ], Screen_String[ 1 ], C ) ;
                Move( Contents^._Attribute_Pointer^.Chars[ Offset_Vc ], Attribute_String[ 1 ], C ) ;
                { Update virtual window }
                Write_Screen( R1, Dummy ) ;
                { Write row to physical screen }
            end ;
        end else { Window is partially occluded }
        begin
            C := Window_P^.Map_Column ; { Physical screen column }
            if( Offset_Vc > High ) then
            begin
                goto Done ;
                { All done if we're already past the last updated character }
            end ;
            if( Offset_Vc + Window_P^.Map_Columns >= Low ) then
            { if we're up to first updated character }
                  while( C <= Screen_Width ) do
                  { Until we hit the right side of the screen }
                  begin
                    if ( Screen.Window.Bytes[ R1, C ] = Wid ) and
                     ( C <= Screen_Root.Screen_Width ) and
                     ( Offset_Vc + C > Low ) and
                     ( Offset_Vc + C - 1 <= High )
                      then {if within the updated area}
                        begin
                          Stop := C ;
                          while ( Screen.Window.Bytes[ R1, Stop ] = Wid ) and
                            ( Stop <= Screen_Root.Screen_Width ) and
                            ( Offset_Vc + Stop > Low ) and
                            ( Offset_Vc + Stop - 1 <= High ) do Inc( Stop ) ;
                            {Find edge of changed window on screen for
                             this line}
                          Dec( Stop ) ; {Don't go past this window}
                          Lc := Stop - C + 1 ;
                          Screen_String[ 0 ] := Chr( Lc ) ;
                          {LC=Number of characters to update}
                          Dummy2 := Vc + C - 1 ; {Screen column}
                          Dummy := Offset + C - Vc ; {Window offset}
                          Move( Contents^._Data_Pointer^.Chars[ Dummy ],
                              Screen_String[ 1 ], Lc ) ;
                          Move( Contents^._Attribute_Pointer^.Chars[ Dummy ],
                              Attribute_String[ 1 ], Lc ) ;
                          if Screen_Root.Model = Model_Unknown then {ANSI.SYS}
                          begin
                              Move( Contents^._Data_Pointer^.Chars[ Dummy ],
                                      Screen.Data.Chars[ R1, Dummy2 ], Lc ) ;
                          end ;
                          Write_Screen( R1, C ) ;
                          C := Stop ;
                        end ;
                        Inc( C ) ;
                  end ;
                  Offset := Offset + Columns
                  { Increment pointer into window }
          end ;
          Offset_Vc := Offset_Vc + Columns ; { Offset into virtual screen }
          Inc( R1 ) ;
    end ; { while( R1 <= Lr ) }
Done:
    Window_P^.Low_Update := 32767 ;
    Window_P^.High_Update := 0 ;
    { Nothing needs updates anymore }

    if( Window_P = Screen_Root.Default_P ) then
    begin
	Set_Cursor( Window_To_Screen_Row( Window_P, Save_R ),
            Window_To_Screen_Column( Window_P, Save_C ) ) ;
            { Restore original cursor position }
    end else
    begin
	Window_P^.Position_Row := Save_R ;
	Window_P^.Position_Column := Save_C ;
    end ;
end ; { Update_Window }


procedure Invalidate_Window( Window_P : PWindow ) ;

begin
    if( Window_P = nil ) then
    begin
        Window_P := Screen_Root.Default_P ; { Get current window }
    end ;
{$IFDEF Debug}
    Check_Window( Window_P ) ;
{$ENDIF}
    Window_P^.High_Update := 32767 ;
    Window_P^.Low_Update := 0 ;
end ;


procedure Update_Screen ; {Update all windows}

var Window_P : PWindow ;

begin
    Window_P := Screen_Root.Next_Window ; { Get first window }
    while( Window_P <> nil ) do { Until end of windows }
    begin
	Update_Window( Window_P ) ;
	Window_P := Window_P^.Next_Window
    end
end ;


procedure Refresh_Window( Window_P : PWindow ) ;

begin
{$IFDEF Debug}
    Check_Window( Window_P ) ;
{$ENDIF}
    if( ( Window_P^.Flags and 2 ) = 2 ) then
        exit ; { Window is occluded }
    Window_P^.Low_Update := 0 ;
    Window_P^.High_Update := Window_P^.Size - 1 ;
    Update_Window( Window_P )
end ;


procedure Refresh_Screen ; { Refresh physical screen }

var P : PWindow ;

begin
    P := Screen_Root.Next_Window ; { Get first window }
    while( P <> nil ) do { for each existing window }
    begin
	Refresh_Window( P ) ;
	P := P^.Next_Window
    end
end ;


procedure Text_Out( Window_P : PWindow ; T : string ; L : integer ) ;

var C : Integer ;
    Attribute : Byte ;
    Ch, Ch1 : Char ;
    Columns : Integer ; { Window column count }
    Contents : PSWindow_Contents ;
    Dummy, Dummy1 : Integer ; { Dummy variables to speed execution }
    Do_Update : Boolean ;
    Offset : Integer ;
    Ord_Ch, Ord_Ch1 : Byte ;
    Out_Of_Bounds : boolean ;
    Rows, Flags : Integer ;
    Size : Integer ;
{
    Window_A_P, Window_D_P : Window_Def_Pointer ; { for speed }
    Wrap : Boolean ;

    procedure Handle_LF ; { Handling of Linefeed }

    var B : integer ;
        Offset1 : Integer ;

    begin
        Do_Update := True ; { Update potentially needed }
        Inc( Window_P^.Position_Row ) ; { Increment row }
        if( Window_P^.Position_Row > Rows ) then { Past end of window }
        begin
            Window_P^.Low_Update := 0 ; { Refresh from beginning of window }
            if( ( Flags and 1 ) = 0 ) then { Scroll}
            begin
                Offset1 := Size - Columns ;
                { Number of bytes from origin to end of next-to-last line }

                Contents^.Copy( Contents, Columns, 0, Offset1 ) ;
                { Do scroll }

                B := Size - Columns ; { Beginning of last line }
                Fill_Data( Window_P, B, Columns, 32 ) ; { Space fill }
                Fill_Attributes( Window_P, B, Columns, Attribute ) ;
                Window_P^.Contents^.Fill_Background_Colors( B, Columns, Window_P^.Background ) ;
                Window_P^.Contents^.Fill_Foreground_Colors( B, Columns, Window_P^.Foreground ) ;
                { Zero last line }

                Window_P^.High_Update := Size - 1 ; { End of window }
                Window_P^.Position_Row := Rows ;
                { Current row is last row of window }
                Offset := B ; { New offset }
                Dec( Input_Row ) ;
                if( Input_Row < 1 ) then
                begin
                    Input_Row := 1 ;
                    Input_Column := 1
                end ;
                Exit ;
            end else { No scroll }
            begin
                Window_P^.Position_Row := 1 ; { To top of window }
                Offset1 := Window_P^.Position_Column - 1 ; { To top row }
                if( Offset1 > Window_P^.High_Update ) then
                begin
                    Window_P^.High_Update := Offset1 ;
                end ;
            end ;
        end else { Not past end of window }
        begin
            Offset := Offset + Columns ; { Move to next line }
        end ;
    end ; { Text_Out.Handle_LF }

begin { Text_Out }
    if( L < 0 ) then { Output entire passed string }
    begin
	L := length( T ) ;
    end ;
    if( L < 1 ) then { Nothing to output }
    begin
	exit ;
    end ;
    if( Window_P = nil ) then
    begin
	write( copy( T, 1, L ) ) ;
	exit ;
    end ;
{$IFDEF Debug}
    Check_Window( Window_P ) ;
{$ENDIF}
    Rows := Window_P^.Rows ;
    Columns := Window_P^.Columns ; { Get columns in window }
    Size := Window_P^.Size ; { Get size of window }
    Flags := Window_P^.Flags ;
    if( ( Flags and 4 ) = 0 ) then
    begin
        Wrap := True ;
    end else
    begin
        Wrap := False ;
    end ;
    Attribute := Window_P^.Attribute ;
    Contents := Window_P^.Contents ;
{
    Window_D_P := Window_P^.Data_Pointer ;
    Window_A_P := Window_P^.Attribute_Pointer ;
}
    Do_Update := False ;

    { Determine offset in window... }
    Offset := ( Window_P^.Position_Row - 1 ) * Columns ;
    Offset := Offset + Window_P^.Position_Column - 1 ;

    C := 1 ;
    while( C <= L ) do { Process each character in output }
    begin
	Ch := T[ C ] ; { Get next character }
	Ord_Ch := Byte( Ch ) ;
	if( Screen_Root.Printer_Mode = 1 ) then { Print mode active }
	begin
	    Ch1 := Char( Printer0^[ Ord_Ch ] ) ; { Translate }
	    if( ( Ord_Ch < 32 ) or ( Ord_Ch > 126 ) ) then { Special character }
	    begin
		if( Printer1^[ Ord_Ch ] = 1 ) then
		begin
		    OS^.Print( #27 ) ;
		    OS^.Print( 'O' ) ;
		end ;
		OS^.Print( Ch1 ) ;
	    end ;
	end ;
	if( Window_P = nil ) then
	begin
	    Write( Ch ) ; { No current window defined }
	end else
	begin
            Out_Of_Bounds :=
              (
                ( Window_P^.Position_Row > Rows )
                or
                ( Window_P^.Position_Column > Columns )
              ) ; { Beyond bounds of window? }
	    Ord_Ch1 := Conversion^[ Ord_Ch ] ;
	    if( Print_Characters[ Ord_Ch ] ) then { if a printing character }
	    begin
		if( Ord_Ch = 9 ) then { Tab }
		begin
		    Dummy := ( Window_P^.Position_Column + 8 ) and ( Not 7 ) ;
		    for Dummy1 := Window_P^.Position_Column To Dummy - 1 do
		    begin
			if( ( Dummy1 <= Columns ) and ( not Out_Of_Bounds ) ) then
			{ if not past right side of window }
			begin
			    Put_Char( Window_P, Offset + Dummy1 -
				Window_P^.Position_Column, 32, Attribute ) ;
			    { Put character on window }
			end ;
		    end ;
		    Offset := Offset + Dummy - Window_P^.Position_Column ;
		    Window_P^.Position_Column := Dummy ;
		    { Move to next tab position }
		end ;
		{ Move to next tab column.  Tab stops are every 8 columns }
		if( Window_P^.Position_Column > Columns ) then
		{ Past right side of window }
		begin
                    if( Wrap ) then { Wrap }
                    begin
                        if( Window_P^.Position_Row > Rows ) then { end of window }
                        begin
                            Window_P^.Position_Column := 0 ;
                        end else
                        begin
                            Window_P^.Position_Column := 1 ;
                        end ;
                        { Go back to first column }
                        Handle_LF ;
                        { Implied linefeed.  This also determines the new
                          offset. }
                    end else
                    begin
                        Window_P^.Position_Column := Columns + 1 ;
                        Offset := ( Window_P^.Position_Row - 1 ) * Columns ;
                    end ;
                    { Don't get more than one position beyond end of window }
		end ;
		if(
                    ( Window_P^.Position_Column <= Columns )
                    and
                    ( Window_P^.Position_Row <= Rows )
                    and
                    ( Ord_Ch <> 9 )
                  ) then { if not past right side of window and not a tab }
                  begin
                      Put_Char( Window_P, Offset, Ord_CH1, Attribute ) ;
                      { Put character on window }
                      Inc( Offset ) ;
                      Inc( Window_P^.Position_Column ) ;
                      { Increment column count for printing characters }
                  end ;
	    end else { Non-printing character }
	    begin
 		if( Ord_Ch = 7 ) then { BELL }
		begin
		    OS^.Beep( 0 ) ;
		end ;
		if( Ord_Ch = 8 ) then { Backspace }
                begin
                    if( Window_P^.Position_Column > 1 ) then { Not at start of line }
                    begin
                        Dec( Offset ) ;
                        Dec( Window_P^.Position_Column ) { Back up one column }
                    end ;
                end ;
		if( Ord_Ch = 254 ) then { Special backspace }
                begin
		    if( Window_P^.Position_Column > 1 ) then { Not at start of line }
                    begin
                        Dec( Offset ) ;
                        Dec( Window_P^.Position_Column ) ;
                        { Back up one column }
                    end else { At start of line }
                    if( Window_P^.Position_Row > 1 ) then { Not at home position }
                    begin
                        Dec( Offset ) ;
                        Dec( Window_P^.Position_Row ) ;
                        Window_P^.Position_Column := Columns ;
                    end ;
                end ;
		if( Ord_Ch = 13 ) then { Return }
		begin
		    Window_P^.Position_Column := 1 ;
		    { Move to first column of current line }
		    Offset := ( Window_P^.Position_Row - 1 ) * Columns ;
		end ;
	    end ;
	    if( Ord_Ch = 10 ) then
	    begin
		Handle_Lf ;
	    end ;
	end ;
	Inc( C ) ;
    end ;
    if( ( Flags and 8 ) = 0 ) then { Allow autoupdate }
    begin
        if( Do_Update ) then { Something changed on the screen }
        begin
            Update_Window( Window_P ) ; { Update if needed }
        end else
        begin
            Set_Cursor( Window_To_Screen_Row( Window_P, Window_P^.Position_Row ),
                Window_To_Screen_Column( Window_P, Window_P^.Position_Column ) ) ;
        end ;
    end ;
end ; { Text_Out }


{$F+}
function Window_Output( var F : Text ) : Integer ; { Output to current window }

var Window_P : PWindow ;

begin
    Window_Output := 0 ;
{$IFDEF WINDOWSORDELPHI}
    with tTextrec( F ) do SetLength( Output_Buffer, Bufpos ) ;
{$ELSE}
    With Textrec( F ) do Output_Buffer[ 0 ] := Chr( Bufpos ) ;
    {Size the output string}
{$ENDIF}
    Window_P := Screen_Root.Default_P ; {Find window}
    if( Window_P = nil ) then
    begin
	write( Output_Buffer ) ;
    end else
    begin
	Text_Out( Window_P, Output_Buffer, length( Output_Buffer ) ) ;
    end ;
{$IFDEF WINDOWSORDELPHI}
    with tTextrec( F ) do Bufpos := 0 ;
{$ELSE}
    with Textrec( F ) do Bufpos := 0 ;
{$ENDIF}
end ; {Window_Output}


function Scrnull( var F : Text ) : Integer ;

begin
    Scrnull := 0 ;
end ;


procedure Assignscr( var F : Text ) ;

begin
{$IFDEF WINDOWSORDELPHI}
    With tTextrec( F ) do
{$ELSE}
    With Textrec( F ) do
{$ENDIF}
    begin
	Mode := Fmclosed ;
	Bufsize := 255 ;
	Bufptr := @Output_buffer[ 1 ] ;
	Openfunc := @Scrnull ;
	Inoutfunc := @Window_Output ;
	Flushfunc := @Window_Output ;
	Closefunc := @Scrnull ;
	Name[ 0 ] := #0 ;
    end ;
end ;
{$F-}


procedure Init_SMU( X, Y, Z, A : Integer ) ;
{This procedure initializes the SMU root, with machine type of X.  if X is
 0, we try to determine machine type.  Y is the printer type.  Z is the number
 of input lines to save (default is 20).  A is 0 for overwrite mode, or 1 for
 insert mode}

var C : integer ;
{$IFNDEF DELPHI}
    I, R : Integer ;
    Reg : Registers ;
{$ENDIF}

begin
    if( Screen_Root.Model <> 0 ) then { Already initialized }
    begin
        exit ;
    end ;
    Getmem( Printer0, Sizeof( Block ) ) ;
    Getmem( Printer1, Sizeof( Block ) ) ;
    Getmem( Conversion, Sizeof( Block ) ) ;
    Getmem( Deconversion, Sizeof( Block ) ) ;
    if( Z = 0 ) then Z := 20 ; { Default is 20 }
    Max_Saved_Input_Lines := Z ;
    Assignscr( Scr ) ;
    Rewrite( Scr ) ; { Setup screen driver }
    if( A = 0 ) then
  	Screen_Root.Insert_Mode := False
  	else
  	Screen_Root.Insert_Mode := True ;
    if( X <> 0 ) then { A specific machine type specified }
    begin
  	 Screen_Root.Model := X ; { Set model to specified type }
    end else
    begin
{$IFDEF DELPHI}
        C := $F8 ; { Assume PC compatible if not real mode }
{$ELSE}
        C := Mem[ $F000 : $Fffe ] ; { Get machine type }
{$ENDIF}
        if ( C = 6 ) or ( C = $94 )
            then
                Screen_Root.Model := Model_Rainbow { Rainbow }
            else
                if C < $F8
                    then
                        Screen_Root.Model := Model_Unknown {Unknown model}
                    else
                        begin {PC or compatible}
{$IFNDEF DELPHI}
                            Reg.Ax := $0F00 ; {Get video mode}
  {$IFDEF WINDOWS}
                            Reg.Al := $13 ;
  {$ELSE}
  			    Intr( $10, Reg ) ; {do BIOS call}
  {$ENDIF}
                            case Reg.Al of {Video type}
                                0..1 : {40 column text}
                                       begin
{$ENDIF}
                                           Screen_Root.Model := Model_PC_Monochrome ;
                                           { Text mode }
{$IFNDEF DELPHI}
                                           Reg.Ax := $0007 ;
                                           {Set mode monochrome}
                                           Intr( $10, Reg )
                                       end ;
                                2..3 : {Text mode CGA/EGA/VGA/SVGA}
                                       begin
                                           Screen_Root.Model := Model_PC_Color_Text ;
                                       end ;
                                4..6, $0D..$13 : {Graphics mode}
                                       begin
                                           Screen_Root.Model := Model_PC_Graphics ;
                                           C := 0 ;
                                           I := Reg.Al ;
  {$IFNDEF WINDOWS}
                                           Initgraph( C, I, '' ) ;
                                           C := Graphresult ;
                                           if C <> grok then
                                           begin
                                               Writeln( Grapherrormsg( C ) ) ;
                                               Halt ;
                                           end ;
                                           Setviewport( 0, 0, Getmaxx,
                                               Getmaxy, True ) ;
  {$ENDIF}
                                       end ;
                                7 : {Monochrome (MDA)}
                                    begin
                                        I := Port[ $3Ba ] and 128 ;
                                        C := 0 ;
                                        while( C < 4096 ) do
                                        begin
                                            R := Port[ $3Ba ] and 128 ;
                                            if( R <> I ) then
                                            begin
                                                C := 4096 ;
                                            end ;
                                            C := C + 1
                                        end ;
                                        Screen_Root.Model := Model_PC_Monochrome ;
                                        {Assume MDA}
                                        if( R <> I ) then
                                        begin {Hercules card}
                                            if( ( Port[ $3B8 ] and 1 ) = 1 ) then
                                            begin {Hercules graphics}
                                                Screen_Root.Model := Model_Hercules ;
                                            end ;
                                        end ;
                                    end ;
                                else Screen_Root.Model := Model_Unknown ;
                                       {Unknown}
                            end ;
{$ENDIF}
                        end ;
    end ;
    Set_Model( Screen_Root.Model ) ;
    Attribute_Default := 0 ;
    Init_Screen ; {Set up depending upon model}
    Screen_Root.Next_Window := nil ; {No windows defined yet}
    Screen_Root.Window_Count := 0 ;
    Screen_Root.Default := 0 ; {No default}
    Screen_Root.Default_P := nil ;
    Screen_Root.Printer_Type := Y ; {Type of printer}
    Init_Printer ;
    Screen_Root.Printer_Mode := 2 ; {Print mode disabled}
    Screen_Root.Input_Buffer_Max := 20 ; {Number of input lines to save}
    Screen_Root.Last_Input := nil ;
    Screen_Root.Insert_Mode := False ; {Overwrite mode}
    Screen_Root.Driver := nil ; {No video driver}
    Fillchar( Screen.Window.Bytes[ 1, 1 ], 1920, 0 ) ; {No windows mapped}
    Fillchar( Screen.Data.Bytes[ 1, 1 ], 1920, 32 ) ; {Space fill screen}
end ; { Init_SMU }


procedure Set_Curs( X, Y : Byte ) ;
{ Display/remove cursor at location X,Y.  1,1 is home position }

{$IFDEF CONSOLE}
var Info : CONSOLE_CURSOR_INFO ;
    Point : COORD ;
{$ENDIF}

begin
{$IFDEF DELPHI}
    {$IFDEF CONSOLE}
        GetConsoleCursorInfo( GetStdHandle( STD_OUTPUT_HANDLE ), Info ) ;
        Info.bVisible := True ;
        SetConsoleCursorInfo( GetStdHandle( STD_OUTPUT_HANDLE ), Info ) ;
        Point.X := Y - 1 ;
        Point.Y := X - 1 ;
        SetConsoleCursorPosition( GetStdHandle( STD_OUTPUT_HANDLE ), Point ) ;
    {$ELSE}
{$IFDEF Debug}
        Check_Window( Screen_Root.Default_P ) ;
{$ENDIF}
        SetCaretPos( Screen_Root.Default_P^.Text_Width * ( Y - 1 ),
            Screen_Root.Default_P^.Text_Height * ( X - 1 ) ) ;
    {$ENDIF}
{$ELSE}
    _Set_Cursor( X, Y, Screen_Root.Foreground, Screen_Root.Background ) ;
{$ENDIF}
end ; { Set_Curs }


procedure Set_Background( const RGB : TRGB ) ;

var Window_P : PWindow ;

begin
    Window_P := Screen_Root.Default_P ; { Get default window }
    if( Window_P = nil ) then { Must be in console mode }
    begin
        exit ;
    end ;
{$IFDEF Debug}
    Check_Window( Window_P ) ;
{$ENDIF}
    Window_P^.Background := RGB ;
end ;


procedure Set_Foreground( const RGB : TRGB ) ;

var Window_P : PWindow ;

begin
    Window_P := Screen_Root.Default_P ; { Get default window }
    if( Window_P = nil ) then { Must be in console mode }
    begin
        exit ;
    end ;
{$IFDEF Debug}
    Check_Window( Window_P ) ;
{$ENDIF}
    Window_P^.Foreground := RGB ;
end ;


procedure Set_Cursor( Vr, Vc : Integer ) ; { Place visible cursor on screen }

var Window_P : PWindow ;

begin
    Window_P := Screen_Root.Default_P ; { Get default window }
    if( Window_P = nil ) then { Must be in console mode }
    begin
        exit ;
    end ;
{$IFDEF Debug}
    Check_Window( Window_P ) ;
{$ENDIF}
    if( ( ( Screen_Root.Flags or Window_P^.Flags ) and 32 ) = 32 ) then
    begin
        exit ; { Nope if cursor disabled }
    end ;
    if( Vr = 0 ) then
    begin
        Vr := Window_To_Screen_Row( Window_P, Window_P^.Position_Row ) ;
     end ;
    if( Vc = 0 ) then
    begin
        Vc := Window_To_Screen_Column( Window_P, Window_P^.Position_Column ) ;
    end ;
{$IFDEF WINDOWS}
    HideCaret( Window_P^.ID ) ;
    DestroyCaret ;
    CreateCaret( Window_P^.ID, 0, 0, Window_P^.Text_Height ) ;
    VC := ( VC - 1 ) * Window_P^.Text_Width + Window_P^.Border ;
    VR := ( VR - 1 ) * Window_P^.Text_Height + Window_P^.Border ;
    SetCaretPos( VC, VR ) ;
    ShowCaret( Window_P^.ID ) ;
{$ELSE}
    Set_Curs( Vr, Vc {- 1} )
{$ENDIF}
end ; { Set_Cursor }


procedure Clear_Cursor( Vr, Vc : Integer ) ; {Blank cursor}

var Window_P : PWindow ;
{$IFDEF CONSOLE}
    Info : CONSOLE_CURSOR_INFO ;
{$ENDIF}

begin
    Window_P := Screen_Root.Default_P ; {Get default window}
{$IFDEF Debug}
    Check_Window( Window_P ) ;
{$ENDIF}
{$IFDEF DELPHI}
    {$IFDEF CONSOLE}
        GetConsoleCursorInfo( GetStdHandle( STD_OUTPUT_HANDLE ), Info ) ;
        Info.bVisible := False ;
        SetConsoleCursorInfo( GetStdHandle( STD_OUTPUT_HANDLE ), Info ) ;
    {$ELSE}
        HideCaret( Window_P^.ID ) ;
    {$ENDIF}
{$ELSE}
    if ( ( ( Screen_Root.Flags or Window_P^.Flags ) and 32 ) = 32 ) then
    begin
        Exit ; { Nope if cursor disabled }
    end ;

    { Handle defaults... }
    if Vr = 0 then Vr := Window_P^.Position_Row ;
    if Vc = 0 then Vc := Window_P^.Position_Column ;

    _Clear_Cursor( VR, VC, Screen_Root.Background, Screen_Root.Foreground,
        Get_Window_Attribute( Vr, Vc - 1 ) ) ;
{$ENDIF}
end ; { Clear_Cursor }


procedure Set_Screen( Model, Printer_Model, Input_Buffers, Mode : Integer ) ;
{Set current screen attributes}

var Input_Pointer, Input_Pointer1 : Input_P ;
    Temp : Integer ;

begin
    if( Mode >= 0 ) then
        Screen_Root.Printer_Mode := Mode ;
    if( Printer_Model > 0 ) then
    begin
	Screen_Root.Printer_Type := Printer_Model ; { Type of printer }
	Init_Printer
    end ;
    if( Input_Buffers > 0 ) { Number of input lines to save }
	then
	    begin
		Temp := Screen_Root.Input_Buffer_Max ;
		if Temp < Input_Buffers then { Reduction in buffers }
		begin
		    Input_Pointer := Screen_Root.Last_Input ;
		    while Input_Pointer <> nil do
		    begin
			Input_Pointer := Input_Pointer^.Last_Input ;
			Dec( Temp ) ;
			if( Temp = 0 ) then { New end of buffers }
			    if( Input_Pointer^.Last_Input <> nil ) then
			    { Others exist }
			    begin
			      Input_Pointer1 := Input_Pointer ;
			      {Save current record}
			      Input_Pointer := Input_Pointer^.Last_Input ;
			      {Go to last input record}
			      Input_Pointer1^.Last_Input := nil ;
			      {Terminate list at new point}
			      while Input_Pointer <> nil do {Until done}
				begin
				    Freemem( Input_Pointer, Sizeof( Input_Record ) ) ;
				    {Delete old input records}
				    Input_Pointer := Input_Pointer^.Last_Input
				end
			    end
		    end
		end ;
		Screen_Root.Input_Buffer_Max := Input_Buffers
	    end ;
    if( Model > 0 ) then { Change model}
    begin
	Screen_Root.Model := Model ;
        Set_Model( Model ) ;
	Init_Screen ; { Set up as requested }
    end ;
end ;


procedure Move_To( Window_P : PWindow ; Row, Column : integer ) ;

begin
    if( Window_P = nil ) then
    begin
	 exit ;
    end ;
{$IFDEF Debug}
    Check_Window( Window_P ) ;
{$ENDIF}
    Window_P^.Position_Row := Row ;
    Window_P^.Position_Column := Column ;
end ;


procedure Set_Window( Window_P : PWindow ; Row, Column, Flag,
		     Attributes : Integer ) ;
{Set window flags/current position}

var Temp : Integer ; {Temporary variable for accumulating attributes}

begin
    { Attribute settings:
			 and 1 = Reverse-video
			 and 2 = High-intensity
			 and 4 = Blink
			 and 8 = Low-intensity
			 and 16 = No blink
			 and 32 = Normal video
			 and 64 = Modify existing flags
			 and 128-16384 = reserved }
    if( Window_P = nil ) then
    begin
	 exit ;
    end ;
{$IFDEF Debug}
    Check_Window( Window_P ) ;
{$ENDIF}
    if( Attributes < 0 ) { if attributes to change }
	then
	    begin
		Attributes := -Attributes ; {Recover bits}
{$IFDEF WINDOWS}
Temp := 0 ;
{$ELSE}
		Case Screen_Root.Model of { Default settings}
		    Model_Unknown : Temp := 0 ;
		    Model_Rainbow : Temp := 14 ;
		    Model_PC_Monochrome..Model_PC_Color_Text : Temp := 7 ;
		end ;
{$ENDIF}
		if ( Attributes and 64 ) = 64 then
		    Temp := Window_P^.Attribute ; {Get current attributes}
{$IFNDEF WINDOWS}
		if Screen_Root.Model = Model_Unknown then Temp := Attributes and 14 ;
		{Unknown}
{$ENDIF}
		if ( Attributes and 1 ) = 1 then {Reverse-video}
                begin
{$IFDEF WINDOWS}
                    Temp := Temp or 1 ;
{$ELSE}
		    case Screen_Root.Model of
			Model_Rainbow : Temp := Temp or 1 ;
			    {Reverse to normal}
			Model_PC_Monochrome..Model_PC_Color_Text : Temp := Temp and ( Not 7 ) or $70 ;
			       {Normal to reverse}
		    end ;
{$ENDIF}
                end ;
		if ( Attributes and 2 ) = 2 then {High-intensity}
		    Case Screen_Root.Model of
			Model_Rainbow : Temp := Temp and ( Not 2 ) ;
			Model_PC_Monochrome..Model_PC_Color_Text : Temp := Temp or 8 ;
		    end ;
		if ( Attributes and 4 ) = 4 then {Blink}
		    Case Screen_Root.Model of
			Model_Rainbow : Temp := Temp and ( Not 4 ) ;
			Model_PC_Monochrome : Temp := Temp or 128 ;
			Model_PC_Color_Text : Temp := Temp or 64 ;
		    end ;
		if ( Attributes and 8 ) = 8 then {Low-intensity}
		    Case Screen_Root.Model of
			Model_Rainbow : Temp := Temp or 2 ;
			Model_PC_Monochrome..Model_PC_Color_Text : Temp := Temp and ( Not 8 ) ;
		    end ;
		if ( Attributes and 16 ) = 16 then {No blink}
		    Case Screen_Root.Model of
			Model_Rainbow : Temp := Temp or 4 ;
			Model_PC_Monochrome : Temp := Temp and ( Not 128 ) ;
			Model_PC_Color_Text : Temp := Temp and ( Not 64 ) ;
		    end ;
		if( ( Attributes and 32 ) = 32 ) then {Normal video}
                begin
{$IFDEF WINDOWS}
                    Temp := Temp and ( not 1 ) ;
{$ELSE}
		    Case Screen_Root.Model of
			Model_Rainbow : Temp := Temp and ( Not 1 ) ;
			    {Reverse to normal}
			Model_PC_Monochrome..Model_PC_Color_Text : Temp := Temp and ( Not $70 ) or 7 ;
		    end ;
{$ENDIF}
                end ;
		Window_P^.Attribute := Temp ; { Now set attributes }
	    end ;
    {Set attributes if flag (negative)}
    if( ( Row < 0 ) or ( Column < 0 ) or ( Flag <> 0 ) ) then
        Window_P^.Flags := Flag ;
    {Flags are always set if row or column are negative}
    if( Row > 0 ) then
        Window_P^.Position_Row := Row ;
    if( Column > 0 ) then
        Window_P^.Position_Column := Column ;
end ;


procedure Select_Window( Window_P : PWindow ) ;

begin
{$IFDEF Debug}
    Check_Window( Window_P ) ;
{$ENDIF}
    if( Screen_Root.Default_P <> nil ) then
    begin
        Clear_Cursor( 0, 0 ) ;
    end ;
    Screen_Root.Default := Window_P^.Id ;
    Screen_Root.Default_P := Window_P ;
    Set_Cursor( Window_To_Screen_Row( Screen_Root.Default_P, Screen_Root.Default_P^.Position_Row ),
        Window_To_Screen_Column( Screen_Root.Default_P, Screen_Root.Default_P^.Position_Column ) ) ;
end ;


procedure Erase_EOW( Window_P : PWindow ) ; {Erase to end of window}

var A, B, Offset : Integer ;

begin
{$IFDEF Debug}
    Check_Window( Window_P ) ;
{$ENDIF}
    Offset := ( Window_P^.Position_Row - 1 ) * Window_P^.Columns +
	Window_P^.Position_Column - 1 ; {Get current offset}
    B := Window_P^.Size - Offset ; {Find size of end of window}
    if( B < 1 ) then {Past end of window}
    begin
	exit ;
    end ;

    A := Offset ;
    while( A < Offset + B ) do
    begin
	if(
            ( Window_P^.Contents^.Get_Data_Byte( A ) <> 32 )
            or
	    ( Window_P^.Contents^.Get_Attribute( A ) <> Window_P^.Attribute )
            or
	    ( not RGBs_Equal( Window_P^.Contents^.Get_Background_Color( A ), Window_P^.Background ) )
            or
	    ( not RGBs_Equal( Window_P^.Contents^.Get_Foreground_Color( A ), Window_P^.Foreground ) )
          ) then
	begin
	    if( Window_P^.Low_Update > A ) then
	    begin
		Window_P^.Low_Update := A ;
	    end ;
	    if( Window_P^.High_Update < A ) then
	    begin
		Window_P^.High_Update := A ;
	    end ;
	end ;
	A := A + 1 ;
    end ;

    Fill_Data( Window_P, Offset, B, 32 ) ; { Space fill }
    Fill_Attributes( Window_P, Offset, B, Window_P^.Attribute ) ; { Null attributes }
    Window_P^.Contents^.Fill_Background_Colors( Offset, B, Window_P^.Background ) ;
    Window_P^.Contents^.Fill_Foreground_Colors( Offset, B, Window_P^.Foreground ) ;
    if ( Window_P^.Flags and 8 ) = 0 {if auto-update mode} then
    begin
	Update_Window( Screen_Root.Default_P ) ;
    end ;
end ; { Erase_EOW }


procedure Clear_Window( Window_P : PWindow ) ; {Clear a window}

begin
{$IFDEF Debug}
    Check_Window( Window_P ) ;
{$ENDIF}
    Window_P^.Position_Row := 1 ;
    Window_P^.Position_Column := 1 ;
    Erase_EOW( Window_P ) ;
end ;


procedure Erase_EOL( Window_P : PWindow ) ;
{ Erase to end of line }

var A, B, Offset : Integer ;

begin
{$IFDEF Debug}
    Check_Window( Window_P ) ;
{$ENDIF}
    A := Window_P^.Position_Column ;
    if( A < 1 ) then
    begin
	A := 1 ;
    end ;
    B := Window_P^.Position_Row ;
    if( B < 1 ) then
    begin
	B := 1 ;
    end ;
    if( ( A > Window_P^.Columns ) or ( B > Window_P^.Rows ) ) then
    begin
        exit ; { Out of range current row and/or column }
    end ;
    Offset := ( B - 1 ) * Window_P^.Columns + A - 1 ; { Get current offset }
    B := Window_P^.Columns - A + 1 ; { Find size of end of line }
    if( B < 1 ) then { Past end of line }
    begin
	exit ;
    end ;

    A := Offset ;
    while( A < Offset + B ) do
    begin
	if(
            ( Window_P^.Contents^.Get_Data_Byte( A ) <> 32 )
            or
	    ( Window_P^.Contents^.Get_Attribute( A ) <> Window_P^.Attribute )
            or
	    ( not RGBs_Equal( Window_P^.Contents^.Get_Background_Color( A ), Window_P^.Background ) )
            or
	    ( not RGBs_Equal( Window_P^.Contents^.Get_Foreground_Color( A ), Window_P^.Foreground ) )
          ) then
	begin
	    if( Window_P^.Low_Update > A ) then
	    begin
		Window_P^.Low_Update := A ;
	    end ;
	    if( Window_P^.High_Update < A ) then
	    begin
		Window_P^.High_Update := A ;
	    end ;
	end ;
	A := A + 1 ;
    end ;

    Fill_Data( Window_P, Offset, B, 32 ) ; { Space fill }
    Fill_Attributes( Window_P, Offset, B, Window_P^.Attribute ) ;
    Window_P^.Contents^.Fill_Background_Colors( Offset, B, Window_P^.Background ) ;
    Window_P^.Contents^.Fill_Foreground_Colors( Offset, B, Window_P^.Foreground ) ;
    { Spaces and null attributes }

    if( ( Window_P^.Flags and 8 ) = 0 ) then { if auto-update mode }
    begin
	Update_Window( Screen_Root.Default_P ) ;
    end ;
end ; { Erase_EOL }


procedure Insert_Line( Window_P : PWindow ; Direction : Byte ) ;
{ Insert line at current position.  Direction of 0=lose top line, otherwise=lose
  bottom line }

var B, Offset : Integer ;

begin
{$IFDEF Debug}
    Check_Window( Window_P ) ;
{$ENDIF}
    Window_P^.Position_Column := 1 ; {Reset cursor to BOL}
    Offset := ( Window_P^.Position_Row - 1 ) * Window_P^.Columns ;
    {Get beginning of current line's offset}
    if Direction = 0 then {Lose top line}
    begin
	if Offset = 0 {Insert line at line 1 does nothing if losing top line}
	    then
		Exit ;
	Window_P^.Low_Update := 0 ;
	if Window_P^.High_Update < Offset + Window_P^.Columns - 1
	    then
		Window_P^.High_Update := Offset + Window_P^.Columns - 1 ;
	Offset := Offset - Window_P^.Columns ; {Blank previous line}
	B := Offset - 1 ; {Number of bytes to move}
        Window_P^.Contents^.Copy( Window_P^.Contents, Window_P^.Columns, 0, B ) ;
	{Scroll up}
    end else {Lose bottom line}
    begin
	if Window_P^.Low_Update >= Offset then Window_P^.Low_Update := Offset ;
	Window_P^.High_Update := Window_P^.Size - 1 ;
	B := Window_P^.Size - Offset - Window_P^.Columns ;
	{Number of bytes to move}
        Window_P^.Contents^.Copy( Window_P^.Contents, Offset, Offset + Window_P^.Columns, B ) ;
	{Scroll down}
    end ;
    Fill_Data( Window_P, Offset, Window_P^.Columns, 32 ) ; { Space fill }
    Fill_Attributes( Window_P, Offset, Window_P^.Columns, Window_P^.Attribute ) ;
    Window_P^.Contents^.Fill_Background_Colors( Offset, Window_P^.Columns,
        Window_P^.Background ) ;
    Window_P^.Contents^.Fill_Foreground_Colors( Offset, Window_P^.Columns,
        Window_P^.Foreground ) ;
    {Spaces and null attributes for newly created line}

    if ( Window_P^.Flags and 8 ) = 0 {if auto-update mode} then
    begin
	Update_Window( Screen_Root.Default_P ) ;
    end ;
end ; { Insert_Line }


procedure Scroll_Window( Window_P : PWindow ; X, Y : integer ; Rect : TRect ) ;

var A, C, Offset : integer ;

begin
    if( Window_P = nil ) then
    begin
	exit ;
    end ;
{$IFDEF Debug}
    Check_Window( Window_P ) ;
{$ENDIF}
    if( Y = 0 ) then
    begin
	exit ;
    end ;

    {Bounds check}
    if( Rect.Top < 1 ) then
    begin
	Rect.Top := 1 ;
    end ;
    if( Rect.Bottom > Window_P^.Rows ) then
    begin
	Rect.Bottom := Window_P^.Rows ;
    end ;

    { Calculate offsets... }
    Offset := ( Rect.Top - 1 ) * Window_P^.Columns ; { Get top line's offset }
    A := ( Rect.Bottom ) * Window_P^.Columns - 1 ; { Get bottom line's offset }
    if( Offset > A ) then { Inverted range }
    begin
	exit ;
    end ;
    C := A - Offset - Window_P^.Columns + 1 ; { Amount to move }
    if( C < 1 ) then { Nothing to scroll }
    begin
	exit ;
    end ;

    { Calculate update region... }
    if( Offset < Window_P^.Low_Update ) then
    begin
	Window_P^.Low_Update := Offset ;
    end ;
    if( A > Window_P^.High_Update ) then
    begin
	Window_P^.High_Update := A ;
    end ;

    { do the scroll }
    while( Y > 0 ) do { Scroll down }
    begin
        Window_P^.Contents.Copy( Window_P^.Contents, Offset, Offset + Window_P^.Columns, C ) ;
        Fill_Data( Window_P, Offset, Window_P^.Columns, 32 ) ; { Space fill }
        Fill_Attributes( Window_P, Offset, Window_P^.Columns, Window_P^.Attribute ) ;
        Window_P^.Contents^.Fill_Background_Colors( Offset, Window_P^.Columns,
            Window_P^.Background ) ;
        Window_P^.Contents^.Fill_Foreground_Colors( Offset, Window_P^.Columns,
            Window_P^.Foreground ) ;
	Y := Y - 1 ;
    end ;
    while( Y < 0 ) do {Scroll up}
    begin
        Window_P^.Contents.Copy( Window_P^.Contents, Offset + Window_P^.Columns, Offset, C ) ;
        Fill_Data( Window_P, A - Window_P^.Columns + 1, Window_P^.Columns, 32 ) ; { Space fill }
        Fill_Attributes( Window_P, A - Window_P^.Columns + 1, Window_P^.Columns, Window_P^.Attribute ) ;
        Window_P^.Contents^.Fill_Background_Colors( A - Window_P^.Columns + 1,
            Window_P^.Columns, Window_P^.Background ) ;
        Window_P^.Contents^.Fill_Foreground_Colors( A - Window_P^.Columns + 1,
            Window_P^.Columns, Window_P^.Foreground ) ;
	Y := Y + 1 ;
    end ;

    if( ( Window_P^.Flags and 8 ) = 0 ) then {Update window}
    begin
	Update_Window( Window_P ) ;
    end ;
end ; { Scroll_Window }


procedure Delete_Line( Window_P : PWindow ; Direction : Byte ) ;
{Delete line at current position.  Direction: 0=gain top line, otherwise=gain
 bottom line}

var A, B, Offset : Integer ;

begin
    if( Window_P = nil ) then
    begin
	exit ;
    end ;
{$IFDEF Debug}
    Check_Window( Window_P ) ;
{$ENDIF}
    Window_P^.Position_Column := 1 ; { Reset cursor to BOL }
    A := Window_P^.Columns ; { Number of columns in row }
    Offset := ( Window_P^.Position_Row - 1 ) * A ;
    { Get beginning of current line's offset }

    if( Direction = 0 ) then { Gain top line }
    begin
	Window_P^.Low_Update := 0 ;
	if Window_P^.High_Update < Offset then
        begin
            Window_P^.High_Update := Offset ;
        end ;

        Window_P^.Contents^.Copy( Window_P^.Contents, 0, A, Offset ) ;
	{ Scroll down }
        Fill_Data( Window_P, 0, A, 32 ) ; { Space fill }
        Fill_Attributes( Window_P, 0, A, Window_P^.Attribute ) ;
        Window_P^.Contents^.Fill_Background_Colors( 0, A,
            Window_P^.Background ) ;
        Window_P^.Contents^.Fill_Foreground_Colors( 0, A,
            Window_P^.Foreground ) ;
	{ Erase first line }
    end else { Gain bottom line }
    begin
	if Window_P^.Low_Update > Offset + A - 1 then
        begin
	    Window_P^.Low_Update := Offset + A - 1 ; {Adjust low update pointer}
	end ;
        Window_P^.High_Update := Window_P^.Size - 1 ; {Adjust high update pointer}
	B := Window_P^.Size - Offset - A ; {Amount of data to scroll}

        Window_P^.Contents^.Copy( Window_P^.Contents, Offset + A, Offset, B ) ;
	{Scroll up}

	B := Window_P^.Size - A ; { Beginning of last line }
        Fill_Data( Window_P, B, A, 32 ) ; { Space fill }
        Fill_Attributes( Window_P, B, A, Window_P^.Attribute ) ;
        Window_P^.Contents^.Fill_Background_Colors( B, A,
            Window_P^.Background ) ;
        Window_P^.Contents^.Fill_Foreground_Colors( B, A,
            Window_P^.Foreground ) ;
	{ Clear last line }
    end ;
    if( ( Window_P^.Flags and 8 ) = 0 ) {if auto-update mode} then
    begin
	Update_Window( Screen_Root.Default_P ) ;
    end ;
end ; { Delete_Line }


procedure Insert_Input( X : string ) ;
{Insert string as most recent input record}

var Count : Integer ;
    Input_Pointer, Input_Pointer1 : Input_P ;

begin
    if( Screen_Root.Input_Buffer_Max = 0 ) then
    begin
        Exit ; { No input saves }
    end ;
    Count := 1 ;
    Getmem( Input_Pointer, Sizeof( Input_Record ) ) ;
    Input_Pointer^.Input_Line := X ;
    Input_Pointer^.Last_Input := Screen_Root.Last_Input ;
    Input_Pointer^.Next_Input := nil ;
    { Set up new record }
    Screen_Root.Last_Input := Input_Pointer ;
    { Link it as the most recent last input }
    if( Input_Pointer^.Last_Input <> nil ) then
    begin
    	Input_Pointer^.Last_Input^.Next_Input := Input_Pointer ;
    end ;
    {Set back pointer for previous last input, if any}
    while Input_Pointer^.Last_Input <> nil do {Until end of list}
    begin
	Input_Pointer := Input_Pointer^.Last_Input ; {backup one record}
	Count := Count + 1 {Count them up}
    end ;
    while Count > Screen_Root.Input_Buffer_Max do {Get rid of extra input}
    begin
	Input_Pointer1 := Input_Pointer ; {Save new oldest record}
	Input_Pointer := Input_Pointer^.Next_Input ; {Get next record}
	Input_Pointer^.Last_Input := nil ;
	Count := Count - 1 ; {One less input record}
	Freemem( Input_Pointer1, Sizeof( Input_Record ) ) ; {Bye bye}
    end ;
end ;


function Get_Window_Attribute( R, C : Integer ) : Byte ;
{ Return attribute of character in current window at row R, column C.  Use
  default positions if R, C, or both are zero (0) }

var Window_P : PWindow ;
    Offset : Integer ;

begin
    Get_Window_Attribute := 0 ;
    Window_P := Screen_Root.Default_P ; { Get current window }
{$IFDEF Debug}
    Check_Window( Window_P ) ;
{$ENDIF}
    if( R = 0 ) then
        R := Window_P^.Position_Row ;
    if( C = 0 ) then
        C := Window_P^.Position_Column ;
    if( R > Window_P^.Rows ) then
        Exit ; {Out of range}
    if( C > Window_P^.Columns ) then
        Exit ; {Ditto}
    if( ( R < 1 ) or ( C < 1 ) ) then
        Exit ; {Ditto ditto}
    Offset := ( R - 1 ) * Window_P^.Columns + C - 1 ;
    Get_Window_Attribute := Window_P^.Contents^.Get_Attribute( Offset ) ;
end ;


procedure Window_Input_Line( Window_P : PWindow ; Flags : Integer ) ;
{ Get a line from the keyboard }

label Loop ;

{$IFNDEF DELPHI}
var B : string ;
{$ELSE}
    {$IFDEF CONSOLE}
var C : DWORD ;
    P : PChar ;
    {$ENDIF}
{$ENDIF}

begin
{$IFDEF Debug}
    Check_Window( Window_P ) ;
{$ENDIF}
{$IFDEF CONSOLE}
    Set_Cursor( 0, 0 ) ;
    if( Window_P = nil ) then
    begin
        SetLength( KB_Buffer, 255 ) ;
        P := PChar( KB_Buffer[ 1 ] ) ;
        if( not ReadConsole( GetStdHandle( STD_INPUT_HANDLE ), P, 255, C, nil ) ) then
        begin
            readln( KB_Buffer ) ;
            C := length( KB_Buffer ) ;
        end ;
        SetLength( KB_Buffer, C ) ;
        while( ( C > 0 ) and ( ( KB_Buffer[ C ] = CR ) or ( Kb_Buffer[ C ] = LF ) ) ) do
        begin
            dec( C ) ;
            SetLength( KB_Buffer, C ) ;
        end ;
    end else
    begin
        SetLength( Window_P^.KB_Buffer, 255 ) ;
        P := PChar( Window_P^.KB_Buffer[ 1 ] ) ;
        if( not ReadConsole( GetStdHandle( STD_INPUT_HANDLE ), P, 255, C, nil ) ) then
        begin
            readln( Window_P^.KB_Buffer ) ;
            C := length( Window_P^.KB_Buffer ) ;
        end ;
        SetLength( Window_P^.KB_Buffer, C ) ;
        while( ( C > 0 ) and ( ( Window_P^.KB_Buffer[ C ] = CR ) or ( Window_P^.Kb_Buffer[ C ] = LF ) ) ) do
        begin
            dec( C ) ;
            SetLength( Window_P^.KB_Buffer, C ) ;
        end ;
    end ;
    Text_Out( Screen_Root.Default_P, CR, 1 ) ;
    Text_Out( Screen_Root.Default_P, LF, 1 ) ;
{$ELSE}
    Window_Async_Input_Line( Window_P, Flags, '' ) ; { Initialize }
Loop:
  {$IFNDEF WINDOWS}
    B := Default_Input.Inkey ; { Get next key }
    if( not Window_Async_Input_Line( Window_P, Flags, B ) ) then
    { Until done with input line }
    begin
        goto Loop ;
    end ;
  {$ENDIF}
{$ENDIF}
end ; { Window_Input_Line }


function Get_KB_Buffer( Window_P : PWindow ) : string ;

begin
    if( Window_P = nil ) then
    begin
        Get_KB_Buffer := KB_Buffer ;
    end else
    begin
{$IFDEF Debug}
        Check_Window( Window_P ) ;
{$ENDIF}
        Get_KB_Buffer := Window_P^.KB_Buffer ;
    end ;
end ;


procedure Set_KB_Buffer( Window_P : PWindow ; const S : string ) ;

begin
    if( Window_P = nil ) then
    begin
        KB_Buffer := S ;
    end else
    begin
{$IFDEF Debug}
        Check_Window( Window_P ) ;
{$ENDIF}
        Window_P^.KB_Buffer := S ;
    end ;
end ;


function Window_Async_Input_Line( Window_P : PWindow ; Flags : integer ;
    Key : string ) : boolean ;

var _Offset, _Offset1 : integer ; { Loop control variables }

    function Line_Format( const KB_Buffer : string ) : string ;

    var A : string ;
        B : Byte ;

    begin
        A := '' ;
        for B := 1 To Length( Kb_Buffer ) do { Build line character by character }
        begin
            if ( Kb_Buffer[ B ] < ' ' ) and ( Kb_Buffer[ B ] <> '   ' ) then
                A := A + '^' + Chr( Ord( Kb_Buffer[ B ] ) + 64 )
                { Control character }
                else A := A + Kb_Buffer[ B ] ;
            if B = Static_Input^.Cursor then {At current cursor position}
                if ( Kb_Buffer[ B ] < ' ' ) and ( Kb_Buffer[ B ] <> '       ' ) then
                Static_Input^.Cursor1 := Length( A ) - 1 else Static_Input^.Cursor1 := Length( A ) ;
        end ;
        Line_Format := A ; {Return value}
    end ;


    procedure Erase_Line ;

    var Dummy : integer ;

    begin
      with Static_Input^ do
      begin
        if( Window_P = nil ) then
        begin
	    Text_Out( Window_P, CRLF, 2 )
        end else
        begin
            Set_Window( Window_P, Input_Row, Input_Column, 0, 0 ) ;
            for Dummy := 1 To Length( Command ) do
            begin
                Text_Out( Window_P, ' ', 1 )
            end ;
            Set_Window( Window_P, Input_Row, Input_Column, 0, 0 ) ;
            Update_Window( Screen_Root.Default_P )
        end ;
        Set_KB_Buffer( Window_P, '' ) ;
        Command := '' ;
        Cursor := 1 ;
        Cursor1 := 1 ;
      end ;
    end ; { Window_Async_Input_Line.Erase_Line }

var Dummy : integer ;
{$IFNDEF DELPHI}
    R : Registers ;
{$ENDIF}
    This_Char : char ;

begin { Window_Async_Input_Line }
  Window_Async_Input_Line := False ;
  with Static_Input^ do
  begin
    if( length( Key ) = 0 ) then { Initializing }
    begin
        Input_Pointer := nil ;
        Command := '' ;
        Cursor := 1 ;
        Cursor1 := 1 ;
        if( Window_P <> nil ) then
        begin
            Backup := Chr( 254 ) ;
            Input_Column := Window_P^.Position_Column ;
            Input_Row := Window_P^.Position_Row ;
            Wid := Window_P^.Id ;
        end else
        begin
            Backup := Chr( 8 ) ;
            Input_Column := 1 ;
            Input_Row := 1 ;
        end ;
        Set_KB_Buffer( Window_P, '' ) ;
    end ;

    if( Window_P <> nil ) then
    begin
{$IFDEF Debug}
      Check_Window( Window_P ) ;
{$ENDIF}
      Offset := Window_P^.Position_Column ;
      Offset1 := Window_P^.Position_Row ;
      if ( Window_P^.Flags and 4 ) = 0 then {Wrap}
	if Offset > Window_P^.Columns then {Past right of window}
	begin
	    Offset := 1 ;
	    Offset1 := Offset1 + 1 ;
	end ;
      if ( Window_P^.Flags and 2 ) = 0 then {Window isn't occluded}
	if ( Offset1 >= Window_P^.View_Row ) and
	  ( Offset1 <= Window_P^.View_Row +
	    Window_P^.Map_Rows - 1 ) then {Within viewport rows}
	      if ( Offset >= Window_P^.View_Column )
		and ( Offset <= Window_P^.View_Column +
		 Window_P^.Map_Columns - 1 ) then
		  {Within viewport columns}
		  begin
		    Vr := Offset1 - Window_P^.View_Row + Window_P^.Map_Row ;
		    Vc := Offset -
		       Window_P^.View_Column + Window_P^.Map_Column + 1 ;
		    if( Screen.Window.Bytes[ Vr, Vc - 1 ] = Wid ) then
                    begin
			Set_Cursor( Vr, Vc ) ; {Place cursor}
                    end ;
		  end ;
    end ; { if Window_P <> nil }
    B := Key ; { Get next key }
    if( length( Key ) = 0 ) then { Doing setup }
    begin
	exit ; { Done with setup }
    end ;
    if( Ord( B[ 1 ] ) > 127 ) then
    begin
	B := Chr( Deconversion^[ Ord( B[ 1 ] ) ] ) ;
    end ;

    if( B = 'LEFT' ) then {Left-arrow}
    begin
	if Cursor > 1 then {Not at beginning of line}
	begin
	    Cursor := Cursor - 1 ;
	    Cursor1 := Cursor1 - 1 ;
	    Text_Out( Window_P, Backup, 1 ) ; {Backup}
            if( Window_P = nil ) then
            begin
                This_Char := KB_Buffer[ Cursor ] ;
            end else
            begin
                This_Char := Window_P^.KB_Buffer[ Cursor ] ;
            end ;
	    if ( This_Char < ' ' ) and ( This_Char <> '     ' ) then
	    {Control}
	    begin
		Text_Out( Window_P, Backup, 1 ) ; {Backup over caret}
		Cursor1 := Cursor1 - 1 ;
	    end ;
	    if( Window_P <> nil ) then
	    begin
		if( This_Char = '      ' ) then {Tab}
		begin
		    Set_Window( Window_P, Input_Row, Input_Column, 0, 0 ) ;
		    Text_Out( Window_P, Copy( Command, 1, Cursor1 - 1 ),
			Cursor1 - 1 ) ;
		    Offset := Window_P^.Position_Row ;
		    Offset1 := Window_P^.Position_Column ;
		    Text_Out( Window_P, Copy( Command, Cursor1, 255 ),
			length( Command ) - Cursor1 + 1 ) ;
		    Set_Window( Window_P, Offset, Offset1, 0, 0 ) ;
		    {Backup}
		end ;
		Update_Window( Screen_Root.Default_P ) ;
	    end ; {if Window_P <> nil}
	end ;
	exit ;
    end ;
    if B = 'RIGHT' then { Right-arrow }
    begin
        if( Window_P = nil ) then
        begin
            Dummy := Length( KB_Buffer ) ;
        end else
        begin
            Dummy := Length( Window_P^.KB_Buffer ) ;
        end ;
	if( Cursor <= Dummy ) then { Before end of line }
	begin
            if( Window_P = nil ) then
            begin
                This_Char := KB_Buffer[ Cursor ] ;
            end else
            begin
                This_Char := Window_P^.KB_Buffer[ Cursor ] ;
            end ;
	    if( ( This_Char < ' ' ) and ( This_Char <> '     ' ) ) then
	    {Control character}
	    begin
		Text_Out( Screen_Root.Default_P, '^', 1 ) ;
		Text_Out( Screen_Root.Default_P,
		    Chr( Ord( This_Char ) + 64 ), 1 ) ;
		Cursor1 := Cursor1 + 1
	    end else
	    begin
		Text_Out( Screen_Root.Default_P, This_Char, 1 ) ;
	    end ;
	    Cursor := Cursor + 1 ;
	    Cursor1 := Cursor1 + 1 ;
	    Update_Window( Screen_Root.Default_P ) ;
	end ;
	exit ;
    end ;

    Ord_B := Ord( B[ 1 ] ) ;

    if( ( B = 'UP' ) or ( Ord_B = 2 ) ) then { Up-arrow or ^B }
    begin
	if( Input_Pointer = nil ) then Input_Pointer := Screen_Root.Last_Input else
	begin
	    if( Input_Pointer^.Last_Input = nil ) then
            begin
                exit ;
            end ;
	    Input_Pointer := Input_Pointer^.Last_Input ;
	end ;
	if( Input_Pointer <> nil ) then { Some input }
	begin
	    Erase_Line ;
            Set_KB_Buffer( Window_P, Input_Pointer^.Input_Line ) ;
            Cursor := Length( Get_KB_Buffer( Window_P ) ) + 1 ;
            Command := Line_Format( Get_KB_Buffer( Window_P ) ) ;
	    Cursor1 := Length( Command ) + 1 ;
	    Text_Out( Screen_Root.Default_P, Command, length( Command ) ) ;
	    Update_Window( Screen_Root.Default_P ) ;
	end ;
	exit ;
    end ;
    if( B = 'DOWN' ) then { Down-arrow }
    begin
	if Input_Pointer = nil then
        begin
            exit ;
        end ;
	Input_Pointer := Input_Pointer^.Next_Input ;
	Erase_Line ;
	if( Input_Pointer <> nil ) then { Some input }
	begin
            Set_KB_Buffer( Window_P, Input_Pointer^.Input_Line ) ;
            Cursor := Length( Get_Kb_Buffer( Window_P ) ) + 1 ;
            Command := Line_Format( Get_Kb_Buffer( Window_P ) ) ;
	    Cursor1 := Length( Command ) + 1 ;
	    Text_Out( Screen_Root.Default_P, Command, length( Command ) ) ;
	    Update_Window( Screen_Root.Default_P ) ;
	end ;
	exit ;
    end ;
    if B = 'F4' then {Toggle print mode}
    begin
	if Screen_Root.Printer_Mode = 0
	    then
		Screen_Root.Printer_Mode := 1
	    else
		if Screen_Root.Printer_Mode = 1
		    then
			Screen_Root.Printer_Mode := 0 ;
	{Toggle between on and off}
	exit ;
    end ;
    if B = 'F6' then {Print screen}
    begin
	if Screen_Root.Printer_Mode <> 2 then {Not disabled}
	    for _Offset := 1 To Max_Row do {for each row}
	    begin
                Offset := _Offset ;
		for _Offset1 := 1 To Screen_Root.Screen_Width do {for each column}
		begin
                    Offset1 := _Offset1 ;
		    Ord_B := Deconversion^[ Get_Screen( Offset, Offset1 ) ] ;
		    {Get physical character}
		    Ch1 := Chr( Printer0^[ Ord_B ] ) ; {Translate}
{$IFNDEF DELPHI}
		    R.Ax := $0500 ; {Printer output}
		    if( ( Ord_B < 32 ) or ( Ord_B > 126 ) ) then {Special character}
                    begin
			if Printer1^[ Ord_B ] = 1 then
			begin
			    OS^.Print( #27 ) ;
			    OS^.Print( 'O' ) ;
			end ;
                    end ;
		    OS^.Print( Ch1 ) ;
{$ENDIF}
		end ;
		Writeln( Printer_File ) ; {end line}
	    end ;
	exit ;
    end ;
    if B = 'ENTER' then
    begin
	B := Chr( 13 ) ;
	Ord_B := 13 ;
    end ;
    if( Length( B ) > 1 ) then {Special key}
    begin
	Write( BEL ) ;
	exit ;
    end ;
    if( Ord_B = 1 ) then {Control-A}
    begin
	Screen_Root.Insert_Mode := Not Screen_Root.Insert_Mode ; {Toggle mode}
	exit ;
    end ;
    if( Ord_B = 5 ) then {^E - goto end of line}
    begin
	if Cursor1 <= Length( Command ) then
	begin
	    Text_Out( Screen_Root.Default_P, Copy( Command, Cursor1, 255 ), -1 ) ;
	end ;
	if( Window_P = nil ) then
        begin
            Cursor := Length( Kb_Buffer ) + 1 ;
        end else
        begin
            Cursor := Length( Window_P^.Kb_Buffer ) + 1 ;
        end ;
        Cursor1 := Length( Command ) + 1 ;
	Update_Window( Screen_Root.Default_P ) ;
	exit ;
    end ;
    if( Ord_B = 18 ) then {Control-R}
    begin
	Refresh_Window( Screen_Root.Default_P ) ;
	exit ;
    end ;
    if( Ord_B = 23 ) then {Control-W}
    begin
	Refresh_Screen ;
	exit ;
    end ;
    if( ( Ord_B = 127 ) or ( Ord_B = 8 ) ) then { Handle delete }
    begin
        if( Window_P = nil ) then
        begin
            Dummy := length( KB_Buffer ) ;
        end else
        begin
            Dummy := length( Window_P^.KB_Buffer ) ;
        end ;
	if( Dummy > 0 ) then { Something to delete }
	begin
            if( Window_P = nil ) then
            begin
                Ord_B := Ord( Kb_Buffer[ Cursor - 1 ] ) ; {Get character to be deleted}
                Kb_Buffer := Copy( Kb_Buffer, 1, Cursor - 2 ) +
                    Copy( Kb_Buffer, Cursor, 255 ) ; {Remove character}

                Old_Length := Length( Command ) ;
                Command := Line_Format( Kb_Buffer ) ;
            end else
            begin
                Ord_B := Ord( Window_P^.Kb_Buffer[ Cursor - 1 ] ) ; {Get character to be deleted}
                Kb_Buffer := Copy( Window_P^.Kb_Buffer, 1, Cursor - 2 ) +
                    Copy( Window_P^.Kb_Buffer, Cursor, 255 ) ; {Remove character}

                Old_Length := Length( Command ) ;
                Command := Line_Format( Window_P^.Kb_Buffer ) ;
            end ;
	    if( Ord_B = 9 ) then {Tab}
	    begin
		{Move to beginning of tab}
		if( Window_P <> nil ) then
		begin
		    Set_Window( Window_P, Input_Row, Input_Column, 0, 0 ) ; {Input start}
		    Text_Out( Window_P, Copy( Command, 1, Cursor1 - 2 ),
			-1 ) ; {Write up to tab}
		    Offset := Window_P^.Position_Row ;
		    Offset1 := Window_P^.Position_Column ;
		    { Save cursor position }
		end else { DOS output }
                begin
		    Text_Out( Window_P, Backup, 1 ) ;
                end ;
		Text_Out( Window_P, Copy( Command, Cursor1 - 1, 255 ), -1 ) ;
		{ Write rest of line }
		Text_Out( Window_P, '        ', -1 ) ;
		if Window_P = nil then
		begin
		    Dummy := Length( Command ) - Cursor1 + 8 ;
		    while Dummy > 0 do
		    begin
			Text_Out( Window_P, Backup, -1 ) ;
			Dec( Dummy ) ;
		    end ;
		end else Set_Window( Window_P, Offset, Offset1, 0, 0 ) ;
		{ Move back to cursor position }
	    end else { Non-tab }
	    begin
		Text_Out( Window_P, Backup + ' ' + Backup, 3 ) ;
		{ Erase old character }
		if( Ord_B < 32 ) then { Deleting a control character }
		begin
		    Text_Out( Window_P, Backup + ' ' + Backup, 3 ) ;
		    Cursor1 := Cursor1 - 1 ;
		end
	    end ;
	    if( Window_P <> nil ) then
	    begin
		Offset := Window_P^.Position_Row ;
		Offset1 := Window_P^.Position_Column ;
	    end ;
	    Cursor := Cursor - 1 ;
	    Cursor1 := Cursor1 - 1 ;
	    Text_Out( Window_P, Copy( Command, Cursor1, 254 ), -1 ) ;
	    Dummy := Old_Length - Length( Command ) + Length( Command ) - Cursor1 + 1 ;

	    while Old_Length > Length( Command ) do
	    begin
		Old_Length := Old_Length - 1 ;
		Text_Out( Window_P, ' ', 1 )
	    end ;
	    { Update line }
	    if( Window_P = nil ) { Normal DOS output }
		then
		    while Dummy > 0 do
		    begin
			Text_Out( Window_P, Backup, 1 ) ;
			Dec( Dummy ) ;
		    end
		else
		    Set_Window( Window_P, Offset, Offset1, 0, 0 ) ; {Set cursor}
	    Update_Window( Screen_Root.Default_P ) ;
	end ;
	exit ;
    end ;
    if( Ord_B = 3 ) then {Control-C}
    begin
	Text_Out( Screen_Root.Default_P, '^C' + CRLF, 4 ) ;
	Set_KB_Buffer( Window_P, '' ) ;
	Window_Async_Input_Line := True ;
	exit ;
    end ;
    if( Ord_B = 25 ) then {Control-Y}
    begin
	Text_Out( Screen_Root.Default_P, '^Y' + CRLF, 4 ) ;
	Set_KB_Buffer( Window_P, '' ) ;
	Window_Async_Input_Line := True ;
	exit ;
    end ;
    if( Ord_B = 26 ) then {Control-Z}
    begin
	Text_Out( Screen_Root.Default_P, '^Z' + CRLF, 4 ) ;
	Set_KB_Buffer( Window_P, Chr( 26 ) ) ;
	Window_Async_Input_Line := True ;
	exit ;
    end ;
    if Ord_B = 21 then {^U - Erase line}
    begin
	Erase_Line ;
	exit ;
    end ;
    if( ( Length( Command ) = 253 ) and ( Ord_B <> 13 ) and ( Ord_B <> 10 ) and
	( Ord_B <> 12 ) and ( Ord_B <> 11 ) and ( Ord_B <> 27 ) ) then
    begin
	OS^.Beep( 0 ) ; { Bell upon full input line }
	exit ;
    end ;
    if( ( ( Ord_B < 9 ) or ( Ord_B > 13 ) ) and ( Ord_B < 32 ) ) then
    begin
	Command0 := '^' + Chr( Ord_B + 64 )
    end else
    begin
        Command0 := B[ 1 ] ;
    end ;
    if( ( ( Ord_B < 10 ) or ( Ord_B > 13 ) ) and ( Ord_B <> 27 ) ) then { Not a delimiter }
    begin
	if( Screen_Root.Insert_Mode ) then { In insert mode }
	begin
	    if( Window_P = nil ) then
            begin
                KB_Buffer := Copy( KB_Buffer, 1, Cursor - 1 ) + B +
		    Copy( KB_Buffer, Cursor, 255 ) ;
            end else
            begin
                KB_Buffer := Copy( Window_P^.KB_Buffer, 1, Cursor - 1 ) + B +
		    Copy( Window_P^.KB_Buffer, Cursor, 255 ) ;
            end ;
	    Command :=
		Copy( Command, 1, Cursor1 - 1 ) + Command0 +
                Copy( Command, Cursor1, 255 ) ;
	end else { In over-write mode }
	begin
	    if( Window_P = nil ) then
            begin
                Old_Length := Ord( Kb_Buffer[ Cursor ] ) ;
                Kb_Buffer[ Cursor ] := B[ 1 ] ;
                if( Length( Kb_Buffer ) < Cursor ) then
                begin
{$IFDEF DELPHI}
                    SetLength( Kb_Buffer, Cursor ) ;
{$ELSE}
                    Kb_Buffer[ 0 ] := Chr( Cursor ) ;
{$ENDIF}
                end ;
                Offset := Cursor1 ;
                Command := Copy( Command, 1, Cursor1 - 1 ) +
                    Line_Format( Copy( Kb_Buffer, Cursor, 255 ) ) ;
            end else
            begin
                Old_Length := Ord( Window_P^.Kb_Buffer[ Cursor ] ) ;
                Window_P^.Kb_Buffer[ Cursor ] := B[ 1 ] ;
                if( Length( Window_P^.KB_Buffer ) < Cursor ) then
                begin
{$IFDEF DELPHI}
                    SetLength( Window_P^.KB_Buffer, Cursor ) ;
{$ELSE}
                    Window_P^.Kb_Buffer[ 0 ] := Chr( Cursor ) ;
{$ENDIF}
                end ;
                Offset := Cursor1 ;
                Command := Copy( Command, 1, Cursor1 - 1 ) +
                    Line_Format( Copy( Window_P^.KB_Buffer, Cursor, 255 ) ) ;
            end ;
	    Cursor1 := Offset ; {Restore position}
	end ;
	Text_Out( Screen_Root.Default_P, Command0, length( Command0 ) ) ;
	{Write character}
	if( Window_P <> nil ) then
	begin
	    Offset := Window_P^.Position_Row ;
	    Offset1 := Window_P^.Position_Column ;
	end ;
	Text_Out( Window_P,
	    Copy( Command, Cursor1 + Length( Command0 ), 255 ), -1 ) ;
	    {Update line}
	if( Old_Length = 9 ) then
	begin
	    Text_Out( Window_P, '       ', 8 ) ;
	    if( Window_P = nil )
		then
		    for Dummy := 1 To 8 do Write( Backup ) ;
	end ;
	if( Window_P = nil ) then
	begin
	    Dummy := Length( Command ) - Cursor1 - Length( Command0 ) + 1 ;
	    while( Dummy > 0 ) do
	    begin
		Write( Backup ) ;
		Dec( Dummy ) ;
	    end ;
	end ;
	Cursor := Cursor + 1 ; {Position in buffer}
	Cursor1 := Cursor1 + Length( Command0 ) ;
	Set_Window( Window_P, Offset, Offset1, 0, 0 ) ;
	Update_Window( Screen_Root.Default_P ) ;
	exit ;
    end ;

    { A delimeter if we get here }
    if( ( ( Window_P = nil ) or ( ( Window_P^.Flags and 128 ) = 0 ) ) and ( ( Flags and 128 ) = 0 ) )
	then {CRLF allowed}
    begin
	Text_Out( Screen_Root.Default_P, CR, 1 ) ;
	if( Ord_B <> 10 ) then
        begin
            Text_Out( Screen_Root.Default_P, LF, 1 ) ;
        end ; { Make sure delimiter causes CRLF }
    end ;
  end ;
    Window_Async_Input_Line := True ; { Done with input }
end ;


procedure Window_Read ; {Read a line}

begin
    Window_Read_Input( 0 ) ; {do standard input}
end ;


procedure Window_Read_Input( Flags : Integer ) ; {Read a line}

var Window_P : PWindow ;

begin
    Window_P := Screen_Root.Default_P ; { Get default window }
    Set_KB_Buffer( Window_P, '' ) ;
    Window_Input_Line( Window_P, Flags ) ;
    if( Length( Get_Kb_Buffer( Window_P ) ) > 0 ) then
    begin
        Insert_Input( Get_Kb_Buffer( Window_P ) ) ; { Save this line }
    end ;
end ;


procedure Load_Driver( X : string ) ;
{Load a screen driver from the file specified in X}

{$IFNDEF DELPHI}
var F : file of string ;
    G : Integer ;
    B : string ;
    U : file ;
{$ENDIF}

begin
{$IFNDEF DELPHI}
    Assign( F, X ) ;
    Assign( U, X ) ;
    Reset( F ) ; {Open driver file}
    G := Filesize( F ) shl 8 ; {Get file size in bytes}
    if( G = 0 ) then
    begin
        exit ; {File is empty}
    end ;
    Getmem( Screen_Root.Driver, G ) ; {Create room for driver}
    Close( F ) ;
    Reset( U ) ; {Open for block reads}
    Blockread( U, Screen_Root.Driver^, G div 128 ) ; {Load driver}
    Close( U ) ;
{$ENDIF}
end ;


function Get_Window_Char( R, C : Integer ) : Byte ;
{ Return character in window at row R, column C.  Use default positions if R, C,
  or both are zero (0). }

var Window_P : PWindow ;
    Offset : Integer ;

begin
    Get_Window_Char := 0 ;
    Window_P := Screen_Root.Default_P ; {Get current window}
{$IFDEF Debug}
    Check_Window( Window_P ) ;
{$ENDIF}
    if( R = 0 ) then R := Window_P^.Position_Row ;
    if( C = 0 ) then C := Window_P^.Position_Column ;
    if( R > Window_P^.Rows ) then Exit ; {Out of range}
    if( C > Window_P^.Columns ) then Exit ; {Ditto}
    if( ( R < 1 ) or ( C < 1 ) ) then Exit ; {Ditto ditto}
    Offset := ( R - 1 ) * Window_P^.Columns + C - 1 ;
    Get_Window_Char := Window_P^.Contents^.Get_Data_Byte( Offset ) ;
end ;


procedure Set_Viewport( Row, Column : Byte ) ; {Set viewport row and column}

var Window_P : PWindow ;

begin
    Window_P := Screen_Root.Default_P ;
{$IFDEF Debug}
    Check_Window( Window_P ) ;
{$ENDIF}
    if( Row > 0 ) then Window_P^.View_Row := Row ;
    if( Column > 0 ) then Window_P^.View_Column := Column ;
    if( ( Row > 0 ) or ( Column > 0 ) ) then
    begin
	Window_P^.Low_Update := 0 ;
	Window_P^.High_Update := 32767 ;
    end ;
    if ( Window_P^.Flags and 8 ) = 0 {if auto-update mode} then
    begin
	Update_Window( Screen_Root.Default_P ) ;
    end ;
end ;


procedure Output_Text( const S : string ) ;

var Window_P : PWindow ;

begin
    Window_P := Screen_Root.Default_P ; { Find window }
    if( Window_P = nil ) then
    begin
	write( S ) ;
    end else
    begin
	Text_Out( Window_P, S, length( S ) ) ;
    end ;
end ;


function Window_To_Screen_Row( Window : PWindow ; Row : integer ) : integer ;

begin
{$IFDEF Debug}
    Check_Window( Window ) ;
{$ENDIF}
    Window_To_Screen_Row := Row + Window^.Map_Row - Window^.View_Row ;
end ;


function Window_To_Screen_Column( Window : PWindow ; Column : integer ) : integer ;

begin
{$IFDEF Debug}
    Check_Window( Window ) ;
{$ENDIF}
    Window_To_Screen_Column := Column + Window^.Map_Column - Window^.View_Column ;
end ;


{$IFDEF Test}
procedure Test_Unit ;

begin
    Init_SMU( 0, 0, 0, 0 ) ;
    { TODO: Complete test suite here... }
    End_SMU ;
end ;
{$ENDIF}


{$IFDEF DELPHI}
initialization
    {$IFDEF CONSOLE}
    assign( scr, 'CON' ) ;
    {$ELSE}
    Assignscr( Scr ) ;
    {$ENDIF}
{$ELSE}
begin
    assignscr( Scr ) ;
{$ENDIF}
    Rewrite( Scr ) ; { Setup screen driver }
    Screen_Root.Insert_Mode := False ;
    Screen_Root.Model := 0 ;
    Attribute_Default := 0 ;
    Screen_Root.Next_Window := nil ; {No windows defined yet}
    Screen_Root.Window_Count := 0 ;
    Screen_Root.Default := 0 ; {No default}
    Screen_Root.Default_P := nil ;
    Screen_Root.Printer_Type := 0 ; {Type of printer}
    Screen_Root.Printer_Mode := 2 ; {Print mode disabled}
    Screen_Root.Input_Buffer_Max := 20 ; {Number of input lines to save}
    Screen_Root.Last_Input := nil ;
    Screen_Root.Insert_Mode := False ; {Overwrite mode}
    Screen_Root.Driver := nil ; {No video driver}
    New( Static_Input ) ;
{$IFDEF WINDOWS}
    Get_Window_Size( nil, Screen_Root.Screen_Width, Screen_Root.Screen_Width ) ;
    Init_SMU( 0, 0, 0, 0 ) ;
{$ENDIF}
{$IFDEF DELPHI}
    Printer0 := nil ;
    
finalization
    End_SMU ;
{$ENDIF}
end.
