{
        Program Name : SStreams
        Package Name : LB
        Purpose      : String Stream objects
        Institution  : Conroy & Conroy Co.
        Date Written :
        Written By   : Alan Conroy
        Version      : 1.0

	Copyright (C) 2003 by Alan Conroy.  All rights reserved.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

}

unit SStreams ;

interface

{$I EDEFINES.INC}

uses _Streams, { TCOM_Stream }
     Streams, { TBase_Stream }
     _UEHDefs ; { TUEC }

type TSString_Stream = object( TBase_Stream )
                           private { Instance data... }
                               Contents : string ;
                               _Position : integer ; { Always just before character to read/write (eg 0=position 1) }

                           public { Overrides... }
                               constructor Init ;
                               destructor Done ; virtual ;

                           public { API... }
                               function At_End : boolean ; virtual ;

                               function Position : integer ; virtual ;

                               procedure Read( var Buffer ; var _Size : longint ) ;
                                   virtual ;

                               procedure Seek( Value : longint ) ; virtual ;

                               function Size : longint ; virtual ;

                               procedure Write( var Buffer ; _Size : longint ) ;
                                   virtual ;

                               procedure Set_String( const P : string ) ; virtual ;
                       end ; { TSString_Stream }
      PSString_Stream = ^TSString_Stream ;

{$IFDEF DELPHI}
type TCOM_String_Stream = class( TCOM_Stream )
                               private { Instance data... }
                                   Contents : string ;
                                   _Position : integer ; { Always just before character to read/write (eg 0=position 1) }

                               public { API... }
                                  function As_String : PChar ;
                                      virtual ; stdcall ;

                                  function At_End : boolean ; override ;

                                  procedure Clear ; virtual ; stdcall ;

                                  { Returns last error code, and resets the error code. }
                                  procedure Last_Error( var UEC : TUEC ) ;
                                      override ;

                                  { Reads the specified number of bytes from the stream.
                                    Size is modified to be the actual bytes transferred. }
                                  procedure Read( var Buffer ;
                                      var _Size : longint ) ; override ;

                                  { Reads one line of input, up to the size of the buffer.
                                    The line is assumed to end at an ASCII 13 code.  The
                                    ASCII 13 code is not included in the returned data.
                                    Size is modified to be the actual bytes transferred. }
                                  procedure Read_Line( var Buffer ;
                                      var _Size : longint ) ; override ;

                                  { Position to specified byte within streamed data. }
                                  procedure Seek( Position : longint ) ;
                                      override ;

                                  { Returns size of stream data, in bytes.  -1 indicate that
                                    the size is unknown or larger than 2^31. }
                                  function Size : longint ; override ;

                                  { Writes the specified buffer, of the specified size in
                                    bytes, to the stream. }
                                  procedure Write( var Buffer ;
                                      _Size : longint ) ; override ;

                                  { Writes the specified null-terminated text to the stream.
                                    An ASCII code 13 is appended to the text on output. }
                                  procedure Write_Line( Buffer : PChar ) ;
                                      override ;

                                  procedure Set_String( P : PChar ) ; virtual ;
                          end ; { TCOM_String_Stream }
{$ENDIF}

{$IFDEF Test}
procedure Test_Unit ;
{$ENDIF}

implementation

uses { Borland... }
{$IFDEF DELPHI}
     SysUtils, { strlen }
{$ENDIF}

     { C&C... }
     ASCIIDef ; { CR }

{ TSString_Stream methods... }

{ Overrides... }

constructor TSString_Stream.Init ;

begin
    inherited Init ;

{$IFDEF DELPHI}
    Initialize( Contents ) ;
{$ENDIF}
    Contents := '' ;
    _Position := 0 ;
end ;


destructor TSString_Stream.Done ;

begin
{$IFDEF DELPHI}
    Finalize( Contents ) ;
{$ENDIF}

    inherited Done ;
end ;


function TSString_Stream.At_End : boolean ;

begin
    At_End := ( _Position >= length( Contents ) ) ;
end ;


function TSString_Stream.Position : integer ;

begin
    Position := _Position ;
end ;


procedure TSString_Stream.Read( var Buffer ; var _Size : longint ) ;

var Len : integer ; { Maximum possible length }

begin
    Len := length( Contents ) - _Position ;
    if( Len < _Size ) then
    begin
        _Size := Len ;
    end ;
    move( Contents[ _Position + 1 ], Buffer, _Size ) ;
    _Position := _Position + _Size ;
end ;


procedure TSString_Stream.Seek( Value : longint ) ;

begin
    _Position := Value ;
    if( _Position > length( Contents ) ) then
    begin
        _Position := length( Contents ) ;
    end ;
end ;


function TSString_Stream.Size : longint ;

begin
    Size := length( Contents ) ;
end ;


procedure TSString_Stream.Write( var Buffer ; _Size : longint ) ;

begin
    setlength( Contents, _Position + _Size ) ;
    move( Buffer, PChar( Contents )[ _Position ], _Size ) ;
    _Position := _Position + _Size ;
end ;


procedure TSString_Stream.Set_String( const P : string ) ;

begin
    Contents := P ;
end ;



{$IFDEF DELPHI}
{ TCOM_String_Stream methods... }

{ API... }

function TCOM_String_Stream.As_String : PChar ;

begin
    As_String := PChar( Contents ) ;
end ;


function TCOM_String_Stream.At_End : boolean ;

begin
    At_End := ( _Position >= length( Contents ) ) ;
end ;


procedure TCOM_String_Stream.Clear ;

begin
    Contents := '' ;
    _Position := 0 ;
end ;


procedure TCOM_String_Stream.Last_Error( var UEC : TUEC ) ;

begin
    fillchar( UEC, sizeof( UEC ), 0 ) ;
end ;


procedure TCOM_String_Stream.Read( var Buffer ; var _Size : longint ) ;

var Len : integer ; { Maximum possible length }

begin
    Len := length( Contents ) - _Position ;
    if( Len < _Size ) then
    begin
        _Size := Len ;
    end ;
    move( Contents[ _Position + 1 ], Buffer, _Size ) ;
    _Position := _Position + _Size ;
end ;


procedure TCOM_String_Stream.Read_Line( var Buffer ; var _Size : longint ) ;

var C : char ;
    Count : integer ;
    S : string ;
    Temp : longint ;

begin
    Count := 0 ;
    while( ( Count <= _Size ) and ( not At_End ) ) do
    begin
        Temp := 1 ;
        Read( C, Temp ) ;
        if( C = CR ) then { End of line }
        begin
            break ;
        end ;
        S := S + C ;
        inc( Count ) ;
    end ;
    _Size := Count ;
    move( S[ 1 ], Buffer, Count ) ;
end ;


procedure TCOM_String_Stream.Seek( Position : longint ) ;

begin
    _Position := Position ;
    if( _Position > length( Contents ) ) then
    begin
        _Position := length( Contents ) ;
    end ;
end ;


function TCOM_String_Stream.Size : longint ;

begin
    Size := length( Contents ) ;
end ;


procedure TCOM_String_Stream.Write( var Buffer ; _Size : longint ) ;

begin
    setlength( Contents, _Position + _Size ) ;
    move( Buffer, PChar( Contents )[ _Position ], _Size ) ;
    _Position := _Position + _Size ;
end ;


procedure TCOM_String_Stream.Write_Line( Buffer : PChar ) ;

begin
    Write( Buffer[ 0 ], strlen( Buffer ) ) ;
end ;


procedure TCOM_String_Stream.Set_String( P : PChar ) ;

begin
    Contents := string( P ) ;
end ;
{$ENDIF}


{$IFDEF Test}
procedure Error( const S : string ) ;

begin
    writeln( S ) ;
    halt ;
end ;


procedure Test_Unit ;

var S : string[ 255 ] ;
    Size : integer ;
    Test : TSString_Stream ;

begin
    Test.Init ;

    if( not Test.At_End ) then
    begin
        Error( 'At_End failure' ) ;
    end ;
    S := '1234567890ABCDEFGHIJKLMNOPQRSTUVWYZ' ;
    Test.Write( S[ 1 ], 10 ) ;
    if( not Test.At_End ) then
    begin
        Error( 'At_End failure' ) ;
    end ;
    Test.Write( S[ 1 ], 1 ) ;
    if( Test.Position <> 11 ) then
    begin
        Error( 'Position invalid' ) ;
    end ;
    Test.Seek( 100 ) ;
    Test.Seek( 0 ) ;
    Size := 10 ;
    Test.Read( S[ 1 ], Size ) ;
    if( Size <> 10 ) then
    begin
        Error( 'Read failure' ) ;
    end ;
    S[ 0 ] := #10 ;
    if( S <> '1234567890' ) then
    begin
        Error( 'Read failure' ) ;
    end ;
    Test.Read( S[ 1 ], Size ) ;
    if( Size <> 1 ) then
    begin
        Error( 'Read failure' ) ;
    end ;

    Test.Done ;
end ;
{$ENDIF}



end.
