{
        Program Name : ScrollPanel
        Package Name : LB
        Purpose      : Scrolling Panel
        Institution  : Conroy & Conroy Co.
        Date Written : 24-Oct-1998
        Written By   : Alan Conroy
        Version      : 1.0

        Copyright (C) 1998 by Alan Conroy.  Released to the Public Domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          TScrolling_Panel implements a panel of controls which can scroll
        left/right if they don't all show.
}

unit ScrollPanel ;

interface

uses // Windows...
     Windows,
     Messages,

     // Delphi...
     SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
     ExtCtrls,
     Buttons, // TSpeedButton

    // C&C...
    Sizable_Speed_Button ; // TSizable_Speed_Button

type
  TScrolling_Panel = class( TPanel )
                         public // Constructors and destructors...
                             constructor Create( AOwner : TComponent ) ; override ;

                         private // Instance data...
                             Left_Button, Right_Button : {TSpeedButton}TSizable_Speed_Button ;
                             Client_Panel, Control_Panel : TPanel ;

                         private // Callbacks...
                             procedure CB_Resize( Sender : TObject ) ;
                             procedure CB_Scroll( Sender : TObject ) ;
                             procedure CB_Start_Scroll( Sender : TObject ;
                                 Button : TMouseButton ; Shift : TShiftState ; X, Y : Integer ) ;
                             procedure CB_End_Scroll( Sender : TObject ;
                                 Button : TMouseButton ; Shift : TShiftState ; X, Y : Integer ) ;

                         private // Utility routines...
                             procedure Adjust_Button_States ;

                         protected // Property handlers...
                             function Get_Color : TColor ;
                             procedure Set_Color( Value : TColor ) ;

                         public // API...
                             procedure Add_Control( Control : TControl ) ;

                             function Child( Index : integer ) : TControl ;

                             function Child_Count : integer ;

                             procedure Set_Images( Left, Right : TBitmap ) ;

                             property Color : TColor
                                 read Get_Color
                                 write Set_Color ;
                     end ;

procedure Register ;

implementation

procedure Register ;

begin
    RegisterComponents( 'LB', [TScrolling_Panel] ) ;
end ;

var Scroll_Rate : integer ;

// TScrolling_Panel class...

// Constructors and destructors...

constructor TScrolling_Panel.Create( AOwner : TComponent ) ;

var LBitmap, RBitmap : TBitmap ;

begin
    // Call inherited constructor
    inherited Create( AOwner ) ;

    // Create subcomponents...
    LBitmap := TBitmap.Create ;
    LBitmap.LoadFromResourceName( 0, 'SCROLL_LEFT' ) ;
    LBitmap.Canvas.Pixels[ 0, LBitmap.Height - 1 ] := clDkGray ;
(*
    Left_Button := TSizable_Speed_Button.Init( AOwner, Bitmap ) ;
    Left_Button.Parent := Self ;
    Left_Button.Align := alLeft ;
    Left_Button.Caption := '' ;
    Left_Button.OnMouseDown := CB_Start_Scroll ;
    Left_Button.OnMouseUp := CB_End_Scroll ;
*)

    RBitmap := TBitmap.Create ;
    RBitmap.LoadFromResourceName( 0, 'SCROLL_RIGHT' ) ;
    RBitmap.Canvas.Pixels[ 0, RBitmap.Height - 1 ] := clDkGray ;
    Set_Images( LBitmap, RBitmap ) ;
(*
    Right_Button := TSizable_Speed_Button.Init( AOwner, Bitmap ) ;
    Right_Button.Parent := Self ;
    Right_Button.Align := alRight ;
    Right_Button.Caption := '' ;
    Right_Button.OnMouseDown := CB_Start_Scroll ;
    Right_Button.OnMouseUp := CB_End_Scroll ;
*)

    Client_Panel := TPanel.Create( AOwner ) ;
    Client_Panel.Parent := Self ;
    Client_Panel.Align := alClient ;
    Client_Panel.BevelOuter := bvNone ;
    Client_Panel.BevelInner := bvNone ;
    Client_Panel.BorderStyle := bsNone ;
    Client_Panel.BorderWidth := 0 ;
    Client_Panel.ParentColor := True ;

    Control_Panel := TPanel.Create( AOwner ) ;
    Control_Panel.Parent := Client_Panel ;
    Control_Panel.Align := alNone ;
    Control_Panel.BevelOuter := bvNone ;
    Control_Panel.BevelInner := bvNone ;
    Control_Panel.Width := 0 ;
    COntrol_Panel.Left := 0 ;
    Control_Panel.BorderStyle := bsNone ;
    Control_Panel.BorderWidth := 0 ;
    Control_Panel.ParentColor := True ;

    // Setup delegation callbacks...
    OnResize := CB_Resize ;
end ; // TScrolling_Panel.Create


// Callbacks...

var Active_Button : TSpeedButton ;
    Scroll_Timer : TTimer ;

procedure TScrolling_Panel.CB_Resize( Sender : TObject ) ;

var Control : TControl ;
    Loop : integer ;
    X : integer ;

begin
    Control_Panel.Height := ClientHeight ;
    Control_Panel.Left := 0 ;
    X := 0 ;
    for Loop := 0 to Control_Panel.ControlCount - 1 do
    begin
        Control := Control_Panel.Controls[ Loop ] ;
        Control.Height := Control_Panel.ClientHeight ;
        Control.Left := X ;
        X := X + Control.Width ;
    end ;
    Control_Panel.ClientWidth := X ;
    Adjust_Button_States ;
end ;


procedure TScrolling_Panel.CB_Scroll( Sender : TObject ) ;

begin
    Scroll_Timer.Enabled := False ;
    if( Left_Button = Active_Button ) then
    begin
        if( Control_Panel.Width + Control_Panel.Left > Client_Panel.Width ) then
        begin
            Control_Panel.Left := Control_Panel.Left - Scroll_Rate ;
        end ;
    end ;
    if( Right_Button = Active_Button ) then
    begin
        if( Control_Panel.Left < 0 ) then
        begin
            Control_Panel.Left := Control_Panel.Left + Scroll_Rate ;
        end ;
    end ;
    Update ;
    Scroll_Timer.Enabled := True ;
end ;


procedure TScrolling_Panel.CB_Start_Scroll( Sender : TObject ;
    Button : TMouseButton ; Shift : TShiftState ; X, Y : Integer ) ;

begin
    Active_Button := TSpeedButton( Sender ) ;
    Scroll_Timer.OnTimer := CB_Scroll ;
    Scroll_Timer.Enabled := True ;
end ;


procedure TScrolling_Panel.CB_End_Scroll( Sender : TObject ;
    Button : TMouseButton ; Shift : TShiftState ; X, Y : Integer ) ;

begin
    Active_Button := nil ;
    Scroll_Timer.Enabled := False ;
end ;


// Utility routines...

procedure TScrolling_Panel.Adjust_Button_States ;

begin
    if( Control_Panel.Width > ClientWidth ) then
    begin
        Left_Button.Visible := True ;
        Right_Button.Visible := True ;
    end else
    begin
        Left_Button.Visible := False ;
        Right_Button.Visible := False ;
        Control_Panel.Left := 0 ;
    end ;
end ;


// API...

// Property handlers...

function TScrolling_Panel.Get_Color : TColor ;

begin
    Result := inherited Color ;
end ;


procedure TScrolling_Panel.Set_Color( Value : TColor ) ;

begin
    inherited Color := Value ;

    TSizable_Speed_Button( Left_Button ).Color := Value ;
    TSizable_Speed_Button( Right_Button ).Color := Value ;
end ;


procedure TScrolling_Panel.Add_Control( Control : TControl ) ;

var X : integer ;

begin
    X := Control_Panel.ClientWidth ;
    Control.Parent := Control_Panel ;
    Control.Left := X ;
    if( Control_Panel.ControlCount = 0 ) then
    begin
        Control_Panel.Width := 0 ;
    end ;
    Control_Panel.ClientWidth := Control.Left + Control.Width ;
    Control.Height := Control_Panel.ClientHeight ;
    Adjust_Button_States ;
end ;


function TScrolling_Panel.Child( Index : integer ) : TControl ;

begin
    Result := Control_Panel.Controls[ Index ] ;
end ;


function TScrolling_Panel.Child_Count : integer ;

begin
    Result := Control_Panel.ControlCount ;
end ;


procedure TScrolling_Panel.Set_Images( Left, Right : TBitmap ) ;

begin
    Left_Button.Free ;
    Left_Button := TSizable_Speed_Button.Init( Owner, Left ) ;
    Left_Button.Parent := Self ;
    Left_Button.Align := alLeft ;
    Left_Button.Caption := '' ;
    Left_Button.OnMouseDown := CB_Start_Scroll ;
    Left_Button.OnMouseUp := CB_End_Scroll ;

    Right_Button.Free ;
    Right_Button := TSizable_Speed_Button.Init( Owner, Right ) ;
    Right_Button.Parent := Self ;
    Right_Button.Align := alRight ;
    Right_Button.Caption := '' ;
    Right_Button.OnMouseDown := CB_Start_Scroll ;
    Right_Button.OnMouseUp := CB_End_Scroll ;
end ;


initialization
    // Setup...
    Active_Button := nil ;
    Scroll_Timer := TTimer.Create( Application ) ;
    Scroll_Timer.Interval := 1 ;

    // Defaults
    Scroll_Rate := 3 ;
end.
