{$N+}
{$E+}
{
        Program Name : _GenVal
        Package Name : Subroutine Library - Layer 0
        Purpose      : Generic Value class definitions.
        Institution  : Conroy & Conroy Co.
        Date Written : 14-February-2005
        Written By   : Alan Conroy
        Version      : 1.0

	Copyright (C) 2005 by Alan Conroy.  Released to the Public Domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          TGeneric_Value is a class that provides a generic value (ie variant).
        There are three intended uses for this (and associated) classes:
          1) RTTI support
          2) Compiler/debugger symbol support
          3) Uber-variant
}

unit _GenVal ;

interface

{$I EDEFINES.INC}

uses _Common, { TObject }
{$IFDEF DELPHI}
     ComInter, { TCommon_COM_Interface }
{$ENDIF}
     _Types ; { int8 }

type TData_Type_Family = ( DTF_Unknown, { Variant }
                           DTF_Undefined, { Binary }
                           DTF_Logical,
                           DTF_Numeric,
                           DTF_Character, { Character/string}
                           DTF_Pointer, { pointer/method/procedure/function }
                           DTF_Composite,
                           DTF_Collection,
                           DTF_User_Defined,
                           DTF_Meta, { Type }
                           DTF_Other
                         ) ;

     TValue_Data_Type = ( VDT_Empty,
                          VDT_Null,
                          VDT_Boolean,
                          VDT_Integer,
                          VDT_Real,
                          VDT_String,
                          VDT_Char,
                          VDT_Pointer,
                          VDT_Value, { TGeneric_Value defining a type }
                          VDT_Code, { Function or procedure }
                          VDT_Scalar,
                          VDT_User_Defined,
                          VDT_Unknown
                        ) ;

     TCall_Mechanism = ( CM_By_Value,
                         CM_By_Reference,
                         CM_By_Name
                       ) ;

     TMethod_Class = ( MC_Private,
                       MC_Protected,
                       MC_Public,
                       MC_Published
                     ) ;

{$IFDEF DELPHI}
type TGeneric_Value = class ;
     TGeneric_Parameter = class( TDelphi_Object ) { Parameter information... }
                              public
                                  function Typ : TGeneric_Value ; virtual ; abstract ;
                                  { Type of parameter }
                                  function Output : boolean ; virtual ; abstract ; 
                                  { Purpose of parameter is for returning data }
                                  function Constant : boolean ; virtual ; abstract ;
                                  { True if constant }
                                  function Mechanism : TCall_Mechanism ;
                                      virtual ; abstract ;
                                  function Default : TGeneric_Value ; virtual ; abstract ;
                                  { Default value of parameter }
                          end ;

     TGeneric_Number = class( TDelphi_Object )
                           public
                               function Fixed : boolean ; virtual ; abstract ; 
                               { True if integer or fixed length real }
                               function Signed : boolean ; virtual ; abstract ;
                               function Mantissa_Size : longint ;
                                   virtual ; abstract ;
                               { In bits, for Floating types }
                               function Exponent_Size : longint ;
                                   virtual ; abstract ;
                               { In bits, for Floating types (0 for non-floating) }
                               function Fractional_Size : longint ;
                                   virtual ; abstract ;
                               { In bits, for fixed-point types (0 for non-fixed). }
                       end ;

     TGeneric_Character = class( TDelphi_Object )
                              public
                                  function Max_Length : longint ;
                                      virtual ; abstract ;
                                  { Maximum allowable length (same as Length except for dynamic-length strings) }
                                  function Fixed : boolean ;
                                      virtual ; abstract ;
                                  { True if fixed length }
                                  function Length : longint ;
                                      virtual ; abstract ; { Length in bytes }
                                  function Logical_Length : longint ;
                                      virtual ; abstract ;
                                  { Length in characters }
                          end ;

     TGeneric_Collection = class( TDelphi_Object )
                               public
                                   function Fixed : boolean ;
                                       virtual ; abstract ;
                                       { False if dynamic array }
                                   function Sparse : boolean ;
                                       virtual ; abstract ;
                                       { True if sparse array }
                                   function Subscripts : longint ;
                                       virtual ; abstract ;
                                       { Number of subscripts }
                                   function Low_Bound( Subscript : longint ) : TGeneric_Value ;
                                       virtual ; abstract ;
                                       { Lower bound of subscript for array }
                                   function High_Bound( Subscript : longint ) : TGeneric_Value ;
                                       virtual ; abstract ;
                                   { Upper bound of subscript for array }
                                   function Get_Subscript( Address : longint ) : TGeneric_Value ;
                                       virtual ; abstract ; 
                                   { Get value from array (viewing it as a flat one-dimensional array) }
                                   function Row_Major : boolean ; virtual ; abstract ; 
                                   { True if array is left-to-right major }
                           end ;

     TGeneric_Composite = class( TDelphi_Object )
                              public
                                  function Ancestor( index : longint ) : TGeneric_Value ;
                                      virtual ; abstract ;
                                  function Ancestor_Count : longint ;
                                      virtual ; abstract ;
                                  { Number of ancestors }
                                  function Ancestor_Class( Index : longint ) : TMethod_Class ;
                                      virtual ; abstract ;
                                  { How Indexth ancestor was inherited }

                                  function Count : longint ;
                                      virtual ; abstract ; { Number of values }
                                  function Get_Item( Index : longint ) : TGeneric_Value ;
                                      virtual ; abstract ; { Return value Index }
                          end ;

     TGeneric_Method = class( TDelphi_Object )
                           public
                               function Is_Abstract : boolean ;
                                   virtual ; abstract ;
                               { True if abstract (pure virtual) }
                               function Is_Constructor : boolean ;
                                   virtual ; abstract ;
                               function Is_Destructor : boolean ;
                                   virtual ; abstract ;
                               function Is_Virtual : boolean ;
                                   virtual ; abstract ; { True if virtual method }
                               function Static : boolean ; virtual ; abstract ; 
                               { True if a static class method }
                               function C_Order : boolean ; virtual ; abstract ; 
                               { True if C calling convention parameter order }
                               function Standard : boolean ;
                                    virtual ; abstract ; 
                               { True if standard (stack) parameter passing }
                               function Execute( Parameters : TGeneric_Value ) : TGeneric_Value ;
                                   virtual ; abstract ; { Execute the method }
                               function Compile_Time : boolean ;
                                   virtual ; abstract ;
                               { True if the routine can be evaluated at
                                 compile-time. }
                               function Count : longint ; virtual ; abstract ;
                               { Number of parameters (run-time) }
                               function Max_Count : longint ;
                                   virtual ; abstract ;
                               { Maximum potential number of parameters }
                               function Parameter( Index : integer ) : TGeneric_Value ;
                                   virtual ; abstract ;
                               { Returns parameter for specified index }
                               function Target : boolean ; virtual ; abstract ;
                               { True if a target function }
                               function Method_Class : TMethod_Class ;
                                   virtual ; abstract ;
                       end ;

     TProperty_Information = class( TDelphi_Object )
                                 public
                                     function Getter : TGeneric_Value ;
                                         virtual ; abstract ;
                                     function Setter : TGeneric_Value ;
                                         virtual ; abstract ;
                                     function Default : boolean ;
                                         virtual ; abstract ;
                             end ;

     TType_Information = class( TDelphi_Object )
                             public
                                 function Is_Composite : boolean ;
                                     virtual ; abstract ;
                                 { True if an array/queue/list/stack of Data_Type }
                                 function Is_Collection : boolean ;
                                     virtual ; abstract ;
                                 { True if an object or record }
                                 function Memory_Type : TMemory_Type ;
                                     virtual ; abstract ;

                                 function Parameter : TGeneric_Parameter ;
                                     virtual ; abstract ;
                                 { nil if not a parameter }
                                 function Numeric_Information : TGeneric_Number ;
                                     virtual ; abstract ;
                                 { nil if not numeric }
                                 function Character_Information : TGeneric_Character ;
                                     virtual ; abstract ;
                                 { nil if not character/string }
                                 function Composite_Information : TGeneric_Composite ;
                                     virtual ; abstract ;
                                 { nil if not an object/record }
                                 function Collection_Information : TGeneric_Collection ;
                                     virtual ; abstract ;
                                 { nil if not a composite (eg array) }
                                 function Method : TGeneric_Method ;
                                     virtual ; abstract ;
                                 { nil if not a method/procedure }
                                 function _Property : TProperty_Information ;
                                     virtual ; abstract ;
                                 { nil if not a property }
                                 function Address : longint ;
                                     virtual ; abstract ;
                                 { 0 if undefined, type, or parameter }

                                 function Range_Limited : boolean ;
                                     virtual ; abstract ;
                                 { True if value is limited to a range of values.
                                   This is always true for scalars. }

                                 { (These return nil if not range limited) }
                                 function Low_Range : TGeneric_Value ;
                                     virtual ; abstract ;
                                 function High_Range : TGeneric_Value ;
                                     virtual ; abstract ;

                                 { For scalars... }
                                 function Value_Name( Value : longint ) : string ;
                                     virtual ; abstract ;
                                 { Name for given ordinal value.  Returns null
                                   string if not a scalara or if Value is out of
                                   range. }
                         end ;

     TGeneric_Value = class( TDelphi_Object )
                          public { API... }
                              { Meta type information... }
                              function Family : TData_Type_Family ;
                                  virtual ; abstract ;
                              function Data_Type : TValue_Data_Type ;
                                  virtual ; abstract ;
                              function User_Data_Type_Name : string ;
                                  virtual ; abstract ;
                              function Name : string ; virtual ; abstract ;
                              { Variable, type, method, or parameter name }
                              function Type_Information : TType_Information ;
                                  virtual ; abstract ;

                              { Size, address, and format information... }
                              function Constant : boolean ; virtual ; abstract ;
                              { True if a constant rather than variable }
                              function Size : longint ; virtual ; abstract ;
                              { Size in bytes }
                              function Total_Size : longint ;
                                  virtual ; abstract ;
                              { Size of all associated data, in bytes.  For
                                instance, Size may return the size of an object
                                but Total_Size also includes the size of any
                                associated data. }

                              { For pointers... }
                              function Dereference : TGeneric_Value ;
                                  virtual ; abstract ;

                              { Conversions/promotions/casts... }
                              function Can_Convert( Typ : TValue_Data_Type ) : boolean ;
                                  virtual ; abstract ;
                              { True if value can be converted to the specified data type }
                              function Convert( Class_Type : TGeneric_Value ;
                                  var Loss : boolean ) : TGeneric_Value ;
                                  virtual ; abstract ;
                              { Return a version of the value converted to the
                                passed type specification.  Returns nil if the
                                conversion cannot be done.  Loss is ignored when
                                passed and is True on return if there was a loss
                                of precision. }

                              function Clone : TGeneric_Value ;
                                  virtual ; abstract ;

                              procedure Simple_Cast( Typ : TValue_Data_Type ) ;
                                  virtual ; abstract ;
                              procedure Cast( Typ : TGeneric_Value ) ;
                                  virtual ; abstract ;

                              { Other... }
                              procedure Deserialize( Value : PChar ) ;
                                  virtual ; abstract ;
                              function Serialize : PChar ; virtual ; abstract ;

                              { Get/Set value methods... }
                              procedure Set_Null ; virtual ; abstract ;
                              procedure Get_Undefined( Buffer : pointer ;
                                  Len : longint ) ; virtual ; abstract ;
                              procedure Set_Undefined( Buffer : pointer ;
                                  Len : longint ) ; virtual ; abstract ;

                              function Get_Boolean : boolean ;
                                  virtual ; abstract ;
                              procedure Set_Boolean( B : boolean ) ;
                                  virtual ; abstract ;

                              function Get_ByteBool : bytebool ;
                                  virtual ; abstract ;
                              procedure Set_ByteBool( B : bytebool ) ;
                                  virtual ; abstract ;

                              function Get_WordBool : wordbool ;
                                  virtual ; abstract ;
                              procedure Set_WordBool( B : wordbool ) ;
                                  virtual ; abstract ;

                              function Get_LongBool : longbool ;
                                  virtual ; abstract ;
                              procedure Set_LongBool( B : longbool ) ;
                                  virtual ; abstract ;

                              function Get_Char : char ; virtual ; abstract ;
                              procedure Set_Char( C : char ) ;
                                  virtual ; abstract ;

                              function Get_String : string ;
                                  virtual ; abstract ;
                              procedure Set_String( const S : string ) ;
                                  virtual ; abstract ;

                              function Get_PChar : PChar ; virtual ; abstract ;
                              procedure Set_PChar( P : PChar ) ;
                                  virtual ; abstract ;

                              function Get_Integer8 : int8 ;
                                  virtual ; abstract ;
                              procedure Set_Integer8( I : int8 ) ;
                                  virtual ; abstract ;

                              function Get_Integer16 : int16 ;
                                  virtual ; abstract ;
                              procedure Set_Integer16( I : int16 ) ;
                                  virtual ; abstract ;

                              function Get_Integer32 : int32 ;
                                  virtual ; abstract ;
                              procedure Set_Integer32( I : int32 ) ;
                                  virtual ; abstract ;

                              function Get_Integer64 : int64 ;
                                  virtual ; abstract ;
                              procedure Set_Integer64( I : int64 ) ;
                                  virtual ; abstract ;

                              function Get_Unsigned_Integer8 : uint8 ;
                                  virtual ; abstract ;
                              procedure Set_Unsigned_Integer8( I : uint8 ) ;
                                  virtual ; abstract ;

                              function Get_Unsigned_Integer16 : uint16 ;
                                  virtual ; abstract ;
                              procedure Set_Unsigned_Integer16( I : uint16 ) ;
                                  virtual ; abstract ;

                              function Get_Unsigned_Integer32 : uint32 ;
                                  virtual ; abstract ;
                              procedure Set_Unsigned_Integer32( I : uint32 ) ;
                                 virtual ; abstract ;

                              function Get_Single : single ;
                                  virtual ; abstract ;
                              procedure Set_Single( S : single ) ;
                                  virtual ; abstract ;

                              function Get_Double : double ;
                                  virtual ; abstract ;
                              procedure Set_Double( D : double ) ;
                                 virtual ; abstract ;

                              function Get_Extended : extended ;
                                  virtual ; abstract ;
                              procedure Set_Extended( E : extended ) ;
                                  virtual ; abstract ;

                              property Bool : boolean
                                  read Get_Boolean
                                  write Set_Boolean ;
                              property Byte_Bool : bytebool
                                  read Get_ByteBool
                                  write Set_ByteBool ;
                              property Word_Bool : wordbool
                                  read Get_WordBool
                                  write Set_WordBool ;
                              property Long_Bool : longbool
                                  read Get_LongBool
                                  write Set_LongBool ;
                              property _Char : char
                                  read Get_Char
                                  write Set_Char ;
                              property _String : string
                                  read Get_String
                                  write Set_String ;
                              property _PChar : PChar
                                  read Get_PChar
                                  write Set_PChar ;
                              property Integer8 : int8
                                  read Get_Integer8
                                  write Set_Integer8 ;
                              property Unsigned_Integer8 : uint8
                                  read Get_Unsigned_Integer8
                                  write Set_Unsigned_Integer8 ;
                              property Integer16 : int16
                                  read Get_Integer16
                                  write Set_Integer16 ;
                              property Unsigned_Integer16 : uint16
                                  read Get_Unsigned_Integer16
                                  write Set_Unsigned_Integer16 ;
                              property Integer32 : int32
                                  read Get_Integer32
                                  write Set_Integer32 ;
                              property Unsigned_Integer32 : uint32
                                  read Get_Unsigned_Integer32
                                  write Set_Unsigned_Integer32 ;
                              property Integer64 : int64
                                  read Get_Integer64
                                  write Set_Integer64 ;
                              property _Single : single
                                  read Get_Single
                                  write Set_Single ;
                              property _Double : double
                                  read Get_Double
                                  write Set_Double ;
                              property _Extended : extended
                                  read Get_Extended
                                  write Set_Extended ;
                      end ;

     TCOM_Generic_Value = class ;
     TCOM_Generic_Parameter = class( TCommon_COM_Interface ) { Parameter information... }
                                  public
                                      function Typ : TCOM_Generic_Value ;
                                          virtual ; stdcall ; abstract ;
                                      { Type of parameter }
                                      function Output : boolean ;
                                          virtual ; stdcall ; abstract ;
                                      { Purpose of parameter is for returning data }
                                      function Constant : boolean ;
                                          virtual ; stdcall ; abstract ;
                                      { True if constant }
                                      function Mechanism : TCall_Mechanism ;
                                          virtual ; stdcall ; abstract ;
                                      function Default : TCOM_Generic_Value ;
                                          virtual ; stdcall ; abstract ;
                                      { Default value of parameter }
                              end ;

     TCOM_Generic_Number = class( TCommon_COM_Interface )
                               public
                                   function Fixed : boolean ;
                                       virtual ; stdcall ; abstract ;
                                   { True if integer or fixed length real }
                                   function Signed : boolean ;
                                       virtual ; stdcall ; abstract ;
                                   function Mantissa_Size : longint ;
                                       virtual ; stdcall ; abstract ;
                                   { In bits, for Floating types }
                                   function Exponent_Size : longint ;
                                       virtual ; stdcall ; abstract ;
                                   { In bits, for Floating types (0 for non-floating) }
                                   function Fractional_Size : longint ;
                                       virtual ; stdcall ; abstract ;
                                   { In bytes, for fixed-point types }
                           end ;

     TCOM_Generic_Character = class( TCommon_COM_Interface )
                                  public
                                      function Max_Length : longint ;
                                          virtual ; stdcall ; abstract ;
                                      { Maximum allowable length (same as Length except for dynamic-length strings) }
                                      function Fixed : boolean ;
                                          virtual ; stdcall ; abstract ;
                                      { True if fixed length }
                                      function Length : longint ;
                                          virtual ; stdcall ; abstract ;
                                      { Length in bytes }
                                      function Logical_Length : longint ;
                                          virtual ; stdcall ; abstract ;
                                      { Length in characters }
                              end ;

     TCOM_Generic_Collection = class( TCommon_COM_Interface )
                                   public
                                       function Fixed : boolean ;
                                           virtual ; stdcall ; abstract ;
                                       { False if dynamic array }
                                       function Sparse : boolean ;
                                           virtual ; stdcall ; abstract ;
                                       { True if sparse array }
                                       function Subscripts : longint ;
                                           virtual ; stdcall ; abstract ;
                                       { Number of subscripts }
                                       function Low_Bound( Subscript : longint ) : TCOM_Generic_Value ;
                                           virtual ; stdcall ; abstract ;
                                       { Lower bound of subscript for array }
                                       function High_Bound( Subscript : longint ) : TCOM_Generic_Value ;
                                           virtual ; stdcall ; abstract ;
                                       { Upper bound of subscript for array }
                                       function Get_Subscript( Address : longint ) : TCOM_Generic_Value ;
                                           virtual ; stdcall ; abstract ;
                                       { Get value from array (viewing it as a flat one-dimensional array) }
                                       function Row_Major : boolean ;
                                           virtual ; stdcall ; abstract ;
                                       { True if array is left-to-right major }
                               end ;

     TCOM_Generic_Composite = class( TCommon_COM_Interface )
                                  public
                                      function Ancestor( index : longint ) : TCOM_Generic_Value ;
                                          virtual ; stdcall ; abstract ;
                                      function Ancestor_Count : longint ;
                                          virtual ; stdcall ; abstract ;
                                      { Number of ancestors }
                                      function Ancestor_Class( Index : longint ) : TMethod_Class ;
                                          virtual ; stdcall ; abstract ;
                                      { How Indexth ancestor was inherited }

                                      function Count : longint ;
                                          virtual ; stdcall ; abstract ;
                                      { Number of values }
                                      function Get_Item( Index : longint ) : TCOM_Generic_Value ;
                                          virtual ; stdcall ; abstract ;
                                      { Return value Index }
                              end ;

     TCOM_Generic_Method = class( TCommon_COM_Interface )
                               public
                                   function Is_Abstract : boolean ;
                                       virtual ; stdcall ; abstract ;
                                   { True if abstract (pure virtual) }
                                   function Is_Constructor : boolean ;
                                       virtual ; stdcall ; abstract ;
                                   function Is_Destructor : boolean ;
                                       virtual ; stdcall ; abstract ;
                                   function Is_Virtual : boolean ;
                                       virtual ; stdcall ; abstract ;
                                   { True if virtual method }
                                   function Static : boolean ;
                                       virtual ; stdcall ; abstract ;
                                   { True if a static class method }
                                   function C_Order : boolean ;
                                       virtual ; stdcall ; abstract ;
                                   { True if C calling convention parameter order }
                                   function Standard : boolean ;
                                       virtual ; stdcall ; abstract ;
                                   { True if standard (stack) parameter passing }
                                   function Execute( Parameters : TCOM_Generic_Value ) : TCOM_Generic_Value ;
                                       virtual ; stdcall ; abstract ;
                                   { Execute the method }
                                   function Compile_Time : boolean ;
                                      virtual ; stdcall ; abstract ;
                                   { True if the routine can be evaluated at
                                     compile-time. }
                                   function Count : longint ;
                                       virtual ; stdcall ; abstract ;
                                   { Number of parameters (run-time) }
                                   function Max_Count : longint ;
                                       virtual ; stdcall ; abstract ;
                                   { Maximum potential number of parameters }
                                   function Parameter( Index : integer ) : TCOM_Generic_Value ;
                                       virtual ; stdcall ; abstract ;
                                   { Returns parameter for specified index }
                                   function Target : boolean ;
                                       virtual ; stdcall ; abstract ;
                                   { True if a target function }
                                   function Method_Class : TMethod_Class ;
                                       virtual ; stdcall ; abstract ;
                           end ;

     TCOM_Property_Information = class( TCommon_COM_Interface )
                                     public
                                         function Getter : TCOM_Generic_Value ;
                                             virtual ; stdcall ; abstract ;
                                         function Setter : TCOM_Generic_Value ;
                                             virtual ; stdcall ; abstract ;
                                         function Default : boolean ;
                                             virtual ; stdcall ; abstract ;
                                 end ;

     TCOM_Type_Information = class( TCommon_COM_Interface )
                                 public
                                     function Is_Composite : boolean ;
                                         virtual ; stdcall ; abstract ;
                                     { True if an array/queue/list/stack of Data_Type }
                                     function Is_Collection : boolean ;
                                         virtual ; stdcall ; abstract ;
                                     { True if an object or record }
                                     function Memory_Type : TMemory_Type ;
                                         virtual ; stdcall ; abstract ;

                                     function Parameter : TCOM_Generic_Parameter ;
                                         virtual ; stdcall ; abstract ;
                                     { nil if not a parameter }
                                     function Numeric_Information : TCOM_Generic_Number ;
                                         virtual ; stdcall ; abstract ;
                                     { nil if not numeric }
                                     function Character_Information : TCOM_Generic_Character ;
                                         virtual ; stdcall ; abstract ;
                                     { nil if not character/string }
                                     function Composite_Information : TCOM_Generic_Composite ;
                                         virtual ; stdcall ; abstract ;
                                     { nil if not an object/record }
                                     function Collection_Information : TCOM_Generic_Collection ;
                                         virtual ; stdcall ; abstract ;
                                     { nil if not a composite (eg array) }
                                     function Method : TCOM_Generic_Method ;
                                         virtual ; stdcall ; abstract ;
                                     { nil if not a method/procedure }
                                     function _Property : TCOM_Property_Information ;
                                         virtual ; stdcall ; abstract ; {
                                     { nil if not a property }
                                     function Address : longint ;
                                         virtual ; stdcall ; abstract ;
                                     { 0 if undefined, type, or parameter }

                                     function Range_Limited : boolean ;
                                         virtual ; stdcall ; abstract ;
                                     { True if value is limited to a range of values.
                                       This is always true for scalars. }

                                     { (These return nil if not range limited) }
                                     function Low_Range : TCOM_Generic_Value ;
                                         virtual ; stdcall ; abstract ;
                                     function High_Range : TCOM_Generic_Value ;
                                         virtual ; stdcall ; abstract ;

                                     { For scalars... }
                                     function Value_Name( Value : longint ) : string ;
                                         virtual ; stdcall ; abstract ;
                                     { Name for given ordinal value.  Returns null
                                       string if not a scalara or if Value is out of
                                       range. }
                             end ;

     TCOM_Generic_Value = class( TCommon_COM_Interface )
                              public { API... }
                                  { Meta type information... }
                                  function Family : TData_Type_Family ;
                                      virtual ; stdcall ; abstract ;
                                  function Data_Type : TValue_Data_Type ;
                                      virtual ; stdcall ; abstract ;
                                  function User_Data_Type_Name : string ;
                                      virtual ; stdcall ; abstract ;
                                  function Name : string ;
                                      virtual ; stdcall ; abstract ;
                                  { Variable, type, method, or parameter name }
                                  function Type_Information : TCOM_Type_Information ;
                                      virtual ; stdcall ; abstract ;

                                  { Size, address, and format information... }
                                  function Constant : boolean ;
                                      virtual ; stdcall ; abstract ;
                                  { True if a constant rather than variable }
                                  function Size : longint ;
                                      virtual ; stdcall ; abstract ;
                                  { Size in bytes }
                                  function Total_Size : longint ;
                                      virtual ; stdcall ; abstract ;
                                  { Size of all associated data, in bytes.  For
                                    instance, Size may return the size of an object
                                    but Total_Size also includes the size of any
                                    associated data. }

                                  { For pointers... }
                                  function Dereference : TCOM_Generic_Value ;
                                      virtual ; stdcall ; abstract ;

                                  { Conversions/promotions/casts... }
                                  function Can_Convert( Typ : TValue_Data_Type ) : boolean ;
                                      virtual ; stdcall ; abstract ; 
                                  { True if value can be converted to the specified data type }
                                  function Convert( Class_Type : TCOM_Generic_Value ;
                                      var Loss : boolean ) : TCOM_Generic_Value ;
                                      virtual ; stdcall ; abstract ;
                                  { Return a version of the value converted to the
                                    passed type specification.  Returns nil if the
                                    conversion cannot be done.  Loss is ignored
                                    when passed and is True on return if there
                                    was a loss of precision. }

                                  function Clone : TCOM_Generic_Value ;
                                      virtual ; stdcall ; abstract ;

                                  procedure Simple_Cast( Typ : TValue_Data_Type ) ;
                                      virtual ; stdcall ; abstract ;
                                  procedure Cast( Typ : TCOM_Generic_Value ) ;
                                      virtual ; stdcall ; abstract ;

                                  { Other... }
                                  procedure Deserialize( Value : PChar ) ;
                                      virtual ; stdcall ; abstract ;
                                  function Serialize : PChar ;
                                      virtual ; stdcall ; abstract ;

                                  { Get/Set value methods... }
                                  procedure Set_Null ;
                                      virtual ; stdcall ; abstract ;
                                  procedure Get_Undefined( Buffer : pointer ;
                                      Len : longint ) ;
                                      virtual ; stdcall ; abstract ;
                                  procedure Set_Undefined( Buffer : pointer ;
                                      Len : longint ) ;
                                      virtual ; stdcall ; abstract ;

                                  function Get_Boolean : boolean ;
                                      virtual ; stdcall ; abstract ;
                                  procedure Set_Boolean( B : boolean ) ;
                                      virtual ; stdcall ; abstract ;

                                  function Get_ByteBool : bytebool ;
                                      virtual ; stdcall ; abstract ;
                                  procedure Set_ByteBool( B : bytebool ) ;
                                      virtual ; stdcall ; abstract ;

                                  function Get_WordBool : wordbool ;
                                      virtual ; stdcall ; abstract ;
                                  procedure Set_WordBool( B : wordbool ) ;
                                      virtual ; stdcall ; abstract ;

                                  function Get_LongBool : longbool ;
                                      virtual ; stdcall ; abstract ;
                                  procedure Set_LongBool( B : longbool ) ;
                                      virtual ; stdcall ; abstract ;

                                  function Get_Char : char ;
                                      virtual ; stdcall ; abstract ;
                                  procedure Set_Char( C : char ) ;
                                      virtual ; stdcall ; abstract ;

                                  function Get_String : string ;
                                      virtual ; stdcall ; abstract ;
                                  procedure Set_String( const S : string ) ;
                                      virtual ; stdcall ; abstract ;

                                  function Get_PChar : PChar ;
                                      virtual ; stdcall ; abstract ;
                                  procedure Set_PChar( P : PChar ) ;
                                      virtual ; stdcall ; abstract ;

                                  function Get_Integer8 : int8 ;
                                      virtual ; stdcall ; abstract ;
                                  procedure Set_Integer8( I : int8 ) ;
                                      virtual ; stdcall ; abstract ;

                                  function Get_Integer16 : int16 ;
                                      virtual ; stdcall ; abstract ;
                                  procedure Set_Integer16( I : int16 ) ;
                                      virtual ; stdcall ; abstract ;

                                  function Get_Integer32 : int32 ;
                                      virtual ; stdcall ; abstract ;
                                  procedure Set_Integer32( I : int32 ) ;
                                      virtual ; stdcall ; abstract ;

                                  function Get_Integer64 : int64 ;
                                      virtual ; stdcall ; abstract ;
                                  procedure Set_Integer64( I : int64 ) ;
                                      virtual ; stdcall ; abstract ;

                                  function Get_Unsigned_Integer8 : uint8 ;
                                      virtual ; stdcall ; abstract ;
                                  procedure Set_Unsigned_Integer8( I : uint8 ) ;
                                      virtual ; stdcall ; abstract ;

                                  function Get_Unsigned_Integer16 : uint16 ;
                                      virtual ; stdcall ; abstract ;
                                  procedure Set_Unsigned_Integer16( I : uint16 ) ;
                                      virtual ; stdcall ; abstract ;

                                  function Get_Unsigned_Integer32 : uint32 ;
                                      virtual ; stdcall ; abstract ;
                                  procedure Set_Unsigned_Integer32( I : uint32 ) ;
                                      virtual ; stdcall ; abstract ;

                                  function Get_Single : single ;
                                     virtual ; stdcall ; abstract ;
                                  procedure Set_Single( S : single ) ;
                                      virtual ; stdcall ; abstract ;

                                  function Get_Double : double ;
                                      virtual ; stdcall ; abstract ;
                                  procedure Set_Double( D : double ) ;
                                      virtual ; stdcall ; abstract ;

                                  function Get_Extended : extended ;
                                      virtual ; stdcall ; abstract ;
                                  procedure Set_Extended( E : extended ) ;
                                      virtual ; stdcall ; abstract ;

                                  property Bool : boolean
                                      read Get_Boolean
                                      write Set_Boolean ;
                                  property Byte_Bool : bytebool
                                      read Get_ByteBool
                                      write Set_ByteBool ;
                                  property Word_Bool : wordbool
                                      read Get_WordBool
                                      write Set_WordBool ;
                                  property Long_Bool : longbool
                                      read Get_LongBool
                                      write Set_LongBool ;
                                  property _Char : char
                                      read Get_Char
                                      write Set_Char ;
                                  property _String : string
                                      read Get_String
                                      write Set_String ;
                                  property _PChar : PChar
                                      read Get_PChar
                                      write Set_PChar ;
                                  property Integer8 : int8
                                      read Get_Integer8
                                      write Set_Integer8 ;
                                  property Unsigned_Integer8 : uint8
                                      read Get_Unsigned_Integer8
                                      write Set_Unsigned_Integer8 ;
                                  property Integer16 : int16
                                      read Get_Integer16
                                      write Set_Integer16 ;
                                  property Unsigned_Integer16 : uint16
                                      read Get_Unsigned_Integer16
                                      write Set_Unsigned_Integer16 ;
                                  property Integer32 : int32
                                      read Get_Integer32
                                      write Set_Integer32 ;
                                  property Unsigned_Integer32 : uint32
                                      read Get_Unsigned_Integer32
                                      write Set_Unsigned_Integer32 ;
                                  property Integer64 : int64
                                      read Get_Integer64
                                      write Set_Integer64 ;
                                  property _Single : single
                                      read Get_Single
                                      write Set_Single ;
                                  property _Double : double
                                      read Get_Double
                                      write Set_Double ;
                                  property _Extended : extended
                                      read Get_Extended
                                      write Set_Extended ;
                          end ;
{$ENDIF}

type PSGeneric_Value = ^TSGeneric_Value ;
     TSGeneric_Parameter = object( TObject ) { Parameter information... }
                               public
                                   function Typ : PSGeneric_Value ; virtual ;
                                   { Type of parameter }
                                   function Output : boolean ; virtual ; 
                                   { Purpose of parameter is for returning data }
                                   function Constant : boolean ; virtual ;
                                   { True if constant }
                                   function Mechanism : TCall_Mechanism ;
                                       virtual ;
                                   function Default : PSGeneric_Value ; virtual ;
                                   { Default value of parameter }
                           end ;
     PSGeneric_Parameter = ^TSGeneric_Parameter ;

     TSGeneric_Number = object( TObject )
                            public
                                function Fixed : boolean ; virtual ; 
                                { True if integer or fixed length real }
                                function Signed : boolean ; virtual ;
                                function Mantissa_Size : longint ; virtual ;
                                { In bits, for Floating types }
                                function Exponent_Size : longint ; virtual ;
                                { In bits, for Floating types (0 for non-floating) }
                                function Fractional_Size : longint ; virtual ;
                                { In bytes, for fixed-point types }
                        end ;
     PSGeneric_Number = ^TSGeneric_Number ;

     TSGeneric_Character = object( TObject )
                               public
                                   function Max_Length : longint ; virtual ;
                                   { Maximum allowable length (same as Length except for dynamic-length strings) }
                                   function Fixed : boolean ; virtual ;
                                   { True if fixed length }
                                   function Length : longint ; virtual ;
                                   { Length in bytes }
                                   function Logical_Length : longint ; virtual ;
                                   { Length in characters }
                           end ;
     PSGeneric_Character = ^TSGeneric_Character ;

     TSGeneric_Collection = object( TObject )
                                public
                                    function Fixed : boolean ; virtual ;
                                    { False if dynamic array }
                                    function Sparse : boolean ; virtual ;
                                    { True if sparse array }
                                    function Subscripts : longint ; virtual ;
                                    { Number of subscripts }
                                    function Low_Bound( Subscript : longint ) : PSGeneric_Value ;
                                        virtual ;
                                    { Lower bound of subscript for array }
                                    function High_Bound( Subscript : longint ) : PSGeneric_Value ;
                                        virtual ;
                                    { Upper bound of subscript for array }
                                    function Get_Subscript( Address : longint ) : PSGeneric_Value ;
                                        virtual ;
                                    { Get value from array (viewing it as a flat one-dimensional array) }
                                    function Row_Major : boolean ; virtual ;
                                    { True if array is left-to-right major }
                            end ;
     PSGeneric_Collection = ^TSGeneric_Composite ;

     TSGeneric_Composite = object( TObject )
                               public
                                   function Ancestor( index : longint ) : PSGeneric_Value ;
                                       virtual ;
                                   function Ancestor_Count : longint ; virtual ;
                                   { Number of ancestors }
                                   function Ancestor_Class( Index : longint ) : TMethod_Class ;
                                       virtual ; { How ancestor was inherited }

                                   function Count : longint ; virtual ;
                                   { Number of values }
                                   function Get_Item( Index : longint ) : PSGeneric_Value ;
                                       virtual ; { Return value Index }
                           end ;
     PSGeneric_Composite = ^TSGeneric_Composite ;

     TSGeneric_Method = object( TObject )
                            public
                                function Is_Abstract : boolean ; virtual ;
                                { True if abstract (pure virtual) }
                                function Is_Constructor : boolean ; virtual ;
                                function Is_Destructor : boolean ; virtual ;
                                function Is_Virtual : boolean ; virtual ;
                                { True if virtual method }
                                function Static : boolean ; virtual ;
                                { True if a static class method }
                                function C_Order : boolean ; virtual ;
                                { True if C calling convention parameter order }
                                function Standard : boolean ; virtual ;
                                { True if standard (stack) parameter passing }
                                function Execute( Parameters : PSGeneric_Value ) : PSGeneric_Value ;
                                    virtual ; { Execute the method }
                                function Compile_Time : boolean ; virtual ;
                                { True if the routine can be evaluated at
                                  compile-time. }
                                function Count : longint ; virtual ;
                                { Number of parameters (run-time) }
                                function Max_Count : longint ; virtual ;
                                { Maximum potential number of parameters }
                                function Parameter( Index : integer ) : PSGeneric_Value ;
                                    virtual ;
                                { Returns parameter for specified index }
                                function Target : boolean ; virtual ;
                                { True if a target function }
                                function Method_Class : TMethod_Class ; virtual ;
                        end ;
     PSGeneric_Method = ^TSGeneric_Method ;

     TSProperty_Information = object( TObject )
                                  public
                                      function Getter : PSGeneric_Value ; virtual ;
                                      function Setter : PSGeneric_Value ; virtual ;
                                      function Default : boolean ; virtual ;
                              end ;
     PSProperty_Information = ^TSProperty_Information ;

     TSType_Information = object( TObject )
                              public
                                  function Is_Composite : boolean ; virtual ;
                                  { True if an array/queue/list/stack of Data_Type }
                                  function Is_Collection : boolean ; virtual ;
                                  { True if an object or record }
                                  function Memory_Type : TMemory_Type ; virtual ;

                                  function Parameter : PSGeneric_Parameter ;
                                  { nil if not a parameter }
                                      virtual ;
                                  function Numeric_Information : PSGeneric_Number ;
                                  { nil if not numeric }
                                      virtual ;
                                  function Character_Information : PSGeneric_Character ;
                                      virtual ; { nil if not character/string }
                                  function Composite_Information : PSGeneric_Composite ;
                                      virtual ; { nil if not an object/record }
                                  function Collection_Information : PSGeneric_Collection ;
                                      virtual ; { nil if not a composite (eg array) }
                                  function Method : PSGeneric_Method ; virtual ;
                                  { nil if not a method/procedure }
                                  function _Property : PSProperty_Information ;
                                      virtual ; { nil if not a property }
                                  function Address : longint ; virtual ;
                                  { 0 if undefined, type, or parameter }

                                  function Range_Limited : boolean ; virtual ;
                                  { True if value is limited to a range of values.
                                    This is always true for scalars. }

                                  { (These return nil if not range limited) }
                                  function Low_Range : PSGeneric_Value ; virtual ;
                                  function High_Range : PSGeneric_Value ; virtual ;

                                  { For scalars... }
                                  function Value_Name( Value : longint ) : string ;
                                      virtual ;
                                  { Name for given ordinal value.  Returns null
                                    string if not a scalara or if Value is out
                                    of range. }
                          end ;
     PSType_Information = ^TSType_Information ;

     TSGeneric_Value = object( TObject )
                           public { API... }
                               { Meta type information... }
                               function Family : TData_Type_Family ; virtual ;
                               function Data_Type : TValue_Data_Type ; virtual ;
                               function User_Data_Type_Name : string ; virtual ;
                               function Name : string ; virtual ;
                               { Variable, type, method, or parameter name }
                               function Type_Information : PSType_Information ;
                                   virtual ;

                               { Size, address, and format information... }
                               function Constant : boolean ; virtual ;
                               { True if a constant rather than variable }
                               function Size : longint ; virtual ;
                               { Size in bytes }
                               function Total_Size : longint ; virtual ;
                               { Size of all associated data, in bytes.  For
                                 instance, Size may return the size of an object
                                 but Total_Size also includes the size of any
                                 associated data. }

                               { For pointers... }
                               function Dereference : PSGeneric_Value ;
                                   virtual ;

                               { Conversions/promotions/casts... }
                               function Can_Convert( Typ : TValue_Data_Type ) : boolean ;
                                   virtual ;
                               { True if value can be converted to the specified data type }

                               function Clone : PSGeneric_Value ; virtual ;

                               function Convert( Class_Type : PSGeneric_Value ;
                                   var Loss : boolean ) : PSGeneric_Value ;
                                   virtual ;
                               { Return a version of the value converted to the
                                 passed type specification.  Returns nil if the
                                 conversion cannot be done.  Loss is ignored
                                 when passed.  On return it is True if there was
                                 a loss of precision. }

                               procedure Simple_Cast( Typ : TValue_Data_Type ) ;
                                   virtual ;
                               procedure Cast( Typ : PSGeneric_Value ) ;
                                   virtual ;

                               { Other... }
                               procedure Deserialize( Value : PChar ) ; virtual ;
                               function Serialize : PChar ; virtual ;

                               { Get/Set value methods... }
                               procedure Set_Null ; virtual ;
                               procedure Get_Undefined( Buffer : pointer ;
                                   Len : longint ) ; virtual ;
                               procedure Set_Undefined( Buffer : pointer ;
                                   Len : longint ) ; virtual ;

                               function Get_Boolean : boolean ; virtual ;
                               procedure Set_Boolean( B : boolean ) ; virtual ;

                               function Get_ByteBool : bytebool ; virtual ;
                               procedure Set_ByteBool( B : bytebool ) ; virtual ;

                               function Get_WordBool : wordbool ; virtual ;
                               procedure Set_WordBool( B : wordbool ) ; virtual ;

                               function Get_LongBool : longbool ; virtual ;
                               procedure Set_LongBool( B : longbool ) ; virtual ;

                               function Get_Char : char ; virtual ;
                               procedure Set_Char( C : char ) ; virtual ;

                               function Get_String : string ; virtual ;
                               procedure Set_String( const S : string ) ; virtual ;

                               function Get_PChar : PChar ; virtual ;
                               procedure Set_PChar( P : PChar ) ; virtual ;

                               function Get_Integer8 : int8 ; virtual ;
                               procedure Set_Integer8( I : int8 ) ; virtual ;

                               function Get_Integer16 : int16 ; virtual ;
                               procedure Set_Integer16( I : int16 ) ; virtual ;

                               function Get_Integer32 : int32 ; virtual ;
                               procedure Set_Integer32( I : int32 ) ; virtual ;

                               function Get_Integer64 : int64 ;
                                   virtual ;
                               procedure Set_Integer64( I : int64 ) ;
                                   virtual ;

                               function Get_Unsigned_Integer8 : uint8 ; virtual ;
                               procedure Set_Unsigned_Integer8( I : uint8 ) ; virtual ;

                               function Get_Unsigned_Integer16 : uint16 ; virtual ;
                               procedure Set_Unsigned_Integer16( I : uint16 ) ; virtual ;

{$IFDEF DELPHI}
                               function Get_Unsigned_Integer32 : uint32 ;
                                   virtual ;
                               procedure Set_Unsigned_Integer32( I : uint32 ) ;
                                  virtual ;
{$ENDIF}

                               function Get_Single : single ; virtual ;
                               procedure Set_Single( S : single ) ; virtual ;

                               function Get_Double : double ; virtual ;
                               procedure Set_Double( D : double ) ; virtual ;

                               function Get_Extended : extended ; virtual ;
                               procedure Set_Extended( E : extended ) ; virtual ;
{$IFDEF DELPHI}
                              property Bool : boolean
                                  read Get_Boolean
                                  write Set_Boolean ;
                              property Byte_Bool : bytebool
                                  read Get_ByteBool
                                  write Set_ByteBool ;
                              property Word_Bool : wordbool
                                  read Get_WordBool
                                  write Set_WordBool ;
                              property Long_Bool : longbool
                                  read Get_LongBool
                                  write Set_LongBool ;
                              property _Char : char
                                  read Get_Char
                                  write Set_Char ;
                              property _String : string
                                  read Get_String
                                  write Set_String ;
                              property _PChar : PChar
                                  read Get_PChar
                                  write Set_PChar ;
                              property Integer8 : int8
                                  read Get_Integer8
                                  write Set_Integer8 ;
                              property Unsigned_Integer8 : uint8
                                  read Get_Unsigned_Integer8
                                  write Set_Unsigned_Integer8 ;
                              property Integer16 : int16
                                  read Get_Integer16
                                  write Set_Integer16 ;
                              property Unsigned_Integer16 : uint16
                                  read Get_Unsigned_Integer16
                                  write Set_Unsigned_Integer16 ;
                              property Integer32 : int32
                                  read Get_Integer32
                                  write Set_Integer32 ;
{$IFDEF DELPHI}
                               property Unsigned_Integer32 : uint32
                                   read Get_Unsigned_Integer32
                                   write Set_Unsigned_Integer32 ;
{$ENDIF}
                              property _Single : single
                                  read Get_Single
                                  write Set_Single ;
                              property _Double : double
                                  read Get_Double
                                  write Set_Double ;
                              property _Extended : extended
                                  read Get_Extended
                                  write Set_Extended ;
{$ENDIF}
                       end ;

implementation

{ TSGeneric_Parameter methods... }

function TSGeneric_Parameter.Typ : PSGeneric_Value ;

begin
    Typ := nil ;
    Abstract_Error ;
end ;


function TSGeneric_Parameter.Output : boolean ;

begin
    Output := False ;
    Abstract_Error ;
end ;


function TSGeneric_Parameter.Constant : boolean ;

begin
    Constant := True ;
    Abstract_Error ;
end ;


function TSGeneric_Parameter.Mechanism : TCall_Mechanism ;

begin
    Mechanism := CM_By_Value ;
    Abstract_Error ;
end ;


function TSGeneric_Parameter.Default : PSGeneric_Value ;

begin
    Default := nil ;
    Abstract_Error ;
end ;



{ TSGeneric_Number methods... }

function TSGeneric_Number.Fixed : boolean ;

begin
    Fixed := True ;
    Abstract_Error ;
end ;


function TSGeneric_Number.Signed : boolean ;

begin
    Signed := True ;
    Abstract_Error ;
end ;


function TSGeneric_Number.Mantissa_Size : longint ;

begin
    Mantissa_Size := 0 ;
    Abstract_Error ;
end ;


function TSGeneric_Number.Exponent_Size : longint ;

begin
    Exponent_Size := 0 ;
    Abstract_Error ;
end ;


function TSGeneric_Number.Fractional_Size : longint ;

begin
    Fractional_Size := 0 ;
    Abstract_Error ;
end ;



{ TSGeneric_Character methods... }

function TSGeneric_Character.Max_Length : longint ;

begin
    Max_Length := 0 ;
    Abstract_Error ;
end ;


function TSGeneric_Character.Fixed : boolean ;

begin
    Fixed := True ;
    Abstract_Error ;
end ;


function TSGeneric_Character.Length : longint ;

begin
    Length := 0 ;
    Abstract_Error ;
end ;


function TSGeneric_Character.Logical_Length : longint ;

begin
    Logical_Length := 0 ;
    Abstract_Error ;
end ;



{ TSGeneric_Collection methods... }

function TSGeneric_Collection.Fixed : boolean ;

begin
    Fixed := True ;
    Abstract_Error ;
end ;


function TSGeneric_Collection.Sparse : boolean ;

begin
    Sparse := False ;
    Abstract_Error ;
end ;


function TSGeneric_Collection.Subscripts : longint ;

begin
    Subscripts := 0 ;
    Abstract_Error ;
end ;


function TSGeneric_Collection.High_Bound( Subscript : longint ) : PSGeneric_Value ;

begin
    High_Bound := nil ;
    Abstract_Error ;
end ;


function TSGeneric_Collection.Low_Bound( Subscript : longint ) : PSGeneric_Value ;

begin
    Low_Bound := nil ;
    Abstract_Error ;
end ;


function TSGeneric_Collection.Get_Subscript( Address : longint ) : PSGeneric_Value ;

begin
    Get_Subscript := nil ;
    Abstract_Error ;
end ;


function TSGeneric_Collection.Row_Major : boolean ;

begin
    Row_Major := True ;
    Abstract_Error ;
end ;



{ TSGeneric_Composite methods... }

function TSGeneric_Composite.Ancestor( index : longint ) : PSGeneric_Value ;

begin
    Ancestor := nil ;
    Abstract_Error ;
end ;


function TSGeneric_Composite.Ancestor_Count : longint ;

begin
    Ancestor_Count := 0 ;
    Abstract_Error ;
end ;


function TSGeneric_Composite.Ancestor_Class( Index : longint ) : TMethod_Class ;

begin
    Ancestor_Class := MC_Private ;
    Abstract_Error ;
end ;


function TSGeneric_Composite.Count : longint ;

begin
    Count := 0 ;
    Abstract_Error ;
end ;


function TSGeneric_Composite.Get_Item( Index : longint ) : PSGeneric_Value ;

begin
    Get_Item := nil ;
    Abstract_Error ;
end ;



{ TSGeneric_Method methods... }

function TSGeneric_Method.Is_Abstract : boolean ;

begin
    Is_Abstract := False ;
    Abstract_Error ;
end ;


function TSGeneric_Method.Is_Constructor : boolean ;

begin
    Is_Constructor := False ;
    Abstract_Error ;
end ;


function TSGeneric_Method.Is_Destructor : boolean ;

begin
    Is_Destructor := False ;
    Abstract_Error ;
end ;


function TSGeneric_Method.Is_Virtual : boolean ;

begin
    Is_Virtual := False ;
    Abstract_Error ;
end ;


function TSGeneric_Method.Static : boolean ;

begin
    Static := False ;
    Abstract_Error ;
end ;


function TSGeneric_Method.C_Order : boolean ;

begin
    C_Order := False ;
    Abstract_Error ;
end ;


function TSGeneric_Method.Standard : boolean ;

begin
    Standard := True ;
    Abstract_Error ;
end ;


function TSGeneric_Method.Execute( Parameters : PSGeneric_Value ) : PSGeneric_Value ;

begin
    Execute := nil ;
    Abstract_Error ;
end ;


function TSGeneric_Method.Compile_Time : boolean ;

begin
    Compile_Time := False ;
    Abstract_Error ;
end ;


function TSGeneric_Method.Count : longint ;

begin
    Count := 0 ;
    Abstract_Error ;
end ;


function TSGeneric_Method.Max_Count : longint ;

begin
    Max_Count := 0 ;
    Abstract_Error ;
end ;


function TSGeneric_Method.Parameter( Index : integer ) : PSGeneric_Value ;

begin
    Parameter := nil ;
    Abstract_Error ;
end ;


function TSGeneric_Method.Target : boolean ;

begin
    Target := False ;
    Abstract_Error ;
end ;


function TSGeneric_Method.Method_Class : TMethod_Class ;

begin
    Method_Class := MC_Public ;
    Abstract_Error ;
end ;



{ TSProperty_Information methods... }

function TSProperty_Information.Getter : PSGeneric_Value ;

begin
    Getter := nil ;
    Abstract_Error ;
end ;


function TSProperty_Information.Setter : PSGeneric_Value ;

begin
    Setter := nil ;
    Abstract_Error ;
end ;


function TSProperty_Information.Default : boolean ;

begin
    Default := False ;
    Abstract_Error ;
end ;



{ TSType_Information methods... }

function TSType_Information.Is_Composite : boolean ;

begin
    Is_Composite := False ;
    Abstract_Error ;
end ;


function TSType_Information.Is_Collection : boolean ;

begin
    Is_Collection := False ;
    Abstract_Error ;
end ;


function TSType_Information.Memory_Type : TMemory_Type ;

begin
    Memory_Type := MT_Unknown ;
    Abstract_Error ;
end ;


function TSType_Information.Parameter : PSGeneric_Parameter ;

begin
    Parameter := nil ;
    Abstract_Error ;
end ;


function TSType_Information.Numeric_Information : PSGeneric_Number ;

begin
    Numeric_Information := nil ;
    Abstract_Error ;
end ;


function TSType_Information.Character_Information : PSGeneric_Character ;

begin
    Character_Information := nil ;
    Abstract_Error ;
end ;


function TSType_Information.Composite_Information : PSGeneric_Composite ;

begin
    Composite_Information := nil ;
    Abstract_Error ;
end ;


function TSType_Information.Collection_Information : PSGeneric_Collection ;

begin
    Collection_Information := nil ;
    Abstract_Error ;
end ;


function TSType_Information.Method : PSGeneric_Method ;

begin
    Method := nil ;
    Abstract_Error ;
end ;


function TSType_Information._Property : PSProperty_Information ;

begin
    _Property := nil ;
    Abstract_Error ;
end ;


function TSType_Information.Address : longint ;

begin
    Address := 0 ;
    Abstract_Error ;
end ;


function TSType_Information.Range_Limited : boolean ;

begin
    Range_Limited := False ;
    Abstract_Error ;
end ;


function TSType_Information.Low_Range : PSGeneric_Value ;

begin
    Low_Range := nil ;
    Abstract_Error ;
end ;


function TSType_Information.High_Range : PSGeneric_Value ;

begin
    High_Range := nil ;
    Abstract_Error ;
end ;


function TSType_Information.Value_Name( Value : longint ) : string ;

begin
    Value_Name := '' ;
    Abstract_Error ;
end ;



{ TSGeneric_Value methods... }

function TSGeneric_Value.Family : TData_Type_Family ;

begin
    Family := DTF_Unknown ;
    Abstract_Error ;
end ;


function TSGeneric_Value.Data_Type : TValue_Data_Type ;

begin
    Data_Type := VDT_Unknown ;
    Abstract_Error ;
end ;


function TSGeneric_Value.User_Data_Type_Name : string ;

begin
    User_Data_Type_Name := '' ;
    Abstract_Error ;
end ;


function TSGeneric_Value.Name : string ; { Variable, type, method, or parameter name }

begin
    Name := '' ;
    Abstract_Error ;
end ;


function TSGeneric_Value.Type_Information : PSType_Information ;

begin
    Type_Information := nil ;
    Abstract_Error ;
end ;


function TSGeneric_Value.Constant : boolean ;

begin
    Constant := True ;
    Abstract_Error ;
end ;


function TSGeneric_Value.Size : longint ; { Size in bytes }

begin
    Size := 0 ;
    Abstract_Error ;
end ;


function TSGeneric_Value.Total_Size : longint ;

begin
    Total_Size := 0 ;
    Abstract_Error ;
end ;


function TSGeneric_Value.Dereference : PSGeneric_Value ;

begin
    Dereference := nil ;
    Abstract_Error ;
end ;


function TSGeneric_Value.Can_Convert( Typ : TValue_Data_Type ) : boolean ;

begin
    Can_Convert := False ;
    Abstract_Error ;
end ;


function TSGeneric_Value.Convert( Class_Type : PSGeneric_Value ;
    var Loss : boolean ) : PSGeneric_Value ;

begin
    Convert := nil ;
    Abstract_Error ;
end ;


function TSGeneric_Value.Clone : PSGeneric_Value ;

begin
    Clone := nil ;
    Abstract_Error ;
end ;


procedure TSGeneric_Value.Simple_Cast( Typ : TValue_Data_Type ) ;

begin
    Abstract_Error ;
end ;


procedure TSGeneric_Value.Cast( Typ : PSGeneric_Value ) ;

begin
    Abstract_Error ;
end ;


procedure TSGeneric_Value.Deserialize( Value : PChar ) ;

begin
    Abstract_Error ;
end ;


function TSGeneric_Value.Serialize : PChar ;

begin
    Serialize := nil ;
    Abstract_Error ;
end ;


procedure TSGeneric_Value.Set_Null ;

begin
    Abstract_Error ;
end ;


procedure TSGeneric_Value.Get_Undefined( Buffer : pointer ; Len : longint ) ;

begin
    Abstract_Error ;
end ;


procedure TSGeneric_Value.Set_Undefined( Buffer : pointer ; Len : longint ) ;

begin
    Abstract_Error ;
end ;


function TSGeneric_Value.Get_Boolean : boolean ;

begin
    Get_Boolean := False ;
    Abstract_Error ;
end ;


procedure TSGeneric_Value.Set_Boolean( B : boolean ) ;

begin
    Abstract_Error ;
end ;


function TSGeneric_Value.Get_ByteBool : bytebool ;

begin
    Get_ByteBool := False ;
    Abstract_Error ;
end ;


procedure TSGeneric_Value.Set_ByteBool( B : bytebool ) ;

begin
    Abstract_Error ;
end ;


function TSGeneric_Value.Get_WordBool : wordbool ;

begin
    Get_WordBool := False ;
    Abstract_Error ;
end ;


procedure TSGeneric_Value.Set_WordBool( B : wordbool ) ;

begin
    Abstract_Error ;
end ;


function TSGeneric_Value.Get_LongBool : longbool ;

begin
    Get_LongBool := False ;
    Abstract_Error ;
end ;


procedure TSGeneric_Value.Set_LongBool( B : longbool ) ;

begin
    Abstract_Error ;
end ;


function TSGeneric_Value.Get_Char : char ;

begin
    Get_Char := #0 ;
    Abstract_Error ;
end ;


procedure TSGeneric_Value.Set_Char( C : char ) ;

begin
    Abstract_Error ;
end ;


function TSGeneric_Value.Get_String : string ;

begin
    Get_String := '' ;
    Abstract_Error ;
end ;


procedure TSGeneric_Value.Set_String( const S : string ) ;

begin
    Abstract_Error ;
end ;


function TSGeneric_Value.Get_PChar : PChar ;

begin
    Get_PChar := nil ;
    Abstract_Error ;
end ;


procedure TSGeneric_Value.Set_PChar( P : PChar ) ;

begin
    Abstract_Error ;
end ;


function TSGeneric_Value.Get_Integer8 : int8 ;

begin
    Get_Integer8 := 0 ;
    Abstract_Error ;
end ;


procedure TSGeneric_Value.Set_Integer8( I : int8 ) ;

begin
    Abstract_Error ;
end ;


function TSGeneric_Value.Get_Integer16 : int16 ;

begin
    Get_Integer16 := 0 ;
    Abstract_Error ;
end ;


procedure TSGeneric_Value.Set_Integer16( I : int16 ) ;

begin
    Abstract_Error ;
end ;


function TSGeneric_Value.Get_Integer32 : int32 ;

begin
    Get_Integer32 := 0 ;
    Abstract_Error ;
end ;


procedure TSGeneric_Value.Set_Integer32( I : int32 ) ;

begin
    Abstract_Error ;
end ;


function TSGeneric_Value.Get_Integer64 : int64 ;

begin
    Get_Integer64 := 0 ;
    Abstract_Error ;
end ;


procedure TSGeneric_Value.Set_Integer64( I : int64 ) ;

begin
    Abstract_Error ;
end ;


function TSGeneric_Value.Get_Unsigned_Integer8 : uint8 ;

begin
    Get_Unsigned_Integer8 := 0 ;
    Abstract_Error ;
end ;


procedure TSGeneric_Value.Set_Unsigned_Integer8( I : uint8 ) ;

begin
    Abstract_Error ;
end ;


function TSGeneric_Value.Get_Unsigned_Integer16 : uint16 ;

begin
    Get_Unsigned_Integer16 := 0 ;
    Abstract_Error ;
end ;


procedure TSGeneric_Value.Set_Unsigned_Integer16( I : uint16 ) ;

begin
    Abstract_Error ;
end ;


{$IFDEF DELPHI}
function TSGeneric_Value.Get_Unsigned_Integer32 : uint32 ;

begin
    Get_Unsigned_Integer32 := 0 ;
    Abstract_Error ;
end ;


procedure TSGeneric_Value.Set_Unsigned_Integer32( I : uint32 ) ;

begin
    Abstract_Error ;
end ;
{$ENDIF}


function TSGeneric_Value.Get_Single : single ;

begin
    Get_Single := 0.0 ;
    Abstract_Error ;
end ;


procedure TSGeneric_Value.Set_Single( S : single ) ;

begin
    Abstract_Error ;
end ;


function TSGeneric_Value.Get_Double : double ;

begin
    Get_Double := 0.9 ;
    Abstract_Error ;
end ;


procedure TSGeneric_Value.Set_Double( D : double ) ;

begin
    Abstract_Error ;
end ;


function TSGeneric_Value.Get_Extended : extended ;

begin
    Get_Extended := 0.0 ;
    Abstract_Error ;
end ;


procedure TSGeneric_Value.Set_Extended( E : extended ) ;

begin
    Abstract_Error ;
end ;



end.
