/* 
        Program Name    : CVT
        Package Name    : LB
        Purpose         : Conversion routines
        Institution     : Conroy & Conroy Co.
        Date written    : 10-Dec-85
        Written by      : Alan Conroy
        Version         : 1.0A

	Copyright (C) 1985-1987 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

      *******************************************************
      *                                                     *
      *      M O D I F I C A T I O N   H I S T O R Y        *
      *                                                     *
      *    DATE        BY            REASON                 *
      *                                                     *
      *  03-Sep-87     EAC           Disallow bases<2       *
      *                                                     *
      *******************************************************

      *******************************************************
      *                                                     *
      *          P R O G R A M   P U R P O S E              *
      *                                                     *
      *******************************************************

        These routines perform various conversions.
*/

#include <string.h>
#include <math.h>
#include <stdio.h>

#include "clib.h"
#include "cvt.h"
#include "instr.h"
#include "nums.h"

int Valid_Base( char *Value, int Base )
{
/* Return true if Value is a valid value for base Base. */

int I, I1 ;
char Temp[ 2 ] ;

    if( strlen( Value ) > 36 )
    {
        return False ;
    } ;
    Temp[ 1 ] = 0 ;
    for( I = 0; I < strlen( Value ); I++ ) /* Verify that the value is valid */
    {
	Temp[ 0 ] = Value[ I ] ;
        I1 = Instr( 1, "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ!\"#$%&?()*+~-",
	    Temp ) ;
        if( I1 == 0 || I1 > Base )
        {
            return False ;
        } ;
    } ;
    return True ;
} ;


/* Convert from any base to any base, provided that the source and destination
   bases are integer values.  Legal base values are two (2) to 49.  Bases
   outside of this range may or may not yield correct results.
*/
char *CVTB( int Base1, int Base2, char *X )
{

int Err ;
int E, I, I1 ;
float X1, Xx, Xy ;
float Power ;
char Buffer[ 256 ], Y[ 256 ] ;
char Base_List[] = { "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ!\"#$%&?()*+~-" } ;
char Temp[ 2 ] ;

    X1 = 0 ;
    Err = False ;
    if( Base1 < 2 || Base2 < 2 || Base1 > 49 || Base2 > 49 )
    {
        puts( "%Numeric error" ) ;
        strcpy( Buffer, "0" ) ;
	return &Buffer ;
    } ;
    if( strlen( X ) == 0 )
    {
	strcpy( Buffer, "0" ) ; /* A null value is assumed to be zero */
	return &Buffer ;
    } ;
    if( !Valid_Base( X, Base1 ) )
    {
        if( !Err )
	{
	    puts( "%Numeric error" ) ;
	} ;
        Err = True ;
    } ;
    if( strlen( X ) > 36 )
    {
    /* FP math will only handle 36 digits of base 2 with accuracy, larger bases
     are even more restricted */
	strcpy( X, &X[ strlen( X ) - 36 ] ) ;
    } ;
    if( Base1 == 10 )
    {
	X1 = atof( X ) ;
	/* If the base to convert from is ten, skip the conversion */
    }
    else
    {
	Temp[ 1 ] = 0 ;
	Power = 1 ;
	for( I = strlen( X ); I >= 1; I-- )
	/* Convert value to a floating point value */
	{
	    Temp[ 0 ] = X[ I - 1 ] ;
	    Xx = Power * ( Instr( 1, Base_List, Temp ) - 1 ) ;
	    X1 = X1 + Xx ;
	    Power = Power * Base1 ;
	    if( Power > 99999999999.0 )
	    {
		if( !Err )
		{
		    puts( "%Numeric error" ) ;
		} ;
		Err = True ;
		break ;
	    } ;
	} ;
    } ;
    if( X1 > 99999999999.0 && !Err )
    {
	puts( "%Numeric error" ) ;
    } ;
    if( Base2 == 10 )
    {
	strcpy( Y, Num1( X1 ) ) ;
    /* If the base to convert to is ten, skip the conversion */
    }
    else
    {
	while( X1 > 99999999999.0 )
	{
	    X1 = floor( X1 / Base2 ) ;
	} ;
	Y[ 0 ] = 0 ;
	I = 0 ;
	Power = 1 ;
	while( Power <= X1 )
	{
	    I = I + 1 ;
	    Power = Power * Base2 ;
	} ;
	for( I = I - 1; I >= 0; I-- )
	{
	    Power = Power / Base2 ;
	    if( Power != 0 )
	    {
		Xx = floor( X1 / Power ) ;
		if( Xx < Base2 )
		{
		    if( Xx != 0 || strlen( Y ) != 0 )
		    {
			Y[ strlen( Y ) + 1 ] = 0 ;
			Y[ strlen( Y ) ] = Base_List[ Xx ] ;
		    } ;
		    if( Xx != 0 )
		    {
			X1 = X1 - Power * Xx ;
		    } ;
		} ;
	    } ;
	} ;
	if( strlen( Y ) == 0 )
	{
	    strcpy( Y, "0" ) ;
	} ;
    } ;
    strcpy( Buffer, Y ) ;
    return &Buffer ;
} ;

long Roman_To_Arabic( char *X ) /* Convert Roman to number */
{

long Result, Intermediate, Temp, Previous ;
int Dummy ;

    Result = 0 ;
    Intermediate = 0 ;
    Dummy = 0 ;
    Previous = 10 ;
    while( X[ Dummy ] != 0 )
    {
	switch( X[ Dummy ] )
	{
	    case 'I' :
		Temp = 1 ;
		break ;
	    case 'V' :
		Temp = 5 ;
		break ;
	    case 'X' :
		Temp = 10 ;
	} ;
	if( Previous < Temp )
	{
	    Intermediate = Temp - Intermediate ;
	}
	else
	{
	    if( Previous >= Temp && Dummy > 0 )
	    {
		Result = Result + Intermediate ;
		Intermediate = Temp ;
	    }
	    else
	    {
		Intermediate = Intermediate + Temp ;
	    } ;
	} ;
	Previous = Temp ;
	Dummy = Dummy + 1 ;
    } ;
    Result = Result + Intermediate ;

    return Result ;
} ;


char *Arabic_To_Roman( int X ) /* Convert number to Roman (up to 49) */
{

static char Buffer[ 128 ] ;

    Buffer[ 0 ] = 0 ;
    while( X > 9 )
    {
	strcat( Buffer, "X" ) ;
	X = X - 10 ;
    } ;
    switch( X )
    {
	case 9 :
	    strcat( Buffer, "IX" ) ;
	    break ;
	case 8 :
	    strcat( Buffer, "VIII" ) ;
	    break ;
	case 7 :
	    strcat( Buffer, "VII" ) ;
	    break ;
	case 6 :
	    strcat( Buffer, "VI" ) ;
	    break ;
	case 5 :
	    strcat( Buffer, "V" ) ;
	    break ;
	case 4 :
	    strcat( Buffer, "IV" ) ;
	    break ;
	case 3 :
	    strcat( Buffer, "III" ) ;
	    break ;
	case 2 :
	    strcat( Buffer, "II" ) ;
	    break ;
	case 1 :
	    strcat( Buffer, "I" ) ;
	    break ;
    } ;
    return Buffer ;
} ;

