{$N+}
{
              Program Name : HTTPFiles
              Package Name : LB
              Purpose      : HTTP access routines
              Institution  : Conroy & Conroy Co.
              Date Written : 23-Feb-2003
              Written By   : Alan Conroy
              Version      : 1.0

              TO THE GLORY OF GOD THROUGH JESUS CHRIST

        Copyright (C) 2003 by Conroy & Conroy.  Released to the public domain.

        *********************************************************
        *                                                       *
        *        M O D I F I C A T I O N   H I S T O R Y        *
        *                                                       *
        *     DATE      BY             REASON                   *
        *                                                       *
        *                                                       *
        *********************************************************

        *********************************************************
        *                                                       *
        *           P R O G R A M   P U R P O S E               *
        *                                                       *
        *********************************************************

          This unit defines the THTTP_File_System object, which provides a file
        system interface to the HyperText Transfer Protocol.
}

unit httpfiles ;

interface

{$I EDEFINES.INC}

uses { Borland... }
     NMHTTP, { TNMHTTP }

     { C&C... }
     Common, { TObject }
     Compatib, { Smallint }
     Heap, { TStore }
     TypeDefs, { Must be before Files }
     Files, { TFile_System }
     UStrings, { Boolean_To_String }
     Collect, { TCollection }
     UEHDefs, { TUEC }
     DebugInt,
     SysUtils ;

type THTTP_File_Rec = class
                         public
                             NMHTTP : TNMHTTP ;
                             Remote_Name : string ;
                             Remote_Path : string ;
                             Local_File : PFile ;
                     end ;

     PHTTP_File_System = ^THTTP_File_System ;

     THTTP_File_System = object( TFile_System )
                        public { Constructors and destructors... }
                            constructor Init ;
                            destructor Done ; virtual ;

                        private { Internal utility routines }
                            function Get_File_Rec( Handle : integer ) : THTTP_File_Rec ;
                            function Get_HTTP( const Name : string ;
                                var Remote_Path, Remote_Name : string ) : TNMHTTP ;

                        protected { Instance support }
                            procedure Filespec_String_Scan( const Name, _Separator : string ;
                                var Info : TFSSS ) ; virtual ;

                        public { API... }
                            { Close an open file via handle. }
                            procedure Close( Handle : integer ) ; virtual ;

                            { Create file and open with specified mode.  If
                              file already exists, it is deleted and re-created.
                              Returns the file handle (0 if an error). }
                            function Create( const Name : string ;
                                Mode : integer ) : integer ; virtual ;

                            { Deletes the specified file. }
                            procedure Erase( const Name : String ) ; virtual ;

                            { Returns name associated with specified file. }
                            function File_Name( Handle : integer ) : string ;
                                 virtual ;

                            { Flush any buffered output for specified file. }
                            procedure Flush( Handle : integer ) ; virtual ;

                            { Returns the attributes of the specified file. }
                            function Get_Attributes( Handle : integer ) : integer ;
                                 virtual ;

                            function Get_File_Object( Mode : integer ) : PFile ;
                                virtual ;

                            { Returns size of specified file, in bytes. }
                            function Get_Size( Handle : integer ) : TStore_Address ;
                                 virtual ;

                            function Get_Time( Handle, Typ : integer ) : Longint ;
                                 virtual ;

                            { Open existing file with specified mode.  Returns
                              the file handle (0 if an error). }
                            function Open( const Name : string ;
                                Mode : integer ) : integer ; virtual ;

                            { Read from specified file at specified position
                              into Buffer, up to Len bytes.  Actual bytes read
                              is returned in Count.  Count is ignored on call. }
                            procedure Read( Handle : integer ;
                                Position : TStore_Address ; Len : TStore_Address ;
                                var Buffer ; var Count : TStore_Address ) ; virtual ;

                            { Renames Old file to New file name. }
                            procedure Rename( const Old, New : String ) ;
                                 virtual ;

                            procedure Set_Attributes( Handle : integer ;
                                Attr : integer ) ; virtual ;

                            procedure Set_Time( Handle : integer ;
                                Typ, Time : Longint ) ; virtual ;

                            { Sets the size of the specified file to the
                              specified size, in bytes. }
                            procedure Set_Size( Handle : integer ;
                                New_Size : TStore_Address ) ; virtual ;

                            { Write to specified file at specified position from
                              Buffer, up to Len bytes.  Actual bytes written is
                              returned in Count.  Count is ignored on call. }
                            procedure Write( Handle : integer ;
                                Position : TStore_Address ; Len : TStore_Address ;
                                var Buffer ; var Count : TStore_Address ) ; virtual ;

                            { Returns last error code. }
                            procedure Last_Error( var Res : TUEC ) ; virtual ;

                            function DeviceSize( Nam : string ) : TStore_Address ;
                                virtual ;

                            function DeviceFree( Nam : string ) : TStore_Address ;
                                virtual ;

                            function MkDir( Nam : string ) : longint ;
                                virtual ;

                            { Returns the attributes of the specified file. }
                            function Get_File_Attributes( const Name : string ) : longint ;
                                 virtual ;

                            { Returns the attributes of the specified file. }
                            procedure Set_File_Attributes( const Name : string ;
                                Value : longint ) ; virtual ;

                            function FindFirst( Nam : string ; Attributes : longint ;
                                var Find_Record : TFind_Record ) : longint ;
                                virtual ;

                            function FindNext( var Find_Record : TFind_Record ) : longint ;
                                virtual ;

                            function Separator : string ; virtual ;

                            function Facility : longint ; virtual ;

                            function Version : longint ; virtual ;
                     end ; { THTTP_File_System }

implementation

uses { C&C... }
     ASCIIDef,
     Instrs, { Instr }
     Standard,
     Num1s,
     Windows,
     Disk,
     ErrMastr ; { Error_Code }

type THTTP_File_System_Server = object( TFile_System_Server )
                                   public { API... }
                                       function Get_File_System( const Nam : string ) : PFile_System ;
                                           virtual ;
     			       end ; { THTTP_File_System_Server }

{ THTTP_File_System_Server methods... }

function THTTP_File_System_Server.Get_File_System( const Nam : string ) : PFile_System ;

var Dummy : integer ;
    S, ID : string ;

begin
    Result := nil ;
    S := Edit( Nam, 8 or 128 ) ; { Trim leading/trailing spaces }
    if( copy( S, 1, 6 ) = 'HTTP://' ) then { An HTTP URL }
    begin
        Dummy := Instr( 7, S + '/', '/' ) ;
        if( Dummy < 8 ) then { Malformed URL }
        begin
            exit ;
        end ;
        ID := copy( S, 1, Dummy - 1 ) ;
        ID := Edit( ID, 128 ) ; { Traim any trailing spaces }
        S := copy( S, Dummy + 1, length( S ) ) ;
        S := Edit( S, 8 ) ; { Trim any leading spaces from path/name }
        if( ( length( S ) = 0 ) or ( copy( S, length( S ), 1 ) = '/' ) ) then { Missing path/name }
        begin
            exit ;
        end ;
        //~~~
    end ;
end ; { THTTP_File_System_Server.Get_File_System }



{ THTTP_File_System methods... }

{ Constructors and destructors... }

constructor THTTP_File_System.Init ;

begin
    inherited Init ;
end ;


destructor THTTP_File_System.Done ;

begin
    inherited Done ;
end ;


{ Internal utility routines... }

{ API... }

function THTTP_File_System.Get_File_Rec( Handle : integer ) : THTTP_File_Rec ;

begin
    if( ( Handle < 0 ) or ( Handle >= File_List.Count ) ) then
    begin
        Get_File_Rec := nil ;
        Set_Error( File_SystemErr_Facility, File_SystemErr_Invalid_Handle ) ;
        exit ;
    end ;
    Get_File_Rec := THTTP_File_Rec( File_List.At( Handle - 1 ) ) ;
    if( Result = nil ) then
    begin
        Set_Error( File_SystemErr_Facility, File_SystemErr_FileNotOpen ) ;
        exit ;
    end ;
end ;


{ Instance support }

procedure THTTP_File_System.Filespec_String_Scan( const Name, _Separator : string ;
    var Info : TFSSS ) ;

var Dummy : integer ;

begin
    if( _Separator <> '/' ) then
    begin
        Files.Filespec_String_Scan( Name, '/', Info ) ; { Rescan with correct separator }
    end else
    begin
        Dummy := pos( '@', Info.Filename ) ;
        if( Dummy <> 0 ) then { A password was included - reparse filename }
        begin
            Info.Password := copy( Info.Filename, 1, Dummy - 1 ) ;
            Info.Filename := copy( Info.Filename, Dummy + 1, length( Info.Filename ) ) ;
            Info.Path := copy( _Name, 1, Dummy ) ;
            Dummy := length( Info.Filename ) ;
            while( ( Dummy > 0 ) and ( Info.Filename[ Dummy ] <> '.' ) ) do
            begin
                dec( Dummy ) ;
            end ;
            if( Dummy = 0 ) then { No extension }
            begin
                Dummy := length( Info.Filename ) + 1 ;
                Info.Flags := Info.Flags or FSSS_Null_Extension ;
            end else
            begin
                Info.Flags := Info.Flags and ( not FSSS_Null_Extension ) ;
            end ;
            Info.Name := copy( Info.Filename, 1, Dummy - 1 ) ;
            Info.Extension := copy( Info.Filename, Dummy + 1, length( Info.Filename ) ) ;
        end ;
    end ;
end ;


{ Close an open file via handle. }
procedure THTTP_File_System.Close( Handle : integer ) ;

var File_Rec : THTTP_File_Rec ;

begin
    { Setup... }
    File_Rec := Get_File_Rec( Handle ) ;
    if( File_Rec = nil ) then
    begin
        exit ;
    end ;

    { Close the file... }
    Flush( Handle ) ;
    File_Rec.NMHTTP.Disconnect ;
    Set_Error( 2, 0 ) ;
    THTTP_File_Rec( File_List.At( Handle - 1 ) ).Free ;
    File_List.AtPut( Handle - 1, nil ) ;

    { Reclaim unused handles at the end of the file list }
    while(
           ( File_List.Count > 0 )
           and
           ( File_List.At( File_List.Count - 1 ) = nil )
         ) do
    begin
        File_List.AtDelete( File_List.Count - 1 ) ;
    end ;
end ;


{ Create file and open with specified mode.  If file already exists, it is
  deleted and re-created.  Returns the file handle (0 if an error). }
function THTTP_File_System.Create( const Name : string ; Mode : integer ) : integer ;

begin
    Erase( Name ) ;
    Create := Open( Name, Mode ) ;
end ; { THTTP_File_System.Create }


function THTTP_File_System.Get_HTTP( const Name : string ;
    var Remote_Path, Remote_Name : string ) : TNMHTTP ;

var Dummy : integer ;
    NMHTTP : TNMHTTP ;
    _Path, S : string ;

begin
    Result := nil ; { Assume failure }
    S := Name ;
    Dummy := pos( '//', S ) ;
    if( Dummy > 0 ) then
    begin
        S := copy( S, Dummy + 2, length( S ) ) ;
    end ;
    Dummy := pos( ':', S ) ;
    NMHTTP.UserID := copy( S, 1, Dummy - 1 ) ;
    S := copy( S, Dummy + 1, length( S ) ) ;
    Dummy := pos( '@', S ) ;
    NMHTTP.Password := copy( S, 1, Dummy - 1 ) ;
    Dummy := pos( '/', S + '/' ) ;
    NMHTTP.Host := copy( S, 1, Dummy - 1 ) ;
    S := copy( S, Dummy + 1, length( S ) ) ;
    try
        NMHTTP.Connect ;
    except
    end ;
    if( not NMHTTP.FConnected ) then
    begin
        Set_Error( File_SystemErr_Facility, File_SystemErr_ProtectionViolation ) ;
        NMHTTP.Free ;
        exit ;
    end ;
    Dummy := length( S ) ;
    while( ( length( S ) > 0 ) and ( S[ Dummy ] <> '/' ) ) do
    begin
        dec( Dummy ) ;
    end ;
    Remote_Path := copy( S, 1, Dummy - 1 ) ;
    Remote_Name := copy( S, Dummy + 1, length( S ) ) ;

    { Position to specified directory }
    try
        NMHTTP.ChangeDir( Remote_Path ) ;
    except
        Set_Error( File_SystemErr_Facility, 15 ) ; { Path not found }
        NMHTTP.Free ;
        exit ;
    end ;
    Result := NMHTTP ;
end ; { THTTP_File_System.Get_HTTP }


{ Deletes the specified file. }
procedure THTTP_File_System.Erase( const Name : String ) ;

var NMHTTP : TNMHTTP ;
    _Path, S : string ;

begin
    NMHTTP := Get_HTTP( Name, _Path, S ) ;
    if( NMHTTP = nil ) then
    begin
        exit ;
    end ;

    { Delete file if it exists }
    try
        NMHTTP.Delete( S ) ;
    except
    end ;

    NMHTTP.Free ;
end ; { THTTP_File_System.Erase }


{ Returns name associated with specified file. }
function THTTP_File_System.File_Name( Handle : integer ) : string ;

var File_Rec : THTTP_File_Rec ;
    Temp : FileRec ;

begin
    { Setup... }
    File_Rec := Get_File_Rec( Handle ) ;
    if( File_Rec = nil ) then
    begin
        exit ;
    end ;

    File_Name := File_Rec.Remote_Path + '/' + File_Rec.Remote_Name ;
end ;


{ Flush any buffered output for specified file. }
procedure THTTP_File_System.Flush( Handle : integer ) ;

var File_Rec : THTTP_File_Rec ;

begin
    { Setup... }
    File_Rec := Get_File_Rec( Handle ) ;
    if( File_Rec = nil ) then
    begin
        exit ;
    end ;

    File_Rec.Local_File.Flush ;
    try
        File_Rec.NMHTTP.Upload( File_Rec.Local_File.FileName, File_Rec.Remote_Name ) ;
        Set_Error( 0, 0 ) ;
    except
        Set_Error( File_SystemErr_Facility, File_SystemErr_ProtectionViolation ) ;
    end ;
end ;


{ Returns the attributes of the specified file. }
function THTTP_File_System.Get_Attributes( Handle : integer ) : integer ;

var File_Rec : THTTP_File_Rec ;

begin
    File_Rec := Get_File_Rec( Handle ) ;
    if( File_Rec = nil ) then
    begin
        Get_Attributes := 0 ;
        exit ;
    end ;

    Get_Attributes := Get_File_Attributes( File_Name( Handle ) ) ;
end ;


function THTTP_File_System.Get_File_Object( Mode : integer ) : PFile ;

begin
    Get_File_Object := new( PFile, Init( @self, ( ( Mode and FSOM_Text ) <> 0 ), 1 ) ) ;
end ;


{ Returns size of specified file, in bytes. }
function THTTP_File_System.Get_Size( Handle : integer ) : TStore_Address ;

var File_Rec : THTTP_File_Rec ;

begin
    { Setup... }
    File_Rec := Get_File_Rec( Handle ) ;
    if( File_Rec = nil ) then
    begin
        Get_Size := -1 ;
        exit ;
    end ;

    Get_Size := File_Rec.Local_File.Get_Size ;
    IO_Error := File_Rec.Local_File.IO_Error ;
end ;


function THTTP_File_System.Get_Time( Handle, Typ : integer ) : Longint ;

var File_Rec : THTTP_File_Rec ;
    Time : longint ;

begin
    { Setup... }
    Time := 0 ;
    File_Rec := Get_File_Rec( Handle ) ;
    if( File_Rec = nil ) then
    begin
        Get_Time := Time ;
        exit ;
    end ;

    {$I-}
{$IFDEF DELPHI}
(*
    Time := FileGetDate( File_Rec^.NTF ) ;
*)
{$ELSE}
    {$IFDEF WINDOWS}
        Windos.Getftime( File_Rec^._File, Time ) ;
    {$ELSE}
        Dos.Getftime( File_Rec^._File, Time ) ;
    {$ENDIF}
{$ENDIF}
    {$I+}
    Set_Error( 2, IOResult ) ;
    Get_Time := Time ;
end ;


{ Open existing file with specified mode.  Returns the file handle (0 if an
  error). }
function THTTP_File_System.Open( const Name : string ; Mode : integer ) : integer ;

var File_Rec : THTTP_File_Rec ;
    Dummy : integer ;
    F : file ;
    NMHTTP : TNMHTTP ;
    _Path, S : string ;
    UEC : TUEC ;

begin
    NMHTTP := Get_HTTP( Name, _Path, S ) ;
    if( NMHTTP = nil ) then
    begin
        exit ;
    end ;

    File_Rec := THTTP_File_Rec.Create ;
    File_Rec.Local_File := Files.Open( Get_Temp_File_Name( 'HTTP' ), FSOM_OUTPUT, UEC ) ;
    Set_Error( UEC.Facility, UEC.Code ) ;
    if( IO_Error.Code <> 0 ) then
    begin
        Open := 0 ; { Open failure }
        File_Rec.Free ;
        NMHTTP.Free ;
    end else
    begin
        File_List.Insert( File_Rec ) ;
        Open := File_List.Count ;
        File_Rec.NMHTTP := NMHTTP ;
        File_Rec.Remote_Name := S ;
        File_Rec.Remote_Path := _Path ;
    end ;
end ; { THTTP_File_System.Open }


{ Read from specified file at specified position into Buffer, up to Len bytes.
  Actual bytes read is returned in Count.  Count is ignored on call. }
procedure THTTP_File_System.Read( Handle : integer ; Position : TStore_Address ;
    Len : TStore_Address ; var Buffer ; var Count : TStore_Address ) ;

var Dummy : integer ;
    File_Rec : THTTP_File_Rec ;
    R : TRecord_Size ;
    UEC : TUEC ;

begin
    { Setup... }
    File_Rec := Get_File_Rec( Handle ) ;
    if( File_Rec = nil ) then
    begin
        exit ;
    end ;

    File_Rec.Local_File.Seek( Position ) ;
    File_Rec.Local_File.BlockRead( Buffer, Len, Count ) ;
    IO_Error := File_Rec.Local_File.IO_Error ;
end ; { THTTP_File_System.Read }


{ Renames Old file to New file name. }
procedure THTTP_File_System.Rename( const Old, New : String ) ;

var NMHTTP : TNMHTTP ;
    _Path, S : string ;

begin
    NMHTTP := Get_HTTP( Old, _Path, S ) ;
    if( NMHTTP = nil ) then
    begin
        exit ;
    end ;

    { Rename }
    try
        NMHTTP.Rename( S, New ) ;
    except
        Set_Error( File_SystemErr_Facility, File_SystemErr_ProtectionViolation ) ; { Path not found }
    end ;
    NMHTTP.Free ;
end ; { THTTP_File_System.Rename }


procedure THTTP_File_System.Set_Attributes( Handle : integer ; Attr : integer ) ;

var File_Rec : THTTP_File_Rec ;

begin
    { Setup... }
    File_Rec := Get_File_Rec( Handle ) ;
    if( File_Rec = nil ) then
    begin
        exit ;
    end ;

    Set_File_Attributes( File_Name( Handle ), Attr ) ;
end ;


procedure THTTP_File_System.Set_Time( Handle : integer ; Typ, Time : Longint ) ;

var File_Rec : THTTP_File_Rec ;

begin
    { Setup... }
    File_Rec := Get_File_Rec( Handle ) ;
    if( File_Rec = nil ) then
    begin
        exit ;
    end ;

    //~~~
end ;


{ Sets the size of the specified file to the specified size, in bytes. }
procedure THTTP_File_System.Set_Size( Handle : integer ; New_Size : TStore_Address ) ;

var File_Rec : THTTP_File_Rec ;

begin
    { Setup... }
    File_Rec := Get_File_Rec( Handle ) ;
    if( File_Rec = nil ) then
    begin
        exit ;
    end ;

    File_Rec.Local_File.Set_Size( New_Size ) ;
end ; { THTTP_File_System.Set_Size }


{ Write to specified file at specified position from Buffer, up to Len bytes.
  Actual bytes written is returned in Count.  Count is ignored on call. }
procedure THTTP_File_System.Write( Handle : integer ; Position : TStore_Address ;
    Len : TStore_Address ; var Buffer ; var Count : TStore_Address ) ;

var File_Rec : THTTP_File_Rec ;

begin
    { Setup... }
    File_Rec := Get_File_Rec( Handle ) ;
    if( File_Rec = nil ) then
    begin
        exit ;
    end ;

    File_Rec.Local_File.Seek( Position ) ;
    File_Rec.Local_File.BlockWrite( Buffer, Len, Count ) ;
    IO_Error := File_Rec.Local_File.IO_Error ;
end ;


{ Returns last error code. }
procedure THTTP_File_System.Last_Error( var Res : TUEC ) ;

begin
    Res := IO_Error ;
end ;


function THTTP_File_System.DeviceSize( Nam : string ) : TStore_Address ;

begin
    DeviceSize := -1 ;
end ;


function THTTP_File_System.DeviceFree( Nam : string ) : TStore_Address ;

begin
    DeviceFree := -1 ;
end ;


function THTTP_File_System.MkDir( Nam : string ) : longint ;

var NMHTTP : TNMHTTP ;
    _Path, S : string ;

begin
    NMHTTP := Get_HTTP( Nam, _Path, S ) ;
    try
        NMHTTP.MakeDirectory( S ) ;
        Set_Error( 0, 0 ) ;
    except
        Set_Error( File_SystemErr_Facility, File_SystemErr_ProtectionViolation ) ;
    end ;
    MkDir := IO_Error.Code ;
end ;


{ Returns the attributes of the specified file. }
function THTTP_File_System.Get_File_Attributes( const Name : string ) : longint ;

var Attr : integer ;

begin
    Attr := GetFileAttributes( PChar( Name ) ) ;
    //~~~
end ; { THTTP_File_System.Get_File_Attributes }


function Convert_Attributes( Value : longint ) : longint ;

var Attr : integer ;

begin
    Attr := 0 ;
    if( ( Value and FSFF_Protected ) <> 0 ) then
    begin
	Attr := Attr or faReadOnly ;
    end ;
    if( ( Value and FSFF_Hidden ) <> 0 ) then
    begin
	Attr := Attr or faHidden ;
    end ;
    if( ( Value and FSFF_System ) <> 0 ) then
    begin
	Attr := Attr or faSysFile ;
    end ;
    if( ( Value and FSFF_VolumeID ) <> 0 ) then
    begin
	Attr := Attr or faVolumeID ;
    end ;
    if( ( Value and FSFF_Directory ) <> 0 ) then
    begin
	Attr := Attr or faDirectory ;
    end ;
    if( ( Value and FSFF_Archive ) <> 0 ) then
    begin
	Attr := Attr or faArchive ;
    end ;
    Convert_Attributes := Attr ;
end ; { Convert_Attributes }


{ Returns the attributes of the specified file. }
procedure THTTP_File_System.Set_File_Attributes( const Name : string ;
    Value : longint ) ;

var Attr : integer ;

begin
    Attr := Convert_Attributes( Value ) ;
    //~~~
    if( SetFileAttributes( PChar( Name ), Attr ) ) then
    begin
        Set_Error( 0, 0 ) ;
    end else
    begin
        Set_Error( 0, GetLastError ) ;
    end ;
end ;


type TFR = record
               Last : integer ;
               HTTP : string[ 255 ] ;
           end ;

     THTTP_Directory_List = class
                               public
                                   Count : integer ;

                               public
                                   procedure Counter( Listing : string ) ;
                           end ;

procedure THTTP_Directory_List.Counter( Listing : string ) ;

begin
    inc( Count ) ;
end ;


function THTTP_File_System.FindFirst( Nam : string ; Attributes : longint ;
    var Find_Record : TFind_Record ) : longint ;

var Attr : integer ;
    Dummy : integer ;
    F : file ;
    FR : TFR ;
    HTTPDirectoryList : THTTP_Directory_List ;
    _Path, S : string ;
    NMHTTP : TNMHTTP ;
    
begin
    if( Find_Record.Initialized <> 0 ) then
    begin
        Initialize_Find_Record( Find_Record ) ;
    end ;
    NMHTTP := Get_HTTP( Nam, _Path, S ) ;
    if( NMHTTP = nil ) then
    begin
        exit ;
    end ;

    HTTPDirectoryList := THTTP_Directory_List.Create ;
    try
        NMHTTP.ParseList := True ;
        NMHTTP.OnListItem := HTTPDirectoryList.Counter ;
        NMHTTP.List ;
    except
        NMHTTP.Free ;
        HTTPDirectoryList.Free ;
        Set_Error( File_SystemErr_Facility, File_SystemErr_ProtectionViolation ) ;
        exit ;
    end ;
    if( HTTPDirectoryList.Count = 0 ) then
    begin
        Set_Error( 0, -1 ) ;
        NMHTTP.Free ;
        HTTPDirectoryList.Free ;
        FindFirst := -1 ; { Directory is empty }
        exit ;
    end ;
    FR.Last := 0 ;
    FR.HTTP := Nam ;
    Find_Record.File_Name := NMHTTP.HTTPDirectoryList.Name[ 0 ] ;
    HTTPDirectoryList.Free ;
    NMHTTP.Free ;
    fillchar( Find_Record.Data, sizeof( Find_Record.Data ), 0 ) ;
    move( FR, Find_Record.Data, sizeof( FR ) ) ;

    Attr := Convert_Attributes( Attributes ) ;
    Set_Error( 0, 0 ) ;
    FindFirst := IO_Error.Code ;
end ; { THTTP_File_System.FindFirst }


function THTTP_File_System.FindNext( var Find_Record : TFind_Record ) : longint ;

var Loop : integer ;
    NMHTTP : TNMHTTP ;
    FR : TFR ;
    HTTPDirectoryList : THTTP_Directory_List ;
    _Path, S : string ;
    
begin
    move( Find_Record.Data, FR, sizeof( FR ) ) ;
    NMHTTP := Get_HTTP( FR.HTTP, _Path, S ) ;
    if( NMHTTP = nil ) then
    begin
        Set_Error( File_SystemErr_Facility, File_SystemErr_ProtectionViolation ) ;
        exit ;
    end ;
    HTTPDirectoryList := THTTP_Directory_List.Create ;
    try
        NMHTTP.ParseList := True ;
        NMHTTP.OnListItem := HTTPDirectoryList.Counter ;
        NMHTTP.List ;
    except
        NMHTTP.Free ;
        HTTPDirectoryList.Free ;
        Set_Error( File_SystemErr_Facility, File_SystemErr_ProtectionViolation ) ;
        exit ;
    end ;
    if( HTTPDirectoryList.Count = 0 ) then
    begin
        Set_Error( 0, -1 ) ;
        NMHTTP.Free ;
        HTTPDirectoryList.Free ;
        FindNext := -1 ; { Directory is empty }
        exit ;
    end ;

    { Find file after last found file }
    for Loop := 0 to HTTPDirectoryList.Count - 1 do
    begin
        if( NMHTTP.HTTPDirectoryList.Name[ Loop ] = Find_Record.File_Name ) then
        begin
            if( Loop >= HTTPDirectoryList.Count - 1 ) then
            begin
                Set_Error( 0, -1 ) ;
            end else
            begin
                Set_Error( 0, 0 ) ;
                Find_Record.File_Name := NMHTTP.HTTPDirectoryList.Name[ Loop + 1 ] ;
                FR.Last := Loop + 1 ;
                fillchar( Find_Record.Data, sizeof( Find_Record.Data ), 0 ) ;
                move( FR, Find_Record.Data, sizeof( FR ) ) ;
            end ;
            HTTPDirectoryList.Free ;
            NMHTTP.Free ;
            exit ;
        end ;
    end ;

    { Last found file not found - must have been deleted.  Return file at last index... }
    if( FR.Last >= HTTPDirectoryList.Count - 1 ) then
    begin
        Set_Error( 0, -1 ) ;
    end else
    begin
        Set_Error( 0, 0 ) ;
        Find_Record.File_Name := NMHTTP.HTTPDIrectoryList.Name[ FR.Last ] ;
    end ;
    HTTPDirectoryList.Free ;
    NMHTTP.Free ;
    fillchar( Find_Record.Data, sizeof( Find_Record.Data ), 0 ) ;
    move( FR, Find_Record.Data, sizeof( FR ) ) ;
end ;


function THTTP_File_System.Separator : string ;

begin
    Separator := '/' ;
end ;


function THTTP_File_System.Facility : longint ;

begin
    Facility := 1 ; { DOS }
end ;


function THTTP_File_System.Version : longint ;

begin
    Version := 10 ; { Version 1.0 }
end ;


var HTTP_FSS : THTTP_File_System_Server ;

// Set up HTTP file system server
begin
    HTTP_FSS.Init ;
    Register_File_System_Server( @HTTP_FSS ) ;
end.
