{$V-}
{
              Program Name : XLATE
              Package Name : LB
              Purpose      : Wildcard handling
              Institution  : Conroy & Conroy Co.
              Date Written : 04-Apr-1992
              Written By   : Alan Conroy
              Version      : 2.0

              Copyright (C) 1992 by Alan Conroy.  Released to the public domain.

		TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *********************************************************
        *                                                       *
        *        M O D I F I C A T I O N   H I S T O R Y        *
        *                                                       *
        *     DATE      BY             REASON                   *
        *                                                       *
        *********************************************************

        *********************************************************
        *                                                       *
        *           P R O G R A M   P U R P O S E               *
        *                                                       *
        *********************************************************

	  These functions handle matching wildcards.
}

unit Xlate ;

interface

{$I EDEFINES.INC}

function WC_Match( Wildcard, Filename : string ) : Boolean ;
{ Compare a filename and a wildcard specification and return TRUE if a match. }

function WC_Xlate( Reference, Filename, Wildcard : string ) : string ;
{ Translate a wildcard specification and an input file to an output spec. }

implementation

uses Wildcard ;

function Resolve_Wildcard( Ref_Spec, In_Spec, Out_Spec : string ) : string ;

var Dummy : Integer ;
    _Result : string ;
    Ref, Outp : string ; { Work variable }

begin
    _Result := '' ;
    if( ( Ref_Spec = '\...' ) or ( Ref_Spec = '\...\' ) ) then
    begin
        Ref_Spec := '\*' ;
    end ;
    { Handle special directory wildcard }

    Dummy := Pos( '*', Ref_Spec + '*' ) ;
    Ref := Copy( Ref_Spec, 1, Dummy - 1 ) ;
    Ref_Spec := Copy( Ref_Spec, Dummy, 254 ) ;
    Dummy := Pos( '*', Out_Spec + '*' ) ;
    Outp := Copy( Out_Spec, 1, Dummy - 1 ) ;
    Out_Spec := Copy( Out_Spec, Dummy, 254 ) ;
    Dummy := 1 ;
    while( Dummy <= Length( Outp ) ) do
    begin
	if( Outp[ Dummy ] = '?' ) then
        begin
            _Result := _Result + Copy( In_Spec, 1, 1 )
        end else
        if( Dummy > Length( Ref ) ) then
        begin
            _Result := _Result + Outp[ Dummy ]
        end else
        if( Ref[ Dummy ] = Copy( In_Spec, 1, 1 ) ) then
        begin
            _Result := _Result + Copy( Outp, Dummy, 1 )
        end else
        begin
            _Result := _Result + Copy( In_Spec, 1, 1 ) ;
        end ;
	if( Dummy <= Length( Ref ) ) then
        begin
	    In_Spec := Copy( In_Spec, 2, 254 ) ;
        end ;
	{ One less character to worry about }
	Dummy := Dummy + 1 ; { Try next character }
    end ;
    if( Out_Spec = '*' ) then
    begin
	_Result := _Result + In_Spec ; { Wildcard rest of name }
    end ;
    Resolve_Wildcard := _Result ; { Return result }
end ;


function Count_Dir( X : string ) : integer ;
{ Count number of directories in path }

var C, D : integer ;

begin
    C := 0 ;
    if( X[ length( X ) ] <> '\' ) then
    begin
        X := X + '\' ; { Directory path must end with backslash }
    end ;
    for D := 1 to length( X ) do
    begin
	if( X[ D ] = '\' ) then
        begin
	    inc( C ) ;
        end ;
    end ;
    Count_Dir := C - 1 ;
end ;


function Resolve_Dir( Ref_Spec, In_Spec, Out_Spec : string ) : string ;

var D, Dummy : Integer ;
    _Result : string ;
    Ref, Inp, Outp : string ; { Work variables }
    In_Count, Out_Count : integer ;
    W : boolean ;

begin
    _Result := '' ;
    In_Count := Count_Dir( In_Spec ) ;
    Out_Count := Count_Dir( Out_Spec ) ;
    Dummy := 2 ; { Start past first backslash }
    W := False ; { No wildcards found thus far }
    while( In_Count < Out_Count ) do
    begin
	if( ( Out_Spec[ Dummy ] = '*' ) or ( Out_Spec[ Dummy ] = '?' ) ) then
        begin
	    W := True ;
        end ;
	if( Out_Spec[ Dummy ] = '\' ) then
	begin
	    D := pos( '\...\', Out_Spec ) ;
	    if( ( D <> 0 ) and ( D < Dummy ) ) then
            begin
		W := True ;
	    end ;
	    if( W ) then
            begin
		Out_Count := In_Count ; { Terminate loop immediately }
            end else
            begin
                _Result := _Result + copy( Out_Spec, 1, Dummy - 1 ) ;
                Out_Spec := copy( Out_Spec, Dummy, 254 ) ;
                Dummy := 1 ;
            end ;
	    dec( Out_Count ) ;
	end ;
	Dummy := Dummy + 1 ;
    end ;

    while( True ) do
    begin
        if(
            ( Length( Ref_Spec ) = 0 )
            or
            ( Length( Out_Spec ) = 0 )
            or
            ( Length( In_Spec ) = 0 )
          ) then
        begin
            if( ( length( Out_Spec ) > 0 ) and ( Out_Spec <> '\...\' ) ) then
            begin
                _Result := _Result + Out_Spec ;
            end ;
            if( _Result[ Length( _Result ) ] <> '\' ) then
            begin
                _Result := _Result + '\' ;
                { Directory paths must end with a backslash }
            end ;
            Resolve_Dir := _Result ;
            Exit ;
        end ;
        if( Ref_Spec = '\...\' ) then
        begin
            Ref := Ref_Spec ;
        end else
        begin
            Dummy := Pos( '\', Copy( Ref_Spec, 2, 254 ) + '\' ) + 1 ;
            Ref := Copy( Ref_Spec, 1, Dummy - 1 ) ; { Current reference directory to use }
            Ref_Spec := Copy( Ref_Spec, Dummy, 254 ) ;
        end ;
        if( Out_Spec = '\...\' ) then
        begin
            Outp := Out_Spec ;
        end else
        begin
            Dummy := Pos( '\', Copy( Out_Spec, 2, 254 ) + '\' ) + 1 ;
            Outp := Copy( Out_Spec, 1, Dummy - 1 ) ; { Current output directory to use }
            Out_Spec := Copy( Out_Spec, Dummy, 254 ) ;
        end ;
        Dummy := Pos( '\', Copy( In_Spec, 2, 254 ) + '\' ) + 1 ;
        Inp := Copy( In_Spec, 1, Dummy - 1 ) ; { Current input directory to use }
        In_Spec := Copy( In_Spec, Dummy, 254 ) ;

        if( Outp = '\...\' ) then { Wildcard subdirectories }
        begin
            _Result := _Result + Inp
        end else
        begin
            _Result := _Result + Resolve_Wildcard( Ref, Inp, Outp ) ;
        end ;
    end ;
end ; { Resolve_Dir }


function WC_Xlate( Reference, Filename, Wildcard : string ) : string ;
{ Translate a wildcard specification and an input file to an output spec. }

var Dummy : Integer ; { Dummy variable }
    Cd : string[ 128 ] ; { Current device/directory }
    _Result : string[ 128 ] ; { Resulting translation }
    Ref_Dev, In_Dev, Out_Dev : string ; { Input/output device }
    Ref_Dir, In_Dir, Out_Dir : string ; { Input/output directory path }
    Ref_Name, In_Name, Out_Name : string ; { Input/output filename }
    Ref_Ext, In_Ext, Out_Ext : string ; { Input/output extension }

begin
    _Result := '' ; { Nothing yet }
    Getdir( 0, Cd ) ; { Get current disk/directory }
    if( Cd[ Length( Cd ) ] <> '\' ) then
    begin
        Cd := Cd + '\' ; { Must terminate with "\" }
    end ;

    { Break reference, filename, and wildcard into component pieces... }
    Split( Reference, Ref_Dir, Ref_Name, Ref_Ext ) ;
    Split( Filename, In_Dir, In_Name, In_Ext ) ;
    Split( Wildcard, Out_Dir, Out_Name, Out_Ext ) ;
    Dummy := Pos( ':', Ref_Dir ) ; { Locate device in filename }
    Ref_Dev := Copy( Ref_Dir, 1, Dummy ) ;
    Ref_Dir := Copy( Ref_Dir, Dummy + 1, 254 ) ;
    Dummy := Pos( ':', In_Dir ) ; { Locate device in filename }
    In_Dev := Copy( In_Dir, 1, Dummy ) ;
    In_Dir := Copy( In_Dir, Dummy + 1, 254 ) ;
    Dummy := Pos( ':', Out_Dir ) ; { Locate device in wildcard }
    Out_Dev := Copy( Out_Dir, 1, Dummy ) ;
    Out_Dir := Copy( Out_Dir, Dummy + 1, 254 ) ;

    if( Out_Dev = '' ) then { No device in wildcard }
    begin
	_Result := Copy( Cd, 1, 2 ) { Default to current }
    end else
    if( Out_Dev = '*' ) then { Wildcard output device }
    begin
        _Result := In_Dev ;
    end else
    begin
        _Result := Out_Dev ;
    end ;
    if( _Result[ 1 ] <> Cd[ 1 ] ) then
    begin
        Getdir( Ord( _Result[ 1 ] ) - 64, Cd ) ;
    end ;
    { Get current directory for output disk }

    if( Out_Dir = '' ) then { No directory in wildcard }
    begin
	_Result := _Result + Copy( Cd, 3, 254 )
    end else
    begin
	_Result := _Result + Resolve_Dir( Ref_Dir, In_Dir, Out_Dir ) ;
    end ;
    if( Out_Name = '' ) then { No filename in wildcard }
    begin
	_Result := _Result + In_Name
    end else
    begin
	_Result := _Result + Resolve_Wildcard( Ref_Name, In_Name, Out_Name ) ;
    end ;
    if( Out_Ext = '' ) then { No extension in wildcard }
    begin
	_Result := _Result + In_Ext
    end else
    begin
	_Result := _Result + Resolve_Wildcard( Ref_Ext, In_Ext, Out_Ext ) ;
    end ;
    WC_Xlate := _Result ; { Return result }
end ;


function Match_Wildcard( Ref_Spec, In_Spec : string ) : Boolean ;

var Dummy : Integer ;
    Ref, Inp : string ; { Work variables }

begin
    if( ( Ref_Spec = '\...' ) or ( Ref_Spec = '\...\' ) ) then
    begin
        Ref_Spec := '\*' ;
    end ;
    { Handle special directory wildcard }

    Dummy := Pos( '*', Ref_Spec + '*' ) ;
    Ref := Copy( Ref_Spec, 1, Dummy - 1 ) ;
    Ref_Spec := Copy( Ref_Spec, Dummy, 254 ) ;

    Dummy := Pos( '*', In_Spec + '*' ) ;
    Inp := Copy( In_Spec, 1, Dummy - 1 ) ;
    In_Spec := Copy( In_Spec, Dummy, 254 ) ;

    if Length( Inp ) < Length( Ref ) then
    begin
	Match_Wildcard := False ;
	Exit
    end ;
    Dummy := 1 ;
    while( Dummy <= Length( Ref ) ) do
    begin
	if( ( Ref[ Dummy ] <> '?' ) and ( Ref[ Dummy ] <> Inp[ Dummy ] ) ) then
	begin
	    Match_Wildcard := False ;
	    Exit
	end ;
	Dummy := Dummy + 1 ; { Try next character }
    end ;
    Inp := Copy( Inp, Length( Ref ) + 1, 254 ) ; { Get remaining characters }
    if( ( Length( Inp ) > 0 ) and ( Ref_Spec <> '*' ) ) then
    begin
	Match_Wildcard := False ;
	Exit ;
    end ;

    Match_Wildcard := True ; { Everything matched }
end ;


function Match_Dir( Ref_Spec, In_Spec : string ) : Boolean ;

var Dummy : Integer ;
    Ref, Inp : string ; { Work variables }

begin
    while( True ) do
    begin
        if(
            (
              ( Length( Ref_Spec ) = 0 )
              or
              ( Ref_Spec = '\...\' )
            )
            and
            ( Length( In_Spec ) = 0 )
          ) then
        begin
            Match_Dir := True ;
            Exit
        end ;
        if( Ref_Spec = '\...\' ) then { Match everything else }
        begin
            Match_Dir := True ;
            Exit
        end ;
        Dummy := Pos( '\', Copy( Ref_Spec, 2, 254 ) + '\' ) + 1 ;
        Ref := Copy( Ref_Spec, 1, Dummy - 1 ) ; { Current reference directory to use }
        Ref_Spec := Copy( Ref_Spec, Dummy, 254 ) ;

        Dummy := Pos( '\', Copy( In_Spec, 2, 254 ) + '\' ) + 1 ;
        Inp := Copy( In_Spec, 1, Dummy - 1 ) ; { Current input directory to use }
        In_Spec := Copy( In_Spec, Dummy, 254 ) ;

        if( not Match_Wildcard( Ref, Inp ) ) then
        begin
            Match_Dir := False ;
            Exit ;
        end ;
    end ; { while( True ) }
end ; { Match_Dir }


function WC_Match( Wildcard, Filename : string ) : Boolean ;
{ Compare a filename and a wildcard specification and return TRUE if a match. }

var Dummy : Integer ; { Dummy variable }
    Ref_Dev, In_Dev : string ; { Input/output device }
    Ref_Dir, In_Dir : string ; { Input/output directory path }
    Ref_Name, In_Name : string ; { Input/output filename }
    Ref_Ext, In_Ext : string ; { Input/output extension }

begin
    { Break filename and wildcard into component pieces... }
    Split( Filename, In_Dir, In_Name, In_Ext ) ;
    Split( Wildcard, Ref_Dir, Ref_Name, Ref_Ext ) ;
    Dummy := Pos( ':', In_Dir ) ; { Locate device in filename }
    In_Dev := Copy( In_Dir, 1, Dummy ) ;
    In_Dir := Copy( In_Dir, Dummy + 1, 254 ) ;
    Dummy := Pos( ':', Ref_Dir ) ; { Locate device in wildcard }
    Ref_Dev := Copy( Ref_Dir, 1, Dummy ) ;
    Ref_Dir := Copy( Ref_Dir, Dummy + 1, 254 ) ;

    if( ( Ref_Dev <> '*:' ) and ( Ref_Dev <> In_Dev ) ) then { Device mismatch }
    begin
	Wc_Match := False ;
	Exit ;
    end ;

    if( not Match_Dir( Ref_Dir, In_Dir ) ) then
    begin
	Wc_Match := False ;
	Exit ;
    end ;
    if( not Match_Wildcard( Ref_Name, In_Name ) ) then
    begin
	Wc_Match := False ;
	Exit ;
    end ;
    if( length( In_Ext ) = 0 ) then
    begin
        In_Ext := '.' ;
    end ;
    if( length( Ref_Ext ) = 0 ) then
    begin
        Ref_Ext := '.' ;
    end ;
    if( not Match_Wildcard( Ref_Ext, In_Ext ) ) then
    begin
	Wc_Match := False ;
	Exit
    end ;

    WC_Match := True { Everything matched }
end ; { WC_Match }



end.
